#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/conf.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/ioctl.h>
#include <sys/map.h>
#include <sys/file.h>
#include <sys/tty.h>
#include <sys/uio.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/errno.h>

#include <sys/krpc.h>
#include <sys/krpcdesc.h>

int 		nkrpc;
struct krpc_d	*krpc_dtab;

extern struct krpcserv krpc_servtab[];

#define Q_INIT(q)	((q).tail = &(q).head)
#define Q_ISEMPTY(q)	((q).tail == &(q).head)

/* 
 * Only network interfaces need to be masked out. 
 */
#define splkrpc	splimp

int	 krpcopen __P((dev_t, int, int, struct proc *));
int	 krpcclose __P((dev_t, int, int, struct proc *));
int	 krpcread __P((dev_t, struct uio *, int));
int	 krpcwrite __P((dev_t, struct uio *, int));
int	 krpcioctl __P((dev_t, int, caddr_t, int, struct proc *));
int	 krpcselect __P((dev_t, int, struct proc *));

void	 krpc_establish __P((int));
int	 krpcupcall __P((int, size_t, u_char *, size_t *, u_char *));
int	 krpcreply __P((int, size_t, u_char *));

static void		 krpc_freed __P((struct krpc_d *));
static void		 krpc_freeq __P((struct krpc_msgq *));
static int 		 krpc_enq __P((struct krpc_msgq *, 
				       struct krpc_msghdr *));
static struct krpc_msghdr  *krpc_deq __P((struct krpc_msgq *));

struct devsw krpcsw = {
	NULL, 
	krpcopen, krpcclose, krpcread, krpcwrite, krpcioctl, krpcselect, 
	nommap, nostrat, nodump, nopsize, 0, nostop
};

/* 
 * Called at boot time, to establish the pseudo-device.
 */
void
krpcattach(n)
	int n;
{
	register int i;
	register struct krpc_d *d;

	krpc_dtab = malloc(n * sizeof(*krpc_dtab), M_DEVBUF, M_WAIT);
	nkrpc = n;
	for (d = krpc_dtab, i = 0; i < n; ++d, ++i)
		d->in_use = 0;
}

int
krpcopen(dev, flag, ifmt, p)
	dev_t dev;
	int flag, ifmt;
	struct proc *p;
{
	register struct krpc_d *d;
	int ks = minor(dev);
	int error = 0;

	if (ks >= nkrpc)
		return (ENXIO);
	
	/*
	 * Each minor can be opened by only one process.  If the requested
	 * minor is in use, return EBUSY.
	 */
	d = &krpc_dtab[ks];
	if (d->in_use)
		return (EBUSY);

	bzero(d, sizeof(*d));
	d->in_use = 1;
	d->ks = ks;
	Q_INIT(d->upq);
	Q_INIT(d->downq);
	
	if (krpc_servtab[ks].krpc_init != 0 && 
	    (error = (*krpc_servtab[ks].krpc_init)()) != 0)
		krpc_freed(d);

	return (error);
}

int
krpcclose(dev, flag, ifmt, p)
	dev_t dev;
	int flag, ifmt;
	struct proc *p;
{
	register struct krpc_d *d = &krpc_dtab[minor(dev)];
	int s;

	if (krpc_servtab[d->ks].krpc_close != 0)
		(*krpc_servtab[d->ks].krpc_close)();

	s = splkrpc();
	krpc_freed(d);
	splx(s);
	

	return (0);
}

static void
krpc_freed(d)
	struct krpc_d *d;
{
	krpc_freeq(&d->upq);
	krpc_freeq(&d->downq);
	d->in_use = 0;
}

static void
krpc_freeq(q)
	struct krpc_msgq *q;
{
	struct krpc_msg_l *m;
	
	if (Q_ISEMPTY(*q))
		return;
	m = q->head;
	while (m != 0) {
		struct krpc_msg_l *tmp = m->next;
		(void)free((caddr_t)m->msg, M_DEVBUF);
		(void)free((caddr_t)m, M_DEVBUF);
		m = tmp;
	}

	/* 
	 * XXX Don't really need this - krpcopen() will do this.  But just for
	 * good measure... 
	 */
	Q_INIT(*q);	
}

int
krpcwrite(dev, uio, ioflag)
	dev_t dev;
	struct uio *uio;
	int ioflag;
{
	int ks = minor(dev);
	register struct krpc_d *d = &krpc_dtab[ks];
	struct krpc_msghdr mh, *hp;
	u_char *msg;
	u_char *result;
	size_t len;
	int error, s;
	
	if (!d->in_use)
		return (ENXIO);

	if (uio->uio_resid < sizeof(struct krpc_msghdr))
		return (EIO);

	/* Get the header */
	error = uiomove((caddr_t)&mh, sizeof(mh), uio);
	if (error != 0)
		return (error);

	/* Just to make sure... may as well check. */
	if (mh.ks != ks)
		return (EIO);

	if (mh.len > KRPC_MAXDATA)
		return (E2BIG);

	if ((msg = malloc(mh.len + sizeof(mh), M_DEVBUF, M_WAITOK)) == 0)
		return (ENOMEM);

	if ((error = uiomove(msg + sizeof(mh), mh.len, uio)) != 0)
		return (error);

	hp = (struct krpc_msghdr *)msg;
	*hp = mh;

	/* This is a reply to an upcall */
     	if (mh.type == KRPC_UPCALL) {
		if ((error = krpc_enq(&d->downq, hp)) == 0) {
			wakeup((caddr_t)&d->downq);
			selwakeup(&d->ks_sel);
		} else 
			(void)free(msg, M_DEVBUF);
		return (error);
	}

	/* This is a downcall */
	s = (*krpc_servtab[ks].krpc_call)(msg + mh.len, mh.len, &result, &len);
	(void)free(msg, M_DEVBUF);

	/* If the response is delayed, return now. */
	if (s & KRPC_NORES)
		return (0);

	/* XXX Should realloc */
	if ((msg = malloc(len + sizeof(mh), M_DEVBUF, M_WAITOK)) == 0)
		return (ENOMEM);
	
	hp = (struct krpc_msghdr *)msg;
	hp->len = len;
	hp->ks = ks;
	hp->type = KRPC_REPLY;
	bcopy(result, msg + sizeof(mh), len);

	if (s & KRPC_FREERES)
		(void)free((caddr_t)result, M_DEVBUF);

	error = krpc_enq(&d->upq, hp);

	return (error);
}

int
krpcread(dev, uio, ioflag)
	dev_t dev;
	struct uio *uio;
	int ioflag;
{
	register struct krpc_d *d = &krpc_dtab[minor(dev)];
	struct krpc_msghdr *hp;
	int error = 0;
	int s;

	s = splkrpc();
	while (Q_ISEMPTY(d->upq))
		sleep((caddr_t)&d->upq, PZERO);
		
	hp = krpc_deq(&d->upq);
	splx(s);

	error = uiomove((caddr_t)hp, sizeof(*hp) + hp->len, uio);
	(void)free((caddr_t)hp, M_DEVBUF);
			
	return (error);
}

/*
 * krpcreply is called as a reply to a downcall if krpc_call() did not return
 * the response. 
 */
int
krpcreply(ks, len, data)
	int ks;
	size_t len;
	u_char *data;
{
	register struct krpc_d *d;
	struct krpc_msghdr mh, *hp;
	u_char *m;
	int error;
	int s;

	s = splkrpc();

	d = &krpc_dtab[ks];
	if (!d->in_use) {
		splx(s);
		return (ENXIO);
	}

	if ((m = malloc(sizeof(mh) + len, M_DEVBUF, M_WAITOK)) == 0)
		return (ENOMEM);

	hp = (struct krpc_msghdr *)m;
	hp->ks = ks;
	hp->len = len;
	hp->type = KRPC_REPLY;
	bcopy(m + sizeof(mh), data, len);

	if ((error = krpc_enq(&d->upq, hp)) != 0) {
		splx(s);
		return (error);
	}

	wakeup((caddr_t)&d->upq);
	selwakeup(&d->ks_sel);
	splx(s);

	return (0);
}

/* 
 * krpcupcall originates an upcall to the user level process.
 * NOTE: If the process is waiting for a reply to a downcall, the results
 * are undefined.
 *
 * Must be called at splXXX. 
 */
int
krpcupcall(ks, len, args, reslen, result)
	int ks;
	size_t len, *reslen;
	u_char *args, *result; 
{
	register struct krpc_d *d;
	struct krpc_msghdr mh, *hp;
	u_char *m;
	int error = 0;
	size_t l;
	
	d = &krpc_dtab[ks];
	if (!d->in_use)
		return (0);

	/* Prepare the message for sending up. */
	if ((m = malloc(sizeof(mh) + len, M_DEVBUF, M_WAITOK)) == 0)
		return (ENOMEM);

	hp = (struct krpc_msghdr *)m;
	hp->ks = ks;
	hp->len = len;
	hp->type = KRPC_UPCALL;
	bcopy(args, m + sizeof(mh), len);

	if ((error = krpc_enq(&d->upq, hp)) != 0) {
		(void)free((caddr_t)m, M_DEVBUF);
		return (error);
	}

	wakeup((caddr_t)&d->upq);
	selwakeup(&d->ks_sel);

	while (Q_ISEMPTY(d->downq))
		sleep((caddr_t)&d->downq, PZERO);

	if ((hp = krpc_deq(&d->downq)) == 0)
		return (-1);
		
	if (hp->len > KRPC_MAXDATA)
		return (-1);

	*reslen = hp->len;
	bcopy((u_char *)hp + sizeof(struct krpc_msghdr), result, hp->len);
	(void)free((caddr_t)hp, M_DEVBUF);

	return (0);
}

int
krpcioctl(dev, cmd, addr, flag, p)
	dev_t dev;
	int cmd;
	caddr_t addr;
	int flag;
	struct proc *p;
{
	/* None yet */
	return (0);
}



/* 
 * Support for select() system call.
 */
int
krpcselect(dev, rw, p)
	register dev_t dev;
	int rw;
	struct proc *p;
{
	register struct krpc_d *d;
	register int s;

	if (rw != FREAD)
		return (0);

	d = &krpc_dtab[minor(dev)];
	
	s = splkrpc();
	if (Q_ISEMPTY(d->upq)) {
		/* There is data ready */
		splx(s);
		return (1);
	}

	/* No data ready. */
	selrecord(p, &d->ks_sel);
	splx(s);
	
	return (0);
}

static int
krpc_enq(q, hp)
	struct krpc_msgq *q;
	struct krpc_msghdr *hp;
{      
	struct krpc_msg_l *ml;
	
	if ((ml = malloc(sizeof(*ml), M_DEVBUF, M_WAITOK)) == 0)
		return (ENOMEM);
	ml->msg = hp;
	ml->next = 0;
	*q->tail = ml;
	q->tail = &ml->next;

	return (0);
}

static struct krpc_msghdr *
krpc_deq(q)
	struct krpc_msgq *q;
{
	struct krpc_msghdr *hp;
	struct krpc_msg_l *tmp;

	/* Assumes that q is not empty! */

	/* Sanity check. */
	if (Q_ISEMPTY(*q)) {
		printf("krpc_deq: q is empty!");
		return (0);
	}
	hp = q->head->msg;
	if (q->tail == &q->head->next)
		Q_INIT(*q);

	tmp = q->head;
	q->head = q->head->next;
	(void)free((caddr_t)tmp, M_DEVBUF);

	return (hp);
}

