/*-
 * Copyright (c) 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI kern_license.c,v 2.2 1996/10/14 19:50:15 prb Exp
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/errno.h>
#include <sys/license.h>

typedef struct {
	license_t	id;		/* The external form of the license */
	int		users;		/* Number of allowed users */
	int		flags;
} ilicense_t;

/*
 * By default we issues a single user DEMO license
 * Note that the license ID is invalid (all zeros)
 */
static ilicense_t license = {
	{ 0, }, 1, LICENSE_DEMO,
};

int
license_users()
{
	return (license.users);
}

int
license_flags()
{
	return (license.flags);
}

license_t *
license_id()
{
	return (&license.id);
}

int
license_set(license_t *lp)
{
	license_t r;
	u_char s1, s2;
	int i, j;
	int f, u;

	if ((lp->data[1] & 0x80) != 0x80)	/* version 1 license check */
		return (EINVAL);
	
	bzero(&r, sizeof(r));

	r.data[0] = lp->data[0];
	for (i = 0; i < 8; ++i) {
		unsigned char c = lp->data[i+1];
		for (j = 0; j < 8; ++j)
			r.data[j+1] |= (c & (1<<(7-j))) ? (1 << (7-i)) : 0;
	}

	s1 = s2 = 0;
	for (i = 0; i < 9; ++i) {
		s1 += r.data[i];
		s2 += s1;
	}
	s1 += -r.data[8];

	if (s1 || s2)
		return (EINVAL);

	f = (r.data[1] & 0x7f) << 8;
	f |= r.data[2];

	switch (f & (LICENSE_BITS)) {
	case LICENSE_UNLIMITED:
		u = LICENSE_UNLIMITED;
		break;
	case LICENSE_64:
		u = 64;
		break;
	case LICENSE_16:
		u = 16;
		break;
	case LICENSE_4:
		u = 4;
		break;
	default:
		return (EINVAL);
	}
	f &= ~LICENSE_BITS;

	license.flags = f;
	license.users = u;
	hostid = (r.data[3] << 24) |
		 (r.data[4] << 16) |
		 (r.data[5] << 8) |
		  r.data[6];
	bcopy(lp, &license.id, sizeof(license_t));

	return (0);
}
