/*-
 * Copyright (c) 1996 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 * BSDI sysv_ipc.c,v 1.3 1996/07/22 21:39:07 mdickson Exp 
 */

/*
 * Copyright (c) 1988 University of Utah.
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * the Systems Programming Group of the University of Utah Computer
 * Science Department. Originally from University of Wisconsin.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * from: Utah $Hdr: uipc_shm.c 1.9 89/08/14$
 *
 *	@(#)sysv_shm.c	7.15 (Berkeley) 5/13/91
 */

/*
 * IPC routines shared by the SYSV IPC system calls.
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <vm/vm.h>
#include <sys/sysctl.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/msg.h>

/*
 * Perform access checking.
 */

int
ipcperm(cred, ipc, mode)
	register struct ucred *cred;
	register struct ipc_perm *ipc;
	int mode;
{

	if (cred->cr_uid == 0)
		return (0);
	/*
	 * Access check is based on only one of owner, group, public.
	 * If not owner, then check group.
	 * If not a member of the group, then check public access.
	 */
	if (cred->cr_uid != ipc->uid && cred->cr_uid != ipc->cuid) {
		if (mode & IPC_M)
			return (EPERM);
		mode >>= 3;
		if (!groupmember(ipc->gid, cred) &&
		    !groupmember(ipc->cgid, cred))
			mode >>= 3;
	}
	if (mode & IPC_M)
		return (0);

	return ((mode & ipc->mode) == mode ? 0 : EACCES);
}


int
ipc_sysctl(name, namelen, oldp, oldlenp, newp, newlen)
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
{
	switch (name[0]) {
	case SYSVIPC_SEMINFO:
		return (sysctl_rdstruct(oldp, oldlenp, newp, &seminfo,
			sizeof(struct seminfo)));
	case SYSVIPC_SHMINFO:
		return (sysctl_rdstruct(oldp, oldlenp, newp, &shminfo,
			sizeof(struct shminfo)));
	case SYSVIPC_MSGINFO:
		return (sysctl_rdstruct(oldp, oldlenp, newp, &msginfo,
			sizeof(struct msginfo)));
	case SYSVIPC_SHMSEGS:
		if (shmsegs == NULL)
			return (ENOTDIR);	/* Not initialized yet */
		return (sysctl_rdstruct(oldp, oldlenp, newp, shmsegs,
			sizeof(*shmsegs) * shminfo.shmmni)); 
	case SYSVIPC_MSQIDS:
		if (msqids == NULL)
			return (ENOTDIR);	/* Not initialized yet */
		return (sysctl_rdstruct(oldp, oldlenp, newp, msqids,
			sizeof(*msqids) * msginfo.msgmni)); 
	case SYSVIPC_SEMIDS:
		if (sema == NULL)
			return (ENOTDIR);	/* Not initialized yet */
		return (sysctl_rdstruct(oldp, oldlenp, newp, sema,
			sizeof(*sema) * seminfo.semmni)); 
	default:
		return (EOPNOTSUPP);
	}
	/* NOTREACHED */
}
