#ifdef KERNEL
#include <sys/param.h>
#include <sys/cdefs.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#define NEW(type)	((type *) malloc(sizeof(type), M_DEVBUF, M_WAIT))

#else	/* ! KERNEL */
#define NEW(type)	((type *) malloc(sizeof(type)))
#define mtod(m, type)	((type *) m)
#define bzero(p, len)	memset(p, 0, len)
#define bcopy(src, dst, len)	memcpy(dst, src, len)
#endif	/* KERNEL */

#include "cdcbq.h"

#define INIT_WEIGHT	10
#define CL_EH_OFFSET	2
#define CL_LEN		4

int CBQ_getqnum __P((CBQueue *, int));
IdQueue* CBQ_selectq __P((CBQueue *));


int
CBQ_classify(m)
	struct mbuf *m;
{
	unsigned int value;
	struct ip *ip;
	struct ether_header *eh = mtod(m, struct ether_header*);
	bcopy(&(eh->ether_dhost[CL_EH_OFFSET]), &value, CL_LEN);
	value = ntohl(value);

#ifdef IP_CHECK
	m = m_pullup(m, sizeof(struct ether_header) + sizeof(struct ip));
	ip = mtod(m, struct ip *);
#endif
	return value;
}


CBQueue*
CBQ_new()
{
	CBQueue *cbq = NEW(CBQueue);
	bzero(cbq, sizeof(CBQueue));
	return cbq;
}


int
CBQ_enque(cbq, m)
	CBQueue* cbq;
	struct mbuf *m;
{
	if (m) {
		int id = CBQ_classify(m);
		IdQueue *q = cbq->cbq_q[CBQ_getqnum(cbq, id)];
		IF_ENQUEUE(q, m);
		cbq->cbq_len++;
	}
	return cbq->cbq_len;
}


struct mbuf*
CBQ_deque(cbq)
	CBQueue* cbq;
{
	struct mbuf *m = 0;
	IdQueue *q = CBQ_selectq(cbq);
	if (q) {
		IF_DEQUEUE(q, m);
		if (m)
			cbq->cbq_len--;
	}
	return m;
}


void
CBQ_monitor(cbq, eh, snr)
	CBQueue *cbq;
	struct ether_header *eh;
	int snr;
{
	IdQueue *q;
	int i;
	unsigned int id;
	bcopy(&(eh->ether_shost[CL_EH_OFFSET]), &id, CL_LEN);
	id = ntohl(id);
	i = CBQ_getqnum(cbq, id);
	cbq->cbq_q[i]->weight = snr;
}


int
CBQ_status(cbq)
	CBQueue *cbq;
{
	int i;
	IdQueue *q;

	for (i = 0;  i < cbq->cbq_numq;  i++) {
		q = cbq->cbq_q[i];
		printf("cbq[%d]\t%d %d\t%x %d\n",
		       i, q->ifq_len, q->ifq_maxlen, q->ifq_drops,
		       q->id, q->weight);
	}
	return i;
}


int CBQ_weight(q)
	IdQueue *q;
{
	return q->ifq_len ? q->ifq_len + q->weight : 0;
}


int CBQ_getqnum(cbq, id)
	CBQueue *cbq;
	int id;
{
	int i;
	IdQueue **q_ = cbq->cbq_q;

	for (i = 0;  i < cbq->cbq_numq;  i++) {
		if (q_[i]->id == id)
			return i;
	}

	if (i == CBQ_MAX - 1) {
		printf("CBQ_getqnum: \t q[%d] = %x\n", i, id);
		if (q_[i])
			return i;
	}
	else
		cbq->cbq_numq++;

	printf("CBQ_getqnum: new\t q[%d] = %x\n", i, id);
	q_[i] = NEW(IdQueue);
	bzero(q_[i], sizeof(IdQueue));
	q_[i]->id = id;
	q_[i]->weight = INIT_WEIGHT;
	return i;
}


IdQueue* CBQ_selectq(cbq)
	CBQueue *cbq;
{
	int i;
	int r, w = 0;
	int numq_ = cbq->cbq_numq;
	IdQueue **q_ = cbq->cbq_q;

	for (i = 0;  i < numq_;  i++)
		w += CBQ_weight(q_[i]);
	if (q_[numq_])
		w += CBQ_weight(q_[numq_]);

	if (w <= 0)
		return q_[numq_];

	r = random() % w;
	for (i = 0;  i < numq_;  i++) {
		w = CBQ_weight(q_[i]);
		r -= w;
		if (r <= 0)
			return q_[i];
	}
	return q_[i];
}
