/*	BSDI if.h,v 2.16 1996/11/09 00:07:50 prb Exp	*/

/*
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)if.h	8.3 (Berkeley) 2/9/95
 */

#ifndef _NET_IF_H_
#define _NET_IF_H_

/*
 * Structures defining a network interface, providing a packet
 * transport mechanism (ala level 0 of the PUP protocols).
 *
 * Each interface accepts output datagrams of a specified maximum
 * length, and provides higher level routines with input datagrams
 * received from its medium.
 *
 * Output occurs when the routine if_output is called, with three parameters:
 *	(*ifp->if_output)(ifp, m, dst, rt)
 * Here m is the mbuf chain to be sent and dst is the destination address.
 * The output routine encapsulates the supplied datagram if necessary,
 * and then transmits it on its medium.
 *
 * On input, each interface unwraps the data received by it, and either
 * places it on the input queue of a internetwork datagram routine
 * and posts the associated software interrupt, or passes the datagram to a raw
 * packet input routine.
 *
 * Routines exist for locating interfaces by their addresses
 * or for locating a interface on a certain network, as well as more general
 * routing and gateway routines maintaining information used to locate
 * interfaces.  These routines live in the files if.c and route.c
 */
#ifndef _TIME_ /*  XXX fast fix for SNMP, going away soon */
#include <sys/time.h>
#endif

#if defined(KERNEL) && !defined(ATSH_ADD)
#include <sys/systm.h>
#endif

#ifdef __STDC__
/*
 * Forward structure declarations for function prototypes [sic].
 */
struct	mbuf;
struct	proc;
struct	rtentry;	
struct	socket;
struct	ether_header;
#endif

/* Externally-visible interface information */
struct if_data {
/* generic interface information */
	u_char	ifi_type;		/* ethernet, tokenring, etc */
	u_char	ifi_addrlen;		/* media address length */
	u_char	ifi_hdrlen;		/* media header length */
	u_long	ifi_mtu;		/* maximum transmission unit */
	u_long	ifi_metric;		/* routing metric (external only) */
	u_long	ifi_baudrate;		/* linespeed */
/* volatile statistics */
	u_quad_t ifi_ipackets;		/* packets received on interface */
	u_quad_t ifi_ierrors;		/* input errors on interface */
	u_quad_t ifi_opackets;		/* packets sent on interface */
	u_quad_t ifi_oerrors;		/* output errors on interface */
	u_quad_t ifi_collisions;	/* collisions on csma interfaces */
	u_quad_t ifi_ibytes;		/* total number of octets received */
	u_quad_t ifi_obytes;		/* total number of octets sent */
	u_quad_t ifi_imcasts;		/* packets received via multicast */
	u_quad_t ifi_omcasts;		/* packets sent via multicast */
	u_quad_t ifi_iqdrops;		/* dropped on input, this interface */
	u_quad_t ifi_noproto;		/* destined for unsupported protocol */
	struct	timeval ifi_lastchange;	/* last updated */
};

/*
 * Structure defining a queue for a network interface
 * or network protocol.
 */
struct ifqueue {
	struct	mbuf *ifq_head;
	struct	mbuf *ifq_tail;
	int	ifq_len;
	int	ifq_maxlen;
	int	ifq_drops;
};

#ifdef SEMANTIC_QUEUE
#include <netinet/tcp.h>
/* 
 * These are needed because of VJ header compression -- once 
 * compressed, it's hard to get ahold of the packet's header fields 
 * without actually uncompressing, which we want to avoid.  
 * Also, semhdr isn't just a tcpiphdr because we save some space 
 * in the queue by keeping only the fields we are interested in.
 */
struct semhdr {
	u_long  src;
	u_long  dst;
	u_short sport;
	u_short dport;
	u_short size;
	tcp_seq ack;
	u_char  proto;
	u_char  flags;
};

struct semantic_mbuf {
	struct semantic_mbuf *sm_nextpkt;
	struct mbuf   *mb;
	struct semhdr sh;
};

struct sifqueue {
	struct	semantic_mbuf *ifq_head;
	struct	semantic_mbuf *ifq_tail;
	int	ifq_len;
	int	ifq_maxlen;
	int	ifq_drops;
};
#endif

#ifdef LINK_EMULATION
/* callout queue for delay emulation */
struct callout_queue {
	struct callout_queue *next;
	long send_time;
};

/* structure to hold desttination addresses for which emulation is to be enabled */
struct ifemuaddr {
	struct sockaddr if_emuaddr;
	struct ifemuaddr *next;
};

#ifdef PDROP
/* queue to keep track of priority drop packets */
struct pdrop_queue {
	struct mbuf *m;
	struct mbuf *m_prev;
	struct pdrop_queue *next;
	struct pdrop_queue *prev;
};
#endif

#endif

/*
 * Structure describing information about an interface
 * which may be of interest to management entities.
 *
 * (Would like to call this struct ``if'', but C isn't PL/1.)
 */

struct ifnet {
	char	*if_name;		/* name, e.g. ``en'' or ``lo'' */
	struct	ifnet *if_next;		/* all struct ifnets are chained */
	struct	ifaddr *if_addrlist;	/* linked list of addresses per if */
        int	if_pcount;		/* number of promiscuous listeners */
	void	*if_bpf;		/* packet filter structure */
	u_short	if_index;		/* numeric abbreviation for this if  */
	short	if_unit;		/* sub-unit for lower level driver */
	short	if_timer;		/* time 'til if_watchdog called */
	u_long	if_flags;		/* up/down, broadcast, etc. */
	struct	if_data if_data;
	struct	timeval if_reference_time;
/* procedure handles */
	int	(*if_init)		/* init routine */
		__P((int));
	int	(*if_output)		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		     struct rtentry *));
	int	(*if_ctloutput)		/* control output routine */
		__P((int op, struct ifnet *ifp, int level,
		    int optname, struct mbuf **m));
	int	(*if_start)		/* initiate output routine */
		__P((struct ifnet *));
	int	(*if_done)		/* output complete routine */
		__P((struct ifnet *));	/* (XXX not used; fake prototype) */
	int	(*if_ioctl)		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t/*, struct proc * */));
	int	(*if_watchdog)		/* timer routine */
		__P((int));

	struct	ifqueue  if_snd; /* output queue */
#ifdef SEMANTIC_QUEUE
	struct  sifqueue sif_snd; /* output queue with semantic (TCP) info */
#endif
	struct	ifnet *if_pif;		/* back pointer to pif */
	struct  ifqueue if_fastq; /* fast output queue */
#ifdef LLPROTO
	struct  ll_cache *ll_cache;
#endif
#ifdef LINK_EMULATION
	struct ifqueue if_slow_waitq; /* queue to hold packets awaiting transmission on
					 the emulated link (bandwidth emulation) */
        struct ifqueue if_fast_waitq; /* waitq for high priority packets */	
        struct ifqueue if_delayq;  /* queue to to hold packets awaiting propagation
				      across the emulated link (delay emulation) */
        struct callout_queue *calq_head; /* head of callout queue used for delay 
					     emulation */		
        struct callout_queue *calq_tail; /* tail of callout queue used for delay 
					     emulation */		
	int bandwidth_emulation_timer_pending;  /* whether the emulation timer is 
						   pending */
	int delay_emulation_timer_pending;  /* whether the emulation timer is pending */
        struct timeval end_busy_time; /* time when busy period will end */
        struct timeval delayq_last_enque_time; /* last enque time for if_delayq	*/	
        int  waiting_for_line;        /* a packet is waiting for the line to become free */		
	int  if_bandwidth;            /* emulated link bandwidth (Kbps) */
	int  if_delay;                /* emulated link delay (ms) */
        int  if_acksize;              /* size of TCP acks to use for the purposes of
				       bandwidth emulation (bytes) */
	struct ifemuaddr *if_emuaddrlist; /* list of dst addr to enable emulation for */
#ifdef PDROP
        /* queues to keep track of priority drop queues in the slow and fast queues */
        struct pdrop_queue *pdrop_slow_head;
        struct pdrop_queue *pdrop_slow_tail;
        struct pdrop_queue *pdrop_fast_head;
        struct pdrop_queue *pdrop_fast_tail;
#endif
#endif
};

#define	if_mtu		if_data.ifi_mtu
#define	if_type		if_data.ifi_type
#define	if_addrlen	if_data.ifi_addrlen
#define	if_hdrlen	if_data.ifi_hdrlen
#define	if_metric	if_data.ifi_metric
#define	if_baudrate	if_data.ifi_baudrate
#define	if_ipackets	if_data.ifi_ipackets
#define	if_ierrors	if_data.ifi_ierrors
#define	if_opackets	if_data.ifi_opackets
#define	if_oerrors	if_data.ifi_oerrors
#define	if_collisions	if_data.ifi_collisions
#define	if_ibytes	if_data.ifi_ibytes
#define	if_obytes	if_data.ifi_obytes
#define	if_imcasts	if_data.ifi_imcasts
#define	if_omcasts	if_data.ifi_omcasts
#define	if_iqdrops	if_data.ifi_iqdrops
#define	if_noproto	if_data.ifi_noproto
#define	if_lastchange	if_data.ifi_lastchange

#define	IFF_UP		0x1		/* interface is up */
#define	IFF_BROADCAST	0x2		/* broadcast address valid */
#define	IFF_DEBUG	0x4		/* turn on debugging */
#define	IFF_LOOPBACK	0x8		/* is a loopback net */
#define	IFF_POINTOPOINT	0x10		/* interface is point-to-point link */
#define	IFF_NOTRAILERS	0x20		/* avoid use of trailers */
#define	IFF_RUNNING	0x40		/* resources allocated */
#define	IFF_NOARP	0x80		/* no address resolution protocol */
#define	IFF_PROMISC	0x100		/* receive all packets */
#define	IFF_ALLMULTI	0x200		/* receive all multicast packets */
#define	IFF_OACTIVE	0x400		/* transmission in progress */
#define	IFF_SIMPLEX	0x800		/* can't hear own transmissions */
#define	IFF_LINK0	0x1000		/* per link layer defined bit */
#define	IFF_LINK1	0x2000		/* per link layer defined bit */
#define	IFF_LINK2	0x4000		/* per link layer defined bit */
#define	IFF_MULTICAST	0x8000		/* supports multicast */
#define IFF_FASTQ       0x10000         /* fastq enabled */
#define IFF_AFIL        0x20000         /* ack filtering */
#define IFF_PDROP       0x40000         /* priority drop */
#define IFF_NOICMPREDIR 0x80000         /* no icmp redirection */

/* flags set internally only: */
#define	IFF_CANTCHANGE \
	(IFF_BROADCAST|IFF_POINTOPOINT|IFF_RUNNING|IFF_OACTIVE|\
	    IFF_SIMPLEX|IFF_MULTICAST)

/*
 * Output queues (ifp->if_snd) and internetwork datagram level (pup level 1)
 * input routines have queues of messages stored on ifqueue structures
 * (defined above).  Entries are added to and deleted from these structures
 * by these macros, which should be called with ipl raised to splimp().
 */
#define	IF_QFULL(ifq)		((ifq)->ifq_len >= (ifq)->ifq_maxlen)
#define	IF_DROP(ifq)		((ifq)->ifq_drops++)

#define	IF_ENQUEUE(ifq, m) { \
	(m)->m_nextpkt = 0; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_head = m; \
	else \
		(ifq)->ifq_tail->m_nextpkt = m; \
	(ifq)->ifq_tail = m; \
	(ifq)->ifq_len++; \
}
#define	IF_PREPEND(ifq, m) { \
	(m)->m_nextpkt = (ifq)->ifq_head; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_tail = (m); \
	(ifq)->ifq_head = (m); \
	(ifq)->ifq_len++; \
}
#define	IF_DEQUEUE(ifq, m) { \
	(m) = (ifq)->ifq_head; \
	if (m) { \
		if (((ifq)->ifq_head = (m)->m_nextpkt) == 0) \
			(ifq)->ifq_tail = 0; \
		(m)->m_nextpkt = 0; \
		(ifq)->ifq_len--; \
	} \
}

#define IF_REMOVE(ifq, m, prev) { \
	if ((m) == (ifq)->ifq_head) { \
                IF_DEQUEUE((ifq), (m)); \
                m_freem((m)); \
	} else { \
	        (prev)->m_nextpkt = (m)->m_nextpkt; \
	        if ((ifq)->ifq_tail == (m)) \
		        (ifq)->ifq_tail = (prev); \
	        m_freem((m)); \
	        (ifq)->ifq_len--; \
	} \
}

#ifdef SEMANTIC_QUEUE
/* 
 * There is really no reason to rewrite ENQUEUE and PREPEND, but
 * we're unfortunately forced to because m_nextpkt cannot be a
 * field in semantic_mbuf.  m_nextpkt has been #define'd in <sys/mbuf.h>
 */
/*
#define	SIF_ENQUEUE(ifq, m) { \
	(m)->sm_nextpkt = 0; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_head = m; \
	else \
		(ifq)->ifq_tail->sm_nextpkt = m; \
	(ifq)->ifq_tail = m; \
	(ifq)->ifq_len++; \
}
*/
#define	SIF_PREPEND(ifq, m) { \
	(m)->sm_nextpkt = (ifq)->ifq_head; \
	if ((ifq)->ifq_tail == 0) \
		(ifq)->ifq_tail = (m); \
	(ifq)->ifq_head = (m); \
	(ifq)->ifq_len++; \
}

/*
#define SIF_DEQUEUE(ifq, m) { \
	if ((ifq)->ifq_head && ((m) = (ifq)->ifq_head->mb)) { \
		if (((ifq)->ifq_head = (ifq)->ifq_head->sm_nextpkt) == 0) \
		        (ifq)->ifq_tail = 0; \
		(m)->m_nextpkt = 0; \
		(ifq)->ifq_len--; \
		free((ifq)->ifq_head, M_PCB); \
	} \
}

#define SIF_REMOVE(ifq, m, prev) { \
	(prev)->sm_nextpkt = (m)->sm_nextpkt; \
	if ((ifq)->ifq_tail == (m)) \
		(ifq)->ifq_tail = (prev); \
	m_freem((m)->mb); \
	free((m), M_PCB); \
}
*/
#endif
	
#ifdef SHORT_IFQ
#define	IFQ_MAXLEN	10
#else
#define	IFQ_MAXLEN	50
#endif

#define IFFASTQ_MAXLEN	32
#define	IFNET_SLOWHZ	1		/* granularity is 1 second */

/*
 * control message types sent for an interface
 */
#define	IFS_FLAGS	1		/* if_flags changed */
#define	IFS_DCD		2		/* data carrier detect changed */

#ifdef	KERNEL
struct protoq {
	struct	wayout pq_wayout;	/* how we get called */
	struct	ifqueue pq_queue;	/* input queue for protocol */
};

#define	PQ_INIT(pq, fn)	{ \
	(pq)->pq_queue.ifq_maxlen = IFQ_MAXLEN; \
	(pq)->pq_wayout.func = (fn); \
	(pq)->pq_wayout.arg = (pq); \
}

/* must be protected by splimp */
#define	PQ_ENQUEUE(pq, m) { \
	IF_ENQUEUE(&(pq)->pq_queue, (m)); \
	if ((pq)->pq_wayout.pending == 0) \
		wayout(&(pq)->pq_wayout); \
}
#endif

/*
 * The ifaddr structure contains information about one address
 * of an interface.  They are maintained by the different address families,
 * are allocated and attached when an address is set, and are linked
 * together so all addresses for an interface can be located.
 */
struct ifaddr {
	struct	sockaddr *ifa_addr;	/* address of interface */
	struct	sockaddr *ifa_dstaddr;	/* other end of p-to-p link */
#define	ifa_broadaddr	ifa_dstaddr	/* broadcast address interface */
	struct	sockaddr *ifa_netmask;	/* used to determine subnet */
	struct	ifnet *ifa_ifp;		/* back-pointer to interface */
	struct	ifaddr *ifa_next;	/* next address for interface */
	void	(*ifa_rtrequest)();	/* check or clean routes (+ or -)'d */
	u_short	ifa_flags;		/* mostly rt_flags for cloning */
	int	ifa_refcnt;		/* # of references */
	int	ifa_metric;		/* cost of going out this interface */
#ifdef notdef
	struct	rtentry *ifa_rt;	/* XXXX for ROUTETOIF ????? */
#endif
};
#define	IFA_ROUTE	RTF_UP		/* route installed */

/*
 * Message format for use in obtaining information about interfaces
 * from sysctl and the routing socket
 */
struct if_msghdr {
	u_short	ifm_msglen;	/* to skip over non-understood messages */
	u_char	ifm_version;	/* future binary compatability */
	u_char	ifm_type;	/* message type */
	int	ifm_addrs;	/* like rtm_addrs */
	int	ifm_flags;	/* value of if_flags */
	u_short	ifm_index;	/* index for associated ifp */
	struct	if_data ifm_data;/* statistics and other data about if */
};

/*
 * Message format for use in obtaining information about interface addresses
 * from sysctl and the routing socket
 */
struct ifa_msghdr {
	u_short	ifam_msglen;	/* to skip over non-understood messages */
	u_char	ifam_version;	/* future binary compatability */
	u_char	ifam_type;	/* message type */
	int	ifam_addrs;	/* like rtm_addrs */
	int	ifam_flags;	/* value of ifa_flags */
	u_short	ifam_index;	/* index for associated ifp */
	int	ifam_metric;	/* value of ifa_metric */
};

/*
 * Interface request structure used for socket
 * ioctl's.  All interface ioctl's must have parameter
 * definitions which begin with ifr_name.  The
 * remainder may be interface specific.
 */
struct	ifreq {
#define	IFNAMSIZ	16
	char	ifr_name[IFNAMSIZ];		/* if name, e.g. "en0" */
	union {
		struct	sockaddr ifru_addr;
		struct	sockaddr ifru_dstaddr;
		struct	sockaddr ifru_broadaddr;
		u_long	ifru_flags;
		int	ifru_metric;
		caddr_t	ifru_data;
#ifdef LINK_EMULATION
		int     ifru_bandwidth;
		int     ifru_delay;
		int     ifru_acksize;
		int     ifru_maxlen;
		struct  sockaddr ifru_emuaddr;
		int     ifru_emuaddr_index;
#endif
	} ifr_ifru;
#define	ifr_addr	ifr_ifru.ifru_addr	/* address */
#define	ifr_dstaddr	ifr_ifru.ifru_dstaddr	/* other end of p-to-p link */
#define	ifr_broadaddr	ifr_ifru.ifru_broadaddr	/* broadcast address */
#define	ifr_flags	ifr_ifru.ifru_flags	/* flags */
#define	ifr_metric	ifr_ifru.ifru_metric	/* metric */
#define	ifr_media	ifr_ifru.ifru_metric	/* media options */
#define	ifr_data	ifr_ifru.ifru_data	/* for use by interface */
#ifdef LINK_EMULATION
#define ifr_bandwidth   ifr_ifru.ifru_bandwidth
#define ifr_delay       ifr_ifru.ifru_delay
#define ifr_acksize     ifr_ifru.ifru_acksize
#define ifr_maxlen      ifr_ifru.ifru_maxlen
#define ifr_emuaddr     ifr_ifru.ifru_emuaddr
#define ifr_emuindex    ifr_ifru.ifru_emuaddr_index
#endif
};

struct ifaliasreq {
	char	ifra_name[IFNAMSIZ];		/* if name, e.g. "en0" */
	struct	sockaddr ifra_addr;
	struct	sockaddr ifra_broadaddr;
	struct	sockaddr ifra_mask;
};

struct ifmediareq {
	char	ifm_name[IFNAMSIZ];		/* if name, e.g. "en0" */
	int	ifm_current;			/* return: current media opts */
	int	ifm_mask;			/* return: current mask */
	int	ifm_status;			/* return: current media stat */
	int	ifm_active;			/* return: active opts */
	int	ifm_count;			/* input: words in ifm_ulist */
	int	*ifm_ulist;			/* output: media words */
};

/*
 * Structure used in SIOCGIFCONF request.
 * Used to retrieve interface configuration
 * for machine (useful for programs which
 * must know all networks accessible).
 */
struct	ifconf {
	int	ifc_len;		/* size of associated buffer */
	union {
		caddr_t	ifcu_buf;
		struct	ifreq *ifcu_req;
	} ifc_ifcu;
#define	ifc_buf	ifc_ifcu.ifcu_buf	/* buffer address */
#define	ifc_req	ifc_ifcu.ifcu_req	/* array of structures returned */
};

#include <net/if_arp.h>

#ifdef KERNEL
#define	IFAFREE(ifa) \
	if ((ifa)->ifa_refcnt <= 0) \
		ifafree(ifa); \
	else \
		(ifa)->ifa_refcnt--;

struct	ifnet	*ifnet;
extern	int if_index;
extern	struct ifaddr **ifnet_addrs;

void	ether_attach __P((struct ifnet *));
void	ether_input __P((struct ifnet *, struct ether_header *, struct mbuf *));
int	ether_output __P((struct ifnet *,
	   struct mbuf *, struct sockaddr *, struct rtentry *));
char	*ether_sprintf __P((u_char *));

void	if_attach __P((struct ifnet *));
void	if_newaddr __P((struct ifnet *, int, caddr_t));
void	if_down __P((struct ifnet *));
void	if_qflush __P((struct ifqueue *));
void	if_slowtimo __P((void *));
void	if_up __P((struct ifnet *));
void	if_statechg __P((struct ifnet *));
#ifdef vax
void	ifubareset __P((int));
#endif
int	ifconf __P((int, caddr_t));
void	ifinit __P((void));
int	ifioctl __P((struct socket *, u_long, caddr_t, struct proc *));
int	ifpromisc __P((struct ifnet *, int));
struct	ifnet *ifunit __P((char *));

struct	ifaddr *ifa_ifwithaddr __P((struct sockaddr *));
struct	ifaddr *ifa_ifwithaf __P((int));
struct	ifaddr *ifa_ifwithdstaddr __P((struct sockaddr *));
struct	ifaddr *ifa_ifwithnet __P((struct sockaddr *));
struct	ifaddr *ifa_ifwithroute __P((int, struct sockaddr *,
					struct sockaddr *));
struct	ifaddr *ifaof_ifpforaddr __P((struct sockaddr *, struct ifnet *));
void	ifafree __P((struct ifaddr *));
void	link_rtrequest __P((int, struct rtentry *, struct sockaddr *));

int	loioctl __P((struct ifnet *, u_long, caddr_t));
void	loopattach __P((int));
int	looutput __P((struct ifnet *,
	   struct mbuf *, struct sockaddr *, struct rtentry *));
void	lortrequest __P((int, struct rtentry *, struct sockaddr *));

#ifdef LINK_EMULATION
void bandwidth_emulation_timer_handler __P((void *));
void delay_emulation_timer_handler __P((void *));
void do_bandwidth_emulation __P((struct ifnet *));
void do_delay_emulation __P((struct ifnet *));
void filteracks __P((struct ifqueue *, struct mbuf *));
int  is_emuaddr __P((struct ifnet *, struct sockaddr *));

#ifdef PDROP
void if_pdrop_enque __P((struct ifnet *, struct ifqueue *, struct mbuf *));
void if_pdrop_deque __P((struct ifnet *, struct ifqueue *, struct mbuf *));
int if_pdrop_remove __P((struct ifnet *, struct ifqueue *));
#endif

#endif
#endif

#endif /* !_NET_IF_H_ */
