/*	BSDI if_ethersubr.c,v 2.11 1996/08/16 04:01:49 dab Exp	*/

/*
 * Copyright (c) 1982, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)if_ethersubr.c	8.2 (Berkeley) 4/4/96
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>

#include <machine/cpu.h>

#include <net/if.h>
#include <net/netisr.h>
#include <net/route.h>
#include <net/if_llc.h>
#include <net/if_dl.h>
#include <net/if_types.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#endif
#include <netinet/if_ether.h>

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#ifdef ISO
#include <netiso/argo_debug.h>
#include <netiso/iso.h>
#include <netiso/iso_var.h>
#include <netiso/iso_snpac.h>
#endif

#ifdef LLC
#include <netccitt/dll.h>
#include <netccitt/llc_var.h>
#endif

#if defined(LLC) && defined(CCITT)
extern struct ifqueue pkintrq;
#endif

struct protoq ether_protoq;

u_char	etherbroadcastaddr[ETHER_ADDR_LEN] =
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
extern	struct ifnet loif;

#define senderr(e) { error = (e); goto bad;}

#ifdef LINK_EMULATION
#include <netinet/tcp.h>
#include <netinet/tcp_seq.h>

#define tv_diff(tv1, tv2) ((((tv1)->tv_sec) - (tv2)->tv_sec)*1000000 + \
			       (((tv1)->tv_usec) - (tv2)->tv_usec))
#define tv_plus_usec(tv, usec) { (tv)->tv_usec += (usec); \
	(tv)->tv_sec += (tv)->tv_usec/1000000; (tv)->tv_usec %= 1000000;}
#define tv_lt(tv1, tv2) ((((tv1)->tv_sec < (tv2)->tv_sec) || (((tv1)->tv_sec == (tv2)->tv_sec) && ((tv1)->tv_usec < (tv2)->tv_usec))) ? 1 : 0 )
#define tv_to_usec(tv) ((u_long) ((tv)->tv_sec*1000000 + (tv)->tv_usec))

#define get_ticks(time) (((time) + 1000000/(2*hz))*hz/1000000)
#define US_PER_TICK 1000000/hz

#endif

/*
 * Ethernet output routine.
 * Encapsulate a packet of type family for the local net.
 * Assumes that ifp is actually pointer to arpcom structure.
 */
int
ether_output(ifp, m0, dst, rt0)
	register struct ifnet *ifp;
	struct mbuf *m0;
	struct sockaddr *dst;
	struct rtentry *rt0;
{
	u_short type;
	int s, error = 0;
 	u_int16_t edst[ETHER_ADDR_LEN/2];
	register struct mbuf *m = m0;
	register struct rtentry *rt;
	struct mbuf *mcopy = (struct mbuf *)0;
	register struct ether_header *eh;
	int len = m->m_pkthdr.len;
	struct arpcom *ac = (struct arpcom *)ifp;

	struct ip *ip = mtod(m, struct ip *);
	struct ifqueue *ifq;

#ifdef LINK_EMULATION
	struct sockaddr ip_dst;
	struct ifqueue *if_waitq;
	/*
	 * Now check for some TCP-specific fields. Before extracting the
	 * fields, make sure that both the IP and the TCP headers fit in
	 * the first mbuf. If they do not, skip the various checks and
	 * print out a warning.
	 */
	if (ip->ip_p == IPPROTO_TCP) { 
		if (m->m_len >= sizeof(struct ip) + sizeof(struct tcphdr)) {
			struct tcphdr *th = 
				(struct tcphdr *)&((int *)ip)[ip->ip_hl];
			/* if this is a pure ack, set the M_TCPACK flag */
			if (th->th_flags & TH_ACK && 
			   (ntohs(ip->ip_len)==(ip->ip_hl<<2)+(th->th_off<<2)))
				m->m_flags |= M_TCPACK;
#ifdef PDROP
			/* if this is a fast start packet, set M_PDROP flag */
			if (th->th_x2 & TH_FS)
				m->m_flags |= M_PDROP;
#endif
		} else {
			printf("Warning: first mbuf too short. Unable to check TCP header of a %d-byte packet\n", m->m_pkthdr.len);
		}
	}
#endif

	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING))
		senderr(ENETDOWN);
	if (rt = rt0) {
		if ((rt->rt_flags & RTF_UP) == 0) {
			if (rt0 = rt = rtalloc1(dst, 1)) {
				rt->rt_refcnt--;
				if (rt->rt_ifp != ifp)
					return (*rt->rt_ifp->if_output)
							(ifp, m0, dst, rt);
			} else 
				senderr(EHOSTUNREACH);
		}
		if ((rt->rt_flags & RTF_GATEWAY) && dst->sa_family != AF_NS) {
			if (rt->rt_gwroute == 0)
				goto lookup;
			if (((rt = rt->rt_gwroute)->rt_flags & RTF_UP) == 0) {
				rtfree(rt); rt = rt0;
			lookup: rt->rt_gwroute = rtalloc1(rt->rt_gateway, 1);
				if ((rt = rt->rt_gwroute) == 0)
					senderr(EHOSTUNREACH);
				/* the "G" test below also prevents rt == rt0 */
				if ((rt->rt_flags & RTF_GATEWAY) ||
				    (rt->rt_ifp != ifp)) {
					rt->rt_refcnt--;
					rt0->rt_gwroute = 0;
					senderr(EHOSTUNREACH);
				}
			}
		}
		if (rt->rt_flags & RTF_REJECT)
			senderr(rt == rt0 ? EHOSTDOWN : EHOSTUNREACH);
	}
	switch (dst->sa_family) {

#ifdef INET
	case AF_INET:
		if (!arpresolve(ac, rt, m, dst, (u_char *)edst))
			return (0);	/* if not yet resolved */
		/* If broadcasting on a simplex interface, loopback a copy */
		if ((m->m_flags & M_BCAST) && (ifp->if_flags & IFF_SIMPLEX))
			mcopy = m_copy(m, 0, (int)M_COPYALL);
		type = htons(ETHERTYPE_IP);
		break;
#endif
#ifdef NS
	case AF_NS:
		type = htons(ETHERTYPE_NS);
 		ETHER_COPY(&(((struct sockaddr_ns *)dst)->sns_addr.x_host),
			    edst);
		if (!bcmp((caddr_t)edst, (caddr_t)&ns_thishost, sizeof(edst)))
			return (looutput(ifp, m, dst, rt));
		/* If broadcasting on a simplex interface, loopback a copy */
		if ((m->m_flags & M_BCAST) && (ifp->if_flags & IFF_SIMPLEX))
			mcopy = m_copy(m, 0, (int)M_COPYALL);
		break;
#endif
#ifdef	ISO
	case AF_ISO: {
		int	snpalen;
		struct	llc *l;
		register struct sockaddr_dl *sdl;

		if (rt && (sdl = (struct sockaddr_dl *)rt->rt_gateway) &&
		    sdl->sdl_family == AF_LINK && sdl->sdl_alen > 0) {
			ETHER_COPY(LLADDR(sdl), edst);
		} else if (error = iso_snparesolve(ifp,
		    (struct sockaddr_iso *)dst, (char *)edst, &snpalen))
			goto bad; /* Not Resolved */
		/* If multicasting on a simplex interface, loopback a copy */
		if (ETHER_IS_MULTICAST((u_char *)edst))
			m->m_flags |= (M_BCAST|M_MCAST);
		if ((m->m_flags & M_BCAST) && (ifp->if_flags & IFF_SIMPLEX) &&
		    (mcopy = m_copy(m, 0, (int)M_COPYALL))) {
			M_PREPEND(mcopy, sizeof (*eh), M_DONTWAIT);
			if (mcopy) {
				eh = mtod(mcopy, struct ether_header *);
				ETHER_COPY(edst, eh->ether_dhost);
				ETHER_COPY(ac->ac_enaddr, eh->ether_shost);
			}
		}
		M_PREPEND(m, 3, M_DONTWAIT);
		if (m == NULL)
			return (0);
		type = htons(m->m_pkthdr.len);
		l = mtod(m, struct llc *);
		l->llc_dsap = l->llc_ssap = LLC_ISO_LSAP;
		l->llc_control = LLC_UI;
		len += 3;
		IFDEBUG(D_ETHER)
			int i;
			printf("unoutput: sending pkt to: ");
			for (i=0; i<ETHER_ADDR_LEN; i++)
				printf("%x ", ((u_char *)edst)[i] & 0xff);
			printf("\n");
		ENDDEBUG
		} break;
#endif /* ISO */
#ifdef	LLC
/*	case AF_NSAP: */
	case AF_CCITT: {
		register struct sockaddr_dl *sdl = 
			(struct sockaddr_dl *) rt -> rt_gateway;

		if (sdl && sdl->sdl_family == AF_LINK
		    && sdl->sdl_alen > 0) {
			ETHER_COPY(LLADDR(sdl), edst);
		} else
			goto bad; /* Not a link interface ? Funny ... */
		if (ETHER_IS_MULTICAST((u_char *)edst) &&
		    (ifp->if_flags & IFF_SIMPLEX) &&
		    (mcopy = m_copy(m, 0, (int)M_COPYALL))) {
			M_PREPEND(mcopy, sizeof (*eh), M_DONTWAIT);
			if (mcopy) {
				eh = mtod(mcopy, struct ether_header *);
				ETHER_COPY(edst, eh->ether_dhost);
				ETHER_COPY(ac->ac_enaddr, eh->ether_shost);
			}
		}
		type = htons((u_short) m->m_pkthdr.len);
#ifdef LLC_DEBUG
		{
			int i;
			register struct llc *l = mtod(m, struct llc *);

			printf("ether_output: sending LLC2 pkt to: ");
			for (i=0; i<ETHER_ADDR_LEN; i++)
				printf("%x ", ((u_char *)edst)[i] & 0xff);
			printf(" len 0x%x dsap 0x%x ssap 0x%x control 0x%x\n", 
			       m->m_pkthdr.len, l->llc_dsap & 0xff,
			       l->llc_ssap &0xff, l->llc_control & 0xff);

		}
#endif /* LLC_DEBUG */
		} break;
#endif /* LLC */	

	case AF_UNSPEC:
		eh = (struct ether_header *)dst->sa_data;
 		ETHER_COPY(eh->ether_dhost, edst);
		type = eh->ether_type;
		break;

	default:
		printf("%s%d: can't handle af%d\n", ifp->if_name, ifp->if_unit,
			dst->sa_family);
		senderr(EAFNOSUPPORT);
	}

	if (mcopy)
		(void) looutput(ifp, mcopy, dst, rt);
	/*
	 * Add local net header.  If no space in first mbuf,
	 * allocate another.
	 */
	M_PREPEND(m, sizeof (struct ether_header), M_DONTWAIT);
	if (m == 0)
		senderr(ENOBUFS);
	eh = mtod(m, struct ether_header *);
#if defined(LLC) || defined(ISO)
	/* If it is not aligned, use bcopy() */
	if (((char *)eh - (char *)0) & 1) {
		bcopy((caddr_t)&type, (caddr_t)&eh->ether_type,
		    sizeof(eh->ether_type));
		bcopy((caddr_t)edst, (caddr_t)eh->ether_dhost, sizeof (edst));
		bcopy((caddr_t)ac->ac_enaddr, (caddr_t)eh->ether_shost,
		    sizeof(eh->ether_shost));
	} else
#endif
		ETHER_FILL_HDR(edst, ac->ac_enaddr, type, eh);

	s = splimp();
	/*
	 * Queue message on interface, and start output if interface
	 * not yet active. 
	 */
	ifq = &ifp->if_snd;
#ifdef FASTQ
	/* XXX temporary */
	if (ip->ip_p == IPPROTO_ICMP)
		ifq = &ifp->if_fastq;
	/* 
	 * If IPTOS_LOWDELAY is set, then use the fast queue,
	 * else use the slow queue.
	 */
        if (ip->ip_tos & IPTOS_LOWDELAY && !IF_QFULL(&ifp->if_fastq))
		ifq = &ifp->if_fastq;
#endif


#ifdef LINK_EMULATION
#ifdef FASTQ
	if (ifq == &ifp->if_fastq && !(ifp->if_flags & IFF_FASTQ))
		ifq = &ifp->if_snd;

#endif
	bzero((caddr_t) &ip_dst, sizeof(ip_dst));
	ip_dst.sa_len = sizeof(ip_dst);
	ip_dst.sa_family = AF_INET;
	bcopy((caddr_t) &ip->ip_dst, (caddr_t) &ip_dst.sa_data[2], sizeof(struct in_addr));
	if ((ifp->if_bandwidth > 0 || ifp->if_delay > 0) && 
	    (is_emuaddr(ifp,&ip_dst) || !ifp->if_emuaddrlist)) {
		if_waitq = &ifp->if_slow_waitq;
#ifdef FASTQ
		/* XXX temporary */
		if (ip->ip_p == IPPROTO_ICMP && (ifp->if_flags & IFF_FASTQ))
			if_waitq = &ifp->if_fast_waitq;

		if ((ip->ip_tos & IPTOS_LOWDELAY) && 
		    (ifp->if_flags & IFF_FASTQ))
			if_waitq = &ifp->if_fast_waitq;
#endif
		/* 
		 * Check if queue is full. We compare the sum of the lengths of
		 * the slow and fast wait queues with ifq_maxlen to determine if
		 * the queue is full. We do not consider the lengths of the
		 * slow and fast interface queues here. There is a small chance
		 * that we do not find the emulated queues full here, but at a
		 * later point (in dod_delay_emulation()) find the interface queues 
		 * full. We consider such an event to be an error.
		 */
		if (ifp->if_snd.ifq_len + ifp->if_fastq.ifq_len +
		    ifp->if_slow_waitq.ifq_len + ifp->if_fast_waitq.ifq_len
		    >= ifq->ifq_maxlen) { 
			IF_DROP(ifq);
#ifdef PDROP
			/* 
			 * If in priority drop mode, search for a packet with the
			 * M_PDROP flag set. If such a packet is found, drop it
			 * (by default, the marked packet closest to the tail is
			 * dropped) and enqueue the new packet. Otherwise, drop
			 * the new packet just as usual.
			 */
			if (ifp->if_flags & IFF_PDROP) {
				if (if_pdrop_remove(ifp, if_waitq)) {
					IF_ENQUEUE(if_waitq, m);
					m = 0;
				}
			}
#endif
			splx(s);
			senderr(ENOBUFS);
		}
		if (ifp->if_flags & IFF_AFIL && m->m_flags & M_TCPACK)
			filteracks(if_waitq, m);
#ifdef PDROP
		/* 
		 * We do the enque to the pdrop_queue before that to the ifq
		 * so that if_pdrop_enque() has access to the previous packet.
		 */
		if (ifp->if_flags & IFF_PDROP) 
			if_pdrop_enque(ifp, if_waitq, m);
#endif
		IF_ENQUEUE(if_waitq, m);
			
		if (!ifp->bandwidth_emulation_timer_pending)
			do_bandwidth_emulation(ifp);
		else if (!ifp->waiting_for_line)
			printf("in ether_output(): inconsistency between bw_emu_timer_pend and waiting_for_line\n");
	} else {
		if (IF_QFULL(ifq)) {
			IF_DROP(ifq);
			splx(s);
			senderr(ENOBUFS);
		}
		IF_ENQUEUE(ifq, m);
	}
#else
	if (IF_QFULL(ifq)) {
		IF_DROP(ifq);
		splx(s);
		senderr(ENOBUFS);
	}
	IF_ENQUEUE(ifq, m);
#endif

	if ((ifp->if_flags & IFF_OACTIVE) == 0)
		(*ifp->if_start)(ifp);
	splx(s);
	ifp->if_obytes += len + sizeof (struct ether_header);
	if (m->m_flags & M_MCAST)
		ifp->if_omcasts++;
	return (error);

bad:
	if (m)
		m_freem(m);
	return (error);
}

#ifdef LINK_EMULATION

#ifdef PDROP
/* Enqueue to tail of pdrop_queue */
void
if_pdrop_enque(ifp, ifq, m)
	struct ifnet *ifp;
	struct ifqueue *ifq;
	struct mbuf *m;
{
	struct pdrop_queue **headp, **tailp;
	struct pdrop_queue *head, *tail, *new;

	if (!(m->m_flags & M_PDROP))
		return;  /* not a priority drop packet */
	if (ifq == &ifp->if_slow_waitq) {
		headp = &ifp->pdrop_slow_head;
		tailp = &ifp->pdrop_slow_tail;
	}
	else if (ifq == &ifp->if_fast_waitq) {
		headp = &ifp->pdrop_fast_head;
		tailp = &ifp->pdrop_fast_tail;
	}
	else {
		printf("Warning: if_pdrop_enque(): insertion attempted into unknown queue. Exiting\n");
		return;
	}
	head = *headp;
	tail = *tailp;
	new = malloc(sizeof(struct pdrop_queue), M_PCB, M_NOWAIT);
	if (!new)
		panic("Unable to allocate pdrop_queue entry");
	new->m = m;
	new->m_prev = ifq->ifq_tail;
	new->next = 0;
	new->prev = tail;
	if (head == 0) { /* queue is empty */
		head = tail = new;
	}
	else {
		tail->next = new;
		tail = new;
	}
	/* update pointers in the ifnet structure */
	*headp = head;
	*tailp = tail;
}


/* Dequeue from head of pdrop_queue */
void
if_pdrop_deque(ifp, ifq, m)
	struct ifnet *ifp;
	struct ifqueue *ifq;
	struct mbuf *m;
{
	struct pdrop_queue **headp, **tailp;
	struct pdrop_queue *head, *tail, *ptr;

	if (!(m->m_flags & M_PDROP))
		return;  /* not a priority drop packet */
	if (ifq == &ifp->if_slow_waitq) {
		headp = &ifp->pdrop_slow_head;
		tailp = &ifp->pdrop_slow_tail;
	}
	else if (ifq == &ifp->if_fast_waitq) {
		headp = &ifp->pdrop_fast_head;
		tailp = &ifp->pdrop_fast_tail;
	}
	else {
		printf("Error: if_pdrop_deque(): deletion attempted from unknown queue. Exiting\n");
		return;
	}
	head = *headp;
	tail = *tailp;
	if (head == 0) {
		printf("Error: if_pdrop_deque(): deletion attempted from empty queue. Exiting\n");
		return;
	}
	else {
		ptr = head;
		head = head->next;
		if (head) 
			head->prev = 0;
		else {
			tail = head;
		}
		if (ptr->m != m) 
			printf("Warning: if_pdrop_deque(): the item being dequeued does not correspond to the the item being dequeued from the ifq\n");
		free(ptr, M_PCB);
	}
	/* update pointers in the ifnet structure */
	*headp = head;
	*tailp = tail;
}

/* 
 * Drop from tail of pdrop_queue. Also, remove the corresponding packet from the ifq.
 * Return 1 if successful and 0 otherwise.
 */
int
if_pdrop_remove(ifp, ifq)
	struct ifnet *ifp;
	struct ifqueue *ifq;
{
	struct pdrop_queue **headp, **tailp;
	struct pdrop_queue *head, *tail, *ptr;

	if (ifq == &ifp->if_slow_waitq) {
		headp = &ifp->pdrop_slow_head;
		tailp = &ifp->pdrop_slow_tail;
	}
	else if (ifq == &ifp->if_fast_waitq) {
		headp = &ifp->pdrop_fast_head;
		tailp = &ifp->pdrop_fast_tail;
	}
	else {
		printf("Warning: if_pdrop_remove(): deletion attempted from unknown queue. Exiting\n");
		return 0;
	}
	head = *headp;
	tail = *tailp;
	if (tail == 0) 
		return 0;
	else {
		ptr = tail;
		tail = tail->prev;
		if (tail)
			tail->next = 0;
		else
			head = tail;
		IF_REMOVE(ifq, ptr->m, ptr->m_prev);
		free(ptr, M_PCB);
	}
	/* update pointers in the ifnet structure */
	*headp = head;
	*tailp = tail;
	return 1;
}

#endif

void
bandwidth_emulation_timer_handler(ifpvoid)
	void *ifpvoid;
{
	int s;
	struct ifnet *ifp;

	s = splimp();
	ifp = (struct ifnet *) ifp;
	ifp->bandwidth_emulation_timer_pending = 0;
	do_bandwidth_emulation(ifp);
	splx(s);
}

void
delay_emulation_timer_handler(ifpvoid)
	void *ifpvoid;
{
	int s;
	struct ifnet *ifp;

	s = splimp();
	ifp = (struct ifnet *) ifp;
	ifp->delay_emulation_timer_pending = 0;
	do_delay_emulation(ifp);
	splx(s);
}

#define SLOW 1
#define FAST 2
	
void
do_bandwidth_emulation(ifp) 
	struct ifnet *ifp;
{
	struct ifqueue *ifq;
	struct ifqueue *if_waitq;
	struct ifqueue *if_delayq;
	struct timeval curtime, endtime, pkt_endtime, tmptime;
	int nbits; 
	int ticks_to_wait;
	int call_delay_emulation = 0;
	long usec; 
	struct mbuf *m1;
	struct callout_queue *calq_ptr, *calq_head, *calq_tail;

	ifq = &ifp->if_snd;
	if_delayq = &ifp->if_delayq;
	calq_head = ifp->calq_head;
	calq_tail = ifp->calq_tail;

	microtime(&curtime);
	endtime = curtime;
	tv_plus_usec(&endtime, 1000000/(2*hz));
	/* 
	 * if the line is currently idle and if no packet has been waiting for
	 * it to get free, advance end_busy_time to the current time
	 */
	if (tv_lt(&ifp->end_busy_time, &curtime) && !ifp->waiting_for_line) 
		ifp->end_busy_time = curtime;
/*	ifp->waiting_for_line = 0;*/
	while (ifp->if_fast_waitq.ifq_head || ifp->if_slow_waitq.ifq_head) {
		if (ifp->if_fast_waitq.ifq_head && ifp->waiting_for_line != SLOW) {
			m1 = ifp->if_fast_waitq.ifq_head;
			if_waitq = &ifp->if_fast_waitq;
		}
		else {
			m1 = ifp->if_slow_waitq.ifq_head;
			if_waitq = &ifp->if_slow_waitq;
		}
		ifp->waiting_for_line = 0;
		if (ifp->if_acksize > 0 && m1->m_flags & M_TCPACK)
			nbits = ifp->if_acksize*8;
		else
			nbits = (m1->m_pkthdr.len-sizeof(struct ether_header))*8;
		/* set transmission time in microsecs */
		if (ifp->if_bandwidth > 0)
			usec = nbits*1000/ifp->if_bandwidth;
		else
			usec = 0;
		pkt_endtime = ifp->end_busy_time;
		tv_plus_usec(&pkt_endtime, usec);
		if (!tv_lt(&pkt_endtime, &endtime)) {
			if (if_waitq == &ifp->if_slow_waitq)
				ifp->waiting_for_line = SLOW;
			else if (if_waitq == &ifp->if_fast_waitq)
				ifp->waiting_for_line = FAST;
			break;
		}
		/* 
		 * Transmission of the packet over the emulated link is now over.
		 * Adjust end_busy_time to relect this, and move the packet over
		 * to the callout queue used for emulating link delay.
		 */
		tv_plus_usec(&ifp->end_busy_time, usec);
		IF_DEQUEUE(if_waitq, m1);
#ifdef PDROP
		if (ifp->if_flags & IFF_PDROP) 
			if_pdrop_deque(ifp, if_waitq, m1);
#endif
		IF_ENQUEUE(if_delayq, m1);
		if (IF_QFULL(if_delayq)) {
			IF_DROP(if_delayq);
			printf("In do_bandwidth_emulation(): Error: delayq is full\n");
			if (m1)
				m_freem(m1);
			continue;
		}
			
		calq_ptr = malloc(sizeof(struct callout_queue), M_PCB, M_NOWAIT);
		if (!calq_ptr)
			panic("Unable to allocate callout queue entry");
		calq_ptr->next = NULL;
		if (calq_tail) {
			/* 
			 * Establish send time with respect to reference time.  
			 * Put on chain.
			 */
			calq_ptr->send_time = 
				tv_diff(&curtime, &ifp->if_reference_time) + 
					ifp->if_delay * 1000;
			calq_tail->next = calq_ptr;
		}
		else {
			/*
			 * Establish new reference time.  Set send time to be 
			 * if_delay ms in the future.  Determine # of ticks to 
			 * wait.  Flag delay emulation.
			 */
			ifp->if_reference_time = curtime;
			calq_ptr->send_time = ifp->if_delay * 1000;
			calq_head = calq_ptr;
			ticks_to_wait = get_ticks(calq_ptr->send_time);
			call_delay_emulation = 1;
		}
		calq_tail = calq_ptr;
	}
	ifp->calq_head = calq_head;
	ifp->calq_tail = calq_tail;
	/* 
	 * If packets were just enqued to if_delayq while it was previously 
	 * empty, schedule/call do_delay_emulation.
	 */
	if (call_delay_emulation) {
		if (ifp->if_delay > 0) {
			timeout(delay_emulation_timer_handler, (void *) ifp, 
				ticks_to_wait);
			ifp->delay_emulation_timer_pending = 1;
		}
		else
			do_delay_emulation(ifp);
	}
	if (!ifp->bandwidth_emulation_timer_pending && 
	    (ifp->if_fast_waitq.ifq_len > 0 || ifp->if_slow_waitq.ifq_len > 0)) {
		usec = tv_diff(&pkt_endtime, &endtime);
		ticks_to_wait = get_ticks(usec);
		timeout(bandwidth_emulation_timer_handler, (void *) ifp, ticks_to_wait);
		ifp->bandwidth_emulation_timer_pending = 1;
	}
}

void
do_delay_emulation(ifp) 
	struct ifnet *ifp;
{
	struct ifqueue *ifq;
	struct ifqueue *if_delayq;
	struct callout_queue *calq_head, *calq_tail, *calq_ptr;
	int ticks_to_wait;
	struct mbuf *m;
	long usec_since_ref;
	int call_if_start = 0;
	struct timeval curtime;

	microtime(&curtime);
	usec_since_ref = tv_diff(&curtime,&ifp->if_reference_time);
	ifq = &ifp->if_snd;   /* XXX need to consider fastq? */
	if_delayq = &ifp->if_delayq;
	calq_head = ifp->calq_head;
	calq_tail = ifp->calq_tail;
	do {
        IF_DEQUEUE(if_delayq, m);
        if (IF_QFULL(ifq)) {
            IF_DROP(ifq);
            printf("In do_delay_emulation(): Error: ifq is full\n");
            if (m)
                m_freem(m);
        }
        else {
            IF_ENQUEUE(ifq, m);
            call_if_start = 1;
        }
        calq_ptr = calq_head;
        calq_head = calq_ptr->next;
        if (calq_ptr == calq_tail)
            calq_tail = calq_ptr->next;
        free(calq_ptr, M_PCB);
	} while (calq_head && (calq_head->send_time - usec_since_ref)<US_PER_TICK);
	if (((ifp->if_flags & IFF_OACTIVE) == 0) && call_if_start) 
		(*ifp->if_start)(ifp);
	if (calq_head) {
		ticks_to_wait = get_ticks(calq_head->send_time - usec_since_ref);
		if (ticks_to_wait < 1)
			printf("Error trap: ticks_to_wait = %d\n",ticks_to_wait);
		timeout(delay_emulation_timer_handler, (void *) ifp, 
			ticks_to_wait);
		ifp->delay_emulation_timer_pending = 1;
	}
	ifp->calq_head = calq_head;
	ifp->calq_tail = calq_tail;
}
		
void filteracks(ifq, newm)
	struct ifqueue *ifq;
	struct mbuf *newm;
{
	struct mbuf *m, *prev;
	u_char *nbuf, *buf;
	u_long nsrc, ndst;
	struct ip *ip, *newip;
	struct tcphdr *newth, *th;
	
	if (newm->m_len < sizeof(struct ether_header) + sizeof(struct ip) + 
	    sizeof(struct tcphdr))
		return;
	nbuf = mtod(newm, u_char *);
	nbuf += sizeof(struct ether_header);
	newip = (struct ip *)nbuf;
	nsrc = newip->ip_src.s_addr;
	ndst = newip->ip_dst.s_addr;
	nbuf += newip->ip_hl << 2;
	newth = (struct tcphdr *) nbuf;
	/*
	 * Traverse queue, compare flows and compare ack sequence #s.
	 * Make sure not to throw away any acks that have non-zero data!
	 */
	for (m = ifq->ifq_head, prev = m; m != 0; ) {
		/* Must be an ack, and mbuf must have enough bytes to go on. */
		if (!(m->m_flags & M_TCPACK) || m->m_len < 
		    sizeof(struct ether_header) + sizeof(struct ip) + 
		    sizeof(struct tcphdr)) {
			prev = m;
			m = m->m_nextpkt;
			continue;
		}
		buf = mtod(m, u_char *);
		buf += sizeof(struct ether_header);
		ip = (struct ip *)buf;
		buf += ip->ip_hl << 2;
		th = (struct tcphdr *) buf;
		/* 
		 * Purge only pure acks (size 0) from queue; addresses
		 * and port numbers must match.  An ack is purged if it has
		 * a lower sequence number, or if it is a window update. 
		 * The second condition is OK because the later window update
		 * subsumes the earlier one anyway.
		 */
		if (ntohs(ip->ip_len)==(ip->ip_hl << 2) + (th->th_off << 2) &&
		    nsrc==ip->ip_src.s_addr && ndst==ip->ip_dst.s_addr && 
		    newth->th_sport == th->th_sport && 
		    newth->th_dport == th->th_dport) {
			if (SEQ_GT(ntohl(newth->th_ack), ntohl(th->th_ack)) ||
			    (newth->th_ack == th->th_ack && 
			     newth->th_win != th->th_win)) { /* wnd update */
/*
printf("removing ack %x from conn %x:%x\n", ntohl(th->th_ack),ndst,ntohs(th->th_dport));
*/
				/* Remove m and patch up pointers */
				IF_REMOVE(ifq, m, prev);
				m = prev->m_nextpkt;
				continue;
			}
		}
		prev = m;
		m = m->m_nextpkt;
	}
}

#endif

/*
 * Process a received Ethernet packet;
 * the packet is in the mbuf chain m without
 * the ether header, which is provided separately.
 */
void
ether_input(ifp, eh, m)
	struct ifnet *ifp;
	register struct ether_header *eh;
	struct mbuf *m;
{
	register struct ifqueue *inq;
	register struct llc *l;
	struct arpcom *ac = (struct arpcom *)ifp;
	int s;

	if ((ifp->if_flags & IFF_UP) == 0 || m->m_flags & M_DISCARD) {
		m_freem(m);
		return;
	}
	ifp->if_ibytes += m->m_pkthdr.len + sizeof (*eh);
	if (ETHER_IS_MULTICAST(eh->ether_dhost)) {
		if (bcmp((caddr_t)etherbroadcastaddr, (caddr_t)eh->ether_dhost,
		    sizeof(etherbroadcastaddr)) == 0)
			m->m_flags |= M_BCAST;
		else
			m->m_flags |= M_MCAST;
	}
	if (m->m_flags & (M_BCAST|M_MCAST))
		ifp->if_imcasts++;

	switch (eh->ether_type) {
#ifdef INET
	case ETHERTYPE_IP:
		schednetisr(NETISR_IP);
		inq = &ipintrq;
		break;
	case ETHERTYPE_ARP:
		schednetisr(NETISR_ARP);
		inq = &arpintrq;
		break;
#endif /* INET */
#ifdef NS
	case ETHERTYPE_NS:
		schednetisr(NETISR_NS);
		inq = &nsintrq;
		break;

#endif /* NS */
	default:
#if defined (ISO) || defined (LLC)
		if (eh->ether_type > ETHERMTU)
			goto dropanyway;
		l = mtod(m, struct llc *);
		switch (l->llc_dsap) {
#ifdef	ISO
		case LLC_ISO_LSAP: 
			switch (l->llc_control) {
			case LLC_UI:
				/* LLC_UI_P forbidden in class 1 service */
				if ((l->llc_dsap == LLC_ISO_LSAP) &&
				    (l->llc_ssap == LLC_ISO_LSAP)) {
					/* LSAP for ISO */
					if (m->m_pkthdr.len > eh->ether_type)
						m_adj(m, eh->ether_type - m->m_pkthdr.len);
					m->m_data += 3;		/* XXX */
					m->m_len -= 3;		/* XXX */
					m->m_pkthdr.len -= 3;	/* XXX */
					M_PREPEND(m, sizeof *eh, M_DONTWAIT);
					if (m == 0)
						return;
					*mtod(m, struct ether_header *) = *eh;
					IFDEBUG(D_ETHER)
						printf("clnp packet");
					ENDDEBUG
					schednetisr(NETISR_ISO);
					inq = &clnlintrq;
					break;
				}
				ifp->if_noproto++;
				goto dropanyway;
				
			case LLC_XID:
			case LLC_XID_P:
				if(m->m_len < 6)
					goto dropanyway;
				l->llc_window = 0;
				l->llc_fid = 9;
				l->llc_class = 1;
				l->llc_dsap = l->llc_ssap = 0;
				/* Fall through to */
			case LLC_TEST:
			case LLC_TEST_P:
			{
				struct sockaddr sa;
				register struct ether_header *eh2;
				int i;
				u_char c = l->llc_dsap;

				l->llc_dsap = l->llc_ssap;
				l->llc_ssap = c;
				if (m->m_flags & (M_BCAST | M_MCAST))
					ETHER_COPY(ac->ac_enaddr,
					      eh->ether_dhost);
				sa.sa_family = AF_UNSPEC;
				sa.sa_len = sizeof(sa);
				eh2 = (struct ether_header *)sa.sa_data;
				for (i = 0; i < ETHER_ADDR_LEN; i++) {
					eh2->ether_shost[i] = c = eh->ether_dhost[i];
					eh2->ether_dhost[i] = 
						eh->ether_dhost[i] = eh->ether_shost[i];
					eh->ether_shost[i] = c;
				}
				ifp->if_output(ifp, m, &sa, NULL);
				return;
			}
			default:
				m_freem(m);
				return;
			}
			break;
#endif /* ISO */
#ifdef LLC
		case LLC_X25_LSAP:
		{
			if (m->m_pkthdr.len > eh->ether_type)
				m_adj(m, eh->ether_type - m->m_pkthdr.len);
			M_PREPEND(m, sizeof(struct sdl_hdr) , M_DONTWAIT);
			if (m == 0)
				return;
			if ( !sdl_sethdrif(ifp, eh->ether_shost, LLC_X25_LSAP,
					    eh->ether_dhost, LLC_X25_LSAP, 6, 
					    mtod(m, struct sdl_hdr *)))
				panic("ETHER cons addr failure");
			mtod(m, struct sdl_hdr *)->sdlhdr_len = eh->ether_type;
#ifdef LLC_DEBUG
				printf("llc packet\n");
#endif /* LLC_DEBUG */
			schednetisr(NETISR_CCITT);
			inq = &llcintrq;
			break;
		}
#endif /* LLC */
		default:
			ifp->if_noproto++;
		dropanyway:
			M_PREPEND(m, sizeof(*eh), M_DONTWAIT);
			if (m == 0)
				return;
			*mtod(m, struct ether_header *) = *eh;

			s = splimp();
			if (IF_QFULL(&ether_protoq.pq_queue)) {
				IF_DROP(&ether_protoq.pq_queue);
				m_freem(m);
			} else
				PQ_ENQUEUE(&ether_protoq, m);
			splx(s);
			return;
		}
#else /* ISO || LLC */
	    M_PREPEND(m, sizeof(*eh), M_DONTWAIT);
	    if (m == 0)
		    return;
	    *mtod(m, struct ether_header *) = *eh;

	    s = splimp();
	    if (IF_QFULL(&ether_protoq.pq_queue)) {
		    IF_DROP(&ether_protoq.pq_queue);
		    m_freem(m);
	    } else
		    PQ_ENQUEUE(&ether_protoq, m);
	    splx(s);
	    return;
#endif /* ISO || LLC */
	}

	s = splimp();
	if (IF_QFULL(inq)) {
#if 1 
		if (inq == &ipintrq &&
		    !strcmp (m->m_pkthdr.rcvif->if_name, "wl"))
			printf("wl inq full %d/%d\n", 
			       inq->ifq_len, inq->ifq_maxlen);
#endif
		IF_DROP(inq);
		m_freem(m);
	} else
#if 1
	{
		struct timeval now;
		microtime(&now);
		if (inq == &ipintrq &&
		    !strcmp (m->m_pkthdr.rcvif->if_name, "wl") &&
		    (inq->ifq_len*100 / inq->ifq_maxlen > 20))
			printf("%d.%d wl inq %d/%d\n", now.tv_sec, now.tv_usec,
			       inq->ifq_len, inq->ifq_maxlen);
#endif
		IF_ENQUEUE(inq, m);
#if 1
	}
#endif

	splx(s);
}

void
ether_rawintr(arg)
	void *arg;
{
	register struct mbuf *m;
	struct ether_header *eh;
	struct sockproto proto;
	struct sockaddr_dl ether_src;
	struct sockaddr_dl ether_dst;

	bzero(&ether_src, sizeof(ether_src));
	bzero(&ether_dst, sizeof(ether_dst));
	proto.sp_family = PF_LINK;
	ether_src.sdl_family = AF_LINK;
	ether_src.sdl_alen = ETHER_ADDR_LEN;
	ether_dst.sdl_family = AF_LINK;
	ether_dst.sdl_alen = ETHER_ADDR_LEN;
	ether_protoq.pq_wayout.pending = 0;
	for (;;) {
		IF_DEQUEUE(&ether_protoq.pq_queue, m);
		if (m == 0)
			return;
		eh = mtod(m, struct ether_header *);
		proto.sp_protocol = m->m_pkthdr.rcvif->if_index;
		bcopy(eh->ether_shost, ether_src.sdl_data, ETHER_ADDR_LEN);
		bcopy(eh->ether_dhost, ether_dst.sdl_data, ETHER_ADDR_LEN);
		m_adj(m, sizeof(struct ether_header));
		raw_input(m, 0, &proto, &ether_src, &ether_dst);
	}
}

/*
 * Convert Ethernet address to printable (loggable) representation.
 */
static char digits[] = "0123456789abcdef";
char *
ether_sprintf(ap)
	register u_char *ap;
{
	register i;
	static char etherbuf[18];
	register char *cp = etherbuf;

	for (i = 0; i < ETHER_ADDR_LEN; i++) {
		*cp++ = digits[*ap >> 4];
		*cp++ = digits[*ap++ & 0xf];
		*cp++ = ':';
	}
	*--cp = 0;
	return (etherbuf);
}

/*
 * Perform common duties while attaching to interface list
 */
void
ether_attach(ifp)
	register struct ifnet *ifp;
{

	ifp->if_output = ether_output;
	ifp->if_type = IFT_ETHER;
	ifp->if_baudrate = 10*1000*1000;	/* Default, drivers may update */
	ifp->if_addrlen = ETHER_ADDR_LEN;
	ifp->if_hdrlen = sizeof(struct ether_header);
	ifp->if_mtu = ETHERMTU;
	if_attach(ifp);
	if_newaddr(ifp, IFT_ETHER, (caddr_t)((struct arpcom *)ifp)->ac_enaddr);
	PQ_INIT(&ether_protoq, ether_rawintr);
}

u_char	ether_ipmulticast_min[ETHER_ADDR_LEN] =
	{ 0x01, 0x00, 0x5e, 0x00, 0x00, 0x00 };
u_char	ether_ipmulticast_max[ETHER_ADDR_LEN] =
	{ 0x01, 0x00, 0x5e, 0x7f, 0xff, 0xff };

/*
 * Add an Ethernet multicast address or range of addresses to the list for a
 * given interface.
 */
int
ether_addmulti(ifr, ac)
	struct ifreq *ifr;
	register struct arpcom *ac;
{
	register struct ether_multi *enm;
	struct sockaddr_in *sin;
	u_char addrlo[ETHER_ADDR_LEN];
	u_char addrhi[ETHER_ADDR_LEN];
	int s = splimp();

	switch (ifr->ifr_addr.sa_family) {

	case AF_UNSPEC:
		ETHER_COPY(ifr->ifr_addr.sa_data, addrlo);
		ETHER_COPY(addrlo, addrhi);
		break;

#ifdef INET
	case AF_INET:
		sin = (struct sockaddr_in *)&(ifr->ifr_addr);
		if (sin->sin_addr.s_addr == INADDR_ANY) {
			/*
			 * An IP address of INADDR_ANY means listen to all
			 * of the Ethernet multicast addresses used for IP.
			 * (This is for the sake of IP multicast routers.)
			 */
			ETHER_COPY(ether_ipmulticast_min, addrlo);
			ETHER_COPY(ether_ipmulticast_max, addrhi);
		} else {
			ETHER_MAP_IP_MULTICAST(&sin->sin_addr, addrlo);
			ETHER_COPY(addrlo, addrhi);
		}
		break;
#endif

	default:
		splx(s);
		return (EAFNOSUPPORT);
	}

	/*
	 * Verify that we have valid Ethernet multicast addresses.
	 */
	if (!ETHER_IS_MULTICAST(addrlo) || !ETHER_IS_MULTICAST(addrhi)) {
		splx(s);
		return (EINVAL);
	}
	/*
	 * See if the address range is already in the list.
	 */
	ETHER_LOOKUP_MULTI(addrlo, addrhi, ac, enm);
	if (enm != NULL) {
		/*
		 * Found it; just increment the reference count.
		 */
		++enm->enm_refcount;
		splx(s);
		return (0);
	}
	/*
	 * New address or range; malloc a new multicast record
	 * and link it into the interface's multicast list.
	 */
	enm = (struct ether_multi *)malloc(sizeof(*enm), M_IFMADDR, M_NOWAIT);
	if (enm == NULL) {
		splx(s);
		return (ENOBUFS);
	}
	ETHER_COPY(addrlo, enm->enm_addrlo);
	ETHER_COPY(addrhi, enm->enm_addrhi);
	enm->enm_ac = ac;
	enm->enm_refcount = 1;
	enm->enm_next = ac->ac_multiaddrs;
	ac->ac_multiaddrs = enm;
	ac->ac_multicnt++;
	splx(s);
	/*
	 * Return ENETRESET to inform the driver that the list has changed
	 * and its reception filter should be adjusted accordingly.
	 */
	return (ENETRESET);
}

/*
 * Delete a multicast address record.
 */
int
ether_delmulti(ifr, ac)
	struct ifreq *ifr;
	register struct arpcom *ac;
{
	register struct ether_multi *enm;
	register struct ether_multi **p;
	struct sockaddr_in *sin;
	u_char addrlo[ETHER_ADDR_LEN];
	u_char addrhi[ETHER_ADDR_LEN];
	int s = splimp();

	switch (ifr->ifr_addr.sa_family) {

	case AF_UNSPEC:
		ETHER_COPY(ifr->ifr_addr.sa_data, addrlo);
		ETHER_COPY(addrlo, addrhi);
		break;

#ifdef INET
	case AF_INET:
		sin = (struct sockaddr_in *)&(ifr->ifr_addr);
		if (sin->sin_addr.s_addr == INADDR_ANY) {
			/*
			 * An IP address of INADDR_ANY means stop listening
			 * to the range of Ethernet multicast addresses used
			 * for IP.
			 */
			ETHER_COPY(ether_ipmulticast_min, addrlo);
			ETHER_COPY(ether_ipmulticast_max, addrhi);
		} else {
			ETHER_MAP_IP_MULTICAST(&sin->sin_addr, addrlo);
			ETHER_COPY(addrlo, addrhi);
		}
		break;
#endif

	default:
		splx(s);
		return (EAFNOSUPPORT);
	}

	/*
	 * Look up the address in our list.
	 */
	ETHER_LOOKUP_MULTI(addrlo, addrhi, ac, enm);
	if (enm == NULL) {
		splx(s);
		return (ENXIO);
	}
	if (--enm->enm_refcount != 0) {
		/*
		 * Still some claims to this record.
		 */
		splx(s);
		return (0);
	}
	/*
	 * No remaining claims to this record; unlink and free it.
	 */
	for (p = &enm->enm_ac->ac_multiaddrs;
	     *p != enm;
	     p = &(*p)->enm_next)
		continue;
	*p = (*p)->enm_next;
	free(enm, M_IFMADDR);
	ac->ac_multicnt--;
	splx(s);
	/*
	 * Return ENETRESET to inform the driver that the list has changed
	 * and its reception filter should be adjusted accordingly.
	 */
	return (ENETRESET);
}
