/*-
 * Copyright (c) 1995, 1996 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI if_link.c,v 2.8 1996/12/31 20:31:42 karels Exp
 */

/*
 * Copyright (c) 1982, 1986, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	from @(#)raw_ip.c	8.2 (Berkeley) 1/4/94
 */

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/protosw.h>
#include <sys/socketvar.h>
#include <sys/errno.h>
#include <sys/systm.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/raw_cb.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_mroute.h>
#include <netinet/in_pcb.h>

extern int
p2p_sockopt __P((int, struct ifnet *, int, int, struct mbuf **));

/*
 * Generic link layer/sockaddr_dl support,
 * plus suppport of PF_LINK raw sockets.
 */
check_linkaddr(sdl)
	struct sockaddr_dl *sdl;
{

	if (sdl->sdl_index && sdl->sdl_index > if_index ||
	    sdl->sdl_nlen + sdl->sdl_alen >=
	    sdl->sdl_len - (sizeof(*sdl) - sizeof(sdl->sdl_data)))
		return (EINVAL);
	return (0);
}

#ifdef notyet
/*
 * Return ifnet pointer for a sockaddr_dl naming an interface.
 * If sdl_index is nonzero, use it; otherwise, do lookup by name.
 * Any link-layer address is ignored.
 */
struct ifnet *
if_ifwithlinkaddr(sdl)
	struct sockaddr_dl *sdl;
{
	char ifname[IFNAMSIZ + 1];

	if (sdl->sdl_index) {
		if (sdl->sdl_index <= if_index)
			return (ifnet_addrs[sdl->sdl_index - 1]->ifa_ifp);
	} else if (sdl->sdl_nlen && sdl->sdl_nlen <= IFNAMSIZ) {
		bcopy(sdl->sdl_data, ifname, sdl->sdl_nlen);
		ifname[sdl->sdl_nlen] = 0;
		return (ifunit(ifname));
	}
	return (NULL);
}
#endif

struct sockaddr_dl null_dl = { sizeof(struct sockaddr_dl), AF_DLI };

/*
 * Raw link-layer input from specified interface;
 * pass to any link-layer listeners.
 * srcaddr and dstaddr are optional per interface (or type);
 * if not null, they are pointers to raw link-level addresses.
 * We primarily filter on the interface index (+ 1), which we use
 * as the protocol number in the sockproto for raw_input.
 * Sockets which have not yet selected an interface have a protocol
 * number of 0 (no interface).
 */
void
link_input(m, ifp, srcaddr, dstaddr)
	struct mbuf *m;
	struct ifnet *ifp;
	void *srcaddr, *dstaddr;
{
	struct sockproto proto;
	struct sockaddr_dl *src, *dst;
	struct sockaddr_dl sdl, ddl;

	proto.sp_family = PF_LINK;
	proto.sp_protocol = ifp->if_index + 1;
	if (srcaddr == NULL)
		src = &null_dl;
	else {
		src = &sdl;
		bcopy(&null_dl, &sdl, sizeof(sdl));
		sdl.sdl_type = ifp->if_type;
		if (ifp->if_addrlen > sizeof(sdl.sdl_data))
			panic("link_input");
		sdl.sdl_alen = ifp->if_addrlen;
		bcopy(srcaddr, LLADDR(&sdl), ifp->if_addrlen);
	}
	if (dstaddr == NULL)
		dst = &null_dl;
	else {
		dst = &ddl;
		bcopy(&null_dl, &ddl, sizeof(ddl));
		ddl.sdl_type = ifp->if_type;
		if (ifp->if_addrlen > sizeof(ddl.sdl_data))
			panic("link_input");
		ddl.sdl_alen = ifp->if_addrlen;
		bcopy(dstaddr, LLADDR(&ddl), ifp->if_addrlen);
	}
	raw_input(m, 0, &proto, (struct sockaddr *)src, (struct sockaddr *)dst);
}

/*
 * Link-layer output.
 * What to do if connected?
 */
struct sockaddr link_unspec = { sizeof(struct sockaddr), AF_UNSPEC };

int
link_output(m, so)
	register struct mbuf *m;
	struct socket *so;
{
	struct rawcb *rp = sotorawcb(so);
	struct ifnet *ifp;

	if (rp == NULL || rp->rcb_proto.sp_protocol == 0)
		return (ENOTCONN);
	ifp = ifnet_addrs[rp->rcb_proto.sp_protocol - 1]->ifa_ifp;
	return (ifp->if_output(ifp, m,
	    rp->rcb_faddr ? rp->rcb_faddr : &link_unspec,
	    (struct rtentry *) NULL));
}

/*
 * Raw link socket option processing.
 */
int
link_ctloutput(op, so, level, optname, m)
	int op;
	struct socket *so;
	int level, optname;
	struct mbuf **m;
{
	struct ifnet *ifp;
	struct ifaddr *ifa;
	struct rawcb *rp = sotorawcb(so);
	register int error = 0;

	if (rp == NULL || rp->rcb_proto.sp_protocol == 0)
		return (ENOTCONN);

	if (rp->rcb_proto.sp_protocol <= 0 ||
	    rp->rcb_proto.sp_protocol > if_index)
		return(EINVAL);

        ifp = ifnet_addrs[rp->rcb_proto.sp_protocol - 1]->ifa_ifp;

	if (ifp && ifp->if_ctloutput != NULL)
		error = ifp->if_ctloutput(op, ifp, level, optname, m);
	else
		error = ENOPROTOOPT;

	if (op == PRCO_SETOPT && *m)
		(void)m_freem(*m);
	return(error);

#ifdef	LINK_AF_code
	if (op == PRCO_SETOPT) {
		if (*m == 0 || (*m)->m_len < sizeof(struct sockaddr_dl))
			error = EINVAL;
		if (ifp = if_ifwithlinkaddr(mtod(m, struct sockaddr_dl *))) {
			rp->rcp_proto.sp_family = PF_LINK;
			rp->rcp_proto.sp_family = ifp->if_index + 1;
		} else
			error = EADDRNOTAVAIL;
		if (*m)
			(void)m_free(*m);
	} else {
		if (rp->rcb_proto.sp_protocol == 0)
			return (ENOTCONN);
		ifa = ifnet_addrs[rp->rcb_proto.sp_protocol - 1];
		if (ifa == NULL)
			return (EADDRNOTAVAIL);
		*m = m_get(M_WAIT, MT_SOOPTS);
		(*m)->m_len = ifa->ifa_addr->sa_len;
		bcopy(ifa->ifa_addr, mtod(*m, void *), ifa->ifa_addr->sa_len);
	}
#endif
	return (error);
}

/* begin XXX */
#include <net/if_types.h>

typedef int ifnetctlfn __P((struct ifnet *, int *, u_int, void *, size_t *,
	    void *, size_t /* , struct proc * */));

ifnetctlfn *ppp_sysctlfn;
ifnetctlfn *pif_sysctlfn;

/*
 * Sysctl for link-layer sockets.
 * XXX This is a placeholder to implement minimal functionality
 * in the absence of an if_sysctl function.
 *
 * Currently net.pf_link.generic(==0) gets us here
 * with the remaining part of the name.
 * The plan is to support net.pf_link.generic.if_type for parameters
 * global to a particular if_type, and net.pf_link.if_index(orig 1).*
 * for parameters specific to a particular interface.
 *
 * For now, net.pf_link.generic takes an if_type as the following
 * component.
 */
int
link_sysctl(name, namelen, oldp, oldlenp, newp, newlen)
	int	*name;
	u_int	namelen;
	void	*oldp;
	size_t	*oldlenp;
	void	*newp;
	size_t	newlen;
{
	register struct ifnet *ifp;
	ifnetctlfn *ctlfn = NULL;
	int type = name[0];

	if (namelen < 2)
		return (EISDIR);

	/*
	 * At this point we know we are doing a generic sysctl
	 * for some link-layer type (if_type); find one of this type.
	 */
	for (ifp = ifnet; ifp; ifp = ifp->if_next)
		if (ifp->if_type == type)
			break;
	if (ifp == NULL)
		return (ENXIO);

	/* XXX use hard-wired functions until we have if_sysctl entry */
	switch (name[0]) {
	case IFT_PPP:
		ctlfn = ppp_sysctlfn;
		break;

	case IFT_PROPMUX:	/* XXX pif */
		ctlfn = pif_sysctlfn;
		break;

	default:
		break;
	}
	if (ctlfn)
		return ((*ctlfn)(ifp, name + 1, namelen - 1, oldp, oldlenp,
		    newp, newlen));
	return (ENXIO);
}
/* end XXX */
