/*-
 * Copyright (c) 1993, 1994, 1995, 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI if_p2pproto.c,v 2.7 1996/11/13 16:59:46 prb Exp
 */

/*
 * Common code for point-to-point synchronous serial links
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/netisr.h>
#include <net/route.h>
#include <net/if_dl.h>
#include <net/if_types.h>

#if INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>		/* XXX for slcompress */
#endif

#include <net/if_p2p.h>
#include <net/bpf.h>


static int none_hdr __P((struct p2pcom *, u_char *, int));
static int none_input __P((struct p2pcom *, struct mbuf *));
static int none_output __P((struct ifnet *, struct mbuf *, struct sockaddr *,
    struct rtentry *));

extern int p2p_hdr();

#ifdef CISCO_HDLC
extern int c_hdlc_attach(), c_hdlc_init(), c_hdlc_output(), c_hdlc_input();
extern void c_hdlc_detach(), c_hdlc_shutdown();
#endif
#ifdef PPP
#include <net/slcompress.h>
extern int ppp_attach(), ppp_init(), ppp_output(), ppp_input(), ppp_ioctl(),
	   ppp_hdr();
extern void ppp_detach(), ppp_modem(), ppp_shutdown();
#endif
#ifdef  FR
extern int fr_attach(), fr_init(), fr_output(), fr_input(), fr_hdr();
extern void fr_detach(), fr_modem(), fr_shutdown();
#endif


/*
 * The table of available point-to-point encapsulation protocols
 */
struct p2pprotosw p2pprotosw[] = {
	/* UNITIALIZED STATE */
{	IFT_NONE,
	0 /*attach*/,	0 /*detach*/,	0 /*init*/,	0 /*shutdown*/,
	none_output,	none_input,	0 /*ioctl*/, 	0 /*modem*/,
	none_hdr,	0 /*ctloutput*/,
	0 /*cskip*/,	DLT_NULL,	0 /*hdrlen*/,
},
#ifdef CISCO_HDLC
{	IFT_PTPSERIAL,
	c_hdlc_attach,	c_hdlc_detach,	c_hdlc_init,	c_hdlc_shutdown,
	c_hdlc_output,	c_hdlc_input,	0,		0,
	p2p_hdr,	0,
	16,		DLT_C_HDLC,	0,
},
#endif
#ifdef PPP
{	IFT_PPP,
	ppp_attach,	ppp_detach,	ppp_init,	ppp_shutdown,
	ppp_output,	ppp_input,	ppp_ioctl,	ppp_modem,
	ppp_hdr,	0,
	120,		DLT_PPP,	SLC_BPFHDR,
},
#endif
#ifdef  FR
{	IFT_FRELAY,
	fr_attach,	fr_detach,	fr_init,	fr_shutdown,
	fr_output,	fr_input,	0,		fr_modem,
	p2p_hdr,	0,
	16,		DLT_NULL,	0,
},
#endif
};

int np2pproto = (sizeof p2pprotosw / sizeof p2pprotosw[0]);

static int
none_hdr(pp, cp, nbytes)
	struct p2pcom *pp;
	u_char *cp;
	int nbytes;
{

	return (nbytes == -1 ? 0 : 1);
}

static int
none_input(pp, m)
	struct p2pcom *pp;
	struct mbuf *m;
{

	if (m)
		m_freem(m->m_next);
	return (0);
}

static int
none_output(ifp, m, dst, rt)
	struct ifnet *ifp;
	struct mbuf *m;
	struct sockaddr *dst;
	struct rtentry *rt;
{

	if (m)
		m_freem(m->m_next);
	return (ENOPROTOOPT);
}
