/*-
 * Copyright (c) 1993, 1994, 1995, 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI if_pif.c,v 2.4 1996/12/31 20:31:42 karels Exp
 */

#include <sys/param.h>
#include <sys/malloc.h>

#include <sys/mbuf.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/proc.h>
#include <sys/queue.h>

#include <net/if.h>
#include <net/route.h>
#include <net/if_pif.h>
#include <net/if_types.h>

#include <netinet/in_systm.h>	/* define "n_long", needed by <netinet/ip.h> */
#include <netinet/in.h>		/* define "in_addr", needed by <netinet/ip.h> */
#include <netinet/ip.h>		/* for IP_MSS */

struct pif_dst {
	struct rtentry *rt;
	struct sockaddr dst;
};

int	npifif = 0;			/* The number of interfaces */

struct pif_softc **pif_softc;

/*
 * Initialize a newly allocated pif_softc as the indicated unit.
 */
void
pifinit(sc, unit)
	register struct pif_softc *sc;
	int unit;
{
	register struct ifnet *ifp;

	bzero(sc, sizeof(*sc));
	pif_softc[unit] = sc;
	sc->pif_flags = PIF_NEXT_IDLE;
	sc->pif_last = 0;
	sc->pif_nif = 0;
	sc->pif_maxif = PIF_MIN_IF;
	sc->pif_list = sc->_pif_list;

	ifp = &sc->pif_if;

	ifp->if_name = "pif";
	ifp->if_unit = unit;
	ifp->if_mtu = IP_MSS;
	ifp->if_flags = IFF_MULTICAST;
	ifp->if_ioctl = pifioctl;
	ifp->if_output = pifoutput;
	ifp->if_type = IFT_PROPMUX;
	ifp->if_start = pifstart;
	ifp->if_hdrlen = 0;
	ifp->if_addrlen = 0;
	ifp->if_snd.ifq_maxlen = 32;
	if_attach(ifp);
}

static void *
pif_scalloc(n)
	int n;
{
	void *p;

	n *= sizeof(*pif_softc);
	p = malloc(n, M_DEVBUF, M_WAIT);
	bzero(p, n);
	return (p);
}

void
pifattach(n)
	int n;
{
	register struct pif_softc *sc;
	register struct ifnet *ifp;
	int i;

	/*
	 * If "npifif" has been patched, use it instead of the
	 * value from the config file.
	 */
	if (npifif)
		n = npifif;
	npifif = n;

	pif_softc = pif_scalloc(n);
	sc = malloc(n * sizeof(*sc), M_DEVBUF, M_WAIT);
	for (i = 0; i < n; sc++, i++)
		pifinit(sc, i);
	ppp_pif_alloc = pif_alloc;
	ppp_pif_free = pif_free;
}

void
pifmore(n)
	int n;
{
	register struct pif_softc *sc;
	void *new_sc, *old_sc;

	new_sc = pif_scalloc(n);
	old_sc = pif_softc;
	bcopy(pif_softc, new_sc, npifif * sizeof(*pif_softc));
	pif_softc = new_sc;
	free(old_sc, M_DEVBUF);
	sc = malloc((n - npifif) * sizeof(*sc), M_DEVBUF, M_WAIT);
	for (; npifif < n; sc++, npifif++)
		pifinit(sc, npifif);
}

/*
 * Allocate a "pif" structure, if needed, and
 * assign an "if" to that "pif".
 */
int
pif_alloc(ifp, pifp)
	struct ifnet *ifp;
	struct ifnet **pifp;
{
	struct	pif_softc *sc;
	int	error, i;

	if (*pifp == NULL) {
		for (i = 0; ; i++) {
			if (i >= npifif)
				return(ENODEV);
			sc = pif_softc[i];
			if (sc->pif_nif == 0)
				break;
		}
		sc->pif_flags = PIF_NEXT_IDLE;
		sc->pif_if.if_mtu = IP_MSS;
		sc->pif_if.if_flags = IFF_MULTICAST;
		sc->pif_if.if_hdrlen = 0;
		sc->pif_if.if_addrlen = 0;
	} else {
		if ((*pifp)->if_type != IFT_PROPMUX)
			return(EINVAL);
		sc = (struct pif_softc *)*pifp;
	}
	error = pif_add(sc, ifp);
	if (error == 0)
		*pifp = &sc->pif_if;
	return(error);
}

int
pifioctl(ifp, cmd, data)
	register struct ifnet *ifp;
	u_long cmd;
	caddr_t data;
{
	int s;
	int error;
	struct	pifreq *pr;
	struct	pif_softc *sc;
	struct	ifnet *subifp;
	struct	ifreq *ifr;
	struct	proc *p = curproc;	/* XXX should be parameter */
	int	namelen, unitlen, i;
	char	tbuf[12], *unitname;


	pr = (struct pifreq *)data;
	sc = (struct pif_softc *)ifp;

	s = splimp();

	switch (cmd) {
	case PIFIOCADD:
	case PIFIOCDEL:
	case PIFIOCGPIF:
		subifp = ifunit(pr->pifr_subname);

		/* For now, we only allow point-to-point	   XXX */
		/* interfaces to be linked to a PIF.		   XXX */
		if (cmd == PIFIOCADD &&				/* XXX */
		    !(subifp->if_flags & IFF_POINTOPOINT)) {	/* XXX */
			error = EOPNOTSUPP;			/* XXX */
			break;					/* XXX */
		}						/* XXX */
		if (subifp == NULL) {
			error = ENXIO;
			break;
		}
	case PIFIOCSFLG:
		if (error = suser(p->p_ucred, &p->p_acflag))
			goto done;
		break;

	default:
		error = 0;
		break;
	}

	switch (cmd) {
	case PIFIOCGFLG:
		pr->pifr_flags = sc->pif_flags;
		break;

	case PIFIOCSFLG:
		sc->pif_flags = pr->pifr_flags;
		break;

	case PIFIOCADD:
		error = pif_add(sc, subifp);
		break;

	case PIFIOCDEL:
		error = pif_remove(sc, subifp);
		break;

	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;
		break;

	case PIFIOCGPIF:
		if ((ifp = subifp->if_pif) == NULL)
			return(ENXIO);
		unitname = &tbuf[sizeof(tbuf)];
		i = ifp->if_unit;
		do {
			*--unitname = '0' + (i % 10);
			i /= 10;
		} while (i);
		unitlen = &tbuf[sizeof(tbuf)] - unitname;

		namelen = strlen(ifp->if_name);
		bcopy(ifp->if_name, pr->pifr_name, namelen);
		bcopy(unitname, &pr->pifr_name[namelen], unitlen);
		pr->pifr_name[namelen + unitlen] = '\0';
		break;

	case SIOCSIFDSTADDR:
	case SIOCSIFFLAGS:
		break;

	/*
	 * These should probably be done at a lower level,
	 * but this should be ok for now.
	 */
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		ifr = (struct ifreq *)data;
		switch (ifr->ifr_addr.sa_family) {
#ifdef INET
		case AF_INET:
			break;
#endif
		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;

	default:
		error = EINVAL;
		break;
	}
done:
	splx(s);
	return (error);
}

typedef int ifnetctlfn __P((struct ifnet *, int *, u_int, void *, size_t *,
	    void *, size_t /* , struct proc * */));

ifnetctlfn pifsysctl;
ifnetctlfn *pif_sysctlfn = pifsysctl;

/*
 * Sysctl for pif.  Should be the if_sysctl function, but there is none
 * yet.
 */
/* ARGSUSED */
int
pifsysctl(ifp, name, namelen, oldp, oldlenp, newp, newlen)
	struct	ifnet *ifp;
	int	*name;
	u_int	namelen;
	void	*oldp;
	size_t	*oldlenp;
	void	*newp;
	size_t	newlen;
{
	int error;

	if (namelen > 1)
		return (ENOTDIR);

	switch (name[0]) {
	case 0: {			/* XXX */
		int val;

		val = npifif;
		error = sysctl_int(oldp, oldlenp, newp, newlen, &val);
		if (error == 0 && newp) {
			if (val < npifif)
				return (EINVAL);
			pifmore(val);
		}
		break;
		}

	default:
		error = EOPNOTSUPP;
		break;
	}
	return (error);
}

int
pifstart(ifp)
	register struct ifnet *ifp;
{
}

int
pif_add(sc, subifp)
	register struct pif_softc *sc;
	register struct ifnet *subifp;
{
	struct ifnet **nlist;
	int n;

	if (subifp->if_pif)
		return(EBUSY);

	if (sc->pif_nif == sc->pif_maxif) {
		n = sc->pif_maxif * 2;
		nlist = malloc(n * sizeof(*sc->pif_list), M_DEVBUF, M_WAIT);
		bcopy(sc->pif_list, nlist, n/2 * sizeof(*sc->pif_list));
		if (sc->pif_list != sc->_pif_list)
			free(sc->pif_list, M_DEVBUF);
		sc->pif_list = nlist;
		sc->pif_maxif = n;
	}

	sc->pif_list[sc->pif_nif++] = subifp;
	subifp->if_done = pif_done;
	subifp->if_pif = &sc->pif_if;

	pif_cominit(sc);

	return (0);
}


int
pif_free(ifp)
	register struct ifnet *ifp;
{
	register struct pif_softc *sc;

	if ((sc = (struct pif_softc *)ifp->if_pif) == NULL)
		return(ENXIO);

	return(pif_remove(sc, ifp));
}

int
pif_remove(sc, ifp)
	register struct pif_softc *sc;
	register struct ifnet *ifp;
{
	register int i;

	for (i = 0; ; i++) {
		if (i >= sc->pif_nif)
			return(ENXIO);
		if (sc->pif_list[i] == ifp)
			break;
	}

	ifp->if_done = NULL;
	ifp->if_pif = NULL;

	for (--sc->pif_nif; i < sc->pif_nif; i++)
		sc->pif_list[i] = sc->pif_list[i+1];

	sc->pif_list[sc->pif_nif] = NULL;

	sc->pif_last = 0;
	if (sc->pif_nif == 0) {
		if_down(&sc->pif_if);
		if (sc->pif_list != sc->_pif_list) {
			free(sc->pif_list, M_DEVBUF);
			sc->pif_list = sc->_pif_list;
			sc->pif_maxif = PIF_MIN_IF;
			bzero(sc->_pif_list, sizeof sc->_pif_list);
		}
	} else
		pif_cominit(sc);

	return (0);
}

void
pif_cominit(sc)
	register struct pif_softc *sc;
{
	register struct ifnet *ifp, *subifp;
	register int i;

	if (sc->pif_nif <= 0)
		return;
	subifp = sc->pif_list[0];

	ifp = &sc->pif_if;

	ifp->if_mtu = subifp->if_mtu;
	ifp->if_flags = subifp->if_flags | IFF_MULTICAST;
	ifp->if_hdrlen = sizeof(struct pif_dst);

	for (i = 1; i < sc->pif_nif; i++) {
		subifp = sc->pif_list[i];
		if (subifp->if_mtu < ifp->if_mtu)
			ifp->if_mtu = subifp->if_mtu;
		if ((subifp->if_flags & IFF_MULTICAST) == 0)
			ifp->if_flags &= ~IFF_MULTICAST;
		if (subifp->if_hdrlen > ifp->if_hdrlen)
			ifp->if_hdrlen = subifp->if_hdrlen;
	}
}

#define	senderr(e)	{ error = (e); goto bad;}

int
pifoutput(ifp, m, dst, rt)
	register struct ifnet *ifp;
	struct mbuf *m;
	struct sockaddr *dst;
	struct rtentry *rt;
{
	int active;
	int error;
	int len = m->m_pkthdr.len;
	register struct pif_softc *sc;
	struct ifnet *subifp;
	int first_if, i, isup = -1;

	error = 0;

	sc = (struct pif_softc *)ifp;

	if (sc->pif_nif <= 0)
		return(ENXIO);

	/*
	 * Find an interface for output.
	 *
	 * If PIF_RESCAN is set, we start our search with the
	 * first interface on the list.  Otherwise we start our
	 * search after the last interface that was used for output.
	 *
	 * If PIF_IGNORE_OACTIVE is not set, we skip over interfaces
	 * that are currently active doing output (IFF_OACTIVE SET).
	 *
	 * As we scan the list of interfaces, we remember the first
	 * interface that is up but not running.  If we can't find
	 * an interface that is running, then we'll use the first
	 * interface that was up but not running.
	 *
	 * NEXT_IDLE: 0
	 *	Finds next interface that is up and running and is
	 *	not doing output, or the next interface that is up
	 *	but not yet running.
	 * NEXT_UP: (PIF_IGNORE_OACTIVE)
	 *	Finds next interface that is up and running.
	 *	or next interface that is up but not yet running.
	 * FIRST_IDLE: (PIF_RESCAN)
	 *	Finds first interface that is up and running and is
	 *	not doing output, or the first interface that is up
	 *	but not yet running.
	 * FIRST_UP: (PIF_IGNORE_OACTIVE | PIF_RESCAN)
	 *	Finds first interface that is up and running,
	 *	or first interface that is up but not yet running.
	 */

	first_if = (sc->pif_flags & PIF_RESCAN) ? 0 : sc->pif_last + 1;
	if (first_if >= sc->pif_nif)
		first_if = 0;
	i = first_if;

	for (;;) {
		subifp = sc->pif_list[i];
		if ((sc->pif_flags & PIF_IGNORE_OACTIVE) == 0 &&
		    (subifp->if_flags & IFF_OACTIVE))
			active++;
		else if (subifp->if_flags & IFF_UP) {
			if (subifp->if_flags & IFF_RUNNING) {
				sc->pif_last = i;
				error =
				    (*subifp->if_output)(subifp, m, dst, rt);
				break;
			} else if (isup < 0)
				isup = i;
		}
		if (++i >= sc->pif_nif)
			i = 0;
		if (i == first_if) {
			if (isup < 0) {
				if (active == 0)
					senderr(ENETDOWN);
				error = pif_enqueue(&sc->pif_if, m, dst, rt);
				break;
			}
			sc->pif_last = isup;
			subifp = sc->pif_list[isup];
			error = (*subifp->if_output)(subifp, m, dst, rt);
			break;
		}
	}

	ifp->if_opackets++;
	ifp->if_obytes += len;

	return(error);
bad:
	if (m)
		m_freem(m);
	return (error);
}

pif_enqueue(ifp, m, dst, rt)
	struct ifnet *ifp;
	register struct mbuf *m;
	struct sockaddr *dst;
	struct rtentry *rt;
{
	struct pif_dst *pif_dst;
	int error;

	if (IF_QFULL(&ifp->if_snd)) {
		IF_DROP(&ifp->if_snd);
		senderr(ENOBUFS);
	}
	M_PREPEND(m, dst->sa_len + sizeof(struct rtentry *), M_WAIT);
	if (m == NULL)
		senderr(ENOBUFS);
	pif_dst = mtod(m, struct pif_dst *);
	pif_dst->rt = rt;
	bcopy((caddr_t)dst, (caddr_t)&pif_dst->dst, dst->sa_len);
	IF_ENQUEUE(&ifp->if_snd, m);
	return (0);
bad:
	if (m)
		m_freem(m);
	return (error);
}

/*
 * Callback routine when interface is ready to send another packet.
 */
int
pif_done(subifp)
	register struct ifnet *subifp;
{
	register struct ifnet *ifp;
	register struct mbuf *m, *m0;
	struct sockaddr *dst;
	struct rtentry *rt;
	int len;
	struct pif_dst *pif_dst;

	ifp = subifp->if_pif;

	IF_DEQUEUE(&ifp->if_snd, m);
	if (m == NULL)
		return;

	pif_dst = mtod(m, struct pif_dst *);
	rt = pif_dst->rt;
	dst = &pif_dst->dst;
	len = sizeof(struct rtentry *) + pif_dst->dst.sa_len;
	m->m_len -= len;
	m->m_pkthdr.len -= len;
	m->m_data += len;

	m0 = NULL;
	if ((m->m_len == 0) && m->m_next) {
		m0 = m;
		m = m->m_next;
		m0->m_next = NULL;
		m->m_pkthdr = m0->m_pkthdr;
		m->m_flags |= M_PKTHDR;
	}
	(void) (*subifp->if_output)(subifp, m, dst, rt);
	if (m0)
		m_freem(m0);
}
