/*-
 * Copyright (c) 1993, 1994, 1995, 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI if_pppmp.c,v 2.1.2.1 1997/03/18 20:51:55 karels Exp
 */

/*
 * PPP Multilink Protocol specific routines
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/ioctl.h>
#include <sys/errno.h>

#include <net/if.h>

#include <machine/cpu.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include <net/slcompress.h>
#include <net/if_p2p.h>
#include <net/if_ppp.h>
#include <net/if_pppioctl.h>
#include <net/if_pif.h>

int	mp_ioctl __P((struct p2pcom *, int, caddr_t));
void	mp_detach __P((struct p2pcom *, struct ppp_sc *));
void	mp_input __P((struct p2pcom *, struct mbuf *));
int	mp_reass __P((struct p2pcom *, struct mbuf *));

/*
 * External function pointers.  These allow a kernel to be
 * built with if_ppp.c, and linked with or without this module.
 */
int	(*ppp_mp_ioctl) __P((struct p2pcom *, int, caddr_t)) = mp_ioctl;
void	(*ppp_mp_detach) __P((struct p2pcom *, struct ppp_sc *)) = mp_detach;
int	(*ppp_mp_reass) __P((struct p2pcom *, struct mbuf *)) = mp_reass;

#ifdef PPP_DEBUG
#define	dprintf(x)	if (pp->p2p_if.if_flags & IFF_DEBUG) printf x
#else
#define dprintf(x)
#endif

/*
 * Control structure and macros for allocating and freeing
 * Multilink fragment headers.
 */
struct	{
	int	pf_max;			/* Max len for free queue */
	struct	frag_head pf_freelist;
	int	pf_len;			/* Length of the free queue */
	int	pf_alloc;		/* # times frag hdr allocated */
	int	pf_reuse;		/* # times frag hdr reused */
	int	pf_fail;		/* # times frag alloc failed */
	int	pf_freed;		/* # times frag freed */
} pppfrag = { 6 };

#define	FRAG_ALLOC(newfrag) do {					\
	(newfrag) = LIST_FIRST(&pppfrag.pf_freelist);			\
	if ((newfrag) != LIST_END(&pppfrag.pf_freelist)) {		\
		LIST_REMOVE((newfrag), frag_next);			\
		--pppfrag.pf_len;					\
		++pppfrag.pf_reuse;					\
	} else {							\
		(newfrag) = malloc(sizeof(struct ppp_frag),		\
		    M_FTABLE, M_NOWAIT);				\
		if ((newfrag) != NULL)					\
			pppfrag.pf_alloc++;				\
		else							\
			pppfrag.pf_fail++;				\
	}								\
} while (0)

#define	FRAG_FREE(frag) do {						\
	if (pppfrag.pf_len < pppfrag.pf_max) {				\
		LIST_INSERT_HEAD(&pppfrag.pf_freelist, frag, frag_next);\
		pppfrag.pf_len++;					\
	} else {							\
		free((frag), M_FTABLE);					\
		pppfrag.pf_freed++;					\
	}								\
} while (0)

/*
 * Macros for joining, allocating and leaving a Multilink session.
 */
#define	MP_JOIN(mp, ppp) do {						\
	(mp)->mp_ref++;							\
	(ppp)->ppp_mp = (mp);						\
	LIST_INSERT_HEAD(&(mp)->mp_bndl, (ppp), ppp_mp_next);		\
} while (0)

#define	MP_ALLOC(mp, ppp) do {						\
	(mp) = malloc(sizeof(struct ppp_mp), M_DEVBUF, M_WAITOK);	\
	bzero((mp), sizeof(struct ppp_mp));				\
	LIST_INIT(&(mp)->mp_frag);					\
	LIST_INIT(&(mp)->mp_bndl);					\
	LIST_INSERT_HEAD(&mp_list, (mp), mp_next);			\
	MP_JOIN(mp, ppp);						\
} while (0)

#define	MP_INUSE(mp) ((mp)->mp_ref > 1)

#define	MP_FREE(mp) do {						\
	if (--(mp)->mp_ref <= 0) {					\
		struct	ppp_frag *frag;					\
		LIST_REMOVE((mp), mp_next);				\
		while ((frag = LIST_FIRST(&(mp)->mp_frag)) !=		\
		    LIST_END(&(mp)->mp_frag)) {				\
			LIST_REMOVE(frag, frag_next);			\
			m_freem(frag->frag_m);				\
			FRAG_FREE(frag);				\
		}							\
		if (LIST_FIRST(&(mp)->mp_bndl) !=			\
		    LIST_END(&(mp)->mp_bndl))				\
			panic("MP_FREE: bundle isn't empty");		\
		free((mp), M_DEVBUF);					\
	}								\
} while (0)

/*
 * These macros are only used by the mp_reass() routine,
 * and assume that "sign" and "mask" have been declared and
 * initialized correctly.
 */
#define	MPSEQ_LT(a, b)	((((a) - (b)) & sign) != 0)
#define	MPSEQ_GE(a, b)	((((a) - (b)) & sign) == 0)
#define	MPSEQ_GT(a, b)	((((b) - (a)) & sign) != 0)
#define	MPSEQ_LE(a, b)	((((b) - (a)) & sign) == 0)
#define	MPSEQ_EQ(a, b)	((((a) - (b)) & mask) == 0)
#define	MPSEQ_NEQ(a, b)	((((a) - (b)) & mask) != 0)
#define	MPSEQ_ADD(a, b)	(((a) + (b)) & mask)
#define	MPSEQ_NEXT(a)	MPSEQ_ADD(a, 1)
#define	MPSEQ_INC(a)	((a) = MPSEQ_NEXT(a))
#define	MPSEQ_SUB(a, b)	(((a) - (b)) & mask)

int
mp_reass(pp, m)
	register struct p2pcom *pp;
	register struct mbuf *m;
{
	struct	ppp_header *ph;
	u_char	*mph;
	struct	ppp_sc *ppp;
	struct	ppp_sc *ppp2;
	struct	ppp_mp *mp;
	int	flags, mph_len, seq;
	struct	ppp_frag *frag, *newfrag;
	int	mask, sign;
	struct	mbuf *m0;

	ppp = (struct ppp_sc *)pp->p2p_private;

	if ((mp = ppp->ppp_mp) == NULL)	/* Not configured for Multilink */
		return (-1);

	/*
	 * Extract the sequence number and flags and strip off
	 * the header.
	 */
	mph_len = (ppp->ppp_flags & PPP_RCV_SSEQ) ?  MP_SSEQ_LEN : MP_LSEQ_LEN;
	if (m->m_len < sizeof(*ph) + mph_len)
		m = m_pullup(m, sizeof(*ph) + mph_len);
	ph = mtod(m, struct ppp_header *);
	mph = (u_char *)(ph + 1);
	flags = *mph >> MP_FLAG_SHIFT;
	if (ppp->ppp_flags & PPP_RCV_SSEQ) {
		ppp_stat.ppps_mp_sseq++;
		seq = (*mph & 0x0f) << 8;
		seq += *++mph;
		mask = MP_SSEQ_MASK;
		sign = MP_SSEQ_SIGN;
	} else {
		ppp_stat.ppps_mp_lseq++;
		seq = *++mph << 16;
		seq += *++mph << 8;
		seq += *++mph;
		mask = MP_LSEQ_MASK;
		sign = MP_LSEQ_SIGN;
	}

	if (!(m->m_flags & M_PKTHDR)) {
		dprintf(("%s%d: M_PKTHDR not set. (m->m_pkthdr.len = %d)\n",
		    pp->p2p_if.if_name, pp->p2p_if.if_unit,
		    m->m_pkthdr.len));
		mph_len = 0;
		for (m0 = m; m0; m0 = m0->m_next)
			mph_len += m0->m_len;
		m->m_pkthdr.len = mph_len;
		m->m_pkthdr.rcvif = &pp->p2p_if;
		m->m_flags |= M_PKTHDR;
	}
	m0 = m->m_next;
	m0->m_pkthdr.len = m->m_pkthdr.len - m->m_len;
	m0->m_pkthdr.rcvif = m->m_pkthdr.rcvif;
	m0->m_flags |= M_PKTHDR;

	m->m_next = NULL;
	m->m_len = 0;
	m->m_pkthdr.len = 0;
	m = m0;

	/*
	 * If this link has the current minimum sequence number,
	 * recalculate the minimum sequence number.  Otherwise
	 * just update our last sequence number.
	 */
	if (ppp->ppp_seq == mp->mp_rcv_seq_min) {
		if (seq == MPSEQ_NEXT(ppp->ppp_seq)) {
			ppp->ppp_seq = mp->mp_rcv_seq_min = seq;
		} else {
			ppp->ppp_seq = seq;
			for (ppp2 = LIST_FIRST(&mp->mp_bndl); ppp2;
			    ppp2 = LIST_NEXT(ppp2, ppp_mp_next))
				if (MPSEQ_LT(ppp2->ppp_seq, seq))
					seq = ppp2->ppp_seq;
			mp->mp_rcv_seq_min = seq;
			seq = ppp->ppp_seq;
		}
	} else
		ppp->ppp_seq = seq;

	/*
	 * Ok, if this fragment is a complete packet, and it is
	 * the next expected sequence number, just increment the
	 * next expected sequence number and pass up the packet.
	 */
	if ((flags & (MP_BEGIN|MP_END)) == MP_BEGIN|MP_END) {
		if (MPSEQ_LE(seq, mp->mp_rcv_seq_next)) {
			if (MPSEQ_EQ(seq, mp->mp_rcv_seq_next))
				MPSEQ_INC(mp->mp_rcv_seq_next);
			ppp_stat.ppps_mp_inseq++;
			mp_input(pp, m);
			goto check_seq;
		}
		ppp_stat.ppps_mp_wholefrag++;
	}

	/*
	 * Packet is either:
	 *	1) A fragment
	 *	2) Out of sequence
	 *	3) Both
	 * We locate a position for it on the resequencing queue,
	 * and merge it with any adjacent pieces.
	 */

	frag = LIST_FIRST(&mp->mp_frag);

	if (frag == LIST_END(&mp->mp_frag)) {
		FRAG_ALLOC(newfrag);
		if (newfrag == NULL)
			goto frag_alloc_err;
		newfrag->frag_m = m;
		newfrag->frag_seq_first = seq;
		newfrag->frag_seq_last = seq;
		newfrag->frag_flags = flags;
		LIST_INSERT_HEAD(&mp->mp_frag, newfrag, frag_next);
		ppp_stat.ppps_mp_onlyfrag++;
		return(0);
	}
	for (;;) {
		if (MPSEQ_EQ(frag->frag_seq_last + 1, seq)) {
			if (((frag->frag_flags & MP_END) == 0) &&
			    (flags & MP_BEGIN) == 0) {
				frag->frag_flags |= flags;
				frag->frag_seq_last = seq;
				frag->frag_m->m_pkthdr.len += m->m_pkthdr.len;
				m_cat(frag->frag_m, m);
				newfrag = frag;
				ppp_stat.ppps_mp_mergefrag++;
			} else {
				FRAG_ALLOC(newfrag);
				if (newfrag == NULL)
					goto frag_alloc_err;
				newfrag->frag_m = m;
				newfrag->frag_seq_first = seq;
				newfrag->frag_seq_last = seq;
				newfrag->frag_flags = flags;
				LIST_INSERT_AFTER(frag, newfrag, frag_next);
			}
			frag = LIST_NEXT(newfrag, frag_next);
			if (frag != LIST_END(&mp->mp_frag) &&
			    MPSEQ_EQ(newfrag->frag_seq_last + 1,
			    frag->frag_seq_first) &&
			    (newfrag->frag_flags & MP_END) == 0 &&
			    (frag->frag_flags & MP_BEGIN) == 0) {
				newfrag->frag_flags |= frag->frag_flags;
				newfrag->frag_m->m_pkthdr.len +=
				    frag->frag_m->m_pkthdr.len;
				m_cat(newfrag->frag_m, frag->frag_m);
				newfrag->frag_seq_last = frag->frag_seq_last;
				LIST_REMOVE(frag, frag_next);
				FRAG_FREE(frag);
				ppp_stat.ppps_mp_mergefrag++;
			}
			break;
		}
		if (MPSEQ_LT(seq, frag->frag_seq_first)) {
			if ((MPSEQ_EQ(seq+1, frag->frag_seq_first)) &&
			    ((flags & MP_END) == 0) &&
			    (frag->frag_flags & MP_BEGIN) == 0) {
				frag->frag_flags |= flags;
				frag->frag_seq_first = seq;
				m->m_pkthdr.len += frag->frag_m->m_pkthdr.len;
				m_cat(m, frag->frag_m);
				frag->frag_m = m;
				ppp_stat.ppps_mp_mergefrag++;
			} else {
				FRAG_ALLOC(newfrag);
				if (newfrag == NULL)
					goto frag_alloc_err;
				newfrag->frag_m = m;
				newfrag->frag_seq_first = seq;
				newfrag->frag_seq_last = seq;
				newfrag->frag_flags = flags;
				LIST_INSERT_BEFORE(frag, newfrag, frag_next);
			}
			break;
		}
		if (LIST_NEXT(frag, frag_next) == LIST_END(&mp->mp_frag)) {
			FRAG_ALLOC(newfrag);
			if (newfrag == NULL) {
			frag_alloc_err:
				return(-1);
			}
			newfrag->frag_m = m;
			newfrag->frag_seq_first = seq;
			newfrag->frag_seq_last = seq;
			newfrag->frag_flags = flags;
			LIST_INSERT_AFTER(frag, newfrag, frag_next);
			ppp_stat.ppps_mp_endfrag++;
			break;
		}
		frag = LIST_NEXT(frag, frag_next);
	}

	/*
	 * Lost fragment detection
	 * Find all the complete packets before the minimum
	 * and process them, any partial packets that are
	 * missing a piece before the minimum are discarded.
	 */
	while ((frag = LIST_FIRST(&mp->mp_frag)) != LIST_END(&mp->mp_frag)) {
		/*
		 * If the packet is complete, and begins before
		 * the minimum, pass it up and update mp_rcv_seq_next.
		 * We also check for a missing END bit, which is
		 * implied by the next packet having the BEGIN
		 * bit and the next sequence number.
		 */
		if ((frag->frag_flags & MP_BEGIN) &&
		    ((frag->frag_flags & MP_END) ||
		    (LIST_NEXT(frag, frag_next) != LIST_END(&mp->mp_frag) &&
		    MPSEQ_EQ(frag->frag_seq_last + 1,
		    LIST_NEXT(frag, frag_next)->frag_seq_first)))) {
			if (MPSEQ_LE(frag->frag_seq_first,
			    mp->mp_rcv_seq_min)) {
				LIST_REMOVE(frag, frag_next);
				mp_input(pp, frag->frag_m);
				mp->mp_rcv_seq_next =
				    MPSEQ_NEXT(frag->frag_seq_last);
				ppp_stat.ppps_mp_oseqfrag +=
				    MPSEQ_SUB(frag->frag_seq_last + 1,
				    frag->frag_seq_first);
				FRAG_FREE(frag);
				continue;
			}
			break;
		}
		/*
		 * We have a partial packet.  If it has the BEGIN bit
		 * and ends before the minimum, or if it has the END
		 * bit but begins before the minimum, toss it.
		 */
		if (((frag->frag_flags & MP_BEGIN) &&
		    MPSEQ_LT(frag->frag_seq_last, mp->mp_rcv_seq_min)) ||
		    ((frag->frag_flags & MP_END) &&
		    MPSEQ_LE(frag->frag_seq_first, mp->mp_rcv_seq_min))) {
			LIST_REMOVE(frag, frag_next);
			m_freem(frag->frag_m);
			mp->mp_rcv_seq_next =
			    MPSEQ_NEXT(frag->frag_seq_last);
			/* If missing END, skip beyond the missing END pkt */
			if ((frag->frag_flags & MP_END) == 0)
				MPSEQ_INC(mp->mp_rcv_seq_next);
			ppp_stat.ppps_mp_dropfrag +=
			    MPSEQ_SUB(frag->frag_seq_last + 1,
			    frag->frag_seq_first);
			FRAG_FREE(frag);
			continue;
		}
		break;
	}

check_seq:
	/*
	 * Pull off all the in sequence and complete packets from
	 * the front of the resequencing queue.  We also pull off
	 * complete packets that are missing the END bit, by noticing
	 * that the next packet starts with the next sequence number
	 * and has the BEGIN bit set.
	 */
	for (;;) {
		frag = LIST_FIRST(&mp->mp_frag);
		if (frag == LIST_END(&mp->mp_frag))
			break;
		if (frag->frag_seq_first != mp->mp_rcv_seq_next)
			break;
		if (frag->frag_flags == MP_BEGIN|MP_END)
			;
		else if ((frag->frag_flags & MP_BEGIN) != MP_BEGIN)
			break;
		else if (LIST_NEXT(frag, frag_next) == LIST_END(&mp->mp_frag))
			break;
		else if (MPSEQ_NEQ(frag->frag_seq_last + 1,
		    LIST_NEXT(frag, frag_next)->frag_seq_first))
			break;

		LIST_REMOVE(frag, frag_next);
		mp_input(pp, frag->frag_m);
		mp->mp_rcv_seq_next = MPSEQ_NEXT(frag->frag_seq_last);
		FRAG_FREE(frag);
		ppp_stat.ppps_mp_inseqfrag++;
	}
	return(0);
}

void
mp_input(pp, m)
	register struct p2pcom *pp;
	register struct mbuf *m;
{
	struct ppp_mp *mp = ((struct ppp_sc *)pp->p2p_private)->ppp_mp;
	int into_data;

	if (m->m_pkthdr.len > mp->mp_mrru + 5) {		/* XXX */
		/* Packet is too big for this bundle... */
		dprintf(("%s%d: packet too big %d\n", pp->p2p_if.if_name,
		    pp->p2p_if.if_unit, m->m_pkthdr.len));
		m_freem(m);
		ppp_stat.ppps_mp_longfrag++;
		return;
	}
	/*
	 * Hand the pieces back to ppp_hdr(), and then
	 * call ppp_input() again.
	 */
	for (;;) {
		into_data = ppp_hdr(pp, m->m_data, m->m_len);
		m = m_free(m);
		if (m == NULL)
			break;
		if (into_data) {
			pp->p2p_ibuf->m_pkthdr.len += m->m_pkthdr.len;
			m_cat(pp->p2p_ibuf, m);
			break;
		}
	}
	ppp_input(pp, NULL);
}

void
mp_detach(pp, ppp)
	struct p2pcom *pp;
	struct ppp_sc *ppp;
{
	struct ppp_mp *mp;

	if ((mp = ppp->ppp_mp) != NULL) {
		ppp->ppp_mp = NULL;
		LIST_REMOVE(ppp, ppp_mp_next);
		MP_FREE(mp);
		if (ppp_pif_free)
			(*ppp_pif_free)(&pp->p2p_if);
	}
}

int
mp_ioctl(pp, cmd, data)
	struct p2pcom *pp;
	int cmd;
	caddr_t data;
{
	struct ppp_sc *ppp = (struct ppp_sc *)pp->p2p_private;
	struct ppp_mp *mp;
	struct ifpppreq *pio = (struct ifpppreq *)data;
	struct ppp_sc *oppp;
	struct ifnet *pif;
	int error, namelen, unitlen, i;
	char tbuf[12], *unitname;

	error = 0;

	switch (cmd) {
	case PPPIOCSMRRU:
		if (ppp->ppp_mp)
			ppp->ppp_mp->mp_mrru = pio->ifrp_mrru;
		else
			error = ENXIO;
		break;

	case PPPIOCGMRRU:
		if (ppp->ppp_mp)
			pio->ifrp_mrru = ppp->ppp_mp->mp_mrru;
		else
			error = ENXIO;
		break;

	case PPPIOCGID:
		if (ppp->ppp_mp) {
			bcopy(&ppp->ppp_mp->mp_ed, &pio->ifrp_ed,
			    sizeof(struct ppp_ed));
			bcopy(&ppp->ppp_mp->mp_auth, &pio->ifrp_auth,
			    sizeof(struct ppp_auth));
		} else
			error = ENXIO;
		break;

	case PPPIOCFID:
		for (mp = LIST_FIRST(&mp_list); ; mp = LIST_NEXT(mp, mp_next)) {
			if (mp == LIST_END(&mp_list)) {
				error = ENODEV;
				break;
			}
			if (bcmp(&pio->ifrp_ed, &mp->mp_ed,
			    sizeof(struct ppp_ed)))
				continue;
			if (bcmp(&pio->ifrp_auth, &mp->mp_auth,
			    sizeof(struct ppp_auth)))
				continue;
			pif = mp->mp_pif;
			goto return_pif;
		}
		break;

	case PPPIOCSID:
		if (ppp_pif_alloc == NULL)
			return(ENODEV);
		/*
		 * First, see if the bundle already exists.
		 */
		for (mp = LIST_FIRST(&mp_list); mp != LIST_END(&mp_list);
		    mp = LIST_NEXT(mp, mp_next)) {
			if (bcmp(&pio->ifrp_ed, &mp->mp_ed,
			    sizeof(struct ppp_ed)))
				continue;
			if (bcmp(&pio->ifrp_auth, &mp->mp_auth,
			    sizeof(struct ppp_auth)))
				continue;
			break;
		}
		/*
		 * If the bundle exists and we are not
		 * already part of a it, join it.
		 */
		if (mp != LIST_END(&mp_list)) {
			if (ppp->ppp_mp == NULL) {
				pif = mp->mp_pif;
				error = (*ppp_pif_alloc)(&pp->p2p_if, &pif);
				if (error)
					break;
				/*
				 * Link into the common TCP compression
				 * information.
				 */
				if (ppp->ppp_ctcp &&
				    --ppp->ppp_ctcp->sls_ref < 0)
					free(ppp->ppp_ctcp, M_DEVBUF);
				oppp = LIST_FIRST(&mp->mp_bndl);
				ppp->ppp_ctcp = oppp->ppp_ctcp;
				if (ppp->ppp_ctcp)
					ppp->ppp_ctcp->sls_ref++;
				MP_JOIN(mp, ppp);
				/* pick up the current common state */
				ppp->ppp_flags |=
					(oppp->ppp_flags & PPP_MLFLAGS);
				goto return_pif;
			} else if (ppp->ppp_mp != mp)
				error = EBUSY;
			break;
		}
		/*
		 * See if we are part of a bundle, and if others
		 * are also part of it, then we can't change it.
		 */
		if ((mp = ppp->ppp_mp) != NULL) {
			if (MP_INUSE(mp)) {
				error = EBUSY;
				break;
			}
			pif = mp->mp_pif;
		} else {
			pif = NULL;
			error = (*ppp_pif_alloc)(&pp->p2p_if, &pif);
			if (error)
				break;
			MP_ALLOC(mp, ppp);
			mp->mp_pif = pif;
		}
		/*
		 * Copy in the new ED/auth information
		 */
		bcopy(&pio->ifrp_ed, &mp->mp_ed, sizeof(mp->mp_ed));
		bcopy(&pio->ifrp_auth, &mp->mp_auth, sizeof(mp->mp_auth));
		goto return_pif;

	case PPPIOCGPIF:
		if ((pif = pp->p2p_if.if_pif) == NULL) {
			error = ENODEV;
			break;
		}

	return_pif:
		unitname = &tbuf[sizeof(tbuf)];
		i = pif->if_unit;
		do {
			*--unitname = '0' + (i % 10);
			i /= 10;
		} while (i);
		unitlen = &tbuf[sizeof(tbuf)] - unitname;

		namelen = strlen(pif->if_name);
		bcopy(pif->if_name, pio->ifrp_pif, namelen);
		bcopy(unitname, &pio->ifrp_pif[namelen], unitlen);
		pio->ifrp_pif[namelen + unitlen] = '\0';
		break;

	case PPPIOCCID:
		mp_detach(pp, ppp);
		break;
	default:
		error = EINVAL;
		break;
	}
	return(error);
}
