#ifndef _NTRBUF_H_
#define _NTRBUF_H_

#define NTR_BUF_SIZE (64*1024)

typedef struct NtrBuf {
	int head;
	int tail;
	int size;
	char *data;
} NtrBuf;


#define NtrBuf_size(bp)  ((bp)->size)


static void
NtrBuf_reset(NtrBuf* bp)
{
	bp->head = bp->tail = 0;
	bp->size = NTR_BUF_SIZE;
	if (bp->data == 0)
		bp->data = (char*) Malloc(NTR_BUF_SIZE+256);
}


/*
   Write the data at the tail of the buffer
   Invariance:  head = tail only when empty
 */
static int
NtrBuf_write(NtrBuf* bp, void* buf, int nbyte)
{
	char *data = (char*) buf;
	int head, tail, avail;
	int s;

	if (nbyte > bp->size) {
		printf("NtrBuf_write: nbyte=%d exceeds buffer size %d\n",
		       nbyte, bp->size); 
		return 0;
	}

	s = splimp();
	head = bp->head;	/* previous head */
	tail = bp->tail;	/* previous tail */

	/* Check for overflow: distance between tail and head < nbyte */
	if (tail+nbyte >= ((head > tail) ? head : head+bp->size)) {
		printf("NtrBuf_write: %d+%d overflow head=%d, data=0x%x\n",
		       tail, nbyte, head, *((u_int*)data));
		/* make head > newtail */
		bp->head = (tail + nbyte + sizeof(u_int32_t)) % bp->size;
	}

	/* Set the new tail */
	bp->tail = (bp->tail + nbyte) % bp->size;
	splx(s);

	avail = bp->size - tail;
	if (avail < nbyte) {
		bcopy(data, bp->data + tail, avail);
		bcopy(data + avail, bp->data, nbyte - avail);
	}
	else bcopy(data, bp->data + tail, nbyte);

#if DEBUG > 1
	printf("NtrBuf_write: %d bytes data[%d]=0x%x\n",
	       nbyte, tail, *((u_int*)(bp->data+tail)));
#endif
	return nbyte;
}


static int
NtrBuf_read(NtrBuf* bp, void* buf, int nbyte)
{
	char *data = (char*) buf;
	int head, tail, avail;
	int s;

	if (bp->head == bp->tail)
		return 0;

	s = splimp();
	head = bp->head;	/* previous head */
	tail = bp->tail;	/* previous tail */

	if ((avail = tail - head) < 0)
		avail += bp->size;
	if (avail < nbyte)
		nbyte = avail;
	bp->head = (bp->head + nbyte) % bp->size;
	splx(s);

	avail = bp->size - head;
	if (avail < nbyte) {
		bcopy(bp->data + head, data, avail);
		bcopy(bp->data, data + avail, nbyte - avail);
	}
	else bcopy(bp->data + head, data, nbyte);
	return nbyte;
}


int
NtrBuf_uiomove(NtrBuf* bp, struct uio* uio, int ioflag)
{
	int nbyte, status;
	int head, tail, avail;
	int s;

	if (bp->head == bp->tail)
		return 0;

	s = splimp();
	head = bp->head;	/* previous head */
	tail = bp->tail;	/* previous tail */

	/* Calculate the transfer length */
	if ((nbyte = tail - head) < 0)
		nbyte += bp->size;
	if (uio->uio_resid < nbyte)
		nbyte = uio->uio_resid;
	bp->head = (bp->head + nbyte) % bp->size;
	splx(s);

	avail = bp->size - head;
	if (avail < nbyte) {
		uiomove(bp->data + head, avail, uio);
		status = uiomove(bp->data, nbyte - avail, uio);
	}
	else status = uiomove(bp->data + head, nbyte, uio);

#ifdef DEBUG
	printf("NtrBuf_uiomove: %d bytes data[%d]=0x%x\n",
	       nbyte, head, *((u_int*)(bp->data+head)));
#endif
	return status;
}	    

#endif
