#ifndef _NTRTRACK_H_
#define _NTRTRACK_H_

#define NTR_NUM_TRACKS 127

typedef struct NtrTrack {
	u_int32_t magic;
	int vlsize;
	struct tr_plist_ent_t *plist;
} NtrTrack;


/*
   Return a pointer to a track
   If track's magic is 0, then no track exist with the same magic
   Otherwise, track->magic == magic
 */
static NtrTrack*
NtrTrack_find(NtrTrack* tracks[], u_int32_t magic)
{
	int bucket;

	bucket = magic % NTR_NUM_TRACKS;
	while (tracks[bucket] && tracks[bucket]->magic != magic)
		bucket = (bucket + 1) % NTR_NUM_TRACKS;
	return tracks[bucket];
}


/*
   Return a pointer to a newly allocated track
   Find an empty track (ie. magic==0) and set the track->magic
   The new magic might be increments of the magic passed in
 */
static NtrTrack*
NtrTrack_new(NtrTrack* tracks[], u_int32_t magic)
{
	int bucket;

	do {
		bucket = magic % NTR_NUM_TRACKS;
		while (tracks[bucket] && tracks[bucket]->magic != magic)
			bucket = (bucket + 1) % NTR_NUM_TRACKS;
	} while (tracks[bucket] && magic++);
	tracks[bucket] = (NtrTrack*) Malloc(sizeof(NtrTrack));
	tracks[bucket]->magic = magic;
	return tracks[bucket];
}


static void
NtrTrack_reset(NtrTrack* tracks[])
{
	int bucket;

	for (bucket = 0;  bucket < NTR_NUM_TRACKS;  bucket++) {
		if (tracks[bucket]) {
			bzero(tracks[bucket], sizeof(NtrTrack));
			tracks[bucket] = 0;
		}
	}
}

#endif
