/* #define DEBUG 2 */

#include <sys/param.h>
#include <sys/cdefs.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>

#include <net/ntrvar.h>
#include <net/ntr_buf.h>
#include <net/ntr_track.h>
#include <net/ntrrec.h>


int ntr_on = 0;			/* off = 0;  on = 1 */
/* list of devices to be traced */
static ntrace_d *ntr_dlist = 0;
static NtrTrack *ntr_tracks[NTR_NUM_TRACKS];

static struct tr_plist_ent_t plist_usr[] = {
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_SEND, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_RECV, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, SOCK_SENDPORT, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, SOCK_RECVPORT, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, PKT_SEQUENCE, 0 },
	{ 0, PLIST_TYPE_UINT32, 8, PKT_SENTTIME, 0 },
	{ 0, 0, 0, 0, 0 }
};

static struct tr_plist_ent_t plist_udp[] = {
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_SEND, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_RECV, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, SOCK_SENDPORT, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, SOCK_RECVPORT, 0 },
	{ 0, 0, 0, 0, 0 }
};

static struct tr_plist_ent_t plist_tcp[] = {
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_SEND, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_RECV, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, SOCK_SENDPORT, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, SOCK_RECVPORT, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, PKT_SEQUENCE, 0 },
	{ 0, 0, 0, 0, 0 }
};

static struct tr_plist_ent_t plist_icmp[] = {
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_SEND, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, ADDR_RECV, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, ICMP_KIND, 0 },
	{ 0, PLIST_TYPE_UINT16, 2, ICMP_ID, 0 },
	{ 0, PLIST_TYPE_UINT32, 4, PKT_SEQUENCE, 0 },
	{ 0, PLIST_TYPE_UINT32, 8, ICMP_PINGTIME, 0 },
	{ 0, 0, 0, 0, 0 }
};


#ifdef	KERNEL
static NtrBuf ntr_buf = {0, 0, 0, 0};
NtrBuf *ntrbufp = &ntr_buf;
#else
#include <stdio.h>
FILE *ntrbufp;
#endif


#ifdef	KERNEL
#define MTOD_IP(ip, m) \
do { \
	ip = mtod(m, struct ip *); \
	if (m->m_len < ip->ip_hl) { \
		ip = (m=m_pullup(m, ip->ip_hl+32)) ? mtod(m, struct ip*) : 0; \
	} \
} while (0)

static u_long get_ipaddr(struct ifnet *ifp)
{
	struct ifaddr *ifa;
#ifdef	__NetBSD__
	for (ifa=ifp->if_addrlist.tqh_first;  ifa;  ifa=ifa->ifa_list.tqe_next)
#else
	for (ifa=ifp->if_addrlist;  ifa;  ifa=ifa->ifa_next)
#endif
	{
		if (ifa->ifa_addr->sa_family == AF_INET)
			return ((struct sockaddr_in *) ifa->ifa_addr)->sin_addr.s_addr;
	}
}

#else  /* not KERNEL */
#define MTOD_IP(ip, m)  ((ip) = (struct ip *) m)

static u_long get_ipaddr(struct ifnet *ifp)
{
	struct hostent *host;
	char hostname[64];
	gethostname(hostname, sizeof(hostname)-1);
	host = (struct hostent *) gethostbyname(hostname);
	return host ? ((struct in_addr *)(host->h_addr))->s_addr : gethostid();
}
#endif


#define NEW_TRACK(tp, new)  (((tp) && (tp)->magic) ? (tp) : ((tp) = (new)))
#define FREE_TRACK(tp) \
do { \
	if (tp) \
		Mfree(tp); \
} while (0)


/*
 * tr_plist_size:
 *	calculate the header size and entry's vlist size in bytes
 *	return the number of entries in plist
 */

static int
tr_plist_size(struct tr_plist_ent_t *plist, int *hdrsize, int *vlsize)
{
	int sz, plsize;

	plsize = *hdrsize = *vlsize = 0;
	for (;  plist->pl_type != 0;  plist++) {
		sz = plist->pl_size + PAD_BYTES(plist->pl_size);
		if (! (plist->pl_flags & PLIST_HEADER_ONLY))
			*vlsize += sz;
		else
			*hdrsize += sz;
		*hdrsize += TR_SIZEOF(struct tr_plist_ent_t);
		plsize++;
	}
	return plsize;
}


/*
 * return size of the value list corresponding to the plist
 */
static int
ntrace_write_plist(struct tr_plist_ent_t *plist)
{
	int sz, vlsize;
	struct tr_plist_ent_t prop[8]; /* buffer to marshal plist */

	vlsize = 0;
	for (;  plist->pl_type != 0;  plist++) {
		sz = plist->pl_size + PAD_BYTES(plist->pl_size);
		if (! (plist->pl_flags & PLIST_HEADER_ONLY)) {
			vlsize += sz;
			sz = 0;
		}
		*prop = *plist;
		HTONS(prop->pl_size);
		HTONL(prop->pl_name);
		NtrBuf_write(ntrbufp, prop, TR_SIZEOF(*plist) + sz);
	}
	return vlsize;
}


/*
 * New_dev_track:  create new device track and save a new tr_dev_hdr_t
 */

static NtrTrack*
New_dev_track(ntrace_d *ndp, struct tr_plist_ent_t *plist)
{
	struct tr_dev_hdr_t dh;
	NtrTrack *tp;
	int hdrsize, plsize;

	if (ndp->ipaddr == 0)
		ndp->ipaddr = (u_long) get_ipaddr(ndp->ifp);
	tp = NtrTrack_new(ntr_tracks, ndp->dev_type);
	tp->plist = plist;
	plsize = tr_plist_size(tp->plist, &hdrsize, &(tp->vlsize));

	dh.dh_magic = htonl(TR_DEV_HDR_MAGIC);
	dh.dh_size = htonl(TR_SIZEOF(dh) + hdrsize);
	NANOTIME(&(dh.dh_timestamp));
	dh.dh_trackmagic = htonl(tp->magic);
	dh.dh_dev_type = htonl(ndp->dev_type);
	dh.dh_ipaddr = ndp->ipaddr;
	dh.dh_plistsz = htonl(plsize);

#ifdef DEBUG
	printf("New_dev_track: 0x%x, dev 0x%x, plsize %d, vlsize %d\n",
	       tp->magic, ndp->dev_type, plsize, tp->vlsize);
#endif
	NtrBuf_write(ntrbufp, &dh, TR_SIZEOF(dh));
	ntrace_write_plist(tp->plist);
	return tp;
}


/*
 * New_pkt_track:  create new packet track and save a new tr_pkt_hdr_t
 */

static NtrTrack*
New_pkt_track(ntrace_d *ndp, struct tr_plist_ent_t *plist, u_int32_t protocol)
{
	struct tr_pkt_hdr_t ph;
	NtrTrack *tp;
	int plsize, hdrsize;

	if (ndp->ipaddr == 0)
		ndp->ipaddr = (u_long) get_ipaddr(ndp->ifp);
	tp = NtrTrack_new(ntr_tracks, ndp->dev_type);
	tp->plist = plist;
	plsize = tr_plist_size(tp->plist, &hdrsize, &(tp->vlsize));

	ph.ph_magic = htonl(TR_PKT_HDR_MAGIC);
	ph.ph_size = htonl(TR_SIZEOF(ph) + hdrsize);
	NANOTIME(&(ph.ph_timestamp));
	ph.ph_trackmagic = htonl(tp->magic);
	ph.ph_dev_type = htonl(ndp->dev_type);
	ph.ph_ipaddr = ndp->ipaddr;
	ph.ph_protocol = htonl(protocol);
	ph.ph_plistsz = htonl(plsize);

#ifdef DEBUG
	printf("New_pkt_track: 0x%x, dev 0x%x, plsz %d, vlsz %d, proto %d\n",
	       tp->magic, ndp->dev_type, plsize, tp->vlsize, protocol);
#endif
	NtrBuf_write(ntrbufp, &ph, TR_SIZEOF(ph));
	ntrace_write_plist(tp->plist);
	return tp;
}


int
ntrace_pkt_ip(ntrace_d *ndp, struct ip *ip)
{
	struct tr_pkt_ent_t pe;
	NtrTrack *tp;
	struct icmp *icp;
	struct udphdr *udp;
	struct tcphdr *tcp;
	u_int32_t *data;
	u_int32_t vlist[8];

	icp = (struct icmp *) ((char *)ip + (ip->ip_hl << 2));
	tcp = (struct tcphdr *) icp;

	switch(ip->ip_p) {
	case IPPROTO_TCP:
		tp = NEW_TRACK(ndp->tcp_tp, New_pkt_track(ndp, plist_tcp,
							  TR_PKT_TCP_MAGIC));
		*(u_int16_t *)(vlist+2) = tcp->th_sport;
		*(u_int16_t *)(vlist+3) = tcp->th_dport;
		vlist[4] = tcp->th_seq;
		break;

	case IPPROTO_UDP:
		udp = (struct udphdr *) icp;
		*(u_int16_t *)(vlist+2) = udp->uh_sport;
		*(u_int16_t *)(vlist+3) = udp->uh_dport;
		data = (u_int32_t *)(udp + 1);
		if (data[0] != htonl(PKT_SENTTIME)) {
			tp = NEW_TRACK(ndp->udp_tp, New_pkt_track(ndp, plist_udp, TR_PKT_UDP_MAGIC));
		}
		else {
			tp = NEW_TRACK(ndp->usr_tp, New_pkt_track(ndp, plist_usr, TR_PKT_UDP_MAGIC));
			/* special UDP stream */
			vlist[4] = data[1];
			vlist[5] = data[2];
			vlist[6] = data[3];
		}
		break;

	case IPPROTO_ICMP:
		tp = NEW_TRACK(ndp->icmp_tp, New_pkt_track(ndp, plist_icmp,
							   TR_PKT_ICMP_MAGIC));
		if (ntohs(ip->ip_len) < ICMP_MINLEN)
			return 0;
		*(u_int16_t *)(vlist+2) = htons((icp->icmp_type << 8) | icp->icmp_code);
		*(u_int16_t *)(vlist+3) = icp->icmp_id;
		if (icp->icmp_type == ICMP_ECHOREPLY || icp->icmp_type == ICMP_ECHO) {
			/* ECHOREPLY data is in host order (ping) */
			struct timeval now, *pingtime;
			microtime(&now);
			pingtime = (struct timeval *) (vlist+5);
			vlist[4] = htonl((u_int32_t) icp->icmp_seq);
			timersub(&now, (struct timeval *)icp->icmp_data, pingtime);
			TR_TIME_HTON(pingtime);
		}
		else {
			vlist[4] = htonl((u_int32_t) ntohs(icp->icmp_seq));
			vlist[5] = vlist[6] = 0;
		}
		break;

	default:
		return 0;
	}
	vlist[0] = ip->ip_src.s_addr;
	vlist[1] = ip->ip_dst.s_addr;

	pe.pe_magic = htonl(tp->magic);
	pe.pe_size = htonl(TR_SIZEOF(pe) + tp->vlsize);
	NANOTIME(&(pe.pe_timestamp));
	pe.pe_flags = 0;
	pe.pe_psize = htonl(ntohs(ip->ip_len));

#ifdef DEBUG
	printf("ntrace_pkt: %x, %d, vl[%d] %x %x %x %x %x %x\n",
	       ndp->dev_type, ip->ip_p, tp->vlsize / sizeof(u_int32_t),
	       ntohl(vlist[0]), ntohl(vlist[1]),
	       ntohl(vlist[2]), ntohl(vlist[3]),
	       ntohl(vlist[4]), ntohl(vlist[6]));
#endif
	NtrBuf_write(ntrbufp, &pe, TR_SIZEOF(pe));
	NtrBuf_write(ntrbufp, vlist, tp->vlsize);
	return htonl(pe.pe_size);
}


#ifdef	KERNEL
/*--------------------------------------------------------------
 * Called from ntr.c
 */

int
ntrace_open(dev_t dev, int flag, int ifmt, struct proc *p)
{
	ntrace_d *ndp;
	int s;

	if (ntr_on)
		return 1;

	for (ndp = ntr_dlist;  ndp;  ndp = ndp->next) {
		ifpromisc(ndp->ifp, 1);
	}
	NtrTrack_reset(ntr_tracks);
	s = splimp();
	NtrBuf_reset(ntrbufp);
	ntr_on = 1;
	splx(s);

	printf("ntrace_open:  bufsize=%d\n", NtrBuf_size(ntrbufp));
	return 0;
}


int
ntrace_close(dev_t dev, int flag, int ifmt, struct proc *p)
{
	ntrace_d *ndp;
	int s;

	s = splimp();
	NtrBuf_reset(ntrbufp);
	ntr_on = 0;
	splx(s);

	for (ndp = ntr_dlist;  ndp;  ndp = ndp->next) {
		ifpromisc(ndp->ifp, 0);
		FREE_TRACK(ndp->dev_tp);
		FREE_TRACK(ndp->icmp_tp);
		FREE_TRACK(ndp->tcp_tp);
		FREE_TRACK(ndp->udp_tp);
		FREE_TRACK(ndp->usr_tp);
		ndp->dev_tp = 0;
		ndp->icmp_tp = ndp->tcp_tp = ndp->udp_tp = ndp->usr_tp = 0;
	}

	printf("ntrace_close\n");
	return 0;
}


int
ntrace_read(dev_t dev, struct uio *uio, int ioflag)
{
	if (!ntr_on)
		return EIO;
	return NtrBuf_uiomove(ntrbufp, uio, ioflag);
}


/*--------------------------------------------------------------*/
/* Call from device drivers */
/*--------------------------------------------------------------*/

ntrace_d*
ntrace_dev_attach(ntrace_d **ndpp, struct tr_plist_ent_t *plist,
		  int dev_type, int period, struct ifnet *ifp, int hlen)
{
	ntrace_d *ndp;
	NtrTrack *tp;

	ndp = (ntrace_d *) Malloc(sizeof(ntrace_d));
	bzero(ndp, sizeof(ntrace_d));
	microtime(&(ndp->timestamp));
	ndp->ipaddr = 0;
	ndp->dev_type = dev_type;
	ndp->period.tv_sec = period / 1000000;
	ndp->period.tv_usec = period % 1000000;
	ndp->ifp = ifp;
	ndp->hlen = hlen;
	ndp->dev_plist = plist;
	*ndpp = ndp;

	ndp->next = ntr_dlist;
	ntr_dlist = ndp;
	printf("ntrace_dev_attach: 0x%x, period %dusec\n", dev_type, period);
}


int
ntrace_dev_ready(ntrace_d *ndp)
{
	struct timeval now;
	struct timeval diff;
	long usec;

	if (!ntr_on)
		return 0;
	microtime(&now);
	timersub(&now, &ndp->timestamp, &diff);
	if (timercmp(&diff, &ndp->period, >)) {
		ndp->timestamp = now;
		return 1;
	}
	return 0;
}


int
ntrace_dev(ntrace_d *ndp, char *vlist)
{
	struct tr_dev_ent_t de;
	NtrTrack *tp;

	if (!ntr_on)
		return 0;
	tp = NEW_TRACK(ndp->dev_tp, New_dev_track(ndp, ndp->dev_plist));

	de.de_magic = htonl(tp->magic);
	de.de_size = htonl(TR_SIZEOF(de) + tp->vlsize);
	NANOTIME(&(de.de_timestamp));

#ifdef DEBUG
	printf("ntrace_dev: 0x%x, vlsize %d\n",
	       ndp->dev_type, ndp->dev_tp->vlsize);
#endif
	NtrBuf_write(ntrbufp, &de, TR_SIZEOF(de));
	NtrBuf_write(ntrbufp, vlist, ndp->dev_tp->vlsize);
	return htonl(de.de_size);
}


int
ntrace_pkt(ntrace_d *ndp, struct mbuf *pm, struct ether_header *eh)
{
	struct ip *ip;
	int    ether_type;

	if (!ntr_on) 
		return 0;

	ether_type = ntohs(eh->ether_type);
	/* For now, we only look at IP packets, nothing else */
	if (ether_type != ETHERTYPE_IP) {
#ifdef DEBUG
		printf("Packet not traced: ether_type 0x%x\n", ether_type);
#endif
		return 0;
	}

	MTOD_IP(ip, pm);
	if (!ip) {
#ifdef DEBUG
		printf("Packet not traced: MTOD_IP fails\n");
#endif
		return 0;
	}

	return ntrace_pkt_ip(ndp, ip);
}


#else	/* not KERNEL */
#include <netdb.h>

FILE*
ntrace_open(char *trfile, char *device, char *desc)
{
	struct tr_header_t *hp;
	struct tr_pkt_hdr_t *php;
	u_int32_t dev_type;
	FILE *fp;

	if ((fp = fopen(trfile, "w")) == 0) {
		perror(trfile);
		exit(1);
	}
	ntrbufp = fp;

	if (!strncmp(device, "wl", 2))
		dev_type = TR_DEV_WL_MAGIC;
	else if (!strncmp(device, "strip", 5))
		dev_type = TR_DEV_MC_MAGIC;
	else if (!strncmp(device, "ppp", 3))
		dev_type = TR_DEV_MC_MAGIC;
	else if (!strncmp(device, "sl", 2))
		dev_type = TR_DEV_SL_MAGIC;
	else if (!strncmp(device, "ee", 2))
		dev_type = TR_DEV_EE_MAGIC;
	else
		dev_type = TR_DEV_EP_MAGIC;

	tr_begin_write(fp);
	hp = tr_header_new(desc);
	tr_write_header(fp, hp);
	php = tr_pkt_hdr_new(TR_PKT_UDP_MAGIC, dev_type, plist_udp);
	tr_write_pkt_hdr(fp, php);
	php = tr_pkt_hdr_new(TR_PKT_TCP_MAGIC, dev_type, plist_tcp);
	tr_write_pkt_hdr(fp, php);
	php = tr_pkt_hdr_new(TR_PKT_ICMP_MAGIC, dev_type, plist_icmp);
	tr_write_pkt_hdr(fp, php);

	ntr_on = 1;
	printf("ntrace_open:  %s\n", trfile);
	return fp;
}


void
ntrace_close(FILE *fp)
{
	if (fp) {
		struct tr_trailer_t *tp;
		tp = (struct tr_trailer_t *) tr_trailer_new();
		tr_write_trailer(fp, tp);
		tr_end_write(fp);
		fclose(fp);
	}

	ntr_on = 0;
	printf("ntrace_close");
	fp = 0;
}


struct tr_pkt_hdr_t*
New_pkt_hdr(ntrace_d *ndp, struct tr_plist_ent_t *plist, int protocol)
{
	struct tr_pkt_hdr_t *php;
	int plsize, hdrsize;
	NtrTrack *tp;

	ndp->ipaddr = (u_long) get_ipaddr(ndp->ifp);
	tp = NtrTrack_new(ntr_tracks, ndp->dev_type);
	tp->plist = plist;
	plsize = tr_plist_size(plist, &hdrsize, &tp->vlsize);

	php = (struct tr_pkt_hdr_t *) Malloc(sizeof(*php));
	php->ph_magic = TR_PKT_HDR_MAGIC;
	php->ph_size = TR_SIZEOF(*php) + plsize;
	NANOTIME(&(php->ph_timestamp));
	php->ph_trackmagic = TR_PKT_HDR_MAGIC + protocol;
	php->ph_dev_type = dev_type;
	php->ph_ipaddr = ndp->ipaddr;
	php->ph_protocol = protocol;
	php->ph_plistsz = plsize;
	php->ph_plist = plist;
}

int
ntrace_pkt(struct ip *ip, char *vlist, int nbyte)
{
	struct tr_pkt_ent_t pe;

	if (! ntr_on)
		return 0;

	pe.pe_magic = htonl(TR_PKT_HDR_MAGIC + ip->ip_p);
	pe.pe_size = htonl(TR_SIZEOF(pe) + nbyte);
	NANOTIME(&(pe.pe_timestamp));
	pe.pe_psize = htonl(ntohs(ip->ip_len));

	NtrBuf_write(ntrbufp, &pe, TR_SIZEOF(pe));
	NtrBuf_write(ntrbufp, vlist, nbyte);
	return htonl(pe.pe_size);
}

#endif	/* KERNEL */
