#ifndef _BLURB_
#define _BLURB_
/*
# 		    Ntrace: Mobile Network Tracing
# 			  Release 1.0-alpha
# 
# Copyright (c) 1996 Carnegie Mellon University, Regents of the
# University of California.  All Rights Reserved
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#  1. Redistributions of source code of this work, derivative works, or
#     modified versions must retain the above copyright notice, this
#     list of conditions and the following disclaimer. 
#  2. Redistributions in binary form of this work, derivative works or
#     modified versions must reproduce the above copyright notice, this
#     list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the
#     distribution.  
#  3. All advertising materials mentioning features or use of this
#     software must display the following acknowledgement:
# 	This product includes software developed by Carnegie Mellon
# 	University, the University of California, Berkeley, and their
# 	contributors.  
#  4. Neither the names of the universities nor the names of their
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
# 
#   NTRACE IS AN EXPERIMENTAL SOFTWARE SYSTEM AND IS KNOWN TO HAVE BUGS,
#   SOME OF WHICH MAY HAVE SERIOUS CONSEQUENCES.  CARNEGIE MELLON, THE
#   REGENTS, AND THEIR CONTRIBUTORS ALLOW FREE USE OF THIS SOFTWARE IN
#   ITS "AS IS" CONDITION.  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
#   BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
#   CARNEGIE MELLON, THE REGENTS, OR THEIR CONTRIBUTORS BE LIABLE FOR ANY
#   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
#   GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
#   IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE OR OF
#   ANY DERIVATIVE WORK, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#   DAMAGE.
# 
# Carnegie Mellon and the Regents encourage users of this software to
# return any improvements or extensions that they make, and to grant
# Carnegie Mellon and the Regents the rights to redistribute these
# changes without encumbrance.
# 
*/
#endif /*_BLURB_*/







#ifndef _NTRACE_H_
#define _NTRACE_H_

/* 
 * ntrace.h
 * 
 * Common data structures, etc. 
 */

/* Default kernel tracing device */
#define NTR_DEV "/dev/ntr0"

/* Definitions we need to pick up from various places */
#ifndef __P                     /* from <sys/cdefs.h> */
 #if defined(__STDC__) || defined(__cplusplus)
  #define __P(protos) protos
 #else
  #define __P(protos) ()
 #endif
#endif

/* Defines u_int32_t, etc. */
#include <sys/types.h>
#if !(defined(__bsdi__) || defined(__NetBSD__) || defined(__linux__))
#if __GNUC__

typedef unsigned long long u_int64_t;
typedef long long          int64_t;
typedef unsigned int       u_int32_t;
typedef int                int32_t;
typedef unsigned short     u_int16_t;
typedef short              int16_t
typedef unsigned char      u_int8_t;
typedef char               int8_t;

#endif
#endif

/* 64-bit integer htonq */
#if BYTE_ORDER == LITTLE_ENDIAN
#ifdef __GNUC__
#define htonq(x) \
           ((((x) & 0xff00000000000000LL) >> 56) | \
	    (((x) & 0x00ff000000000000LL) >> 40) | \
	    (((x) & 0x0000ff0000000000LL) >> 24) | \
	    (((x) & 0x000000ff00000000LL) >>  8) | \
	    (((x) & 0x00000000ff000000LL) <<  8) | \
	    (((x) & 0x0000000000ff0000LL) << 24) | \
	    (((x) & 0x000000000000ff00LL) << 40) | \
	    (((x) & 0x00000000000000ffLL) << 56))
#define ntohq(x) htonq(x)    
/* XXX - need some way to do this w/out long long */
#endif
#else
#define htonq(x)    (x)
#define ntohq(x)    (x)
#endif

#define NTOHQ(x)     ((x) = ntohq(x))
#define HTONQ(x)     NTOHQ(x)

/* Defines htonl/htons */
#include <netinet/in.h>
#ifndef HTONL
 #define HTONL(x) ((x) = htonl(x))
#endif
#ifndef HTONS
 #define HTONS(x) ((x) = htons(x))
#endif


/* Booleans */
typedef enum {
    false = 0,
    true = 1
} tr_bool_t;

/* Disk representations of time. */

struct tr_time_t {
    u_int32_t   tt_sec;       /* seconds */
    u_int32_t   tt_nsec;      /* nanoseconds */
};

#define NSECIFY(usecs)    ((usecs) *= 1000)
#define USECIFY(nsecs)    ((nsecs) /= 1000)
#define TR_TIME_NTOH(tp)  TR_TIME_HTON(tp)
#define TR_TIME_HTON(tp) \
do { \
    HTONL(((struct tr_time_t *) tp) ->tt_sec); \
    HTONL(((struct tr_time_t *) tp) ->tt_nsec); \
} while (0)


/* Sized strings */
struct tr_string_t {
    u_int32_t      s_size;
    unsigned char  *s_string;
};

#define TR_STRING_HTON(sp) \
do { \
    HTONL((sp)->s_size); \
} while (0)

#define TR_STRING_NTOH(sp)       TR_STRING_HTON(sp)

/* Operations on timevals.
 * Compatibility: code taken from NetBSD's /usr/include/sys/time.h
 */
#ifndef timercmp
#define timercmp(tvp, uvp, cmp)                                         \
        (((tvp)->tv_sec == (uvp)->tv_sec) ?                             \
            ((tvp)->tv_usec cmp (uvp)->tv_usec) :                       \
            ((tvp)->tv_sec cmp (uvp)->tv_sec))
#endif

#ifndef timeradd
#define timeradd(tvp, uvp, vvp)                                         \
        do {                                                            \
                (vvp)->tv_sec = (tvp)->tv_sec + (uvp)->tv_sec;          \
                (vvp)->tv_usec = (tvp)->tv_usec + (uvp)->tv_usec;       \
                if ((vvp)->tv_usec >= 1000000) {                        \
                        (vvp)->tv_sec++;                                \
                        (vvp)->tv_usec -= 1000000;                      \
                }                                                       \
        } while (0)
#endif

#ifndef timersub
#define timersub(tvp, uvp, vvp)                                         \
        do {                                                            \
                (vvp)->tv_sec = (tvp)->tv_sec - (uvp)->tv_sec;          \
                (vvp)->tv_usec = (tvp)->tv_usec - (uvp)->tv_usec;       \
                if ((vvp)->tv_usec < 0) {                               \
                        (vvp)->tv_sec--;                                \
                        (vvp)->tv_usec += 1000000;                      \
                }                                                       \
        } while (0)
#endif

#endif /* _NTRACE_H_ */
