#ifndef _NTRVAR_H_
#define _NTRVAR_H_

#if	defined(_KERNEL) && defined(__NetBSD__)
#define KERNEL
#endif


struct ether_header;
struct NtrTrack;

typedef struct ntrace_d {
	/* initialized during boot to be used later to create tracks */
	struct ifnet *ifp;	/* interface data */
	int hlen;		/* length of the link-level header */
	u_int dev_type;		/* device type magic number */
	struct tr_plist_ent_t *dev_plist; /* device's plist, to create track */
	struct timeval period;	/* time between consecutive device reading */

	/* following fields are initialized for each trace */
	struct timeval timestamp; /* when the device was last read */
	u_long ipaddr;		  /* IP-address of the interface */

	struct NtrTrack *dev_tp;  /* track data for device*/
	struct NtrTrack *icmp_tp; /* track data for ICMP */
	struct NtrTrack *tcp_tp;  /* track data for TCP */
	struct NtrTrack *udp_tp;  /* track data for UDP */
	struct NtrTrack *usr_tp;  /* track data for user specific UDP */

	struct ntrace_d *next;	/* next ntrace device on the list */
} ntrace_d;


#ifdef	KERNEL

/* 
 * Some declarations to keep everyone honest
 */
int ntrace_open  __P((dev_t, int, int, struct proc *));
int ntrace_close __P((dev_t, int, int, struct proc *));
int ntrace_read  __P((dev_t, struct uio *, int));
ntrace_d* ntrace_dev_attach __P((ntrace_d **, struct tr_plist_ent_t*,
				 int, int, struct ifnet*, int));
int ntrace_dev_ready __P((struct ntrace_d *));
int ntrace_dev	__P((struct ntrace_d *, char *));
int ntrace_pkt	__P((struct ntrace_d *, struct mbuf *, struct ether_header *));

#define Malloc(sz)	malloc(sz, M_DEVBUF, M_WAIT)
#define Mfree(ptr)	free(ptr, M_DEVBUF)

#else	/* not kernel */
#define Malloc(sz)	malloc(sz)
#define MFree(ptr)	free(ptr)
#define microtime(tv)	gettimeofday(tv, 0)

#define ip_print(p1, l)	ip_trace(p1, l)
#define icmp_print(p1, p2)	icmp_trace(p1, p2)
#define tcp_print(p1, l, p2)	tcp_trace(p1, l, p2)
#define udp_print(p1, l, p2)	udp_trace(p1, l, p2)
#endif


#define NANOTIME(tv) \
{ \
	microtime((struct timeval *) tv); \
	((struct timeval *) tv) ->tv_usec *= 1000; \
	TR_TIME_HTON(tv); \
}

#endif	/* _NTRVAR_H_ */
