/*-
 * Copyright (c) 1995 Berkeley Software Design, Inc. All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI raw_usrreq.c,v 2.5 1996/06/11 19:14:33 jch Exp
 */


/*
 * Copyright (c) 1980, 1986, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)raw_usrreq.c	8.1 (Berkeley) 6/10/93
 */

#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/domain.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/socketvar.h>
#include <sys/errno.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <net/netisr.h>
#include <net/raw_cb.h>

/*
 * Initialize raw connection block q.
 */
void
raw_init()
{

	rawcb.rcb_next = rawcb.rcb_prev = &rawcb;
}


/*
 * Raw protocol input routine.  Find the socket
 * associated with the packet(s) and move them over.  If
 * nothing exists for this packet, drop it.
 */
/*
 * Raw protocol interface.
 */
void
raw_input(m, c, proto, src, dst)
	register struct mbuf *m;
	register struct mbuf *c;
	register struct sockproto *proto;
	struct sockaddr *src, *dst;
{
	register struct rawcb *rp;
	register int sockets = 0;
	struct socket *last;

	last = 0;
	for (rp = rawcb.rcb_next; rp != &rawcb; rp = rp->rcb_next) {
		if (rp->rcb_proto.sp_family != proto->sp_family)
			continue;
		if (rp->rcb_proto.sp_protocol  &&
		    rp->rcb_proto.sp_protocol != proto->sp_protocol)
			continue;
		/*
		 * We assume the lower level routines have
		 * placed the address in a canonical format
		 * suitable for a structure comparison.
		 *
		 * Note that if the lengths are not the same
		 * the comparison will fail at the first byte.
		 */
#define	equal(a1, a2) \
  (bcmp((caddr_t)(a1), (caddr_t)(a2), a1->sa_len) == 0)
		if (rp->rcb_laddr && !equal(rp->rcb_laddr, dst))
			continue;
		if (rp->rcb_faddr && !equal(rp->rcb_faddr, src))
			continue;
		if (last) {
			struct mbuf *n = 0;
			struct mbuf *d = 0;
			if (m && (n = m_copy(m, 0, (int)M_COPYALL)) == NULL ||
			    c && (d = m_copy(c, 0, (int)M_COPYALL)) == NULL) {
					printf("dropped raw link packet\n");
					m_freem(n);
					m_freem(d);
			} else if (sbappendaddr(&last->so_rcv, src, n, d) == 0){
				m_freem(n);
				m_freem(d);
			} else {
				sorwakeup(last);
				sockets++;
			}
		}
		last = rp->rcb_socket;
	}
	if (last) {
		if (sbappendaddr(&last->so_rcv, src, m, c) == 0) {
			m_freem(m);
			m_freem(c);
		} else {
			sorwakeup(last);
			sockets++;
		}
	} else {
		if (m)
			m->m_pkthdr.rcvif->if_noproto++;
		m_freem(m);
		m_freem(c);
	}
}

/*ARGSUSED*/
void
raw_ctlinput(cmd, arg)
	int cmd;
	struct sockaddr *arg;
{

	if (cmd < 0 || cmd > PRC_NCMDS)
		return;
	/* INCOMPLETE */
}

/*ARGSUSED*/
int
raw_usrreq(so, req, m, nam, control)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *control;
{
	register struct rawcb *rp = sotorawcb(so);
	register int error = 0;
	int len;

	/*
	 * For the moment, we use raw network sockets (e.g. raw IP)
	 * for link-layer address/type changes; this will move when
	 * PF_LINK become real.
	 */
	if (req == PRU_CONTROL)
		return (link_control(so, (int)m, (caddr_t)nam,
		    (struct ifnet *)control));		/* XXX temporary */
	if (control && control->m_len) {
		m_freem(control);
		if (m && req != PRU_SENSE && req != PRU_RCVOOB)
			m_freem(m);
		return (EOPNOTSUPP);
	}
	if (rp == 0 && req != PRU_ATTACH) {
		if (m && req != PRU_SENSE && req != PRU_RCVOOB)
			m_freem(m);
		return (EINVAL);
	}
	switch (req) {

	/*
	 * Allocate a raw control block and fill in the
	 * necessary info to allow packets to be routed to
	 * the appropriate raw interface routine.
	 */
	case PRU_ATTACH:
		if ((so->so_state & SS_PRIV) == 0) {
			error = EACCES;
			break;
		}
		error = raw_attach(so, (int)nam);
		break;

	/*
	 * Destroy state just before socket deallocation.
	 * Flush data or not depending on the options.
	 */
	case PRU_DETACH:
		if (rp == 0) {
			error = ENOTCONN;
			break;
		}
		raw_detach(rp);
		break;

	/*
	 * If a socket isn't bound to a single address,
	 * the raw input routine will hand it anything
	 * within that protocol family (assuming there's
	 * nothing else around it should go to). 
	 */
	case PRU_CONNECT:
		if (rp->rcb_faddr)
			raw_disconnect(rp);
		nam = m_copym(nam, 0, M_COPYALL, M_WAIT);
		rp->rcb_faddr = mtod(nam, struct sockaddr *);
		soisconnected(so);
		break;

	case PRU_BIND:
		if (rp->rcb_laddr) {
			error = EINVAL;			/* XXX */
			break;
		}
		error = raw_bind(so, nam);
		break;

	case PRU_CONNECT2:
		error = EOPNOTSUPP;
		break;

	case PRU_DISCONNECT:
		if (rp->rcb_faddr == 0) {
			error = ENOTCONN;
			break;
		}
		raw_disconnect(rp);
		soisdisconnected(so);
		break;

	/*
	 * Mark the connection as being incapable of further input.
	 */
	case PRU_SHUTDOWN:
		socantsendmore(so);
		break;

	/*
	 * Ship a packet out.  The appropriate raw output
	 * routine handles any massaging necessary.
	 */
	case PRU_SEND:
		if (nam) {
			if (rp->rcb_faddr) {
				error = EISCONN;
				break;
			}
			rp->rcb_faddr = mtod(nam, struct sockaddr *);
		}
#ifdef notdef
		else if (rp->rcb_faddr == 0) {
			error = ENOTCONN;
			break;
		}
#endif
		error = (*so->so_proto->pr_output)(m, so);
		m = NULL;
		if (nam)
			rp->rcb_faddr = 0;
		break;

	case PRU_ABORT:
		/* this looks broken */
		raw_disconnect(rp);
		soisdisconnected(so);
		sofree(so);
		break;

	case PRU_SENSE:
		/*
		 * stat: don't bother with a blocksize.
		 */
		return (0);

	/*
	 * Not supported.
	 */
	case PRU_RCVOOB:
	case PRU_RCVD:
		return(EOPNOTSUPP);

	case PRU_LISTEN:
	case PRU_ACCEPT:
	case PRU_SENDOOB:
		error = EOPNOTSUPP;
		break;

	case PRU_SOCKADDR:
		if (rp->rcb_laddr == 0) {
			error = EINVAL;
			break;
		}
		len = rp->rcb_laddr->sa_len;
		bcopy((caddr_t)rp->rcb_laddr, mtod(nam, caddr_t), (unsigned)len);
		nam->m_len = len;
		break;

	case PRU_PEERADDR:
		if (rp->rcb_faddr == 0) {
			error = ENOTCONN;
			break;
		}
		len = rp->rcb_faddr->sa_len;
		bcopy((caddr_t)rp->rcb_faddr, mtod(nam, caddr_t), (unsigned)len);
		nam->m_len = len;
		break;

	default:
		panic("raw_usrreq");
	}
	if (m != NULL)
		m_freem(m);
	return (error);
}

/*
 * This routine should be in if_link.c with support for PF_LINK sockets,
 * and will be before too long...
 */
/* ARGSUSED */
link_control(so, cmd, data, ifp)
	struct socket *so;
	int cmd;
	caddr_t data;
	register struct ifnet *ifp;
{
	struct ifaddr *ifa;
	struct sockaddr_dl *sdl;
	struct ifreq *ifr = (struct ifreq *)data;
	int error;
	int type;

	switch (cmd) {

	case SIOCSIFADDR:
		/*
		 * We only pick off the requests that are trying to set
		 * the link type.  These are then forwarded off to the
		 * device, if possible.
		 */
		if (ifr->ifr_addr.sa_family != AF_LINK)
			return (EAFNOSUPPORT);

		for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next)
			if ((sdl = (struct sockaddr_dl *)ifa->ifa_addr) &&
			    sdl->sdl_family == AF_LINK)
				break;
		if (ifa == NULL || ifp->if_ioctl == NULL)
			return (ENOPROTOOPT);
		
		type = sdl->sdl_type;
		sdl->sdl_type =
		    ((struct sockaddr_dl *)&(ifr->ifr_addr))->sdl_type;

		/*
		 * Internal-use version of SIOCSIFADDR for AF_LINK only;
		 * used to prevent type from changing accidentally
		 * if driver has default case for setting protocol addr
		 * with SIOCSIFADDR.
		 */
		if (error = (*ifp->if_ioctl)(ifp, _SIOCSLINK, (caddr_t)ifa))
			sdl->sdl_type = type;

		return (error);

	default:
		return (EOPNOTSUPP);
	}
	/* NOTREACHED */
}
