/*	BSDI route.c,v 2.12.2.1 1997/03/14 22:08:13 jch Exp	*/

/*
 * Copyright (c) 1980, 1986, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)route.c	8.4 (Berkeley) 4/4/96
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/domain.h>
#include <sys/protosw.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/route.h>
#include <net/raw_cb.h>

#define	SA(p) ((struct sockaddr *)(p))

int	rttrash;		/* routes not in table but not freed */
struct	sockaddr wildcard;	/* zero valued cookie for wildcard searches */
int	rt_netclone = 1;	/* Clone network routes for local use */
int	rtt_idletimeout = 10*60;/* 10 minute timeout for idle cloned routes */
int	rt_timer_interval = 5;	/* Run timers every 5 seconds */

CIRCLEQ_HEAD(rt_timers, rttimer) rt_timers;

void
rtable_init(table)
	void **table;
{
	struct domain *dom;
	for (dom = domains; dom; dom = dom->dom_next)
		if (dom->dom_rtattach)
			dom->dom_rtattach(&table[dom->dom_family],
			    dom->dom_rtoffset);
}

void
route_init()
{
	rn_init();	/* initialize all zeroes, all ones, mask table */
	rtable_init((void **)rt_tables);
	rt_timer_init();
}

/*
 * Packet routing routines.
 */
void
rtalloc(ro)
	register struct route *ro;
{
	if (ro->ro_rt && ro->ro_rt->rt_ifp && (ro->ro_rt->rt_flags & RTF_UP))
		return;				 /* XXX */
	ro->ro_rt = rtalloc1(&ro->ro_dst, 1);
}

/*
 * rtcalloc() is just like rtalloc(), except that if the route
 * that gets resolved is a network route, it will clone a host
 * level route (if rt_netclone is set...)
 */

void
rtcalloc(ro)
	register struct route *ro;
{
	if (ro->ro_rt && ro->ro_rt->rt_ifp && (ro->ro_rt->rt_flags & RTF_UP))
		return;				 /* XXX */
	ro->ro_rt = rtalloc1(&ro->ro_dst, rt_netclone ? RTF_CLONING : 1);
}

struct rtentry *
rtalloc1(dst, report)
	register struct sockaddr *dst;
	int report;
{
	register struct radix_node_head *rnh = rt_tables[dst->sa_family];
	register struct rtentry *rt;
	register struct radix_node *rn;
	struct rtentry *newrt = 0;
	struct rt_addrinfo info;
	int  s = splnet(), err = 0, msgtype = RTM_MISS;

	if (rnh && (rn = rnh->rnh_matchaddr((caddr_t)dst, rnh)) &&
	    ((rn->rn_flags & RNF_ROOT) == 0)) {
		newrt = rt = (struct rtentry *)rn;
		if (report &&
		    (((report|rt->rt_flags) & (RTF_CLONING|RTF_HOST)) ==
		    RTF_CLONING)) {
			err = rtrequest(RTM_RESOLVE, dst, SA(0),
					      SA(0), 0, &newrt);
			if (err) {
				newrt = rt;
				rt->rt_refcnt++;
				goto miss;
			}
			if ((rt = newrt) && (rt->rt_flags & RTF_XRESOLVE)) {
				msgtype = RTM_RESOLVE;
				goto miss;
			}
		} else
			rt->rt_refcnt++;
	} else {
		rtstat.rts_unreach++;
	miss:	if (report) {
			bzero((caddr_t)&info, sizeof(info));
			info.rti_info[RTAX_DST] = dst;
			rt_missmsg(msgtype, &info, 0, err);
		}
	}
	splx(s);
	return (newrt);
}

void
rtfree(rt)
	register struct rtentry *rt;
{
	register struct ifaddr *ifa;

	if (rt == 0)
		panic("rtfree");
	rt->rt_refcnt--;
	if (rt->rt_refcnt > 0)
		return;
	if ((rt->rt_flags & RTF_UP) == 0) {
		if (rt->rt_nodes->rn_flags & (RNF_ACTIVE | RNF_ROOT))
			panic ("rtfree 2");
		rttrash--;
		if (rt->rt_refcnt < 0) {
			printf("rtfree: %x not freed (neg refs)\n", rt);
			return;
		}
		ifa = rt->rt_ifa;
		IFAFREE(ifa);
		if (LIST_FIRST(&rt->rt_timer) != LIST_END(&rt->rt_timer))
			rt_timer_remove_all(rt);
		Free(rt_key(rt));
		Free(rt);
		return;
	}

	/* Turn on inactivity timer for cloned routes */
	if (rtt_idletimeout && (rt->rt_flags &
	    (RTF_MODIFIED|RTF_DYNAMIC|RTF_STATIC|RTF_LLINFO|RTF_CLONED)) ==
	    RTF_CLONED)
		rt_timer_add(rt, time.tv_sec + rtt_idletimeout, NULL, NULL);
}

void
ifafree(ifa)
	register struct ifaddr *ifa;
{
	if (ifa == NULL)
		panic("ifafree");
	if (ifa->ifa_refcnt == 0)
		free(ifa, M_IFADDR);
	else
		ifa->ifa_refcnt--;
}

void
rtredirtimeo(rt, rtt)
	struct rtentry *rt;
	struct rttimer *rtt;
{
	struct rt_addrinfo info;

	bzero((caddr_t)&info, sizeof(info));
	info.rti_info[RTAX_DST] = rt_key(rt);
	info.rti_info[RTAX_GATEWAY] = rt->rt_gateway;
	info.rti_info[RTAX_NETMASK] = rt_mask(rt);
	rt_missmsg(RTM_REDIRECT, &info, rt->rt_flags, 0);

	(void) rtrequest(RTM_DELETE, rt_key(rt), 0, 0, 0, NULL);
	Free(rtt);
}

/*
 * Force a routing table entry to the specified
 * destination to go through the given gateway.
 * Normally called as a result of a routing redirect
 * message from the network layer.
 *
 * N.B.: must be called at splnet
 *
 */
int
rtredirect(dst, gateway, netmask, flags, src, timeo)
	struct sockaddr *dst, *gateway, *netmask, *src;
	int flags;
	int timeo;
{
	register struct rtentry *rt;
	int error = 0;
	u_quad_t *stat = 0;
	struct rt_addrinfo info;
	struct ifaddr *ifa;

	/* verify the gateway is directly reachable */
	if ((ifa = ifa_ifwithnet(gateway)) == 0) {
		error = ENETUNREACH;
		goto out;
	}
	rt = rtalloc1(dst, 0);
	/*
	 * If the redirect isn't from our current router for this dst,
	 * it's either old or wrong.  If it redirects us to ourselves,
	 * we have a routing loop, perhaps as a result of an interface
	 * going down recently.
	 */
#define	equal(a1, a2) (bcmp((caddr_t)(a1), (caddr_t)(a2), (a1)->sa_len) == 0)
	if (!(flags & RTF_DONE) && rt &&
	     (!equal(src, rt->rt_gateway) || rt->rt_ifa != ifa))
		error = EINVAL;
	else if (ifa_ifwithaddr(gateway))
		error = EHOSTUNREACH;
	if (error)
		goto done;
	/*
	 * Create a new entry if we just got back a wildcard entry
	 * or the the lookup failed.  This is necessary for hosts
	 * which use routing redirects generated by smart gateways
	 * to dynamically build the routing tables.
	 */
	if ((rt == 0) || (rt_mask(rt) && rt_mask(rt)->sa_len < 2))
		goto create;
	/*
	 * Don't listen to the redirect if it's
	 * for a route to an interface. 
	 */
	if (rt->rt_flags & RTF_GATEWAY) {
		if (((rt->rt_flags & RTF_HOST) == 0) && (flags & RTF_HOST)) {
			struct rtentry *nrt;

			/*
			 * Changing from route to net => route to host.
			 * Create new route, rather than smashing route to net.
			 */
		create:
			if (rt) {
				rtfree(rt);
				rt = NULL;
			}
			flags |=  RTF_GATEWAY | RTF_DYNAMIC;
			error = rtrequest((int)RTM_ADD, dst, gateway,
				    netmask, flags, &nrt);
			rt = nrt;
			if (rt && timeo)
				rt_timer_add(rt, time.tv_sec + timeo,
				    rtredirtimeo, NULL);
			stat = &rtstat.rts_dynamic;
		} else {
			/*
			 * Smash the current notion of the gateway to
			 * this destination.  Should check about netmask!!!
			 */
			if (timeo && ((rt->rt_flags &
			    (RTF_STATIC|RTF_LLINFO|RTF_CLONED)) == RTF_CLONED))
				rt_timer_add(rt, time.tv_sec + timeo,
				    rtredirtimeo, NULL);
			rt->rt_flags |= RTF_MODIFIED;
			flags |= RTF_MODIFIED;
			stat = &rtstat.rts_newgateway;
			rt_setgate(rt, rt_key(rt), gateway);
		}
	} else
		error = EHOSTUNREACH;
done:
	if (rt)
		rtfree(rt);
out:
	if (error)
		rtstat.rts_badredirect++;
	else if (stat != NULL)
		(*stat)++;
	bzero((caddr_t)&info, sizeof(info));
	info.rti_info[RTAX_DST] = dst;
	info.rti_info[RTAX_GATEWAY] = gateway;
	info.rti_info[RTAX_NETMASK] = netmask;
	info.rti_info[RTAX_AUTHOR] = src;
	rt_missmsg(RTM_REDIRECT, &info, flags, error);
}

/*
* Routing table ioctl interface.
*/
int
rtioctl(req, data, p)
	u_long req;
	caddr_t data;
	struct proc *p;
{
	return (EOPNOTSUPP);
}

struct ifaddr *
ifa_ifwithroute(flags, dst, gateway)
	int flags;
	struct sockaddr	*dst, *gateway;
{
	register struct ifaddr *ifa;
	register struct ifnet *ifp;

	if ((flags & RTF_GATEWAY) == 0) {
		/*
		 * If we are adding a route to an interface,
		 * and the interface is a pt to pt link
		 * we should search for the destination
		 * as our clue to the interface.  Otherwise
		 * we can assume that the gateway is the
		 * address of a non-pt to pt link.
		 */
		ifa = 0;
		if (flags & RTF_HOST) 
			ifa = ifa_ifwithdstaddr(dst);
		if (ifa == 0) {
			for (ifp = ifnet; ifp; ifp = ifp->if_next)
			    if ((ifp->if_flags & IFF_POINTOPOINT) == 0)
				for (ifa = ifp->if_addrlist; ifa;
				    ifa = ifa->ifa_next)
					if (ifa->ifa_addr->sa_family ==
					    gateway->sa_family
					    && equal(gateway, ifa->ifa_addr))
					    goto found;
		}
	} else {
		/*
		 * If we are adding a route to a remote net
		 * or host, the gateway may still be on the
		 * other end of a pt to pt link.
		 */
		ifa = ifa_ifwithdstaddr(gateway);
	}
	if (ifa == 0)
		ifa = ifa_ifwithnet(gateway);
	if (ifa == 0) {
		struct rtentry *rt = rtalloc1(dst, 0);
		if (rt == 0)
			return (0);
		rt->rt_refcnt--;
		/* the gateway must be local... */
		if (rt->rt_flags & RTF_GATEWAY)
			return (0);
		if ((ifa = rt->rt_ifa) == 0)
			return (0);
	}
 found:
	if (ifa->ifa_addr->sa_family != dst->sa_family) {
		struct ifaddr *oifa = ifa;
		ifa = ifaof_ifpforaddr(dst, ifa->ifa_ifp);
		if (ifa == 0)
			ifa = oifa;
	}
	return (ifa);
}

int
rt_rmclone(rn, nrt)
	struct radix_node *rn;
	struct rtentry *nrt;
{
	struct rtentry *rt = (struct rtentry *)rn;
	struct sockaddr *sa;
	register unsigned char c;
	register int masklen;

	/*
	 * If this is a network route, then we want to skip it
	 * and everything under it.  We do that by returning the
	 * length of the mask so that the rn_walkbranch() can
	 * figure out what to skip!
	 */
	if ((rt->rt_flags & RTF_HOST) == 0) {
		if ((sa = rt_mask(rt)) == NULL)
			return (0);
		if (sa->sa_len == 0)
			return (0);

		masklen = sa->sa_len - 1;
		c = *(((unsigned char *)sa) + masklen);
		if (c != 0xff) {
			/* mask ends in the middle of a byte */
			for (masklen *= 8; c; c <<= 1)
				masklen++;
		} else
			masklen = (masklen + 1) * 8;
		return (masklen);
	}
	if (nrt == (struct rtentry *)(-1))
		return (-1);	/* special case... */

	/* We are only interested in cloned host routes! */
	if ((rt->rt_flags & RTF_CLONED) == 0)
		return (-1);

	/*
	 * If we are adding a route, then if the interface,
	 * address and the gateway are the same, leave it alone.
	 */
	if (nrt && nrt->rt_ifp == rt->rt_ifp &&
	    nrt->rt_ifa == rt->rt_ifa &&
	    nrt->rt_gwroute == rt->rt_gwroute)
		return (-1);

	(void) rtrequest(RTM_DELETE, rt_key(rt), 0, 0, 0, NULL);
	return (-1);
}

#define ROUNDUP(a) (a>0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))

int
rtrequest(req, dst, gateway, netmask, flags, ret_nrt)
	int req, flags;
	struct sockaddr *dst, *gateway, *netmask;
	struct rtentry **ret_nrt;
{
	int s = splnet(); int error = 0;
	register struct rtentry *rt, *crt;
	register struct radix_node *rn;
	register struct radix_node_head *rnh;
	struct ifaddr *ifa;
	struct sockaddr *ndst;
	struct rt_addrinfo info;
#define senderr(x) { error = x ; goto bad; }

	if ((rnh = rt_tables[dst->sa_family]) == 0)
		senderr(ESRCH);
	if (flags & RTF_HOST)
		netmask = 0;
	switch (req) {
	case RTM_DELETE:
		if (rnh->rnh_walkbranch) {
			/*
			 * We need to look up the route first and
			 * see if it's a network route.  If it is,
			 * remove up any cloned routes underneath it.
			 */
			if ((rn = rnh->rnh_lookup(dst, netmask, rnh)) == 0)
				senderr(ESRCH);
			rt = (struct rtentry *)rn;
			if ((rt->rt_flags & RTF_HOST) == 0) {
				/* Clean up any cloned routes.  */
				rnh->rnh_walkbranch(rn, rt_rmclone, 0);
			}
		}
		if ((rn = rnh->rnh_deladdr(dst, netmask, rnh)) == 0)
			senderr(ESRCH);
		if (rn->rn_flags & (RNF_ACTIVE | RNF_ROOT))
			panic ("rtrequest delete");
		rt = (struct rtentry *)rn;
		if (rt->rt_gwroute) {
			rt = rt->rt_gwroute; RTFREE(rt);
			(rt = (struct rtentry *)rn)->rt_gwroute = 0;
		}
		rt->rt_flags &= ~RTF_UP;
		if ((ifa = rt->rt_ifa) && ifa->ifa_rtrequest)
			ifa->ifa_rtrequest(RTM_DELETE, rt, SA(0));
		rttrash++;
		if (ret_nrt)
			*ret_nrt = rt;
		else if (rt->rt_refcnt <= 0) {
			rt->rt_refcnt++;
			rtfree(rt);
		}
		break;

	case RTM_RESOLVE:
		if (ret_nrt == 0 || (rt = *ret_nrt) == 0)
			senderr(EINVAL);
		ifa = rt->rt_ifa;
		flags = (rt->rt_flags | RTF_CLONED) & ~(RTF_CLONING|RTF_STATIC);
		gateway = rt->rt_gateway;
		if ((netmask = rt->rt_genmask) == 0)
			flags |= RTF_HOST;
		goto makeroute;

	case RTM_ADD:
		if ((ifa = ifa_ifwithroute(flags, dst, gateway)) == 0)
			senderr(ENETUNREACH);
	makeroute:
		R_Malloc(rt, struct rtentry *, sizeof(*rt));
		if (rt == 0)
			senderr(ENOBUFS);
		Bzero(rt, sizeof(*rt));
		rt->rt_flags = RTF_UP | flags;
		if (rt_setgate(rt, dst, gateway)) {
			Free(rt);
			senderr(ENOBUFS);
		}
		ndst = rt_key(rt);
		if (netmask) {
			rt_maskedcopy(dst, ndst, netmask);
		} else
			Bcopy(dst, ndst, dst->sa_len);
		rn = rnh->rnh_addaddr((caddr_t)ndst, (caddr_t)netmask,
					rnh, rt->rt_nodes);
		if (rn == NULL && (crt = rtalloc1(dst, 0))) {
			if (crt->rt_flags & (RTF_CLONED|RTF_DYNAMIC) &&
			    rtrequest(RTM_DELETE,
			        (struct sockaddr *)rt_key(crt),
				crt->rt_gateway,
				rt_mask(crt), crt->rt_flags, NULL) == 0) {
				if (crt->rt_flags & RTF_DYNAMIC) {
					bzero((caddr_t)&info, sizeof(info));
					info.rti_info[RTAX_DST] = rt_key(rt);
					info.rti_info[RTAX_GATEWAY] =
					    rt->rt_gateway;
					info.rti_info[RTAX_NETMASK] =
					    rt_mask(rt);
					rt_missmsg(RTM_DELETE, &info,
					    rt->rt_flags, 0);
				}
				rn = rnh->rnh_addaddr((caddr_t)ndst,
				    (caddr_t)netmask, rnh, rt->rt_nodes);
			}
			RTFREE(crt);
		}
		if (rn == NULL) {
			Free(rt_key(rt));
			Free(rt);
			senderr(EEXIST);
		}
		ifa->ifa_refcnt++;
		rt->rt_ifa = ifa;
		rt->rt_ifp = ifa->ifa_ifp;
		rt->rt_rmx.rmx_mtu = ifa->ifa_ifp->if_mtu;
		if (req == RTM_RESOLVE) {
			rt->rt_rmx = (*ret_nrt)->rt_rmx; /* copy metrics */
			rt->rt_rmx.rmx_expire = 0;	/* but not timer */
		}
		if (ifa->ifa_rtrequest)
			ifa->ifa_rtrequest(req, rt, SA(ret_nrt ? *ret_nrt : 0));
		if (ret_nrt) {
			*ret_nrt = rt;
			rt->rt_refcnt++;
		}

		if (rnh->rnh_walkbranch && (rt->rt_flags & RTF_HOST) == 0) {
			/* clean up any cloned routes */
			rnh->rnh_walkbranch(rn, rt_rmclone, rt);
		}
		break;
	}
bad:
	splx(s);
	return (error);
}

int
rt_setgate(rt0, dst, gate)
	struct rtentry *rt0;
	struct sockaddr *dst, *gate;
{
	caddr_t new, old;
	int dlen = ROUNDUP(dst->sa_len), glen = ROUNDUP(gate->sa_len);
	register struct rtentry *rt = rt0;

	if (rt->rt_gateway == 0 || glen > ROUNDUP(rt->rt_gateway->sa_len)) {
		old = (caddr_t)rt_key(rt);
		R_Malloc(new, caddr_t, dlen + glen);
		if (new == 0)
			return (1);
		rt->rt_nodes->rn_key = new;
	} else {
		new = rt->rt_nodes->rn_key;
		old = 0;
	}
	Bcopy(gate, (rt->rt_gateway = (struct sockaddr *)(new + dlen)), glen);
	if (old) {
		Bcopy(dst, new, dlen);
		Free(old);
	}
	if (rt->rt_gwroute) {
		rt0 = rt->rt_gwroute; rt->rt_gwroute = 0;
		rt = rt0; RTFREE(rt);
	}
	return (0);
}

void
rt_maskedcopy(src, dst, netmask)
	struct sockaddr *src, *dst, *netmask;
{
	register u_char *cp1 = (u_char *)src;
	register u_char *cp2 = (u_char *)dst;
	register u_char *cp3 = (u_char *)netmask;
	u_char *cplim = cp2 + *cp3;
	u_char *cplim2 = cp2 + *cp1;

	*cp2++ = *cp1++; *cp2++ = *cp1++; /* copies sa_len & sa_family */
	cp3 += 2;
	if (cplim > cplim2)
		cplim = cplim2;
	while (cp2 < cplim)
		*cp2++ = *cp1++ & *cp3++;
	if (cp2 < cplim2)
		bzero((caddr_t)cp2, (unsigned)(cplim2 - cp2));
}

/*
 * Set up a routing table entry, normally
 * for an interface.
 */
int
rtinit(ifa, cmd, flags)
	register struct ifaddr *ifa;
	int cmd, flags;
{
	register struct rtentry *rt;
	register struct sockaddr *dst;
	register struct sockaddr *deldst;
	struct mbuf *m = 0;
	struct rtentry *nrt = 0;
	int error;

	dst = flags & RTF_HOST ? ifa->ifa_dstaddr : ifa->ifa_addr;
	if (cmd == RTM_DELETE) {
		if ((flags & RTF_HOST) == 0 && ifa->ifa_netmask) {
			m = m_get(M_WAIT, MT_SONAME);
			deldst = mtod(m, struct sockaddr *);
			rt_maskedcopy(dst, deldst, ifa->ifa_netmask);
			dst = deldst;
		}
		if (rt = rtalloc1(dst, 0)) {
			rt->rt_refcnt--;
			if (rt->rt_ifa != ifa) {
				if (m)
					(void) m_free(m);
				return (flags & RTF_HOST ?
				    EHOSTUNREACH : ENETUNREACH);
			}
		}
	}
	error = rtrequest(cmd, dst, ifa->ifa_addr, ifa->ifa_netmask,
	    flags | ifa->ifa_flags, &nrt);
	if (m)
		(void) m_free(m);
	if (cmd == RTM_DELETE && error == 0 && (rt = nrt)) {
		rt_newaddrmsg(cmd, ifa, error, nrt);
		if (rt->rt_refcnt <= 0) {
			rt->rt_refcnt++;
			rtfree(rt);
		}
	}
	if (cmd == RTM_ADD && error == 0 && (rt = nrt)) {
		rt->rt_refcnt--;
		if (rt->rt_ifa != ifa) {
			if (rt->rt_ifa->ifa_rtrequest)
			    rt->rt_ifa->ifa_rtrequest(RTM_DELETE, rt, SA(0));
			IFAFREE(rt->rt_ifa);
			rt->rt_ifa = ifa;
			rt->rt_ifp = ifa->ifa_ifp;
			rt->rt_rmx.rmx_mtu = ifa->ifa_ifp->if_mtu;
			ifa->ifa_refcnt++;
			if (ifa->ifa_rtrequest)
			    ifa->ifa_rtrequest(RTM_ADD, rt, SA(0));
		}
		rt_newaddrmsg(cmd, ifa, error, nrt);
	}
	return (error);
}

/*
 * Route table timers.
 *
 * Routing table timers supports an arbitrary number of timers
 * for any routing table entry.  Each timer has an absolute time
 * of when the timer goes off, and a function to call (with the
 * rtentry as an argument) when the timer goes off.  If no function
 * was specified, it is assumed that the route is to be deleted.
 *
 * Each rtentry has a pointer to the timers associated with it,
 * which are linked together via the rtt_link field.  The expire
 * value in rt->rt_rmx.rmx_expire represents the first timer that
 * is scheduled to expire.  When the first timer goes off, the
 * rmx_expire field is updated with the value from the next timer.
 */

void
rt_timer_init()
{
	CIRCLEQ_INIT(&rt_timers);
	timeout(rt_timer_timer, NULL, 1);
}

/*
 * Remove all the timers associated with a particular
 * route, used when a route is deleted.
 */
void
rt_timer_remove_all(rt)
	struct rtentry *rt;
{
	struct rttimer *rtt;

	while ((rtt = LIST_FIRST(&rt->rt_timer)) != LIST_END(&rt->rt_timer)) {
		CIRCLEQ_REMOVE(&rt_timers, rtt, rtt_next);
		LIST_REMOVE(rtt, rtt_link);
	}
	rt->rt_rmx.rmx_expire = 0;
}

/*
 * Add/change a timer to call function "func" at time "timeout"
 * for route "rt".  If "expire_at" is zero, it means to cancel
 * the timer for this function.
 */
int
rt_timer_add(rt, expire_at, func, rtt)
	struct rtentry *rt;
	time_t expire_at;
	void (*func) __P((struct rtentry *, struct rttimer *));
	struct rttimer *rtt;
{
	struct rttimer *rtt2;

	/*
	 * If there is an existing timer for this function,
	 * then pull that timer out of the existing chains.
	 */
	for (rtt2 = LIST_FIRST(&rt->rt_timer); rtt2 != LIST_END(&rt->rt_timer);
	    rtt2 = LIST_NEXT(rtt2, rtt_link)) {
		if (rtt2->rtt_func != func)
			continue;
		LIST_REMOVE(rtt2, rtt_link);
		CIRCLEQ_REMOVE(&rt_timers, rtt2, rtt_next);
		rt->rt_rmx.rmx_expire =
		    (LIST_FIRST(&rt->rt_timer) == LIST_END(&rt->rt_timer)) ? 0 :
		    LIST_FIRST(&rt->rt_timer)->rtt_expire;
		if (rtt)
			Free(rtt);
		rtt = rtt2;
		break;
	}

	/*
	 * If we are cancelling the timer, then just free up the
	 * resources and return.
	 */
	if (expire_at == 0) {
		if (rtt)
			Free(rtt);
		return (0);
	}

	/*
	 * If we are aren't reusing an old timer, allocate the
	 * data structure and fill in the necessary info.
	 */
	if (rtt == NULL) {
		R_Malloc(rtt, struct rttimer *, sizeof(struct rttimer));
		if (rtt == NULL)
			return (ENOBUFS);

		rtt->rtt_rt = rt;
		rtt->rtt_func = func;
	}

	rtt->rtt_expire = expire_at;

	rtt2 = LIST_FIRST(&rt->rt_timer);
	if (rtt2 == LIST_END(&rt->rt_timer) || (expire_at < rtt2->rtt_expire)) {
		rt->rt_rmx.rmx_expire = rtt->rtt_expire;
		LIST_INSERT_HEAD(&rt->rt_timer, rtt, rtt_link);
	} else {
		while (LIST_NEXT(rtt2, rtt_link)) {
			if (LIST_NEXT(rtt2, rtt_link)->rtt_expire > expire_at)
				break;
			rtt2 = LIST_NEXT(rtt2, rtt_link);
		}
		LIST_INSERT_AFTER(rtt2, rtt, rtt_link);
	}

	if ((CIRCLEQ_FIRST(&rt_timers) == CIRCLEQ_END(&rt_timers)) ||
	    (rtt->rtt_expire < CIRCLEQ_FIRST(&rt_timers)->rtt_expire)) {
		CIRCLEQ_INSERT_HEAD(&rt_timers, rtt, rtt_next);
		return (0);
	}

	if (rtt->rtt_expire >= CIRCLEQ_LAST(&rt_timers)->rtt_expire) {
		CIRCLEQ_INSERT_TAIL(&rt_timers, rtt, rtt_next);
		return (0);
	}

	if (CIRCLEQ_LAST(&rt_timers)->rtt_expire - rtt->rtt_expire >
	    rtt->rtt_expire - CIRCLEQ_FIRST(&rt_timers)->rtt_expire) {
		for (rtt2 = CIRCLEQ_FIRST(&rt_timers);
		    rtt2 != CIRCLEQ_END(&rt_timers);
		    rtt2 = CIRCLEQ_NEXT(rtt2, rtt_next)) {
			if (rtt2->rtt_expire > rtt->rtt_expire)
				break;
		}
		CIRCLEQ_INSERT_BEFORE(&rt_timers, rtt2, rtt, rtt_next);
	} else {
		for (rtt2 = CIRCLEQ_LAST(&rt_timers);
		    rtt2 != CIRCLEQ_END(&rt_timers);
		    rtt2 = CIRCLEQ_PREV(rtt2, rtt_next)) {
			if (rtt2->rtt_expire <= rtt->rtt_expire)
				break;
		}
		CIRCLEQ_INSERT_AFTER(&rt_timers, rtt2, rtt, rtt_next);
	}
	return (0);
}

void
rt_timer_timer(unused)
	void *unused;
{
	struct rttimer *rtt;
	struct rtentry *rt;

	for (;;) {
		rtt = CIRCLEQ_FIRST(&rt_timers);
		if (rtt == CIRCLEQ_END(&rt_timers))
			break;
		if (rtt->rtt_expire > time.tv_sec)
			break;

		/*
		 * Remove the timer from the lists it is on,
		 * and reset the rt_rmx.rmx_expire value.
		 */
		CIRCLEQ_REMOVE(&rt_timers, rtt, rtt_next);
		LIST_REMOVE(rtt, rtt_link);
		rt = rtt->rtt_rt;
		rt->rt_rmx.rmx_expire =
		    (LIST_FIRST(&rt->rt_timer) == LIST_END(&rt->rt_timer)) ? 0 :
		    LIST_FIRST(&rt->rt_timer)->rtt_expire;

		if (rtt->rtt_func) {
			(*rtt->rtt_func)(rtt->rtt_rt, rtt);
		} else {
			if (rt->rt_refcnt <= 0) {
				(void) rtrequest(RTM_DELETE, rt_key(rt),
				    0, 0, 0, NULL);
			}
			Free(rtt);
		}
	}
	timeout(rt_timer_timer, NULL, hz*rt_timer_interval);
}
