/*-
 * Copyright (c) 1995, 1996 Berkeley Software Design, Inc.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *	BSDI slcompress.c,v 2.9 1996/11/13 22:30:05 prb Exp
 */

/*-
 * Copyright (c) 1989, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)slcompress.c	8.2 (Berkeley) 4/16/94
 */

/*
 * Routines to compress and uncompess tcp packets (for transmission
 * over low speed serial lines.
 *
 * Van Jacobson (van@helios.ee.lbl.gov), Dec 31, 1989:
 *	- Initial distribution.
 *
 * static char rcsid[] =
 * "Header: slcompress.c,v 1.19 89/12/31 08:52:59 van Exp";
 */

#include <sys/param.h>
#include <sys/mbuf.h>

#include <sys/socket.h>
#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include <net/slcompress.h>
#ifdef	PPP
#include <net/if_ppp.h>
#ifdef	MULTILINK
#include <net/if_pppioctl.h>
#include <net/if_p2p.h>
#endif
#endif

#ifndef SL_NO_STATS
#define INCR(counter) ++comp->counter;
#else
#define INCR(counter)
#endif

#define BCMP(p1, p2, n) bcmp((char *)(p1), (char *)(p2), (int)(n))
#define BCOPY(p1, p2, n) bcopy((char *)(p1), (char *)(p2), (int)(n))
#ifndef KERNEL
#define ovbcopy bcopy
#endif

#define dprintf(x)      if (vjc->vjc_ifp->if_flags & IFF_DEBUG) printf x

void
sl_compress_init(comp)
	struct slcompress *comp;
{
	register u_int i;
	register struct cstate *tstate = comp->tstate;

	i = comp->sls_ref;
	bzero((char *)comp, sizeof(*comp));
	comp->sls_ref = i;
	for (i = MAX_STATES - 1; i > 0; --i) {
		tstate[i].cs_id = i;
		tstate[i].cs_next = &tstate[i - 1];
	}
	tstate[0].cs_next = &tstate[MAX_STATES - 1];
	tstate[0].cs_id = 0;
	comp->last_cs = &tstate[0];
	comp->last_recv = 255;
	comp->last_xmit = 255;
	comp->flags = SLF_TOSS;
}


/* ENCODE encodes a number that is known to be non-zero.  ENCODEZ
 * checks for zero (since zero has to be encoded in the long, 3 byte
 * form).
 */
#define ENCODE(n) { \
	if ((u_short)(n) >= 256) { \
		*cp++ = 0; \
		cp[1] = (n); \
		cp[0] = (n) >> 8; \
		cp += 2; \
	} else { \
		*cp++ = (n); \
	} \
}
#define ENCODEZ(n) { \
	if ((u_short)(n) >= 256 || (u_short)(n) == 0) { \
		*cp++ = 0; \
		cp[1] = (n); \
		cp[0] = (n) >> 8; \
		cp += 2; \
	} else { \
		*cp++ = (n); \
	} \
}

#define DECODEL(f) { \
	if (*cp == 0) {\
		(f) = htonl(ntohl(f) + ((cp[1] << 8) | cp[2])); \
		cp += 3; \
	} else { \
		(f) = htonl(ntohl(f) + (u_long)*cp++); \
	} \
}

#define DECODES(f) { \
	if (*cp == 0) {\
		(f) = htons(ntohs(f) + ((cp[1] << 8) | cp[2])); \
		cp += 3; \
	} else { \
		(f) = htons(ntohs(f) + (u_long)*cp++); \
	} \
}

#define DECODEU(f) { \
	if (*cp == 0) {\
		(f) = htons((cp[1] << 8) | cp[2]); \
		cp += 3; \
	} else { \
		(f) = htons((u_long)*cp++); \
	} \
}

u_int
sl_compress_tcp(mp, ip, comp, compress_cid, cbuf)
	struct mbuf **mp;
	register struct ip *ip;
	struct slcompress *comp;
	int compress_cid;
	struct slcompress_buf *cbuf;
{
	register struct cstate *cs = comp->last_cs->cs_next;
	register u_int hlen = ip->ip_hl;
	register struct tcphdr *oth;
	register struct tcphdr *th;
	register u_int deltaS, deltaA;
	register u_int changes = 0;
	u_char new_seq[16];
	register u_char *cp = new_seq;
	struct mbuf *m = *mp;
	int type;

	/*
	 * Bail if this is an IP fragment or if the TCP packet isn't
	 * `compressible' (i.e., ACK isn't set or some other control bit is
	 * set).  (We assume that the caller has already made sure the
	 * packet is IP proto TCP).
	 */
	if ((ip->ip_off & htons(0x3fff)) || m->m_len < 40) {
		type = TYPE_IP;
		goto out;
	}

	th = (struct tcphdr *)&((int *)ip)[hlen];
	if ((th->th_flags & (TH_SYN|TH_FIN|TH_RST|TH_ACK)) != TH_ACK) {
		type = TYPE_IP;
		goto out;
	}

	/*
	 * Packet is compressible -- we're going to send either a
	 * COMPRESSED_TCP or UNCOMPRESSED_TCP packet.  Either way we need
	 * to locate (or create) the connection state.  Special case the
	 * most recently used connection since it's most likely to be used
	 * again & we don't have to do any reordering if it's used.
	 */
	INCR(sls_packets)
	if (ip->ip_src.s_addr != cs->cs_ip.ip_src.s_addr ||
	    ip->ip_dst.s_addr != cs->cs_ip.ip_dst.s_addr ||
	    *(int *)th != ((int *)&cs->cs_ip)[cs->cs_ip.ip_hl]) {
		/*
		 * Wasn't the first -- search for it.
		 *
		 * States are kept in a circularly linked list with
		 * last_cs pointing to the end of the list.  The
		 * list is kept in lru order by moving a state to the
		 * head of the list whenever it is referenced.  Since
		 * the list is short and, empirically, the connection
		 * we want is almost always near the front, we locate
		 * states via linear search.  If we don't find a state
		 * for the datagram, the oldest state is (re-)used.
		 */
		register struct cstate *lcs;
		register struct cstate *lastcs = comp->last_cs;

		do {
			lcs = cs; cs = cs->cs_next;
			INCR(sls_searches)
			if (ip->ip_src.s_addr == cs->cs_ip.ip_src.s_addr
			    && ip->ip_dst.s_addr == cs->cs_ip.ip_dst.s_addr
			    && *(int *)th == ((int *)&cs->cs_ip)[cs->cs_ip.ip_hl])
				goto found;
		} while (cs != lastcs);

		/*
		 * Didn't find it -- re-use oldest cstate.  Send an
		 * uncompressed packet that tells the other side what
		 * connection number we're using for this conversation.
		 * Note that since the state list is circular, the oldest
		 * state points to the newest and we only need to set
		 * last_cs to update the lru linkage.
		 */
		INCR(sls_misses)
		comp->last_cs = lcs;
		hlen += th->th_off;
		hlen <<= 2;
		goto uncompressed;

	found:
		/*
		 * Found it -- move to the front on the connection list.
		 */
		if (cs == lastcs)
			comp->last_cs = lcs;
		else {
			lcs->cs_next = cs->cs_next;
			cs->cs_next = lastcs->cs_next;
			lastcs->cs_next = cs;
		}
	}

	/*
	 * Make sure that only what we expect to change changed. The first
	 * line of the `if' checks the IP protocol version, header length &
	 * type of service.  The 2nd line checks the "Don't fragment" bit.
	 * The 3rd line checks the time-to-live and protocol (the protocol
	 * check is unnecessary but costless).  The 4th line checks the TCP
	 * header length.  The 5th line checks IP options, if any.  The 6th
	 * line checks TCP options, if any.  If any of these things are
	 * different between the previous & current datagram, we send the
	 * current datagram `uncompressed'.
	 */
	oth = (struct tcphdr *)&((int *)&cs->cs_ip)[hlen];
	deltaS = hlen;
	hlen += th->th_off;
	hlen <<= 2;

	if (((u_short *)ip)[0] != ((u_short *)&cs->cs_ip)[0] ||
	    ((u_short *)ip)[3] != ((u_short *)&cs->cs_ip)[3] ||
	    ((u_short *)ip)[4] != ((u_short *)&cs->cs_ip)[4] ||
	    th->th_off != oth->th_off ||
	    (deltaS > 5 &&
	     BCMP(ip + 1, &cs->cs_ip + 1, (deltaS - 5) << 2)) ||
	    (th->th_off > 5 &&
	     BCMP(th + 1, oth + 1, (th->th_off - 5) << 2)))
		goto uncompressed;

	/*
	 * Figure out which of the changing fields changed.  The
	 * receiver expects changes in the order: urgent, window,
	 * ack, seq (the order minimizes the number of temporaries
	 * needed in this section of code).
	 */
	if (th->th_flags & TH_URG) {
		deltaS = ntohs(th->th_urp);
		ENCODEZ(deltaS);
		changes |= NEW_U;
	} else if (th->th_urp != oth->th_urp)
		/* argh! URG not set but urp changed -- a sensible
		 * implementation should never do this but RFC793
		 * doesn't prohibit the change so we have to deal
		 * with it. */
		 goto uncompressed;

	if (deltaS = (u_short)(ntohs(th->th_win) - ntohs(oth->th_win))) {
		ENCODE(deltaS);
		changes |= NEW_W;
	}

	if (deltaA = ntohl(th->th_ack) - ntohl(oth->th_ack)) {
		if (deltaA > 0xffff)
			goto uncompressed;
		ENCODE(deltaA);
		changes |= NEW_A;
	}

	if (deltaS = ntohl(th->th_seq) - ntohl(oth->th_seq)) {
		if (deltaS > 0xffff)
			goto uncompressed;
		ENCODE(deltaS);
		changes |= NEW_S;
	}

	switch(changes) {

	case 0:
		/*
		 * Nothing changed. If this packet contains data and the
		 * last one didn't, this is probably a data packet following
		 * an ack (normal on an interactive connection) and we send
		 * it compressed.  Otherwise it's probably a retransmit,
		 * retransmitted ack or window probe.  Send it uncompressed
		 * in case the other side missed the compressed version.
		 */
		if (ip->ip_len != cs->cs_ip.ip_len &&
		    ntohs(cs->cs_ip.ip_len) == hlen)
			break;

		/* (fall through) */

	case SPECIAL_I:
	case SPECIAL_D:
		/*
		 * actual changes match one of our special case encodings --
		 * send packet uncompressed.
		 */
		goto uncompressed;

	case NEW_S|NEW_A:
		if (deltaS == deltaA &&
		    deltaS == ntohs(cs->cs_ip.ip_len) - hlen) {
			/* special case for echoed terminal traffic */
			changes = SPECIAL_I;
			cp = new_seq;
		}
		break;

	case NEW_S:
		if (deltaS == ntohs(cs->cs_ip.ip_len) - hlen) {
			/* special case for data xfer */
			changes = SPECIAL_D;
			cp = new_seq;
		}
		break;
	}

	deltaS = ntohs(ip->ip_id) - ntohs(cs->cs_ip.ip_id);
	if (deltaS != 1) {
		ENCODEZ(deltaS);
		changes |= NEW_I;
	}
	if (th->th_flags & TH_PUSH)
		changes |= TCP_PUSH_BIT;
	/*
	 * Grab the cksum before we overwrite it below.  Then update our
	 * state with this packet's header.
	 */
	deltaA = ntohs(th->th_sum);
	BCOPY(ip, &cs->cs_ip, hlen);

	/*
	 * We want to use the original packet as our compressed packet.
	 * (cp - new_seq) is the number of bytes we need for compressed
	 * sequence numbers.  In addition we need one byte for the change
	 * mask, one for the connection id and two for the tcp checksum.
	 * So, (cp - new_seq) + 4 bytes of header are needed.  hlen is how
	 * many bytes of the original packet to toss so subtract the two to
	 * get the new packet size.
	 */
	deltaS = cp - new_seq;

	cp = (u_char *)ip;
	if (compress_cid == SL_COMP_NOCID || comp->last_xmit != cs->cs_id) {
		compress_cid = SL_COMP_NOCID;
		comp->last_xmit = cs->cs_id;
		cbuf->chlen = deltaS + 4;
	} else
		cbuf->chlen = deltaS + 3;
	cbuf->uhlen = hlen;

	/*
	 * If we are doing BPF then we need to place the compressed
	 * header in a separate buffer (along with room for the link
	 * level header and BPF header)
	 */
	if (cbuf->bpflen) {
		M_PREPEND(m, SLC_BPFHDR, M_DONTWAIT);
		*mp = m;
		if (m == 0)
			return (0);
		cp = mtod(m, u_char *) + cbuf->lllen + cbuf->bpflen;
	} else {
		hlen -= cbuf->chlen + cbuf->lllen;
		m->m_len -= hlen;
		m->m_pkthdr.len -= hlen;
		m->m_data += hlen;
		cp = mtod(m, u_char *) + cbuf->lllen;
	}

	if (compress_cid == SL_COMP_NOCID) {
		*cp++ = changes | NEW_C;
		*cp++ = cs->cs_id;
	} else
		*cp++ = changes;

	*cp++ = deltaA >> 8;
	*cp++ = deltaA;
	BCOPY(new_seq, cp, deltaS);
	INCR(sls_compressed)
	return (TYPE_COMPRESSED_TCP);

	/*
	 * Update connection state cs & send uncompressed packet ('uncompressed'
	 * means a regular ip/tcp packet but with the 'conversation id' we hope
	 * to use on future compressed packets in the protocol field).
	 */
uncompressed:
	BCOPY(ip, &cs->cs_ip, hlen);
	if (cbuf->bpflen) {
		M_PREPEND(m, SLC_BPFHDR, M_DONTWAIT);
		*mp = m;
		if (m == 0)
			return(0);
		cp = mtod(m, u_char *) + cbuf->lllen + cbuf->bpflen;
		*cp++ = TYPE_UNCOMPRESSED_TCP;
		*cp = cs->cs_id;
		cbuf->chlen = 2;
		comp->last_xmit = cs->cs_id;
		return(TYPE_UNCOMPRESSED_TCP);
	}
	ip->ip_p = cs->cs_id;
	comp->last_xmit = cs->cs_id;
	type = TYPE_UNCOMPRESSED_TCP;
out:
	if (cbuf->bpflen)
		hlen = SLC_BPFHDR;
	else
		hlen = cbuf->lllen;
	if (hlen) {
		M_PREPEND(m, hlen, M_DONTWAIT);
		*mp = m;
		if (m == 0)
			return (0);
	}
	return (type);
}

#if 0	/* currently unused */
int
sl_uncompress_tcp(bufp, len, type, comp)
	u_char **bufp;
	int len;
	u_int type;
	struct slcompress *comp;
{
	register u_char *cp;
	register u_int hlen, changes;
	register struct tcphdr *th;
	register struct cstate *cs;
	register struct ip *ip;

	switch (type) {

	case TYPE_UNCOMPRESSED_TCP:
		ip = (struct ip *) *bufp;
		if (ip->ip_p >= MAX_STATES)
			goto bad;
		cs = &comp->rstate[comp->last_recv = ip->ip_p];
		comp->flags &=~ SLF_TOSS;
		ip->ip_p = IPPROTO_TCP;
		hlen = ip->ip_hl;
		hlen += ((struct tcphdr *)&((int *)ip)[hlen])->th_off;
		hlen <<= 2;
		BCOPY(ip, &cs->cs_ip, hlen);
		cs->cs_ip.ip_sum = 0;
		cs->cs_hlen = hlen;
		INCR(sls_uncompressedin)
		return (len);

	default:
		goto bad;

	case TYPE_COMPRESSED_TCP:
		break;
	}
	/* We've got a compressed packet. */
	INCR(sls_compressedin)
	cp = *bufp;
	changes = *cp++;
	if (changes & NEW_C) {
		/* Make sure the state index is in range, then grab the state.
		 * If we have a good state index, clear the 'discard' flag. */
		if (*cp >= MAX_STATES)
			goto bad;

		comp->flags &=~ SLF_TOSS;
		comp->last_recv = *cp++;
	} else {
		/* this packet has an implicit state index.  If we've
		 * had a line error since the last time we got an
		 * explicit state index, we have to toss the packet. */
		if (comp->flags & SLF_TOSS) {
			INCR(sls_tossed)
			return (0);
		}
	}
	cs = &comp->rstate[comp->last_recv];
	hlen = cs->cs_ip.ip_hl << 2;
	th = (struct tcphdr *)&((u_char *)&cs->cs_ip)[hlen];
	th->th_sum = htons((*cp << 8) | cp[1]);
	cp += 2;
	if (changes & TCP_PUSH_BIT)
		th->th_flags |= TH_PUSH;
	else
		th->th_flags &=~ TH_PUSH;

	switch (changes & SPECIALS_MASK) {
	case SPECIAL_I:
		{
		register u_int i = ntohs(cs->cs_ip.ip_len) - cs->cs_hlen;
		th->th_ack = htonl(ntohl(th->th_ack) + i);
		th->th_seq = htonl(ntohl(th->th_seq) + i);
		}
		break;

	case SPECIAL_D:
		th->th_seq = htonl(ntohl(th->th_seq) + ntohs(cs->cs_ip.ip_len)
				   - cs->cs_hlen);
		break;

	default:
		if (changes & NEW_U) {
			th->th_flags |= TH_URG;
			DECODEU(th->th_urp)
		} else
			th->th_flags &=~ TH_URG;
		if (changes & NEW_W)
			DECODES(th->th_win)
		if (changes & NEW_A)
			DECODEL(th->th_ack)
		if (changes & NEW_S)
			DECODEL(th->th_seq)
		break;
	}
	if (changes & NEW_I)
		DECODES(cs->cs_ip.ip_id)
	else
		cs->cs_ip.ip_id = htons(ntohs(cs->cs_ip.ip_id) + 1);

	/*
	 * At this point, cp points to the first byte of data in the
	 * packet.  If we're not aligned on a 4-byte boundary, copy the
	 * data down so the ip & tcp headers will be aligned.  Then back up
	 * cp by the tcp/ip header length to make room for the reconstructed
	 * header (we assume the packet we were handed has enough space to
	 * prepend 128 bytes of header).  Adjust the length to account for
	 * the new header & fill in the IP total length.
	 */
	len -= (cp - *bufp);
	if (len < 0)
		/* we must have dropped some characters (crc should detect
		 * this but the old slip framing won't) */
		goto bad;

	if ((int)cp & 3) {
		if (len > 0)
			(void) ovbcopy(cp, (caddr_t)((int)cp &~ 3), len);
		cp = (u_char *)((int)cp &~ 3);
	}
	cp -= cs->cs_hlen;
	len += cs->cs_hlen;
	cs->cs_ip.ip_len = htons(len);
	BCOPY(&cs->cs_ip, cp, cs->cs_hlen);
	*bufp = cp;

	/* recompute the ip header checksum */
	{
		register u_short *bp = (u_short *)cp;
		for (changes = 0; hlen > 0; hlen -= 2)
			changes += *bp++;
		changes = (changes & 0xffff) + (changes >> 16);
		changes = (changes & 0xffff) + (changes >> 16);
		((struct ip *)cp)->ip_sum = ~ changes;
	}
	return (len);
bad:
	comp->flags |= SLF_TOSS;
	INCR(sls_errorin)
	return (0);
}
#endif

/*
 * The same as sl_compress_tcp but operates on
 * the chain of mbufs instead of continuous buffer.
 * Returns 0 on success.
 * Certain presumptions are made:
 *
 *  The first mbuf contains only the link level header and compressed header
 *
 *  The second mbuf can have a full uncompressed header prepended to it
 *  without allocating a new mbuf.  If there is no second mbuf, we allocate it.
 */
sl_muncompress_tcp(mp, type, comp, llen)
	struct mbuf **mp;
	u_int type;
	struct slcompress *comp;
	u_int llen;
{
	register u_char *cp, *cq;
	register u_int hlen, changes;
	register struct tcphdr *th;
	register struct cstate *cs;
	register struct ip *ip;
	struct mbuf *m = *mp;
	struct mbuf *m1 = m->m_next;
	int len;

	switch (type) {

	case TYPE_UNCOMPRESSED_TCP:
		if (m1 == 0)
			goto bad;
		if ((m->m_pkthdr.len - m->m_len) < sizeof(struct ip))
			goto bad;
		ip = mtod(m1, struct ip *);
		if (ip->ip_p >= MAX_STATES)
			goto bad;

		cp = mtod(m, u_char *);
		cp += llen;
		*cp++ = *mtod(m1, u_char *) | TYPE_UNCOMPRESSED_TCP;
		*cp = ip->ip_p;
		m->m_len += 2;
		m->m_pkthdr.len += 2;

		cs = &comp->rstate[comp->last_recv = ip->ip_p];
		comp->flags &=~ SLF_TOSS;
		ip->ip_p = IPPROTO_TCP;
		hlen = ip->ip_hl << 2;
		if (hlen > (m->m_pkthdr.len - m->m_len))
			goto bad;
		hlen += ((struct tcphdr *)&((u_char *)ip)[hlen])->th_off << 2;
		if (hlen > MAX_HDR || hlen > (m->m_pkthdr.len - m->m_len))
			goto bad;
		m_copydata(m1, 0, hlen, &cs->cs_ip);
		cs->cs_ip.ip_sum = 0;
		cs->cs_hlen = hlen;
		INCR(sls_uncompressedin)
		return (0);

	default:
		goto bad;

	case TYPE_COMPRESSED_TCP:
		break;
	}
	/* We've got a compressed packet. */
	INCR(sls_compressedin)
	cp = mtod(m, u_char *) + llen;
	changes = *cp++;
	if (changes & NEW_C) {
		/* Make sure the state index is in range, then grab the state.
		 * If we have a good state index, clear the 'discard' flag. */
		if (*cp >= MAX_STATES)
			goto bad;

		comp->flags &=~ SLF_TOSS;
		comp->last_recv = *cp++;
	} else {
		/* this packet has an implicit state index.  If we've
		 * had a line error since the last time we got an
		 * explicit state index, we have to toss the packet. */
		if (comp->flags & SLF_TOSS) {
			INCR(sls_tossed)
			return (EINVAL);
		}
	}
	cs = &comp->rstate[comp->last_recv];
	hlen = cs->cs_ip.ip_hl << 2;
	th = (struct tcphdr *)&((u_char *)&cs->cs_ip)[hlen];
	th->th_sum = htons((*cp << 8) | cp[1]);
	cp += 2;
	if (changes & TCP_PUSH_BIT)
		th->th_flags |= TH_PUSH;
	else
		th->th_flags &=~ TH_PUSH;

	switch (changes & SPECIALS_MASK) {
	case SPECIAL_I:
		{
		register u_int i = ntohs(cs->cs_ip.ip_len) - cs->cs_hlen;
		th->th_ack = htonl(ntohl(th->th_ack) + i);
		th->th_seq = htonl(ntohl(th->th_seq) + i);
		}
		break;

	case SPECIAL_D:
		th->th_seq = htonl(ntohl(th->th_seq) + ntohs(cs->cs_ip.ip_len)
				   - cs->cs_hlen);
		break;

	default:
		if (changes & NEW_U) {
			th->th_flags |= TH_URG;
			DECODEU(th->th_urp)
		} else
			th->th_flags &=~ TH_URG;
		if (changes & NEW_W)
			DECODES(th->th_win)
		if (changes & NEW_A)
			DECODEL(th->th_ack)
		if (changes & NEW_S)
			DECODEL(th->th_seq)
		break;
	}
	if (changes & NEW_I)
		DECODES(cs->cs_ip.ip_id)
	else
		cs->cs_ip.ip_id = htons(ntohs(cs->cs_ip.ip_id) + 1);

	/*
	 * The first mbuf only contains link level and compressed header
	 * information.  Skip it.
	 * If we do not have a second mbuf, then allocate one to place
	 * the uncompressed header into.
	 */
	if (m1 == 0) {
		MGETHDR(m1, M_DONTWAIT, MT_DATA)
		if (m1 == NULL)
			goto bad;
		m1->m_pkthdr.rcvif = m->m_pkthdr.rcvif;
		m->m_next = m1;
		if (cs->cs_hlen > MHLEN) {
			MCLGET(m1, M_DONTWAIT);
			if ((m1->m_flags & M_EXT) == 0)
				goto bad;
		}
		m1->m_len = 0;
		m1->m_data += cs->cs_hlen;
	}
	/*
	 * Carve out room for the header at the start of the second
	 * mbuf, this one has enough room for it.
	 */
	m1->m_len += cs->cs_hlen;
	m1->m_data -= cs->cs_hlen;
	m->m_pkthdr.len += cs->cs_hlen;
	cs->cs_ip.ip_len = htons(m->m_pkthdr.len - m->m_len);
	BCOPY(&cs->cs_ip, mtod(m1, caddr_t), cs->cs_hlen);

	/* recompute the ip header checksum */
	{
		register u_short *bp = mtod(m1, u_short *);
		for (changes = 0; hlen > 0; hlen -= 2)
			changes += *bp++;
		changes = (changes & 0xffff) + (changes >> 16);
		changes = (changes & 0xffff) + (changes >> 16);
		mtod(m1, struct ip *)->ip_sum = ~ changes;
	}
	return (0);
bad:
	comp->flags |= SLF_TOSS;
	INCR(sls_errorin)
	return (EINVAL);
}

int
vjc_hdr(vjc, cp, nbytes)
	vjc_t *vjc;
	u_char *cp;
	int nbytes;
{
	struct mbuf *m;
	struct mbuf *mh;


	/*
	 * The first mbuf in the chain will leave 3 bytes at the start
	 * for a potential BPF header.  The filled in data will be exactly
	 * the link level header (probably 1 byte) plus any compressed IP
	 * header.  This assures the data in compressed IP packets will be
	 * aligned (as they will start in the next mbuf) and the IP header
	 * and data will be aligned in all other packets.
	 */
	mh = m = vjc->vjc_mbuf;

	if (nbytes == -1) {
		if (m->m_next)
			m_freem(m->m_next);
		m->m_next = 0;
		m->m_len = 0;
		vjc->vjc_hstate = vjc->vjc_basement;
		return (0);
	}

	if (m->m_len == 0 && m->m_next == 0) {
		m->m_data = m->m_pktdat;
		m->m_data += SLC_BPFHDRLEN;	/* Room for BPF header */
		m->m_pkthdr.len = 0;
		m->m_pkthdr.rcvif = vjc->vjc_ifp;
		vjc->vjc_hstate = vjc->vjc_basement;
		vjc->vjc_type = 0;
		vjc->vjc_space = 0;
		vjc->vjc_ibuf = m;
	} else
		m = vjc->vjc_ibuf;

	while (nbytes-- > 0) {
		register u_char c = *cp++;

		switch (vjc->vjc_hstate) {
		case PH_SLINK:
			if ((c & 0xf0) == (IPVERSION << 4)) {
				vjc->vjc_hstate = PH_DATA;
				vjc->vjc_type = TYPE_IP;
				goto LPH_DATA;
			}
			if (c & 0x80) {
				vjc->vjc_type = TYPE_COMPRESSED_TCP;
				vjc->vjc_hstate = PH_HDR;
				goto LPH_HDR;
			} else if ((c & 0xf0) == TYPE_UNCOMPRESSED_TCP) {
				c &= 0x4f; /* XXX */
				vjc->vjc_hstate = PH_DATA;
				vjc->vjc_type = TYPE_UNCOMPRESSED_TCP;
				goto LPH_DATA;
			}
			dprintf(("%s%d: Invalid packet: %x\n",
			    vjc->vjc_ifp->if_name,
			    vjc->vjc_ifp->if_unit, c));
			m = vjc->vjc_mbuf;
			m->m_len = 0;
			if (m->m_next) {
				m_freem(m->m_next);
				m->m_next = 0;
			}
			return (-1);
#ifdef	PPP
		case PH_LINK:
			if (c == PPP_ADDRESS) {
				m->m_data[m->m_len++] = c;
				mh->m_pkthdr.len++;
				vjc->vjc_hstate = PH_LINK1;
			} else if ((vjc->vjc_type = c) & 0x01) {
				m->m_data[m->m_len++] = PPP_ADDRESS;
				m->m_data[m->m_len++] = PPP_CONTROL;
				m->m_data[m->m_len++] = 0;
				m->m_data[m->m_len++] = c;
				mh->m_pkthdr.len += 4;
				goto LL_END;
			} else {
				m->m_data[m->m_len++] = PPP_ADDRESS;
				m->m_data[m->m_len++] = PPP_CONTROL;
				m->m_data[m->m_len++] = c;
				mh->m_pkthdr.len += 3;
				vjc->vjc_hstate = PH_TYPE1;
			}
			break;
		case PH_LINK1:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			if (c != PPP_CONTROL) {
				dprintf(("%s%d: Invalid field: %x %x\n",
				    vjc->vjc_ifp->if_name,
				    vjc->vjc_ifp->if_unit,
				    m->m_data[0] & 0xff,
				    m->m_data[1] & 0xff));
				m = vjc->vjc_mbuf;
				m->m_len = 0;
				if (m->m_next) {
					m_freem(m->m_next);
					m->m_next = 0;
				}
				return (-1);
			}
			vjc->vjc_hstate = PH_TYPE;
			break;
		case PH_TYPE:
			vjc->vjc_type = c;
			if (c & 0x01) {
				m->m_data[m->m_len++] = 0;
				m->m_data[m->m_len++] = c;
				mh->m_pkthdr.len += 2;
				goto LL_END;
			}
			vjc->vjc_hstate = PH_TYPE1;
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			break;
		case PH_TYPE1:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_type <<= 8;
			vjc->vjc_type |= c;
		LL_END:
			vjc->vjc_llen = m->m_len;
			vjc->vjc_hstate = (vjc->vjc_type == PPP_VJC) ? PH_HDR
#ifdef	MULTILINK
					  : (vjc->vjc_type == PPP_MP) ? PH_MP
#endif	/* MULTILINK */
								     : PH_DATA;
			break;
#ifdef	MULTILINK
		/*
		 * The Multi Link header is either 2 or 4 bytes long,
		 * depending on whether we are using short or long
		 * sequence numbers.
		 */
		case PH_MP:
			if (((struct ppp_sc *)(((struct p2pcom *)(vjc->vjc_ifp))
			    ->p2p_private))->ppp_flags & PPP_RCV_SSEQ)
				vjc->vjc_hstate += 2;
			/* FALL THROUGH */
		case PH_MP2:
		case PH_MP3:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_hstate++;
			break;

		case PH_MP4:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_hstate = PH_DATA;
			break;
#endif	/* MULTILINK */
#endif
		LPH_HDR:
		case PH_HDR:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_changes = c;
			vjc->vjc_hstate = (c & NEW_C) ? PH_CID : PH_CSUM1;
			break;
		case PH_CID:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_hstate = PH_CSUM1;
			break;
		case PH_CSUM1:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_hstate = PH_CSUM2;
			break;
		case PH_CSUM2:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_nvar = 0;
			switch (vjc->vjc_changes & SPECIALS_MASK) {
			case SPECIAL_I:
			case SPECIAL_D:
				break;
			default:
				if (vjc->vjc_changes & NEW_U)
					vjc->vjc_nvar++;
				if (vjc->vjc_changes & NEW_W)
					vjc->vjc_nvar++;
				if (vjc->vjc_changes & NEW_A)
					vjc->vjc_nvar++;
				if (vjc->vjc_changes & NEW_S)
					vjc->vjc_nvar++;
				break;
			}
			if (vjc->vjc_changes & NEW_I)
				vjc->vjc_nvar++;
			vjc->vjc_hstate = vjc->vjc_nvar ? PH_NUM : PH_DATA;
			break;
		case PH_NUM:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			if (c)
				vjc->vjc_hstate = --(vjc->vjc_nvar) ? PH_NUM
								    : PH_DATA;
			else
				vjc->vjc_hstate = PH_ZNUM1;
			break;

		case PH_ZNUM1:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_hstate = PH_ZNUM2;
			break;

		case PH_ZNUM2:
			m->m_data[m->m_len++] = c;
			mh->m_pkthdr.len++;
			vjc->vjc_hstate = --(vjc->vjc_nvar) ? PH_NUM : PH_DATA;
			break;
			
		LPH_DATA:
		case PH_DATA:
			/*
			 * If there is no input buffer or no space in
			 * the current buffer allocate a new buffer.
			 */
			if (vjc->vjc_space == 0) {
				if (vjc->vjc_mbuf == vjc->vjc_ibuf)
					MGETHDR(m, M_DONTWAIT, MT_DATA)
				else
					MGET(m, M_DONTWAIT, MT_DATA)
				if (m == 0) {	/* No space available */
					dprintf(("%s%d: no mbufs left\n",
					    vjc->vjc_ifp->if_name,
					    vjc->vjc_ifp->if_unit));
					m = vjc->vjc_mbuf;
					if (m->m_next)
						m_freem(m->m_next);
					m->m_len = 0;
					m->m_next = 0;
					return (-1);
				}
				MCLGET(m, M_DONTWAIT)
				if (vjc->vjc_mbuf == vjc->vjc_ibuf) {
					if ((m->m_flags & M_EXT) == 0) {
						dprintf(("%s%d: out of clusters\n",
						    vjc->vjc_ifp->if_name,
						    vjc->vjc_ifp->if_unit));
						m_freem(m);
						m = vjc->vjc_mbuf;
						if (m->m_next)
							m_freem(m->m_next);
						m->m_len = 0;
						m->m_next = 0;
						return (-1);
					}
					m->m_pkthdr.rcvif = vjc->vjc_ifp;
					m->m_data += 120;
					vjc->vjc_space = MCLBYTES - 120;
				} else
					vjc->vjc_space = (m->m_flags & M_EXT)
					    ? MCLBYTES : MLEN;
				m->m_len = 0;
				vjc->vjc_ibuf->m_next = m;
				vjc->vjc_ibuf = m;
			}
			if (nbytes > 0) {
				int n = nbytes < vjc->vjc_space ?
				    nbytes + 1 : vjc->vjc_space;
				bcopy(cp - 1,
				    mtod(vjc->vjc_ibuf, u_char *) + m->m_len,
				    n);

				m->m_len += n;
				mh->m_pkthdr.len += n;
				vjc->vjc_space -= --n;
				nbytes -= n;
				cp += n;
			} else {
				m->m_data[m->m_len++] = c;
				mh->m_pkthdr.len++;
				vjc->vjc_space--;
			}
			break;
		}
	}
	return (vjc->vjc_hstate == PH_DATA ? 1 : 0);
}
