/*
 * Copyright (c) 1995, 1996 Carnegie Mellon University.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted (including for commercial or
 * for-profit use), provided that both the copyright notice and this
 * permission notice appear in all copies of the software, derivative
 * works, or modified versions, and any portions thereof, and that
 * both notices appear in supporting documentation, and that credit
 * is given to Carnegie Mellon University in all publications reporting
 * on direct or indirect use of this code or its derivatives.
 * 
 * THIS IMPLEMENTATION OF MOBILE IP IS EXPERIMENTAL AND IS KNOWN TO
 * HAVE BUGS, SOME OF WHICH MAY HAVE SERIOUS CONSEQUENCES.  CARNEGIE
 * MELLON PROVIDES THIS SOFTWARE IN ITS "AS IS" CONDITION, AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Carnegie Mellon encourages (but does not require) users of this
 * software to return any improvements or extensions that they make,
 * and to grant Carnegie Mellon the rights to redistribute these
 * changes without encumbrance.   
 *
 *
 * Version 0.9 1996/02/12 - David B. Johnson, dbj@cs.cmu.edu.
 *	Initial beta test release.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/kernel.h>
#include <sys/syslog.h>
#include <sys/malloc.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <net/raw_cb.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_icmp.h>
#include <netinet/mobileip.h>

#ifndef bsdi
#include "loop.h"
struct	ifnet loif[NLOOP];
#else
extern struct ifnet loif;
#endif

LIST_HEAD(hl_head, hl_entry) hl_head;
LIST_HEAD(vl_head, vl_entry) vl_head;

static struct sockaddr afinet_dlinfo = {
	sizeof(struct sockaddr), AF_INET
};

mobileip_init()
{
	LIST_INIT(&hl_head);
	LIST_INIT(&vl_head);
}

/*
 * This routine is called first for all packets of IP protocol
 * IPPROTO_IPIP (protocol 4).  If the IP destination address of
 * the inner (encapsulated) packet is a multicast destination,
 * we hand the packet to mroute_ipip_input.  Otherwise, we hand
 * it to mobileip_input.
 */
ipip_input(m, hlen)
	register struct mbuf *m;
	register int hlen;
{
	register struct ip *ip0, *ip;

#ifdef	MOBILEIP_DEBUG
	printf("ipip_input: m 0x%x, hlen %d, m_data 0x%x, m_len %d\n",
	    m, hlen, m->m_data, m->m_len);
#endif

	if (m->m_len < hlen + sizeof (struct ip) &&
	    (m = m_pullup(m, hlen + sizeof (struct ip))) == 0) {
		ipstat.ips_toosmall++;
		return;
	}

	ip0 = mtod(m, struct ip *);
	ip = (struct ip *)(((char *)ip0) + hlen);

	if (IN_MULTICAST(ip->ip_dst.s_addr)) {
#ifdef	MROUTING
		mroute_ipip_input(m, hlen);
#else
		ipstat.ips_cantforward++;
		m_freem(m);
#endif
	} else {
		mobileip_input(m, hlen);
	}
}

/*
 * This routine is called for all packets of IP protocol IPPROTO_IPIP
 * (protocol 4) for which the inner (encapsulated) packet is *not*
 * addressed to a multicast destination address.  It thus *should* be
 * a packet tunneled by Mobile IP (or at least we treat it that way).
 *
 * This routine is also called for all packets of IP protocol
 * IPPROTO_MINENC (protocol 55).  We decide how to decapsulate it
 * by looking at the IP protocol number, and thus share all the
 * rest of the common code for what to do with it once we've
 * decapsulated it.
 */
mobileip_input(m, hlen)
	register struct mbuf *m;
	int hlen;
{
	register struct ip *ip0;
	register struct ip *ip;
	struct vl_entry *vlp;
	struct ifnet *ifp;
	struct mbuf *m0;
	struct ip *mhip;
	int error;
	int len;
	int off;
	int mhlen, firstlen;
	struct mbuf **mnext;
	struct sockaddr *dlinfo;
	struct minenc *minenc;
	struct in_addr dst;
	int melen;

#ifdef	MOBILEIP_DEBUG
	printf("mobileip_input: m %x, hlen %d\n", m, hlen);
	printf("m 0x%x len %d data 0x%x, flags 0x%x, pkthdr.len %d, next 0x%x\n",
	    m, m->m_len, m->m_data, m->m_flags, m->m_pkthdr.len, m->m_next);
	if (m0 = m->m_next)
		printf("m0 0x%x len %d data 0x%x, flags 0x%x\n",
		    m0, m0->m_len, m0->m_data, m0->m_flags);
#endif	/* MOBILEIP_DEBUG */

	ip0 = mtod(m, struct ip *);
#ifdef	MOBILEIP_DEBUG
	printf("ip0 0x%x, m_data 0x%x, m_len %d, m_pkthdr.len %d\n",
	    ip0, m->m_data, m->m_len, m->m_pkthdr.len);
	dump_ip_header(ip0, "outer header");
#endif	/* MOBILEIP_DEBUG */

	switch (ip0->ip_p) {
	case IPPROTO_IPIP:
		if (m->m_len < hlen + sizeof(struct ip) &&
		    (m = m_pullup(m, hlen + sizeof(struct ip))) == 0) {
			ipstat.ips_toosmall++;
			return;
		}
		ip0 = mtod(m, struct ip *);
		ip = (struct ip *)((char *)ip0 + hlen);
#ifdef	MOBILEIP_DEBUG
		dump_ip_header(ip, "inner header");
#endif	/* MOBILEIP_DEBUG */
		dst = ip->ip_dst;
		break;

	case IPPROTO_MINENC:
#ifdef	MOBILEIP_DEBUG
		dump_minenc_header((char *)ip0 + hlen, "minenc header");
#endif	/* MOBILEIP_DEBUG */
		minenc = NULL;	/* ??? */
		melen = sizeof(struct minenc);
		if (m->m_len < hlen + melen &&
		    (m = m_pullup(m, hlen + melen)) == 0) {
#ifdef	MOBILEIP_DEBUG
			printf("1st pullup too small\n");
#endif	/* MOBILEIP_DEBUG */
			/* can't get whole header.  try for the
			   shorter header and make sure S is set */
			melen -= sizeof(struct in_addr);
			if (m->m_len < hlen + melen &&
			    (m = m_pullup(m, hlen + melen)) == 0) {
#ifdef	MOBILEIP_DEBUG
				printf("2nd pullup too small\n");
#endif	/* MOBILEIP_DEBUG */
				ipstat.ips_toosmall++;
				return;
			}
			ip0 = mtod(m, struct ip *);
			minenc = (struct minenc *)((char *)ip0 + hlen);
			if ((minenc->me_flags & MINENC_SBIT) == 0) {
#ifdef	MOBILEIP_DEBUG
				printf("2nd pullup no S bit\n");
#endif	/* MOBILEIP_DEBUG */
				ipstat.ips_toosmall++;
				return;
			}
		} else {
			ip0 = mtod(m, struct ip *);
			minenc = (struct minenc *)((char *)ip0 + hlen);
		}
#ifdef	MOBILEIP_DEBUG
		dump_ip_header(ip0, "outer header");
		dump_minenc_header(minenc, "minenc header");
#endif	/* MOBILEIP_DEBUG */
		dst = minenc->me_dst;
		break;

	default:
		printf("unknown tunneling protocol %d\n", ip0->ip_p);
		m_freem(m);
		return;
	}

	vlp = find_vl_entry(dst.s_addr);
	if (vlp == NULL) {
		register struct in_ifaddr *ia;

#ifdef	MOBILEIP_DEBUG
		printf("mobileip_input %d.%d.%d.%d not in visitor list\n",
		    IP_ADDR_FORMAT(dst.s_addr));
#endif	/* MOBILEIP_DEBUG */
		for (ia = in_ifaddr; ia; ia = ia->ia_next)
			if (dst.s_addr == ia->ia_addr.sin_addr.s_addr)
				goto ours;

		m_freem(m);   /* icmp or tunnel to home agent??? */
		return;
ours:
#ifndef bsdi
		ifp = &loif[0];
#else
		ifp = &loif;
#endif
		dlinfo = &afinet_dlinfo;

	} else {
#ifdef	MOBILEIP_DEBUG
		printf("mobileip_input found vlp 0x%x for %d.%d.%d.%d\n",
		    vlp, IP_ADDR_FORMAT(dst.s_addr));
#endif	/* MOBILEIP_DEBUG */
		ifp = vlp->vl_ifp;
		dlinfo = &vlp->vl_dlinfo;
	}
#ifdef	MOBILEIP_DEBUG
	printf("ifp 0x%x, name %s%d, index %d, mtu %d\n",
	    ifp, ifp->if_name, ifp->if_unit,
	    ifp->if_index, ifp->if_mtu);
#endif	/* MOBILEIP_DEBUG */

	switch (ip0->ip_p) {
	case IPPROTO_IPIP:
#ifdef	MOBILEIP_DEBUG
		printf("IPIP decap\n");
#endif	/* MOBILEIP_DEBUG */
		ip->ip_ttl = ip0->ip_ttl;

		m->m_data += hlen;
		m->m_len -= hlen;
		m->m_pkthdr.len -= hlen;

		/* do this here since we don't call ip_forward */
		if (ip->ip_ttl <= IPTTLDEC) {
			icmp_error(m, ICMP_TIMXCEED, ICMP_TIMXCEED_INTRANS,
			    0, (struct ifnet *)0);
			return;
		}
		ip->ip_ttl -= IPTTLDEC;
		NTOHS(ip->ip_len);
		break;
	case IPPROTO_MINENC:
#ifdef	MOBILEIP_DEBUG
		printf("MINENC decap\n");
#endif	/* MOBILEIP_DEBUG */
		ip0->ip_len += hlen;
		ip0->ip_ttl -= IPTTLDEC;
		ip0->ip_dst = minenc->me_dst;
		if (minenc->me_flags & MINENC_SBIT)
			ip0->ip_src = minenc->me_src;
		ip0->ip_p = minenc->me_p;
#ifdef	MOBILEIP_DEBUG
		printf("ip len %d, melen %d\n", ip0->ip_len, melen);
#endif	/* MOBILEIP_DEBUG */
		ip0->ip_len -= melen;
		HTONS(ip0->ip_id);
		HTONS(ip0->ip_off);
		ip0->ip_sum = 0;
		ip0->ip_sum = in_cksum(m, hlen);
		bcopy(m->m_data, m->m_data + melen, hlen);
		m->m_data += melen;
		m->m_len -= melen;
		m->m_pkthdr.len -= melen;
		ip = mtod(m, struct ip *);
		/* do this here since we don't call ip_forward */
		if (ip->ip_ttl <= IPTTLDEC) {
			icmp_error(m, ICMP_TIMXCEED, ICMP_TIMXCEED_INTRANS,
			    0, (struct ifnet *)0);
			return;
		}
		break;
	default:
		panic("???");
	}

	/*
	 * If small enough for interface, can just send directly.
	 */
	if ((u_int16_t)ip->ip_len <= ifp->if_mtu) {
#ifdef	MOBILEIP_DEBUG
		printf("sending without fragmenting: ip_len %d\n",
		    ip->ip_len);
#endif	/* MOBILEIP_DEBUG */
		HTONS(ip->ip_len);
		ip->ip_sum = 0;
		ip->ip_sum = in_cksum(m, hlen);
		error = (*ifp->if_output)(ifp, m, dlinfo, NULL);
#ifdef	MOBILEIP_DEBUG
		printf("if_output returned error %d\n", error);
#endif	/* MOBILEIP_DEBUG */
		/* icmp ??? */
		return;
	}
#ifdef	MOBILEIP_DEBUG
	printf("must do fragmentation: ip_len %d\n", ip->ip_len);
#endif	/* MOBILEIP_DEBUG */

	NTOHS(ip->ip_off);

	/*
	 * Too large for interface; fragment if possible.
	 * Must be able to put at least 8 bytes per fragment.
	 */
	if (ip->ip_off & IP_DF) {
		ipstat.ips_cantfrag++;
		m_freem(m);
		/* icmp ??? */
		return;
	}
	len = (ifp->if_mtu - hlen) &~ 7;
	if (len < 8) {
		m_freem(m);
		/* icmp ??? */
		return;
	}

	/*
	 * Loop through length of segment after first fragment,
	 * make new header and copy data of each part and link onto chain.
	 */
	firstlen = len;
	mnext = &m->m_nextpkt;
	m0 = m;
	mhlen = sizeof (struct ip);
	for (off = hlen + len; off < (u_int16_t)ip->ip_len; off += len) {
		MGETHDR(m, M_DONTWAIT, MT_HEADER);
		if (m == 0) {
			error = ENOBUFS;
			ipstat.ips_odropped++;
			goto sendorfree;
		}
		*mnext = m;
		mnext = &m->m_nextpkt;
		m->m_data += max_linkhdr;
		mhip = mtod(m, struct ip *);
		*mhip = *ip;
		if (hlen > sizeof (struct ip)) {
			mhlen = ip_optcopy(ip, mhip) + sizeof (struct ip);
			mhip->ip_hl = mhlen >> 2;
		}
		m->m_len = mhlen;
		mhip->ip_off = ((off - hlen) >> 3) + (ip->ip_off & ~IP_MF);
		if (ip->ip_off & IP_MF)
			mhip->ip_off |= IP_MF;
		if (off + len >= (u_int16_t)ip->ip_len)
			len = (u_int16_t)ip->ip_len - off;
		else
			mhip->ip_off |= IP_MF;
		mhip->ip_len = htons((u_int16_t)(len + mhlen));
		m->m_next = m_copy(m0, off, len);
		if (m->m_next == 0) {
			error = ENOBUFS;	/* ??? */
			ipstat.ips_odropped++;
			goto sendorfree;
		}
		m->m_pkthdr.len = mhlen + len;
		m->m_pkthdr.rcvif = (struct ifnet *)0;
		mhip->ip_off = htons((u_int16_t)mhip->ip_off);
		mhip->ip_sum = 0;
		mhip->ip_sum = in_cksum(m, mhlen);
		ipstat.ips_ofragments++;
	}
	/*
	 * Update first fragment by trimming what's been copied out
	 * and updating header, then send each fragment (in order).
	 */
	m = m0;
	m_adj(m, hlen + firstlen - (u_int16_t)ip->ip_len);
	m->m_pkthdr.len = hlen + firstlen;
	ip->ip_len = htons((u_int16_t)m->m_pkthdr.len);
	ip->ip_off = htons((u_int16_t)(ip->ip_off | IP_MF));
	ip->ip_sum = 0;
	ip->ip_sum = in_cksum(m, hlen);
sendorfree:
	for (m = m0; m; m = m0) {
		m0 = m->m_nextpkt;
		m->m_nextpkt = 0;
		if (error == 0)
			error = (*ifp->if_output)(ifp, m, dlinfo, NULL);
		else
			m_freem(m);
	}

	if (error == 0)
		ipstat.ips_fragmented++;
}

/*
 * To add a home list entry:
 *	set RTF_HOME
 *	pass dst as sockaddr_in home IP address
 *	pass gateway as sockaddr_in care-of address
 *
 * To add visitor list entry:
 *	set RTF_VISITOR
 *	pass dst as sockaddr_in home IP address
 *	pass gateway as sockaddr AF_UNSPEC MAC header from bpf
 *	pass ifpaddr as sockaddr_dl with interface name
 */
int
mobileip_rtrequest(req, addrs, dst, gateway, ifpaddr, flags)
	int req;
	int addrs;
	struct sockaddr *dst;
#define	in_dst		((struct sockaddr_in *)dst)
	struct sockaddr *gateway;
#define	in_gateway	((struct sockaddr_in *)gateway)
#define	dl_gateway	((struct sockaddr_dl *)gateway)
	struct sockaddr	*ifpaddr;
#define	dl_ifpaddr	((struct sockaddr_dl *)ifpaddr)
	int flags;
{
	struct hl_entry	*hlp;
	struct vl_entry	*vlp;

#ifdef	MOBILEIP_DEBUG
	printf("mobileip_rtrequest: req %x flags %x\n", req, flags);
	printf("dst len %d family %d port %d addr %d.%d.%d.%d\n",
	    in_dst->sin_len, in_dst->sin_family,
	    in_dst->sin_port, IP_ADDR_FORMAT(in_dst->sin_addr.s_addr));
#endif	/* MOBILEIP_DEBUG */

	if (dst->sa_family != AF_INET ||
	    dst->sa_len != sizeof(struct sockaddr_in))
		return (EINVAL);

	if (flags & RTF_HOME) {
		/*
		 * Manipulate the home list
		 */
#ifdef	MOBILEIP_DEBUG
		printf("doing home list\n");
		printf("gateway len %d family %d port %d addr %d.%d.%d.%d\n",
		    in_gateway->sin_len, in_gateway->sin_family,
		    in_gateway->sin_port,
		    IP_ADDR_FORMAT(in_gateway->sin_addr.s_addr));
#endif	/* MOBILEIP_DEBUG */

		if ((addrs & (RTA_DST|RTA_GATEWAY)) != (RTA_DST|RTA_GATEWAY))
			return (EINVAL);
		if (gateway->sa_family != AF_INET ||
		    gateway->sa_len != sizeof(struct sockaddr_in))
			return (EINVAL);

		hlp = find_hl_entry(in_dst->sin_addr.s_addr);

		switch (req) {
		case RTM_MIPADD:
			if (hlp != NULL) {
#ifdef	MOBILEIP_DEBUG
				printf("returning EEXIST\n");
#endif	/* MOBILEIP_DEBUG */
				return (EEXIST);
			}
			hlp = malloc(sizeof(struct hl_entry),
			    M_RTABLE, M_WAITOK);
			hlp->hl_home = *dst;
			hlp->hl_coa = *gateway;
			hlp->hl_flags = in_dst->sin_port;	/* ??? */
			LIST_INSERT_HEAD(&hl_head, hlp, hl_link);
#ifdef	MOBILEIP_DEBUG
			printf("entry added: hlp %x\n", hlp);
			printf("home %d.%d.%d.%d -> coa %d.%d.%d.%d\n",
			    IP_ADDR_FORMAT((u_long)(&((hlp->hl_home).sa_data[2]))),
			    IP_ADDR_FORMAT((u_long)(&((hlp->hl_coa).sa_data[2]))));
#endif	/* MOBILEIP_DEBUG */
			if (in_dst->sin_addr.s_addr == last_hl_not_found)
				last_hl_not_found = 0;
			break;
		case RTM_MIPDELETE:
			if (hlp == NULL) {
#ifdef	MOBILEIP_DEBUG
				printf("returning ENOENT\n");
#endif	/* MOBILEIP_DEBUG */
				return (ENOENT);
			}
			LIST_REMOVE(hlp, hl_link);
			free(hlp, M_RTABLE);
#ifdef	MOBILEIP_DEBUG
			printf("entry deleted: hlp %x\n", hlp);
#endif	/* MOBILEIP_DEBUG */
			break;
		case RTM_MIPCHANGE:
			if (hlp == NULL) {
#ifdef	MOBILEIP_DEBUG
				printf("returning ENOENT\n");
#endif	/* MOBILEIP_DEBUG */
				return (ENOENT);
			}
			hlp->hl_coa = *gateway;
			hlp->hl_flags = in_dst->sin_port;	/* ??? */
#ifdef	MOBILEIP_DEBUG
			printf("entry changed: hlp %x\n", hlp);
#endif	/* MOBILEIP_DEBUG */
			break;
		}
	} else {
		/*
		 * Manipulate visitor list
		 */
		char name[IFNAMSIZ];
#ifdef	MOBILEIP_DEBUG
		int i;
		char *cp;

		printf("doing visitor list\n");

		printf("gateway len %d family %d\n",
		    gateway->sa_len, gateway->sa_family);

		printf("ifpaddr len %d family %d index %d type %x\n",
		    dl_ifpaddr->sdl_len, dl_ifpaddr->sdl_family,
		    dl_ifpaddr->sdl_index, dl_ifpaddr->sdl_type);
		printf("ifpaddr nlen %d alen %d\n",
		    dl_ifpaddr->sdl_nlen, dl_ifpaddr->sdl_alen);
		cp = dl_ifpaddr->sdl_data;
		printf("name '");
		for (i = 0; i < dl_ifpaddr->sdl_nlen; i++)
			printf("%c", *cp++);
		printf("'\n");
#endif	/* MOBILEIP_DEBUG */

		if ((addrs & (RTA_DST|RTA_GATEWAY|RTA_IFP)) !=
		    (RTA_DST|RTA_GATEWAY|RTA_IFP))
			return (EINVAL);

		if (gateway->sa_family != AF_UNSPEC)
			return (EINVAL);

		if (ifpaddr->sa_family != AF_LINK ||
		    dl_ifpaddr->sdl_nlen == 0 ||
		    dl_ifpaddr->sdl_nlen >= IFNAMSIZ)
			return (EINVAL);

		vlp = find_vl_entry(in_dst->sin_addr.s_addr);

		bcopy(dl_ifpaddr->sdl_data, name, dl_ifpaddr->sdl_nlen);
		name[dl_ifpaddr->sdl_nlen] = '\0';

		switch (req) {
		case RTM_MIPADD:
			if (vlp != NULL) return (EEXIST);
			vlp = malloc(sizeof(struct vl_entry),
			    M_RTABLE, M_WAITOK);
			vlp->vl_home = *dst;
			vlp->vl_dlinfo = *gateway;
			vlp->vl_ifp = ifunit(name);
			if (vlp->vl_ifp == NULL) {
				free(vlp, M_RTABLE);
				return (ENXIO);
			}
			LIST_INSERT_HEAD(&vl_head, vlp, vl_link);
#ifdef	MOBILEIP_DEBUG
			printf("entry added: vlp %x\n", vlp);
#endif	/* MOBILEIP_DEBUG */
			break;
		case RTM_MIPDELETE:
			if (vlp == NULL) return (ENOENT);
			LIST_REMOVE(vlp, vl_link);
			free(vlp, M_RTABLE);
#ifdef	MOBILEIP_DEBUG
			printf("entry deleted: vlp %x\n", vlp);
#endif	/* MOBILEIP_DEBUG */
			break;
		case RTM_MIPCHANGE:
			return (EINVAL);	/* ??? */
		}
	}

	return (0);
}

int
mobileip_output(m, hlp, flags)
	struct mbuf *m;
	struct hl_entry *hlp;
	int flags;
{
	int error;

	error = mobileip_tunnel(m, hlp, flags);
	return (error);
}

mobileip_forward(m, hlp, flags)
	struct mbuf *m;
	struct hl_entry *hlp;
	int flags;
{
	register struct ip *ip = mtod(m, struct ip *);
	int error;

	/*
	 * Check the TTL here so we know it is safe to decrement below.
	 */
	if (ip->ip_ttl <= IPTTLDEC) {
		icmp_error(m, ICMP_TIMXCEED, ICMP_TIMXCEED_INTRANS,
		    0, (struct ifnet *)0);
		return;
	}
	ip->ip_ttl -= IPTTLDEC;

	error = mobileip_tunnel(m, hlp, flags);
#ifdef	MOBILEIP_DEBUG
	printf("mobileip_forward: tunnel returned %d\n", error);
#endif	/* MOBILEIP_DEBUG */
}

/*
 * ip_len and ip_off are in host byte order.
 * otherwise, the packet is complete and in network byte order.
 * To call ip_output, ip_len and ip_off must still be in host byte order.
 */
int
mobileip_tunnel(m, hlp, flags)
	struct mbuf *m;
	struct hl_entry *hlp;
	int flags;
{
	register struct ip *ip = mtod(m, struct ip *);
	struct ip *ip0;
	struct mbuf *m0;
	int error, type = 0, code;
	struct mbuf *mcopy = NULL;
	int hlen, melen;
	struct minenc *minenc;

#ifdef	MOBILEIP_DEBUG
	printf("mobileip_tunnel m %x, hlp %x, flags %x\n", m, hlp, flags);
	printf(">>> TUNNEL %d.%d.%d.%d -> %d.%d.%d.%d\n",
	    IP_ADDR_FORMAT((u_long)(&((hlp->hl_home).sa_data[2]))),
	    IP_ADDR_FORMAT((u_long)(&((hlp->hl_coa).sa_data[2]))));
	printf("m %x m_len %d pkthdr len %d\n",
	    m, m->m_len, m->m_pkthdr.len);
#endif	/* MOBILEIP_DEBUG */

	MGETHDR(m0, M_DONTWAIT, m->m_type);
	if (m0 == 0) {
		printf("mobileip_tunnel MGETHDR returned NULL!\n");
		m_freem(m);
		return (0);		/* return error ??? */
	}
	m->m_flags &= ~M_PKTHDR;
	m0->m_next = m;

	if (ip->ip_src.s_addr == INADDR_ANY)
	    ip->ip_src = in_ifaddr->ia_addr.sin_addr; /*???*/

	if ((hlp->hl_flags & HL_MINENC) == 0 ||
	    (ip->ip_off & (IP_MF|IP_OFFMASK))) /* already a fragment */ {
		MH_ALIGN(m0, sizeof(struct ip));
		m0->m_len = sizeof(struct ip);
		m0->m_pkthdr.len = sizeof(struct ip) + m->m_pkthdr.len;

		ip0 = mtod(m0, struct ip *);
#ifdef	MOBILEIP_DEBUG
		printf("ip %x, ip0 %x\n", ip, ip0);
#endif	/* MOBILEIP_DEBUG */
		ip0->ip_v = IPVERSION;
		ip0->ip_hl = sizeof(*ip0) >> 2;
		ip0->ip_tos = ip->ip_tos;
		ip0->ip_len = ip->ip_len + sizeof(*ip0);
		ip0->ip_id = htons(ip_id++);
		ip0->ip_off = ip->ip_off & IP_DF;
		ip0->ip_ttl = ip->ip_ttl;
		ip0->ip_p = IPPROTO_IPIP;
		ip0->ip_sum = 0;
		ip0->ip_src = in_ifaddr->ia_addr.sin_addr; /*???*/
		ip0->ip_dst.s_addr =  (u_long)(&((hlp->hl_coa).sa_data[2]));

		HTONS(ip->ip_len);
		HTONS(ip->ip_off);
		ip->ip_sum = 0;
		ip->ip_sum = in_cksum(m, ip->ip_hl << 2);
	} else {
		hlen = ip->ip_hl << 2;
		MH_ALIGN(m0, hlen);
		m0->m_len = hlen;
		m0->m_pkthdr.len = m->m_pkthdr.len;

		ip0 = mtod(m0, struct ip *);
#ifdef	MOBILEIP_DEBUG
		printf("ip %x, ip0 %x\n", ip, ip0);
#endif	/* MOBILEIP_DEBUG */
		bcopy((char *)ip, (char *)ip0, hlen);

		/* could check ifwithaddr??? */
		if (ip0->ip_src.s_addr ==
		    in_ifaddr->ia_addr.sin_addr.s_addr) {
			/* don't include the original sender */
			melen = sizeof(struct minenc) - sizeof(struct in_addr);
#ifdef	MOBILEIP_DEBUG
			printf("don't include: melen %d\n", melen);
#endif	/* MOBILEIP_DEBUG */
			m->m_len -= (hlen - melen);
			m->m_data += (hlen - melen);
			minenc = mtod(m, struct minenc *);
			minenc->me_flags = 0;
		} else {
			/* do include the origial sender */
			melen = sizeof(struct minenc);
#ifdef	MOBILEIP_DEBUG
			printf("do include: melen %d\n", melen);
#endif	/* MOBILEIP_DEBUG */
			m->m_len -= (hlen - melen);
			m->m_data += (hlen - melen);
			minenc = mtod(m, struct minenc *);
			minenc->me_flags = MINENC_SBIT;
			minenc->me_src = ip0->ip_src;
			ip0->ip_src = in_ifaddr->ia_addr.sin_addr;
		}
		m0->m_pkthdr.len += melen;
		minenc->me_p = ip0->ip_p;
		minenc->me_dst = ip0->ip_dst;
		minenc->me_sum = 0;
		minenc->me_sum = in_cksum(m, melen);
		ip0->ip_p = IPPROTO_MINENC;
		ip0->ip_dst.s_addr = (u_long)(&((hlp->hl_coa).sa_data[2])); 

		ip0->ip_len += melen;
#ifdef	MOBILEIP_DEBUG
		printf(">>> minenc done\n");
		printf("m0 0x%x len %d data 0x%x, flags 0x%x, pkthdr.len %d, next 0x%x\n",
		    m0, m0->m_len, m0->m_data, m0->m_flags,
		    m0->m_pkthdr.len, m0->m_next);
		dump_ip_header(ip0, "ip header");
		printf("m 0x%x len %d data 0x%x, flags 0x%x\n",
		    m, m->m_len, m->m_data, m->m_flags);
		dump_minenc_header(minenc, "minenc header");
#endif	/* MOBILEIP_DEBUG */
	}

	/*
	 * Save at most 64 bytes of the packet in case
	 * we need to generate an ICMP message to the src.
	 */
	if (flags & IP_FORWARDING)
		mcopy = m_copy(m0, 0, imin((int)ip->ip_len, 64));

	error = ip_output(m0, (struct mbuf *)0, (struct route *)0,
	    flags | IP_RAWOUTPUT, 0);
#ifdef	MOBILEIP_DEBUG
	printf("mobileip_tunnel ip_output returned error %d\n", error);
#endif	/* MOBILEIP_DEBUG */
done:
	if ((flags & IP_FORWARDING) == 0)
		return (error);

	if (error)
		ipstat.ips_cantforward++;
	else {
		ipstat.ips_forward++;
		if (mcopy)
			m_freem(mcopy);
		return (0);
	}

	if (mcopy == NULL)
		return (error);

	switch (error) {

	case 0:
		return (0);

	case ENETUNREACH:
	case EHOSTUNREACH:
	case ENETDOWN:
	case EHOSTDOWN:
	default:
		type = ICMP_UNREACH;
		code = ICMP_UNREACH_HOST;
		break;

	case EMSGSIZE:
		type = ICMP_UNREACH;
		code = ICMP_UNREACH_NEEDFRAG;
		ipstat.ips_cantfrag++;
		break;

	case ENOBUFS:
		type = ICMP_SOURCEQUENCH;
		code = 0;
		break;
	}
	icmp_error(mcopy, type, code, 0, (struct ifnet *)0);

	return (error);
}

struct hl_entry *
find_hl_entry(addr)
	register u_int32_t addr;
{
	register struct hl_entry *hlp;

	for (hlp = hl_head.lh_first; hlp != NULL; hlp = hlp->hl_link.le_next)
		if (((u_long)(&((hlp->hl_home).sa_data[2]))) == addr) {
#ifdef	MOBILEIP_DEBUG
			printf("home entry found: hlp %x\n", hlp);
#endif	/* MOBILEIP_DEBUG */
			return (hlp);
		}

	last_hl_not_found = addr;
	return (NULL);
}

struct vl_entry *
find_vl_entry(addr)
	u_int32_t addr;
{
	register struct vl_entry *vlp;

	for (vlp = vl_head.lh_first; vlp != NULL; vlp = vlp->vl_link.le_next)
		if ((u_long)(&((vlp->vl_home).sa_data[2])) == addr) {
#ifdef	MOBILEIP_DEBUG
			printf("visitor entry found: hlp %x\n", vlp);
#endif	/* MOBILEIP_DEBUG */
			return (vlp);
		}

	return (NULL);
}

#ifdef	MOBILEIP_DEBUG
dump_ip_header(ip, label)
	register struct ip *ip;
	char *label;
{
	printf("0x%x %s:\n", ip, label);

	printf("v %d, hl %d, tos 0x%x, len %d, id 0x%x, off 0x%x, ttl %d\n",
	    ip->ip_v, ip->ip_hl, ip->ip_tos, ip->ip_len, ip->ip_id,
	    ip->ip_off, ip->ip_ttl);
	printf("p %d, sum 0x%x, src %d.%d.%d.%d, dst %d.%d.%d.%d\n",
	    ip->ip_p, ip->ip_sum, IP_ADDR_FORMAT(ip->ip_src.s_addr),
	    IP_ADDR_FORMAT(ip->ip_dst.s_addr));
}

dump_minenc_header(minenc, label)
	register struct minenc *minenc;
	char *label;
{
	printf("0x%x %s:\n", minenc, label);

	printf("p %d, flags 0x%x, sum 0x%x, dest %d.%d.%d.%d",
	    minenc->me_p, minenc->me_flags, minenc->me_sum,
	    IP_ADDR_FORMAT(minenc->me_dst.s_addr));
	if (minenc->me_flags & MINENC_SBIT)
		printf(", src %d.%d.%d.%d",
		    IP_ADDR_FORMAT(minenc->me_src.s_addr));
	printf("\n");
}
#endif	/* MOBILEIP_DEBUG */
