/*
 * Copyright (c) 1995, 1996 Carnegie Mellon University.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted (including for commercial or
 * for-profit use), provided that both the copyright notice and this
 * permission notice appear in all copies of the software, derivative
 * works, or modified versions, and any portions thereof, and that
 * both notices appear in supporting documentation, and that credit
 * is given to Carnegie Mellon University in all publications reporting
 * on direct or indirect use of this code or its derivatives.
 * 
 * THIS IMPLEMENTATION OF MOBILE IP IS EXPERIMENTAL AND IS KNOWN TO
 * HAVE BUGS, SOME OF WHICH MAY HAVE SERIOUS CONSEQUENCES.  CARNEGIE
 * MELLON PROVIDES THIS SOFTWARE IN ITS "AS IS" CONDITION, AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Carnegie Mellon encourages (but does not require) users of this
 * software to return any improvements or extensions that they make,
 * and to grant Carnegie Mellon the rights to redistribute these
 * changes without encumbrance.   
 *
 *
 * Version 0.9 1996/02/12 - David B. Johnson, dbj@cs.cmu.edu.
 *	Initial beta test release.
 */

#include <sys/queue.h> /* hodes */


#define	IPPROTO_MINENC		55	/* Minimal encapsulation */

/*
 * Minimal encapsulation header format.
 */
struct minenc {
	u_char	me_p;			/* original IP protocol number */
	u_char	me_flags;		/* flags (below) */
	u_short	me_sum;			/* minenc header checksum */
	struct in_addr me_dst;		/* orig. destination (mobile node) */
	struct in_addr me_src;		/* orig. source address
					   (only present if MINENC_SBIT) */
};

#define	MINENC_SBIT	0x80		/* sender address present */

/*
 * Home list table entry.
 */
struct hl_entry {
	LIST_ENTRY(hl_entry) hl_link;
	struct sockaddr	hl_home;	/* Home address */
	struct sockaddr	hl_coa;		/* care-of address */
	u_long hl_flags;		/* flags (below) */
};

#define	HL_MINENC	0x1		/* can use minimal encapsulation */
#define	HL_GRE		0x2		/* can use GRE encapsulation
					   (not implemented yet) */

/*
 * Visitor list table entry.
 */
struct vl_entry {
	LIST_ENTRY(vl_entry) vl_link;
	struct sockaddr	vl_home;	/* home address */
	struct sockaddr	vl_dlinfo;	/* link-layer address info */
	struct ifnet	*vl_ifp;	/* interface registered on o*/
};

u_int32_t last_hl_not_found;		/* one-entry negative cache */

struct hl_entry *find_hl_entry __P((u_int32_t));
struct vl_entry *find_vl_entry __P((u_int32_t));

/*
 * Define additional routing socket message types.
 *
 * (The type field in a message is a u_char, and the highest type
 * code assigned by the normal NetBSD code is 0xe.  Hopefully, these
 * won't overlap with other current or future types.)
 */
#define RTM_MIPADD		0x31	/* Add home/visitor list entry */
#define RTM_MIPDELETE		0x32	/* Delete home/visitor list entry */
#define RTM_MIPCHANGE		0x33	/* Change home/visitor list entry */

#define	RTF_HOME	RTF_PROTO1	/* Message for home list entry */
#define	RTF_VISITOR	RTF_PROTO2	/* Message for visitor list entry */

#ifdef	MOBILEIP_DEBUG
#define IP_ADDR_FORMAT(addr) \
	((addr) & 0xff), (((addr) >> 8) & 0xff), \
	(((addr) >> 16) & 0xff), (((addr) >> 24) & 0xff)
#endif
