#ifdef STP
#include <sys/param.h>
#include <sys/systm.h> 
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/errno.h>
 
#include <net/if.h>
#include <net/route.h>
   
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/stp.h>
#include <netinet/stp_var.h>


void    
stp_input(m, iphlen)
	register struct mbuf *m;
	int iphlen;
{
	struct stpcb *sp = 0;
	struct stpiphdr *si;
	struct inpcb *inp;
	struct socket *so;
	u_long seqno;
	u_char pkttype;
	u_long timestamp;
	u_long *dataptr;
	u_short pdu_instance;
	flag fPoll = FALSE;
#if ERROR_DETECTION
	flag fError = FALSE;
#endif
	

	/* 
	 * Get IP and STP header together in first mbuf (IP leaves IP hdr
	 * in first mbuf)
	 */
	si = mtod(m, struct stpiphdr *);
	/* strip any IP options */
	if (iphlen > sizeof (struct ip))
		ip_stripoptions(m, (struct mbuf *) 0);
	/* if the number of bytes in the 1st mbuf < 36, call mpullup */
	if (m->m_len < sizeof(struct stpiphdr)) {
		if (( m = m_pullup(m, sizeof(struct stpiphdr))) == 0) {
			return;
		}
		si = mtod(m, struct stpiphdr *);
	}

	/* Checksum check */
	if (si->si_sum = stp_cksum(m, ((struct ip *) si)->ip_len)) { 
#if ERROR_DETECTION
		fError = TRUE;
#else
#ifdef STP_DEBUG
		printf("Checksum failure\n");
#endif
		goto drop;
#endif
	}

	seqno = ntohl(si->si_seq);
	/* Extract type field */ 
	pkttype = (u_char) (seqno >> 24);	
	/* Zero out highest 8 bits of sequence number field */
	seqno = ((seqno << 8) >> 8);
	pdu_instance = ntohs(si->si_inc);
	/* 
	 * Now that we have the 24 bit sequence number and the pkttype is 
	 * stored separately, overwrite the field in the mbuf with the 
	 * (host order) value, for later use.
	 */
	si->si_seq = seqno;

	/* Get pcb for segment. We don't implement the one-behind cache */
	inp = in_pcblookup(&scb, si->si_src, si->si_sport, si->si_dst, 
	    si->si_dport, INPLOOKUP_WILDCARD);

	/* 
	 * If no pcb, we're in CLOSED state.  Generate a RST for the 
	 * received packet unless the packet is an END (ack it) or ENDAK 
	 * (ignore it) or RST (ignore it). 
	 */
	if (inp == 0) {
#if ERROR_DETECTION
		if (fError) {
			/* Don't take action on checksum-failed packet */
#ifdef STP_DEBUG
			printf("Checksum failure\n");
#endif
			goto drop;
		}
#endif
		if (pkttype == STP_PDU_END) {
			/* ENDAK may have been lost */
			stp_reset(m, si, STP_PDU_ENDAK); 
			return;
		} else if ((pkttype == STP_PDU_ENDAK) || 
		    (pkttype == STP_PDU_RST)) {
			/* Ignore */
			m_freem(m);
			return;
		} else {
			stp_reset(m, si, STP_PDU_RST); 
			return;
		}
	} 
	sp = intostpcb(inp);

#if ERROR_DETECTION
	if (fError) {
		/* if any of the conditions below are true, drop the PDU */
		if (!(sp == 0 || sp->s_state == STP_CLOSED ||
			(pkttype != STP_PDU_SD && pkttype != STP_PDU_SDP) || 
			(sp->s_instance != pdu_instance) ) ) 
			/* if not, process this packet as an errored SD_PDU */
			stp_rcverror(sp, seqno);
#ifdef STP_DEBUG
			printf("Checksum failure\n");
#endif
		goto drop;
	} 
#endif /* ERROR_DETECTION */
	
	if (sp == 0) {
		stp_reset(m, si, STP_PDU_RST); 
		return;
	}
	if (sp->s_state == STP_CLOSED) {
		/* Ignore */
		m_freem(m);
		return;
	}
	/* Drop if broadcast or multicast */
	if (m->m_flags & (M_BCAST | M_MCAST) ||
		IN_MULTICAST(si->si_dst.s_addr)) {
		m_freem(m);
		return;
	}

	/* Create new socket if BGN PDU arrives for listening socket */
	so = inp->inp_socket;
	if ((so->so_options & SO_ACCEPTCONN) && (sp->s_state == STP_LISTEN) &&
	    ((pkttype == STP_PDU_BGN) || (pkttype == STP_PDU_BGN_SD) ||
	    (pkttype == STP_PDU_BGN_SD_END))) {
		so = sonewconn(so,0);
		if (so == 0)
			goto drop;

		inp = (struct inpcb *) so->so_pcb;
		inp->inp_laddr = si->si_dst;
		inp->inp_lport = si->si_dport;
			
		/* if input datagram had source route, this fetches it */
		inp->inp_options = ip_srcroute();
	
		sp = intostpcb(inp);
		/* state of new socket set to LISTEN, and s_instance is set */
		sp->s_state = STP_LISTEN;
		sp->s_instance = ntohs(si->si_inc);
		/* start any connection maintenance timers here */
	} else {
		/* 
		 * Make sure packet belongs to this instance of connection.
		 * Don't respond to a RST packet for a nonexistent connection 
		 */
		if ((sp->s_instance != pdu_instance)&& pkttype != STP_PDU_RST) {
			/* Check for simultaneous open XXX rewrite this check */
			if ((sp->s_state == STP_BGN_SENT || 
			    sp->s_state == STP_BGN_DATA_SENT || 
			    sp->s_state == STP_BGN_DATA_END_SENT) && 
			    (pkttype & 0x10))
				; /* Nothing */
			else 	
				stp_reset(m, si, STP_PDU_RST); 
				return;
		}
	}

#ifdef STP_DEBUG
	if ( (u_short) random() > STP_THRESHOLD ) {
		m_freem(m);
		return;
	}
#endif
	sp->s_idle = 0;
	/* Handle packet.  This calls "rcv" functions except for common cases */
#ifdef STP_STATS
	sp->total_bytes_recv += m->m_len;
	sp->pkts_recv++;
#endif
	switch (pkttype) {
	case STP_PDU_SDP:
		fPoll = TRUE;
		dataptr = (u_long *) (mtod(m, caddr_t) + 
		    sizeof(struct stpiphdr));
		timestamp = ntohl(*dataptr);
		/* FALL THROUGH */

	case STP_PDU_SD:
		/* Adjust pointers */
		m->m_data += sizeof(struct stpiphdr); 
		m->m_len -= sizeof(struct stpiphdr);
		if (sp->oob_pending)
			/* Keep oob_pending > rcv buffer length */ 
			so->so_oobmark += m->m_len; 
		if (STP_PDU_SDP == pkttype)
			m_adj(m, 4);  	/* Trim off timestamp */
		if ((sp->s_state == STP_EST) || 
		    (sp->s_state == STP_HALF_CLOSE_PENDING) ||
		    (sp->s_state == STP_HALF_CLOSED)) {
			if ( (seqno == sp->rcv_max) &&
			 	(sp->rcv_max == sp->rcv_nxt) && 
			 	(si->si_len <= sbspace(&so->so_rcv)) &&
			 	(sp->rcv_wnd) 	) {
				/* Main (fast path) processing */
				sbappend(&so->so_rcv, m);
				sorwakeup(so);
				sp->rcv_max = SEQ_INC(sp->rcv_max);	
				sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);	
				if (fPoll)
					stp_sndstat(sp, timestamp);
				break; 
			} 
		} 
		/* not in sequence; slow path processing */
		stp_rcvsd(si, so, sp, m, seqno, fPoll);
		if (fPoll)
			stp_sndstat(sp, timestamp);
		break;
	
	case STP_PDU_SDP_URG:
		/* 
		 * Check if reception of packet is legal in this state 
		 * with the current receive window
		 */ 
		if ((sp->s_state == STP_EST || 
		    sp->s_state == STP_HALF_CLOSE_PENDING ||
		    sp->s_state == STP_HALF_CLOSED) &&
		    (seq_subtract(seqno, sp->rcv_nxt) < sp->rcv_wnd)) {

			/* Trim off header and poll information */
			dataptr = (u_long *)(mtod(m, caddr_t) + 
			    sizeof(struct stpiphdr));
			timestamp = ntohl(*dataptr);
			m->m_data += sizeof(struct stpiphdr); 
			m->m_len -= sizeof(struct stpiphdr);
			m_adj(m, 4);  	/* Trim off timestamp */

			/* Fix so_oobmark */
			sp->rcv_up = seqno;
			if (seqno == sp->rcv_nxt) {
				so->so_oobmark = so->so_rcv.sb_cc;
			} else if (seq_gt(seqno, sp->rcv_nxt)) {
				so->so_oobmark = so->so_rcv.sb_cc + 
				    seq_subtract(seqno, sp->rcv_nxt) * 
				    sp->s_maxseg;
			}
			if (so->so_oobmark == 0)
				so->so_state |= SS_RCVATMARK;
			if (!sp->oob_pending) {
				/* 
				 * Must notify process of OOB data for 
				 * the first time.
				 */
				sohasoutofband(so);
				sp->s_oobflags &= ~(STPOOB_HAVEDATA | 
				    STPOOB_HADDATA);
			} 
			/* 
			 * If inline, pull character out of band 
			 * into s_iobc 
			 */
			if ((so->so_options & SO_OOBINLINE) == 0)
				stp_pulloutofband(so, si, m);
			/* 
			 * Even if the character has been stripped, we 
			 * still need to process this as an SD since 
		 	 * it has consumed a sequence number
			 */
			stp_rcvsd(si, so, sp, m, seqno, TRUE);
			stp_sndstat(sp, timestamp);
			sp->oob_pending = FALSE;
		}
		break;

	case STP_PDU_POLL_URG:
		/* 
		 * This is a bit of a hack.  The POLL indicates that 
		 * urgent data is on its way, but we need to indicate 
		 * the byte offset in the socket's receive buffer 
		 * (so_oobmark).  But we don't know how far in advance 
		 * the urgent data will be, so set so_oobmark to value 
		 * higher than the amount of data in the receive 
		 * buffer, to prevent it from reading right now.
		 * When the actual SDP_URG_PDU comes, we can adjust it 
		 * to the correct value.
		 */ 
		so->so_oobmark = so->so_rcv.sb_cc + (sp->s_peermaxseg * 
		    seq_subtract(sp->rcv_max, sp->rcv_nxt)) + 1;
		sohasoutofband(so);
		sp->s_oobflags &= ~(STPOOB_HAVEDATA | STPOOB_HADDATA);
		sp->oob_pending = TRUE;
		/* FALL THROUGH */

	case STP_PDU_POLL:
		stp_rcvpoll(sp, m, seqno);
		m_freem(m);
		break;

	case STP_PDU_STAT:
		stp_rcvstat(so, sp, m, seqno);
		m_freem(m);
		break;

	case STP_PDU_USTAT:
		stp_rcvustat(so, sp, m, seqno);
		m_freem(m);
		break;

	case STP_PDU_BGN:
		stp_rcvbgn(so, sp, m, inp, TRUE);
		m_freem(m);
		break;

	case STP_PDU_BGN_SD:
		stp_rcvbgn(so, sp, m, inp, TRUE);
		/* 
		 * Now, trim off window and timestamp, and treat 
		 * like an SD
		 */
		m->m_data += sizeof(struct stpiphdr); 
		m->m_len -= sizeof(struct stpiphdr);
		m_adj(m, 8);
		sbappend(&so->so_rcv, m);
		sorwakeup(so);
		sp->rcv_max = SEQ_INC(sp->rcv_max);	
		sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);	
		break;

	case STP_PDU_BGN_SD_END:
		if (sp->s_state == STP_LISTEN) 
			/* don't respond with a BGNAK */
			stp_rcvbgn(so, sp, m, inp, FALSE); 
		else if (sp->s_state == STP_BGN_SENT || sp->s_state == 
		    STP_BGN_DATA_SENT || sp->s_state == STP_BGN_DATA_END_SENT) 
			/* 
			 * Collision of BGN PDUs.  Since this is not the 
			 * common case, yet not a protocol error, handle the
			 * PDUs as if they were received separately.  
			 * Ignore for now, as it is not handled correctly.
			 */
			break;
		else 
			/* Protocol error */
			break;
		/* Next, receive the data */
		m->m_data += sizeof(struct stpiphdr); 
		m->m_len -= sizeof(struct stpiphdr);
		m_adj(m, 8); /* Strip off BGN header information */
		sbappend(&so->so_rcv, m);
		sorwakeup(so);
		sp->rcv_max = SEQ_INC(sp->rcv_max);	
		sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);	
		/* 
		 * Do not handle END; information that END was
		 * received is already captured in the current state.
		 */
		break;

	case STP_PDU_BGNAK:
		stp_rcvbgnak(so, sp, m);
		m_freem(m);
		break;

	case STP_PDU_BGNAK_SD:
		stp_rcvbgnak(so, sp, m);
		m->m_data += sizeof(struct stpiphdr); 
		m->m_len -= sizeof(struct stpiphdr);
		m_adj(m, 8);
		sbappend(&so->so_rcv, m);
		sorwakeup(so);
		sp->rcv_max = SEQ_INC(sp->rcv_max);	
		sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);	
		break;

	case STP_PDU_END:
		stp_rcvend(so, sp, seqno);
		m_freem(m);
		break;

	case STP_PDU_ENDAK_SD_END:
		/* 
		 * First, we must treat the PDU as an implicit BGNAK, 
		 * and initialize receive variables.
		 * Next, receieve the endak (which will kill timer), 
		 * then fall through.
		 */
		sp->rcv_max = sp->rcv_irs = sp->rcv_nxt = seqno;
		stp_rcvendak(so, sp, seqno);        
		/* FALL THROUGH */

	case STP_PDU_SD_END:
		/* receive the data before receiving the END */
		m->m_data += sizeof(struct stpiphdr); 
		m->m_len -= sizeof(struct stpiphdr);
		if (sp->s_state == STP_EST || 
		    sp->s_state == STP_HALF_CLOSE_PENDING ||
		    sp->s_state == STP_HALF_CLOSED) {
			if ( (seqno == sp->rcv_max) &&
		 	    (sp->rcv_max == sp->rcv_nxt) && 
			    (si->si_len <= sbspace(&so->so_rcv)) &&
			    (sp->rcv_wnd) ) {
				/* Main (fast path) processing */
				sbappend(&so->so_rcv, m);
				sorwakeup(so);
				sp->rcv_max = SEQ_INC(sp->rcv_max);	
				sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);	
				/* Now process END */
				stp_rcvend(so, sp, SEQ_INC(seqno));
				break;
			} 
		} 
		/* 
		 * Not in sequence.  Process SD, but don't send ENDAK 
		 * yet 
		 * Note:  We do not check if data beyond this seqno is 
		 * subsequently received 
		 */
		stp_rcvsd(si, so, sp, m, seqno, fPoll);
		sp->end_received = TRUE;
		break;

	case STP_PDU_ENDAK:
		stp_rcvendak(so, sp, seqno);
		m_freem(m);
		break;

	case STP_PDU_RST:
		stp_rcvrst(so, sp, m);
		m_freem(m);
		break;

	default:
#ifdef STP_DEBUG
		printf("Unknown packet type %x\n", pkttype);
#endif
		m_freem(m);
		break;
	} /* end switch(pkttype) */

	return;
	
drop:
	m_freem(m);
#ifdef STP_DEBUG
	printf("Dropped\n");
#endif
	return;
}

/*
 * Pull out of band byte out of a segment so
 * it doesn't appear in the user's data queue.
 * It is still reflected in the segment length for
 * sequencing purposes.
 * Note:  Unlike TCP, the OOB byte is the first and only data byte of the mbuf.
 */
void
stp_pulloutofband(so, si, m)
	struct socket *so;
	struct stpiphdr *si;
	register struct mbuf *m;
{
	char *cp = mtod(m, caddr_t);
	struct stpcb *sp = sotostpcb(so);
	
	sp->s_iobc = *cp;
	sp->s_oobflags |= STPOOB_HAVEDATA;
	m->m_len--;

}

int 
acked_byte_count(sp, seqno)
	register struct stpcb *sp;	
	u_long seqno;
{
	int wrap = 0;
	int seqno_mod, snd_una_mod;
	int count = 0;
	int i;

	/* 
	 * seqno is the next in sequence expected (by the receiver), so 
	 * (seqno - 1) is actually the last one acked.  
	 * snd_una is the first one acked. 
	 */
	seqno_mod = seqno % MAXBUFFER;
	snd_una_mod = sp->snd_una % MAXBUFFER;

	if (seqno_mod < snd_una_mod) 
		wrap = 1; 	/* wrap around modulo buffer */

	if (!wrap) {
		for (i = snd_una_mod; i < seqno_mod; i++)	
			count += sp->trans_buffer[i].num_bytes;
	} else {
		for (i = snd_una_mod; i < MAXBUFFER; i++)
			count += sp->trans_buffer[i].num_bytes;
		for (i = 0; i < seqno_mod; i++)
			count += sp->trans_buffer[i].num_bytes;
	}
	return count;
}

/*
 * This function places a received packet in the reassembly chain, and passes
 * any (newly) resequenced, contiguous data to the user process.  It returns
 * a value of 1 if an insertion was made, and a value of 0 if no insertion 
 * was made.
 */
int  
stp_reass(sp, si, m)
	struct stpcb *sp;
	struct stpiphdr *si;
	struct mbuf *m;
{
	u_long seqno = si->si_seq;
	int seqno_mod = si->si_seq % MAXBUFFER;
	struct stpiphdr *q;
	struct socket *so = sp->s_inpcb->inp_socket;
	
	/*
	 * Call with si==0 after becoming established to 
	 * force pre-ESTABLISHED data up to user socket.
	 */
	if (si==0)
		goto present;

	/* 
	 * Find a packet with a sequence number >= this one
	 */
	for (q = sp->seg_next; q != (struct stpiphdr *) sp; 
	    q = (struct stpiphdr *) q->si_next) {
		if (seq_gt(q->si_seq, seqno))
			break;
	}
	if (q->si_seq == seqno) { /* received duplicate */
		/* 
	 	 * Unlike TCP, fast path processing may fail (rcv_max > rcv_nxt)
		 * even if there is nothing in the reassembly buffer.  
		 * e.g. if first packet gets lost, and POLL advances rcv_max
		 */
		if (sp->seg_next == sp->seg_prev) {
			if ((seqno == sp->rcv_nxt) && 
			    (si->si_len <= sbspace(&so->so_rcv))
			    && sp->rcv_wnd) {
				sbappend(&so->so_rcv, m);
				sorwakeup(so);
				sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);
				if (seq_geq(seqno, sp->rcv_max))
					sp->rcv_max = SEQ_INC(sp->rcv_max);
				return (1);
			}
		} else {
#ifdef STP_DEBUG
			printf("Received duplicate retransmission\n");
#endif
			m_freem(m);
			return (0);
		}
	}
	
	/* Want to avoid doing m_pullup on incoming packets, so use this macro*/
	STP_REASS_MBUF(si) = m;
	insque(si, q->si_prev);  /* Stick new packet in its place.*/
	sp->recv_buffer[seqno_mod].seqno = seqno;
	sp->recv_buffer[seqno_mod].received = 1;
	sp->recv_buffer[seqno_mod].num_bytes = m->m_len;

present:
	/*
	 * Present data to user, advancing rcv_nxt through
	 * completed sequence space.
	 */
	if (sp->s_state < STP_EST)
		return (1);
	si = sp->seg_next;
	/* if the seqno of the first segment on list != rcv_nxt, return */
	if (si == (struct stpiphdr *) sp || si->si_seq != sp->rcv_nxt)
		return (1);
	do {
		sp->rcv_nxt = SEQ_INC(sp->rcv_nxt);
		remque(si);
		m = STP_REASS_MBUF(si);
		si = (struct stpiphdr *) si->si_next;
		if (so->so_state & SS_CANTRCVMORE || (m->m_len == 0))
			m_freem(m);
		else
			sbappend(&so->so_rcv,m);
	} while (si != (struct stpiphdr *) sp && si->si_seq == sp->rcv_nxt);
	sorwakeup(so);
	if (sp->ustat_pending) {
		if (seq_geq(sp->rcv_nxt, sp->ustat_list2))
			sp->ustat_pending = 0;
	}
	return (1); 
}

/* 
 * Note that this implementation does not make use of the implicit 
 * acknowledgments between acks to free data from send buffer.
 * Call current_time() once here for the whole list.
 * maxrtt is the (liberal) estimate for the link's current rtt
 */
void
stp_process_list(sp, dataptr, num_elements, maxrtt)
	struct stpcb *sp;
	u_long *dataptr;
	int num_elements;
	u_long maxrtt;
{
	u_long curtime = current_time();
	u_long list1, list2;
	int i_mod;
	while (num_elements > 1) {
		dataptr++;
		list1 = ntohl(*dataptr);
		dataptr++;
		list2 = ntohl(*dataptr);
		num_elements -= 2;
		if (seq_lt(list1, list2) && seq_geq(list1, sp->snd_una) &&
		    seq_leq(list2, sp->snd_nxt))
			stp_retransmit_list(sp, list1, list2, STAT, maxrtt, 
			    curtime);
		else {
#ifdef STP_DEBUG
			printf("bad elements l1 %x l2 %x una %x nxt %x\n", 
			    list1, list2, sp->snd_una, sp->snd_nxt);
#endif
		}
	}
}

void
stp_retransmit_list(sp, list1, list2, retransmit_type, maxrtt, curtime)
	struct stpcb *sp;
	u_long list1;
	u_long list2;
	int retransmit_type;
	u_long maxrtt;
	u_long curtime;
{
	u_long i;
	int i_mod;
	u_long temp;
	u_char retransmitted = 0;

	/* Always retransmit upon a USTAT, unless packet already retransmitted*/
	if (retransmit_type == USTAT) {
		for (i = list1; seq_lt(i,list2); i = SEQ_INC(i))  {
			i_mod = i % MAXBUFFER;
			if (1 == sp->trans_buffer[i_mod].num_trans) {
				(void) stp_retransmit(sp,i);
				retransmitted = 1;
			}
		}
	} else {
		for (i = list1; seq_lt(i,list2); i = SEQ_INC(i)) {
			i_mod = i % MAXBUFFER;
			temp = curtime - sp->trans_buffer[i_mod].timesent;
			if (temp > ( (maxrtt + (2 * (sp->s_rttvar >> 
			    SRTT_SHIFT)) ) << sp->snd_backoff) ) {
				(void) stp_retransmit(sp,i);
				retransmitted = 1;
			} 
		}
	}
	if (retransmitted) {
		/* Close the congestion window and reset send timer intervals */
		stp_closecwnd(sp);
		stp_set_output(sp);
	}
}

/*
 * We enter this routine if fast path processing fails, or if data was
 * tacked on to a connection control PDU; cases:
 * i) state not one in which data can be received 
 * ii) seqno == sp->rcv_nxt; check buffer/window; present data to user
 * iii) seqno >= sp->rcv_max; make sure buffer and window available
 * iv) rcv_nxt < seqno < rcv_max; insert
 * v) seqno > rcv_max; check buffer and window  
 */
void            
stp_rcvsd(si, so, sp, m, seqno, fPoll)
	struct stpiphdr *si;
	struct socket *so;  
	struct stpcb *sp;       
	struct mbuf *m;  
	u_long seqno;   
	flag fPoll;   
{
	int seqno_mod = seqno % MAXBUFFER;
	u_long i, i_mod;
	int ooblen = 0;
        
	/* check if arrival of packet is legal in this state */
	if (sp->s_state != STP_EST) {
		if (!(sp->s_state & (STP_WAIT_FOR_DATA | 
		    STP_WAIT_FOR_DATA_CLOSE | STP_HALF_CLOSE_PENDING |
		    STP_HALF_CLOSED | STP_FULL_CLOSE_PENDING | 
		    STP_FULL_CLOSED))) {
			m_freem(m);
			return;
		}
	}
	if (seqno == sp->rcv_max) {
		/* Packet is next expected non-missing packet */
		if (seq_subtract(seqno, sp->rcv_nxt) < sp->rcv_wnd) {
			/* sufficient window to accept it */
			sp->rcv_max = SEQ_INC(sp->rcv_max);
			if (!stp_reass(sp, si, m)) {
				/* Do nothing */;
#ifdef STP_DEBUG
				printf("FAILED TO INSERT seqno %x 
				    sp->rcv_nxt %x\n", seqno, sp->rcv_nxt);
#endif
			}
			/* Check if USTAT pending */
			if (sp->ustat_pending)
				sp->ustat_pending++;
			if ((sp->ustat_pending == NUM_DUP_ACKS) && !fPoll) {
				if (stp_ustat_verify(sp))
					stp_sndustat(sp);
				sp->ustat_pending = 0;
			}
		} else {
			/* Overshot receiver buffer space */
#ifdef STP_DEBUG
			printf("Packet exceeds window: seqno %x rcv_nxt %x,
			    rcv_wnd %d\n", seqno, sp->rcv_nxt, sp->rcv_wnd);
#endif
			m_freem(m);
			return;
		}
	} else if (seqno == sp->rcv_nxt) {
		/* 
		 * Received the next in sequence; pass to
		 * reassembly buffer, which will update vars.
		 */
		if (!stp_reass(sp, si, m)) {
			/* Do nothing */;
#ifdef STP_DEBUG
			printf("FAILED TO INSERT seqno %x sp->rcv_nxt %x\n", 
			    seqno, sp->rcv_nxt);
#endif
		}
	} else if (seq_gt(seqno, sp->rcv_max)) {
		if (sp->ustat_pending && !fPoll && stp_ustat_verify(sp)) 
			stp_sndustat(sp);
		/* Next, mark a new ustat pending or send ustat*/
		sp->ustat_list1 = sp->rcv_max;
		sp->ustat_list2 = seqno;
		if (NUM_DUP_ACKS > 1) 
        		sp->ustat_pending = 1;
		else {
			stp_sndustat(sp);
			sp->ustat_pending = 0;
		}

		/* Clear out "received" flag for missing packets */
		for (i = sp->rcv_max; seq_lt(i, seqno); i = SEQ_INC(i)) {
			i_mod = i % MAXBUFFER;
			sp->recv_buffer[i_mod].seqno = i;
			sp->recv_buffer[i_mod].received = 0;
		}
		/* Mark this one as received */
		if (!stp_reass(sp, si, m)) {
			/* Do nothing */;
#ifdef STP_DEBUG
			printf("FAILED TO INSERT seqno %x sp->rcv_nxt %x\n", 
			    seqno, sp->rcv_nxt);
#endif
		}
		sp->rcv_max = SEQ_INC(seqno);
	} else if (seq_gt(seqno, sp->rcv_nxt)) {
		/* 
		 * sp->rcv_nxt < seqno < sp->rcv_max 
		 * First, may need to update so->so_oobmark. 
		 */ 
		if (seq_lt(seqno, sp->rcv_up)) {
			for (i = sp->rcv_nxt; seq_lt(i, sp->rcv_up); 
			    i = SEQ_INC(i)) {
				i_mod = i % MAXBUFFER;
				if (sp->recv_buffer[i_mod].received)
					ooblen += 
					    sp->recv_buffer[i_mod].num_bytes;
				else
					ooblen += sp->s_peermaxseg;
			}
			so->so_oobmark = so->so_rcv.sb_cc + ooblen;
		}
		if (!stp_reass(sp, si, m)) { /* New pkt for buffer*/
			/* Do nothing */;
#ifdef STP_DEBUG
			printf("FAILED TO INSERT (probably duplicate) seqno %x 
			    sp->rcv_nxt %x\n", seqno, sp->rcv_nxt);
#endif
		}
	} else {
		/* seqno < rcv_nxt */
#ifdef STP_DEBUG
		printf("Dropping it; seqno %x < rcv_nxt %x\n",seqno, 
		    sp->rcv_nxt);
#endif
		m_freem(m);
	}
}   

/* If we've been waiting to send a USTAT, verify that it is still needed */
int
stp_ustat_verify(sp)
	struct stpcb *sp;
{
	int i, i_mod, gap_broken;

	gap_broken = 0;
	for (i = sp->ustat_list1; seq_lt(i, sp->ustat_list2); i = SEQ_INC(i)) {
		i_mod = i % MAXBUFFER;
		if (1 == sp->recv_buffer[i_mod].received)
			gap_broken = 1;
	}
	return (!gap_broken);
}

void 
stp_rcvpoll(sp, m, seqno)
	struct stpcb *sp;
	struct mbuf *m;
	u_long seqno;
{
	struct socket *so;
	u_long *dataptr;
	u_long i, i_mod, rcv_max_old;

	/* Get pointer to payload of POLL packet */
	dataptr = (u_long *) (mtod(m,caddr_t) + sizeof(struct stpiphdr));
	so = sp->s_inpcb->inp_socket;
	switch (sp->s_state) {
	case STP_HALF_CLOSE_PENDING:
	case STP_HALF_CLOSED:
	case STP_FULL_CLOSE_PENDING:
	case STP_FULL_CLOSED:
	case STP_EST:{
		u_long timestamp;
		timestamp = ntohl(*dataptr);

		if (sp->ustat_pending)
			sp->ustat_pending = 0;

		if(seqno != sp->rcv_max) {
			/* If seqno is within allowable receive window
		         * set rcv_max to it; otherwise, set to top of 
			 * window. 
			 */
			rcv_max_old = sp->rcv_max;
			if (seq_lt(seqno, 
			    seq_add(sp->rcv_nxt, sp->rcv_wnd))) 
				sp->rcv_max = seqno;
			else 
				sp->rcv_max = 
				    seq_add(sp->rcv_nxt, sp->rcv_wnd);
			/* 
			 * Clear out "received" flag for missing 
			 * packets
			 */
			for (i = rcv_max_old; seq_lt(i, sp->rcv_max); 
			    i = SEQ_INC(i)) {
				i_mod = i % MAXBUFFER;
				sp->recv_buffer[i_mod].seqno = i;
				sp->recv_buffer[i_mod].received = 0;
			}
		}
		if ((sp->rcv_max == sp->rcv_nxt) && sp->end_received)
			/* Now done; respond to previously sent END */
			stp_rcvend(so, sp, sp->rcv_nxt);
		else
			stp_sndstat(sp, timestamp);
		break;
	}
	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
	case STP_BGN_DATA_END_SENT:
	case STP_LISTEN:
	case STP_WAIT_FOR_DATA:
	case STP_WAIT_FOR_DATA_CLOSE:
	case STP_CLOSE_RECV:
	case STP_LAST_ACK:
		break;
	default:
		/* Protocol error */
		stp_drop(sp, EINVAL);
		break;
	} /* switch */
}

void 
stp_rcvstat(so, sp, m, seqno)
	struct socket *so;
	struct stpcb *sp;
	struct mbuf *m;
	u_long seqno;
{
	u_long *dataptr, timestamp, elapsed, maxrtt;
	int acked, slen;
	struct stpiphdr *si;
	int num_elements = 0;

	/* Note:  The use of slen helps in protecting against buggy 3com cards*/
	si = mtod(m, struct stpiphdr *);
	slen = ((struct ip *) si)->ip_len;
	switch (sp->s_state) {
	case STP_EST:
	case STP_CLOSE_RECV:

		/* Get pointer to payload of STAT packet */
		dataptr = (u_long *) (mtod(m,caddr_t) + 
		    sizeof(struct stpiphdr));
		timestamp = ntohl(*dataptr);
		dataptr++;
		elapsed = time_elapsed(timestamp);
		/* Make sure we're not responding to an old STAT */
		if (sp->last_poll_tstamp >= timestamp)  
			break;
		/* 
		 * maxrtt holds the maximum of s_srtt and elapsed 
		 * The use of maxrtt helps to prevent spurious 
		 * retransmissions if the RTT of the connection 
		 * suddenly jumps.
		 */
		maxrtt = (elapsed < (sp->s_srtt >> SRTT_SHIFT) ? 
		    (u_long) (sp->s_srtt >> SRTT_SHIFT) : elapsed);
		stp_updatertt(sp, time_elapsed(timestamp));
		sp->last_poll_tstamp = timestamp;
		sp->snd_wnd = ntohl(*dataptr);
		if (sp->snd_wnd && (sp->s_timer & STP_TIMER_PERSIST)) 
			/* 
			 * credit to send more data now. set_output() 
			 * called below if data is pending, which will 
			 * turn on SEND timer and POLL.
			 */
			stp_cancel_timer(sp, STP_TIMER_PERSIST);

		if (seq_gt(seqno, sp->snd_una)) { /* New data acked */
			acked = acked_byte_count(sp, seqno);
			if (acked < 0 || acked > (sp->s_maxseg * 
			    seq_subtract(seqno, sp->snd_una))) {
				/* Do nothing */;
#ifdef STP_DEBUG
				printf("acked problem %d seqno %x 
				    snd_una %x\n", acked, seqno, 
				    sp->snd_una);
#endif
			}
			sbdrop(&so->so_snd, acked);
			sp->snd_off -= acked;
			if (sp->snd_up > 0) {
				sp->snd_up -= acked;
				if (sp->snd_up < 0)
					sp->snd_up = 0;
			}
			stp_opencwnd(sp, seq_subtract(seqno, 
			    sp->snd_una));
			if (so->so_snd.sb_flags & SB_NOTIFY)
				sowwakeup(so);
		}
		sp->snd_una = seqno;

		/* Find out how many list elements */
		num_elements = (slen - sizeof(struct stphdr) - 8) / 4;
		if (num_elements == 1) {
#ifdef STP_DEBUG
			printf("Dropping due to num_elements =1\n");
#endif
			stp_drop(sp, EINVAL);
			break;
		}
		/* Sanity check the number of elements (not done yet) */
		if (num_elements) {
			stp_process_list(sp, dataptr, num_elements, maxrtt);
		}

		/* check if we want to reduce POLLing frequency */
		if ((sp->snd_nxt == sp->snd_una) && 
		    (so->so_snd.sb_cc == 0)) {
			sp->poll_pending = 0;
			sp->send_pending = 0;
			stp_cancel_timer(sp, STP_TIMER_SEND);
			if (sp->disc_pending) 
				stp_disconnect(sp);
			if (sp->shutdown_pending)
				stp_shutdown(sp);
		} else {
			stp_set_output(sp);
		}	
		break;

	case STP_HALF_CLOSE_PENDING:
	case STP_HALF_CLOSED:
	case STP_FULL_CLOSE_PENDING:
	case STP_FULL_CLOSED:
	case STP_LAST_ACK:
		/* 
		 * Although we have sent END, old POLLs may still be 
		 * being answered, so ignore them. 
		 */
		break;

	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
		/* 
		 * If BGAK has been lost and more than one packet has 
		 * been sent, this may be a legal reception.  Ignore 
		 * it (wait for BGNAK).
		 */
		break;

	case STP_BGN_DATA_END_SENT:
	case STP_LISTEN:
	case STP_WAIT_FOR_DATA:
	case STP_WAIT_FOR_DATA_CLOSE:
	default:
		/* No POLLs should be outstanding in these states */
		stp_drop(sp, EINVAL);
		break;
	} /* switch */
}

void 
stp_rcvustat(so, sp, m, seqno)
	struct socket *so;
	struct stpcb *sp;
	struct mbuf *m;
	u_long seqno;
{
	u_long *dataptr, list1, list2;
	int acked;

	/* Get pointer to payload of USTAT packet */
	dataptr = (u_long *) (mtod(m,caddr_t) + sizeof(struct stpiphdr));

	switch (sp->s_state) {
	case STP_HALF_CLOSE_PENDING:
	case STP_HALF_CLOSED:
	case STP_FULL_CLOSE_PENDING:
	case STP_FULL_CLOSED:
	case STP_EST:
		list1 = ntohl((u_long) *dataptr);
		dataptr++;
		list2 = ntohl((u_long) *dataptr);
		stp_retransmit_list(sp, list1, list2, USTAT, 0, 0);     
		if (seq_gt(seqno, sp->snd_una)) { /* New data acked */
			acked = acked_byte_count(sp, seqno);		
			sbdrop(&so->so_snd, acked);
			sp->snd_off -= acked;
			/* opencwnd here?? */
			stp_opencwnd(sp, seqno - sp->snd_una);
		}
		sp->snd_una = seqno;
		break;

	case STP_LAST_ACK:
	case STP_CLOSE_RECV:
		/* 
		 * Although peer sender has closed, USTAT may have 
		 * been misordered by the network
		 */
		break;

	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
		/* 
		 * If BGAK has been lost and more than one packet has 
		 * been sent, this may be a legal reception.  Ignore 
		 * it (wait for BGNAK).
		 */
		break;
	
	case STP_BGN_DATA_END_SENT:
	case STP_LISTEN:
	case STP_WAIT_FOR_DATA:
	case STP_WAIT_FOR_DATA_CLOSE:
	default:
		/* This packet is an illegal reception */
		stp_drop(sp, EINVAL);
		break;
	} /* switch */
}

void 
stp_rcvbgn(so, sp, m, inp, fRespond)
	struct socket *so;
	struct stpcb *sp;
	struct mbuf *m;
	struct inpcb *inp;
	flag fRespond;
{
	u_long *dataptr;
	u_long seqno;
	struct stpiphdr *si;
	struct in_addr laddr;

	/* Get pointers to stpiphdr and data */
	si = mtod(m, struct stpiphdr *);
	seqno = si->si_seq;
	dataptr = (u_long *) (mtod(m,caddr_t) + sizeof(struct stpiphdr));

	switch (sp->s_state) {
 	case STP_LISTEN:{
		u_long timestamp;
		struct mbuf *am;
		struct sockaddr_in *sin;

		/* Extract payload fields */
		sp->snd_wnd = ntohl(*dataptr);
		dataptr++;
		timestamp = ntohl(*dataptr);

		/*
		 * get mbuf for client's address and port.  set local
		 * address in pcb.  connect pcb to peer.  Allocate and
		 * initialize IP and TCP header in template.  initialize
		 * pcb.  Ack the BGN and change state.
		 */
		am = m_get(M_DONTWAIT, MT_SONAME);
		am->m_len = sizeof(struct sockaddr_in);
		sin = mtod(am, struct sockaddr_in *);
		sin->sin_family = AF_INET;
		sin->sin_len = sizeof(*sin);
		sin->sin_addr = si->si_src;
		sin->sin_port = si->si_sport;
		bzero((caddr_t) sin->sin_zero, sizeof(sin->sin_zero));
		
		laddr = inp->inp_laddr;
		if (inp->inp_laddr.s_addr == INADDR_ANY)
			inp->inp_laddr = si->si_dst;
		if (in_pcbconnect(inp, am)) {
			inp->inp_laddr = laddr;
			(void) m_free(am);
			m_freem(m);
#ifdef STP_DEBUG
			printf("Connection dropped\n");
#endif
			return;
		}
		(void) m_free(am);
			
		sp->s_template = stp_template(sp);
		if (sp->s_template == 0) {
			sp = stp_drop(sp, ENOBUFS);
			m_freem(m);
#ifdef STP_DEBUG
			printf("Connection dropped\n");
#endif
			return;
		}
		/* connect socket */
		soisconnected(so);
		stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
		
		/* initialize sp */
		sp->s_instance = ntohs(si->si_inc);
		sp->rcv_max = sp->rcv_irs = sp->rcv_nxt = seqno;
		sp->snd_wnd = ntohl(*dataptr);

		if (fRespond) {
			if (WAIT_FOR_DATA) {
				sp->s_state = STP_WAIT_FOR_DATA;
				stp_start_timer(sp, STP_TIMER_WAIT_BGNAK, NULL);
				sp->bgnak_timestamp = timestamp;
			} else {
				sp->s_state = STP_EST;
				sp->bgnak_timestamp = timestamp;
				stp_sndcc_open(sp, STP_PDU_BGNAK);
			}
		} else {
			/* BGN_SD_END received */
			if (WAIT_FOR_DATA) {		
				sp->s_state = STP_WAIT_FOR_DATA_CLOSE;
				stp_start_timer(sp, STP_TIMER_WAIT_BGNAK, NULL);
				sp->bgnak_timestamp = timestamp;
			} else 
				sp->s_state = STP_EST;
		}
#ifdef STP_STATS
		sp->starttime = timestamp;
#endif
		break;
	}
	case STP_WAIT_FOR_DATA:
 	case STP_WAIT_FOR_DATA_CLOSE:
 	case STP_EST:{
		/* Peer must have lost our BGNAK */
		u_long timestamp;

		/* Extract payload fields */
		sp->snd_wnd = ntohl(*dataptr);
		dataptr++;
		timestamp = ntohl(*dataptr);

		/* Send BGNAK */
		if (fRespond) {
			sp->bgnak_timestamp = timestamp;
			stp_sndcc_open(sp, STP_PDU_BGNAK);
		}
		break;
	}
	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
	case STP_BGN_DATA_END_SENT:
		/* 
		 * Simultaneous open; adopt the lower of the two 
		 * connection IDs
		 */
		if (sp->s_instance > ntohs(si->si_inc))
			sp->s_instance = ntohs(si->si_inc);
		stp_rcvbgnak(so, sp, m);
		break;
	default:
		/* closing states; do nothing */
		break;
	} /* switch */
}


void 
stp_rcvbgnak(so, sp, m)
	struct socket *so;
	struct stpcb *sp;
	struct mbuf *m;
{
	u_long *dataptr;
	struct stpiphdr *si;
	u_long timestamp;
	u_long seqno;
	int acked;

	/* Get pointer to stpiphdr */
	si = mtod(m, struct stpiphdr *);
	seqno = si->si_seq;
	dataptr = (u_long *) (mtod(m,caddr_t) + sizeof(struct stpiphdr));

	switch (sp->s_state) {
	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
		/* Initialize receive variables */
		sp->rcv_max = sp->rcv_irs = sp->rcv_nxt = seqno;
		sp->snd_wnd = ntohl(*dataptr);
		sp->s_state = STP_EST;
		/* connect socket */
		soisconnected(so);
		stp_cancel_timer(sp, STP_TIMER_BGN);
		sp->cc_count = 0;

		dataptr++; 
		timestamp = ntohl(*dataptr);
		/* Note: this first RTT estimate may be long */  
		stp_updatertt(sp, time_elapsed(timestamp));

		if (sp->s_state == STP_BGN_DATA_SENT) {
			/* Implicit ACK of first packet */
			acked = acked_byte_count(sp, seqno);
			sbdrop(&so->so_snd, acked);
			sp->snd_off -= acked;
			stp_opencwnd(sp, 1);
			sp->snd_una = SEQ_INC(sp->snd_una);
		}

		stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
		stp_set_output(sp);
#ifdef STP_STATS
		sp->datatime = current_time();
#endif
		break;

	case STP_BGN_DATA_END_SENT:
	case STP_LISTEN:
	case STP_WAIT_FOR_DATA:
	case STP_WAIT_FOR_DATA_CLOSE:
#ifdef STP_DEBUG
		printf("protocol error, bgnak recv'd in %x state\n",
		    sp->s_state);
#endif
		break;
	default:
		/* states LISTEN, WAIT_FOR_DATA, EST, closing states */
		break;
	} /* switch */
}

void 
stp_rcvend(so, sp, seqno)
	struct socket *so;
	struct stpcb *sp;
	u_long seqno;
{
	/* Drop if haven't yet received all of the data */
	if (seqno != sp->rcv_nxt)
		return;
	switch (sp->s_state) {
	case STP_EST:
		/* Mark socket as write-only */ 
		socantrcvmore(so);
		sp->s_state = STP_CLOSE_RECV;
		stp_sndcc_close(sp, STP_PDU_ENDAK);
		break;

	case STP_HALF_CLOSE_PENDING:
		socantrcvmore(so);
		/* FALL THROUGH */

	case STP_FULL_CLOSE_PENDING:
		/* Have sent END, now receive END (end collision) */
		sp->s_state = STP_LAST_ACK;
		stp_sndcc_close(sp, STP_PDU_ENDAK);
		break;

	case STP_FULL_CLOSED:
	case STP_HALF_CLOSED:
		/* Connection is now completely closed */	
		stp_sndcc_close(sp, STP_PDU_ENDAK);
		sp = stp_close(sp);
		break;

	case STP_CLOSE_RECV:
	case STP_LAST_ACK:
		/* ENDAK was lost?? */
		stp_sndcc_close(sp, STP_PDU_ENDAK);
		break;

	case STP_LISTEN:
		/* BGN lost? Just ignore it */
		break;
	case STP_WAIT_FOR_DATA:
		sp->s_state = STP_WAIT_FOR_DATA_CLOSE;
		break;
	case STP_WAIT_FOR_DATA_CLOSE:
		/* Ignore it.  We're waiting for process to send data */
		break;

	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
	case STP_BGN_DATA_END_SENT:
		/* this is a protocol error */
		stp_drop(sp, EINVAL);
		break;
	default:
#ifdef STP_DEBUG
		printf("End recieved, not in EST state %d\n", sp->s_state);
#endif
		break;
	} /* switch */
}

void 
stp_rcvendak(so, sp, seqno)
	struct socket *so;
	struct stpcb *sp;
	u_long seqno;
{
	int acked;
	switch (sp->s_state) {
	case STP_LAST_ACK:
		sp = stp_close(sp);
		break;
	case STP_FULL_CLOSE_PENDING:
		soisdisconnected(so);
		stp_cancel_timer(sp, STP_TIMER_END);
		sp->cc_count = 0;
		sp->s_state = STP_FULL_CLOSED;
		stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
		break;

	case STP_HALF_CLOSE_PENDING:
		stp_cancel_timer(sp, STP_TIMER_END);
		sp->cc_count = 0;
		sp->s_state = STP_HALF_CLOSED;
		stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
		break;

	case STP_BGN_DATA_END_SENT:
		/* 
		 * ACK data.  We've sent one segment numbered 
		 * "sp->snd_una" 
		 */
		acked = acked_byte_count(sp, SEQ_INC(sp->snd_una));
		sbdrop(&so->so_snd, acked);
		sp->snd_off -= acked;
		stp_cancel_timer(sp, STP_TIMER_BGN);
		stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
		sp->s_state = STP_HALF_CLOSED;
		break;

	case STP_FULL_CLOSED:
	case STP_HALF_CLOSED:
		/* Might be retransmission of ENDAK */
		break;
	case STP_EST:
	case STP_CLOSE_RECV:
		if (sp->shutdown_pending || sp->disc_pending) {
			/* Could perform additional sanity checks here*/
			/* Stop data transfer timers */
			sp->poll_pending = 0;
			sp->send_pending = 0;
			stp_cancel_timer(sp, STP_TIMER_SEND);
			if (sp->disc_pending) {
				/* Check this */
				soisdisconnected(so);
			}
			if (sp->s_state == STP_EST) {
				stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
				sp->s_state = STP_FULL_CLOSED;
			} else { 
				/* Peer has closed */
				sp = stp_close(sp);
			}
			break;	
		}
		/* FALL THROUGH */

	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
	case STP_LISTEN:
	case STP_WAIT_FOR_DATA:
	case STP_WAIT_FOR_DATA_CLOSE:
		/* Haven't sent END */
#ifdef STP_DEBUG
		printf("Received ENDAK but haven't sent END\n");
#endif
		break;
	default:
		break;
	} /* switch */
}

/* Mimics TCP  */
void 
stp_rcvrst(so, sp, m)
	struct socket *so;
	struct stpcb *sp;
	struct mbuf *m;
{
	switch (sp->s_state) {
	case STP_BGN_SENT:
	case STP_BGN_DATA_SENT:
	case STP_BGN_DATA_END_SENT:
		so->so_error = ECONNREFUSED;
		goto close;
	case STP_WAIT_FOR_DATA:
	case STP_WAIT_FOR_DATA_CLOSE:
	case STP_EST:
	case STP_HALF_CLOSE_PENDING:
	case STP_HALF_CLOSED:
	case STP_FULL_CLOSE_PENDING:
	case STP_FULL_CLOSED:
		so->so_error = ECONNRESET;
	  close:
		sp->s_state = STP_CLOSED;
		sp = stp_close(sp);
		break;
	case STP_LAST_ACK:
	case STP_CLOSE_RECV:
		sp = stp_close(sp);
	}
}
#if ERROR_DETECTION
void
stp_rcverror(sp, seqno)
	struct stpcb *sp;
	u_long seqno;
{
	/* 
	 * Not yet defined.  The plan is to save a flag in the receiver array, 
	 * such as "received_in_error."  Then, when it is time to send a STAT 
	 * or USTAT message and report the gap, if all of the packets 
	 * corresponding to a gap in the STAT or USTAT message have this flag 
	 * set, the 25th bit in the list element is set to 1 on the list 
	 * elements that correspond to the gap.  At the sender, retransmission 
	 * occurs _without_ closing the congestion window.
	 *
	 * The main difficulty of this is guarding against sequence numbers that
	 * have been corrupted but still look valid.  For example, if the STP
	 * receiver could end up requesting packets that have not been sent, 
	 * causing protocol error.  So additional state variables, and logic
	 * protecting against protocol errors, still need to be worked out.
	 */
	  
	/* First, sanity check the received sequence number */
	if (seq_lt(seqno, sp->rcv_nxt) || 	
	    seq_geq(seqno, seq_add(sp->rcv_wnd, sp->rcv_nxt))) {
		/* Do nothing */ 
		return;
	}
	/* check if arrival of packet is legal in this state */
	if (sp->s_state != STP_EST) {
		if (!(sp->s_state & (STP_WAIT_FOR_DATA | 
		    STP_WAIT_FOR_DATA_CLOSE | STP_HALF_CLOSE_PENDING | 
		    STP_HALF_CLOSED | STP_FULL_CLOSE_PENDING | 
		    STP_FULL_CLOSED))) 
			return;
		
	}
	/* 
	 * Process packet by setting the appropriate flag in the receiver array 
	 * and updating state variables accordingly (NOT YET DEFINED).
	 */
	return;	
}
#endif /* ERROR_DETECTION */

#endif
