/*
 * Copyright (c) 1996, 1997 Berkeley Software Design, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that this notice is retained,
 * the conditions in the following notices are met, and terms applying
 * to contributors in the following notices also apply to Berkeley
 * Software Design, Inc.
 *
 *	BSDI tcp_output.c,v 2.11 1997/01/16 14:06:34 karels Exp
 */

/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993, 1995, 1996, 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tcp_output.c	8.4 (Berkeley) 5/24/95
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/errno.h>

#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#define	TCPOUTFLAGS
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>
#include <netinet/tcp_debug.h>

#ifdef notyet
extern struct mbuf *m_copypack();
#endif

#ifdef TCP_STATS
#include <netinet/tcp_stats.h>
extern       struct timeval time;
#endif

#ifdef SMART
extern int smart_enable;
#endif

#ifdef PERFECT_ELN
extern int perfect_eln_enable;
#endif

#if defined(SACK) || defined(ER)
extern int tcprexmtthresh;
#endif

#ifdef DPC
#include <net/if.h>

extern int advt_large_mss;
#endif

#ifdef TCP_SESSION
#include <netinet/tcp_session.h>
#endif

/*#define MAX_TCPOPTLEN	32*/	/* max # bytes that go in options */
#define MAX_TCPOPTLEN 40	/* need 40 at least for 3 SACKs + TIMESTAMP */

/*
 * Tcp output routine: figure out what should be sent and send it.
 */

#ifdef SACK
void
tcp_print_holes(struct tcpcb *tp)
{
	struct sackhole *p = tp->snd_holes;
	if (p == 0)
		return;
	while (p) {
		printf("\t%x--%x d %d r %x ",  p->start, p->end, p->dups, 
		       p->rxmit);
		p = p->next;
	}
	printf("\n");
}



#endif

#ifdef LIMIT_BURST
extern int hz;

void
tcp_output_dummy(void *arg)
{
	struct tcpcb *tp;
	struct timeval curtime;
		
	microtime(&curtime);
	tp = (struct tcpcb *) arg;
	tp->burst_pending = 0;
	tcp_output(tp);
}

/*
 * XXX tcp_sched_burst() is not effective when data from the socket layer arrives
 * in small chunks (i.e., when len is small). The correct long-term solution is
 * to keep track of the time when each packet is due to complete transmission
 * based on the connection rate of cwnd/rtt. Basically, the code will be very 
 * similar to the bandwidth emulation code under LINK_EMULATION.
 *
 * XXX todo: better solution -- have an array to keep track of the times at
 * which the last N packets were sent.
 */
int
tcp_sched_burst(struct tcpcb *tp, int len) 
{
	u_long ticks;    /* ticks to next burst */
	int burstlen; /* in segments */
	struct timeval curtime;

	if (tp->t_xflags & TXF_IGNORE_BURSTTIMER) {
		burstlen = (!len || len % tp->t_maxseg) ? (len/tp->t_maxseg + 1) : (len/tp->t_maxseg);
		return(burstlen);
	}
	microtime(&curtime);
	/* 
	 * cancel timer if pending
	 */
	if (tp->burst_pending) {
		untimeout(tcp_output_dummy, (void *) tp);
		tp->burst_pending = 0;
	}
	/* if amount to send is less than maxburst */
	if (len <= tp->maxburst*tp->t_maxseg) { 
		burstlen = (!len || len % tp->t_maxseg) ? (len/tp->t_maxseg + 1) : (len/tp->t_maxseg);
	}
	/* 
	 * amount to send more than maxburst, so send maxburst now and 
	 * schedule the rest for later.
	 *
	 * XXX should we consider snd_wnd when scheduling the time for the 
	 * next burst?
	 */
	else { 
		ticks = 
	((tp->t_srtt_exact)*tp->maxburst*tp->t_maxseg)/tp->snd_cwnd/(EXACT_HZ/hz);

		if (ticks == 0) 
			/* XXX we choose not to ignore maxburst in such cases */
			ticks = 1;

		tp->burst_pending = 1;
		timeout(tcp_output_dummy, (void *) tp, ticks);
		burstlen = tp->maxburst;
	}
	return (burstlen);
} 

int
tcp_output_force(tp)
	register struct tcpcb *tp;
{
	int retval;

	tp->t_xflags |= TXF_IGNORE_BURSTTIMER;
	retval = tcp_output(tp);
	tp->t_xflags &= ~TXF_IGNORE_BURSTTIMER;
	return(retval);
}

#endif

#ifdef SACK
struct sackhole *
tcp_sack_output(tp, sendalot)
	register struct tcpcb *tp;
	int sendalot;
{
	struct sackhole *p;
	/* 
	 * If we're explicitly trying to send out new data (when sendalot
	 * is 1 or when in ER_ACTIVE mode), bypass the hole-processing 
	 * steps that send out old segment(s).
	 */
	if (tp->sack_disable || sendalot)
		return 0;
#ifdef ER
	if (tp->t_xflags & TXF_ER_ACTIVE)
		return 0;
#endif
	p = tp->snd_holes;
	while (p) {
		if (p->dups >= tcprexmtthresh && SEQ_LT(p->rxmit, p->end)) {
			if (SEQ_LT(p->rxmit, tp->snd_una)) {/* old SACK hole */
				p = p->next;
				continue;
			}
#ifdef DEBUG
			if (p)
				tcp_print_holes(tp);
#endif
			return p;
		}
		p = p->next;
	}
	return 0;
}
#endif

int
tcp_output(tp)
	register struct tcpcb *tp;
{
	register struct socket *so = tp->t_inpcb->inp_socket;
	register long len, win;
	int off, flags, error;
	register struct mbuf *m;
	register struct tcpiphdr *ti;
	u_char opt[MAX_TCPOPTLEN];
	unsigned optlen, hdrlen;
	int idle, sendalot = 0;
	struct rtentry *rt;
#ifdef SACK
	int i, sack_rxmit = 0;
	struct sackhole *p;
#endif

#ifdef LIMIT_BURST
	/* 
	 * XXX todo: even if the burst timer is not pending, check to see if we
	 * sent data in a previous call to tcp_output() in the very recent past.
	 * This is needed because the application may do writes in small chunks
	 * (specifically, smaller than maxburst), resulting in multiple calls to 
	 * tcp_output in a short space of time.
	 */
	if (tp->burst_pending && 
	    !(tp->t_xflags & TXF_IGNORE_BURSTTIMER)) {
		struct timeval curtime;
		
		microtime(&curtime);
		/* 
		 * We have a choice between cancelling a scheduled burst and
		 * not doing so.
		 */
		if (tp->t_xflags & TXF_CANCEL_BURST)
			untimeout(tcp_output_dummy, (void *) tp);
		else
			return (0);
	}
#endif

	/*
	 * Determine length of data that should be transmitted,
	 * and flags that will be used.
	 * If there is some data or critical controls (SYN, RST)
	 * to send, then transmit; otherwise, investigate further.
	 */
	idle = (tp->t_flags & TF_WASIDLE) ? 1 : (tp->snd_max == tp->snd_una);
	if (idle) {
#ifdef FAST_START
		/* 
		 * TCP fast start tries to avoid incurring slow start overhead
		 * each time a connection restarts after an idle period. The
		 * old value to snd_cwnd (adjusted as described below) together
		 * with a fine-grained software timer is used to clock out
		 * packets during the initial RTT after fast start. Precautions
		 * are taken to guard against performance degradation in case
		 * the old value of snd_cwnd is inappropriate for the present
		 * network conditions. These precautions include:
		 * (a) Packets sent during the fast start phase are assigned
		 *     a high *drop* priority (i.e., low priority). The 
		 *     TH_FS flag in the TCP header, when set,  denotes low 
		 *     priority.
		 * (b) Special loss recovery procedures are in force during
		 *     the fast start phase. This enables teh sender to quickly
		 *     terminate a failed fast start attempt and fall back to
		 *     slow start. The TXF_FLR and TXF_FRT flags in the tcpcb
		 *     are used for this purpose.
		 *
		 * For questions/comments, please contact:
		 * Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
		 * http://www.cs.berkeley.edu/~padmanab
		 */

		/*
		 * Check to see if the connection is restarting after an idle 
		 * time. If fast start is not enabled, it is sufficient to do 
		 * the check (t_idle > t_rxtcur). However, if fast start is 
		 * enabled, we also need to do the check 
		 * (t_snd_idle > t_rxtcur), i.e. check if we have sent any 
		 * data in the recent past. This additional check is needed 
		 * because fast start can cause several packets to be sent out
		 * before the first among them is acked, and we would like to 
		 * prevent packets other than the first one from going through
		 * the "restart after idle time" code below.
		 */

		/* 
		 * XXX bug fix: it is wrong to check t_idle because it is updated
		 * whenever a packet is *received*, but that doesn't tell us
		 * anything about the forward/data path. So we only do a check
		 * on t_snd_idle.
		 */
		if (tp->t_snd_idle >= tp->t_rxtcur) {
			/* check if fast start is enabled */
			if (FS_ENABLED(tp)) {
#ifdef TCP_STANDARD
				u_long min_cwnd = tp->t_maxseg;
#else
				u_long min_cwnd = 2*tp->t_maxseg;
#endif
				/* 
				 * We set snd_cwnd to a "safe" value: snd_cwnd/2
				 * if the connection was in slow start before 
				 * the start of the idle period, and 
				 * (snd_cwnd - t_maxseg) if it was in congestion
				 * avoidance phase. The new value of cwnd_ is 
				 * then the last congestion window size at which
				 * we were able to transmit an entire window's 
				 * worth of data without experiencing a loss.
				 */
				/* slow start phase */
				if (tp->snd_cwnd <= tp->snd_ssthresh+1) 
					tp->snd_cwnd = 
						max(tp->snd_cwnd/2, min_cwnd);
				else  /* cong. avoidance phase */
					tp->snd_cwnd = 
						max(tp->snd_cwnd-tp->t_maxseg, 
						    min_cwnd);
				/* 
				 * Set snd_cwnd to at least the most recent
				 * successful congestion window size. Then 
				 * update the latter.
				 */
				tp->snd_cwnd = max(tp->snd_cwnd, tp->fs_cwnd);
				tp->fs_cwnd = tp->snd_cwnd;
				/* save ssthresh for use if fast start fails */
				tp->fs_ssthresh = tp->snd_ssthresh; 
				/* 
				 * The fast start phase, demarcated by 
				 * fs_startseq and fs_endseq extends for one 
				 * window of data. However, the first two 
				 * segments aren't considered part of the fast 
				 * start phase because they would have been sent
				 *  even without fast start. 
				 */
				tp->fs_startseq = tp->snd_una + min_cwnd;
				tp->fs_endseq = tp->snd_una + 
					min(tp->snd_wnd, tp->snd_cwnd);
				tp->t_xflags |= TXF_FS_MODE;
				/* 
				 * If TXF_FLR and TXF_FRT are enabled, enque 
				 * this connection's PCB to the tcp_faststart 
				 * queue and set the TCPT_RESET timer for this 
				 * connection.
				 */
				if (FS_RESET_PEND(tp)) {
					if (tp->t_timer[TCPT_RESET] == 0)
						tcp_set_reset_timer(tp);
				}
				else if (FS_FLR_FRT(tp)) {
					TAILQ_INSERT_TAIL(&tcp_faststart, tp, 
							  t_faststart);
					if (tcp_faststart.tqh_first && 
    (tcp_faststart.tqh_first == tcp_faststart.tqh_first->t_faststart.tqe_next))
						printf("something wrong in tcp_output()\n");
					tp->t_xflags |= TXF_FS_RESET_PEND;
					tcp_set_reset_timer(tp);
				}
			}
			else {
#ifdef TCP_STANDARD
				tp->snd_cwnd = tp->t_maxseg;
#else
				tp->snd_cwnd = 2*tp->t_maxseg;
#endif
			}
#ifdef TCP_STATS
			if ((tp->t_flags & TF_ENABLE_STATS)&&
			    (tcp_stats_flags & 
			     (TCP_STATS_SND_CWND | TCP_STATS_SND_MAX)))
				tcp_stats_record(TCP_STATS_SND_CWND|
						 TCP_STATS_SND_MAX,tp);
#endif
		}
#else		
		if (tp->t_idle >= tp->t_rxtcur) {
			/*
			 * We have been idle for "a while" and no acks are
			 * expected to clock out any data we send --
			 * slow start to get ack "clock" running again.
			 */
#ifdef TCP_STANDARD
			tp->snd_cwnd = tp->t_maxseg;
#else
			tp->snd_cwnd = 2*tp->t_maxseg;
#endif

#ifdef TCP_STATS
			if ((tp->t_flags & TF_ENABLE_STATS) && 
			    (tcp_stats_flags & (TCP_STATS_SND_CWND | TCP_STATS_SND_MAX)))
				tcp_stats_record(TCP_STATS_SND_CWND|TCP_STATS_SND_MAX,tp);
#endif
		}
#endif
		if (somoretosend(so)) {
			tp->t_flags |= TF_WASIDLE;
			idle = 0;
			goto again;
		}
	}
	tp->t_flags &= ~TF_WASIDLE;

 again:
	off = tp->snd_nxt - tp->snd_una;
	win = min(tp->snd_wnd, tp->snd_cwnd);
	flags = tcp_outflags[tp->t_state];

#ifdef TCP_SESSION
	if (SESSION(tp) && tp->session && tp->t_state != TCPS_TIME_WAIT &&
	    !tp->t_force) {
		/* 
		 * if called from session_output(), send at most 1 segment 
		 * of data 
		 */
		if (SESSION_OUTPUT(tp)) {
			win = min(tp->snd_wnd, off+tp->t_maxseg);
			if (tp->session->max_bytes_snd)
				win = min(win, off+tp->session->max_bytes_snd);
		}
		/*
		 * else if SIN/FIN/RST flag is set, call session_output_oneseg()
		 * to force a segment to be sent on *this* connection
		 */
		else if ((flags&(TH_SYN|TH_FIN|TH_RST)) != 0) {
			session_output_oneseg(tp->session,tp,0);
			return (0);
		}
		/* 
		 * else call session_output() to pick the connection(s) on which
		 * to send segments
		 */
		else {
			session_output(tp->session);
			return(0);
		}
	}
#endif
	
#ifdef SACK
	if (p = tcp_sack_output(tp, sendalot)) {
		off = p->rxmit - tp->snd_una;
		sack_rxmit = 1;
		len = min(tp->t_maxseg, p->end - p->rxmit);
	}

#endif

#ifndef LIMIT_BURST
	sendalot = 0;
#endif
	/*
	 * If in persist timeout with window of 0, send 1 byte.
	 * Otherwise, if window is small but nonzero
	 * and timer expired, we will send what we can
	 * and go to transmit state.
	 */
	if (tp->t_force) {
		if (win == 0) {
			/*
			 * If we still have some data to send, then
			 * clear the FIN bit.  Usually this would
			 * happen below when it realizes that we
			 * aren't sending all the data.  However,
			 * if we have exactly 1 byte of unset data,
			 * then it won't clear the FIN bit below,
			 * and if we are in persist state, we wind
			 * up sending the packet without recording
			 * that we sent the FIN bit.
			 *
			 * We can't just blindly clear the FIN bit,
			 * because if we don't have any more data
			 * to send then the probe will be the FIN
			 * itself.
			 */
			if (off < so->so_snd.sb_cc)
				flags &= ~TH_FIN;
			win = 1;
		} else {
			tp->t_timer[TCPT_PERSIST] = 0;
			tp->t_rxtshift = 0;
		}
	}

#ifdef SACK
	if (!sack_rxmit)
#endif
	{
		if (win < so->so_snd.sb_cc) {
			len = win - off;
			if (idle) {
				tp->t_flags |= TF_WASIDLE;
				idle = 0;
			}
		} else {
			len = so->so_snd.sb_cc - off;
#ifdef TCP_VERBOSE
/*
			if (len > tp->t_maxseg)
				printf("win > buffer:  len %d win %d rnxt %d runa %d 
						sb_cc %d off %d\n", len, win, tp->snd_nxt, 
						tp->snd_una, so->so_snd.sb_cc, off);
*/
#endif TCP_VERBOSE
		}
	}

	if (len < 0) {
		/*
		 * If FIN has been sent but not acked,
		 * but we haven't been called to retransmit,
		 * len will be -1.  Otherwise, window shrank
		 * after we sent into it.  If window shrank to 0,
		 * cancel pending retransmit, pull snd_nxt back
		 * to (closed) window, and set the persist timer
		 * if it isn't already going.  If the window didn't
		 * close completely, just wait for an ACK.
		 */
		len = 0;
		if (win == 0) {
			tp->t_timer[TCPT_REXMT] = 0;
#ifdef FAST_START
			tp->t_timer[TCPT_RESET] = 0;
#endif
			tp->t_rxtshift = 0;
			tp->snd_nxt = tp->snd_una;
			if (tp->t_timer[TCPT_PERSIST] == 0)
				tcp_setpersist(tp);
		}
	}

	if (len > tp->t_maxseg) {
#ifdef LIMIT_BURST
		if (!tp->burst_pending)
			sendalot = tp->maxburst ? tcp_sched_burst(tp, len)-1:1;
#else
		sendalot = 1;
#endif
		len = tp->t_maxseg;
	}
	if (SEQ_LT(tp->snd_nxt + len, tp->snd_una + so->so_snd.sb_cc))
		flags &= ~TH_FIN;

	win = sbspace(&so->so_rcv);

	/*
	 * Sender silly window avoidance.  If connection is idle
	 * and can send all data or a maximum segment,
	 * or are forced, do it; otherwise don't bother.
	 * If peer's buffer is tiny, then send
	 * when window is at least half open.
	 * If retransmitting (possibly after persist timer forced us
	 * to send into a small window), then must resend.
	 */
	if (len) {
		if (len == tp->t_maxseg)
			goto send;
		if ((idle || tp->t_flags & TF_NODELAY) &&
		    len + off >= so->so_snd.sb_cc)
			goto send;
		if (tp->t_force)
			goto send;
		if (len >= tp->max_sndwnd / 2)
			goto send;
		if (SEQ_LT(tp->snd_nxt, tp->snd_max))
			goto send;
#ifdef SACK
		if (!tp->sack_disable && sack_rxmit)
			goto send;
#endif

#ifdef ER
		if (tp->t_xflags & TXF_ER_ACTIVE)
			goto send;
#endif
	}

	/*
	 * Compare available window to amount of window
	 * known to peer (as advertised window less
	 * next expected input).  If the difference is at least two
	 * max size segments, or at least 50% of the maximum possible
	 * window, then want to send a window update to peer.
	 *
	 * Of course, if we've already received the FIN, there's
	 * no point in sending out a window update.
	 */
	if (win > 0 && !TCPS_HAVERCVDFIN(tp->t_state)) {
		/* 
		 * "adv" is the amount we can increase the window,
		 * taking into account that we are limited by
		 * TCP_MAXWIN << tp->rcv_scale.
		 */
		long adv = min(win, (long)TCP_MAXWIN << tp->rcv_scale) -
			(tp->rcv_adv - tp->rcv_nxt);
		u_short t_maxseg = tp->t_maxseg;

#ifdef DPC
		if (tp->peer_mss) {
			t_maxseg = tp->peer_mss;
		}
#endif

#ifdef ACC
		if (adv >= (long) (tp->delack * t_maxseg)) 
			goto send;
#else
		if (adv >= (long) (2 * t_maxseg))
			goto send;
#endif
		if (2 * adv >= (long) so->so_rcv.sb_hiwat)
			goto send;
	}

#ifdef TCP_STANDARD
#ifdef ACC
	if (tp->num_segs_since_ack >= tp->delack)
		goto send;
#else
	if (tp->num_segs_since_ack >= 2)
		goto send;
#endif
#endif

	/*
	 * Send if we owe peer an ACK.
	 */
	if (tp->t_flags & TF_ACKNOW) {
		goto send;
	}
	if (flags & (TH_SYN|TH_RST))
		goto send;
	if (SEQ_GT(tp->snd_up, tp->snd_nxt))
		goto send;
	/*
	 * If our state indicates that FIN should be sent
	 * and we have not yet done so, or we're retransmitting the FIN,
	 * then we need to send.
	 */
	if (flags & TH_FIN && ((tp->t_flags & TF_SENTFIN) == 0 ||
	    SEQ_LT(tp->snd_nxt, tp->snd_max)))
		goto send;

	/*
	 * TCP window updates are not reliable, rather a polling protocol
	 * using ``persist'' packets is used to insure receipt of window
	 * updates.  The three ``states'' for the output side are:
	 *	idle			not doing retransmits or persists
	 *	persisting		to move a small or zero window
	 *	(re)transmitting	and thereby not persisting
	 *
	 * tp->t_timer[TCPT_PERSIST]
	 *	is set when we are in persist state.
	 * tp->t_force
	 *	is set when we are called to send a persist packet.
	 * tp->t_timer[TCPT_REXMT]
	 *	is set when we are retransmitting
	 * The output side is idle when both timers are zero.
	 *
	 * If send window is too small, there is data to transmit, and no
	 * retransmit or persist is pending, then go to persist state.
	 * If nothing happens soon, send when timer expires:
	 * if window is nonzero, transmit what we can,
	 * otherwise force out a byte.
	 */
	if (so->so_snd.sb_cc && tp->t_timer[TCPT_REXMT] == 0 &&
	    tp->t_timer[TCPT_PERSIST] == 0
#ifdef FAST_START
	    && (!FS_FLR_FRT(tp) || tp->t_timer[TCPT_RESET] == 0)
#endif
		) {
		tp->t_rxtshift = 0;
		tcp_setpersist(tp);
	}

	/*
	 * No reason to send a segment, just return.
	 */
	return (0);

send:

	/*
	 * Before ESTABLISHED, force sending of initial options
	 * unless TCP set not to do any options.
	 * NOTE: we assume that the IP/TCP header plus TCP options
	 * always fit in a single mbuf, leaving room for a maximum
	 * link header, i.e.
	 *	max_linkhdr + sizeof (struct tcpiphdr) + optlen <= MHLEN
	 */
	optlen = 0;
	hdrlen = sizeof (struct tcpiphdr);

	if (flags & TH_SYN) {
		tp->snd_nxt = tp->iss;
		if ((tp->t_flags & TF_NOOPT) == 0) {
			u_short mss;

			opt[0] = TCPOPT_MAXSEG;
			opt[1] = 4;
			mss = htons((u_short) tcp_send_mss(tp->t_inpcb));

#ifdef DPC
			/*
			 * Advertise a large enough mss so that the peer's
			 * choice is not restricted. This is useful when
			 * the forward and reverse paths are asymmetric.
			 *
			 * (XXXX temporary)
			 */
			if (advt_large_mss)
			    mss = htons(1500);
#endif
			bcopy((caddr_t)&mss, (caddr_t)(opt + 2), sizeof(mss));
			optlen = 4;
	 
#ifdef SACK
			if (!tp->sack_disable && (tp->t_flags&TF_SACK_PERMIT)&&
			    ((flags & TH_ACK) == 0 ||
			     (tp->t_flags & TF_SACK_PERMIT))) {
				*((u_long *) (opt + optlen)) = 
					htonl(TCPOPT_SACK_PERMIT_HDR);
				optlen += 4;
			}
#endif
			if (tp->t_flags & TF_USE_SCALE) {
				*((u_long *) (opt + optlen)) = htonl(
					TCPOPT_NOP << 24 |
					TCPOPT_WINDOW << 16 |
					TCPOLEN_WINDOW << 8 |
					tp->request_r_scale);
				optlen += 4;
			}
		}
 	}

#ifdef MOBILE_HOST 
#ifdef SMART

	/*
	 * If smart_enable flag has been turned on -and- the connection is in
	 * TCPS_ESTABLISHED state -and- there is a gap between the last
	 * in-sequence packet received and the last packet received, then
	 * we set the TF_SMART option in the packet that we are just sending 
	 * out.
	 */
	if (smart_enable && (tp->t_state == TCPS_ESTABLISHED) &&
	    (SEQ_GT(tp->smart_start_last_rcv, tp->rcv_nxt))) {
		u_long *lp = (u_long *)(opt + optlen);
		
		/* 
		 * TF_SMART option just like tstamp option, 2 leading NOPs.
		 */
		*lp++ = htonl(TCPOPT_SMART_HDR);
		*lp++ = htonl(tp->smart_start_last_rcv);
		*lp = htonl(tp->smart_end_last_rcv);
		
		optlen += TCPOLEN_SMART_TOTAL;
	}
#endif /* SMART */
#endif /* MOBILE_HOST */

	/*
	 * Send a timestamp and echo-reply if this is a SYN and our side 
	 * wants to use timestamps (TF_SEND_TSTMP is set) or both our side
	 * and our peer have sent timestamps in our SYN's.
 	 */
#ifdef DPC
	if (tp->t_xflags & TXF_NO_TSTAMP)
		goto tsdone;
#endif
 	if ((tp->t_flags & TF_SEND_TSTMP) && (flags & TH_RST) == 0) {
		u_long *lp = (u_long *)(opt + optlen);
 
 		/* Form timestamp option as shown in appendix A of RFC 1323. */
 		*lp++ = htonl(TCPOPT_TSTAMP_HDR);
#ifdef FINE_GRAINED_TSTAMP
		{
			struct timeval curtime;
		
			microtime(&curtime);
			*lp++ = htonl(EXACT_TIME(&curtime));
		}
#else
 		*lp++ = htonl(tcp_now);
#endif
 		*lp   = htonl(tp->ts_recent);
 		optlen += TCPOLEN_TSTAMP_APPA;
 	}
#ifdef DPC
 tsdone:
#endif

#ifdef SACK
	/* 
	 * Send SACKs if necessary.  This should be the last option processed.
	 * Only as many SACKs are sent as are permitted by the maximum options
	 * size.  No more than three SACKs are sent.
	 */
	if (!tp->sack_disable && tp->t_state == TCPS_ESTABLISHED && 
	    (tp->t_flags & (TF_SACK_PERMIT|TF_NOOPT)) == TF_SACK_PERMIT &&
	    tp->rcv_numsacks) {
		u_long *lp = (u_long *) (opt + optlen);
		u_long *olp = lp++;
		int count = 0;	/* actual number of SACKs inserted (<= 3) */
		int maxsack = (MAX_TCPOPTLEN - (optlen + 4))/TCPOLEN_SACK;

		maxsack = min(maxsack, TCP_MAX_SACK);
		for (i=0; (i < tp->rcv_numsacks && count < maxsack); i++) {
		        struct sackblk sack = tp->sackblks[i];
			if (sack.start == 0 && sack.end == 0)
				continue;
			*lp++ = htonl(sack.start);
			*lp++ = htonl(sack.end);
			count++;
		}
		*olp = htonl(TCPOPT_SACK_HDR|(TCPOLEN_SACK*count+2));
		optlen += TCPOLEN_SACK*count + 4; /* including leading NOPs */
	}
#endif /* SACK */

#ifdef ACC
	/* 
	 * Use the TCPOPT_PEERWIN option to tell the peer what our window
	 * size (measured in segments) is. We don't set this option in a
	 * pure ACK packet.
	 *
	 * XXX to do: make this option conditional on the TH_DELACKS_OK flag
	 * being set during the SYN exchange.
	 */
	if (len > 0) {
		u_short *sp = (u_short *) (opt + optlen);
		/* 
		 * We do NOT take the min of snd_cwnd and and snd_wnd. 
		 * The receiver takes the min of the peerwin we send and
		 * the size of the of its receiver buffer for deciding
		 * how much to delay acks. The justification for this is
		 * that if in fact win is limited for us because the
		 * receiver's buffer is full, end-to-end throughput is not
		 * really impacted if the receiver waits, either until its 
		 * advertised window advances enough (as determined by
		 * "delack") or the delayed ack timer expires, before
		 * sending an new ack.
		 */
		long w = tp->snd_cwnd;
		*sp++ = htons(TCPOPT_PEERWIN_HDR);
		/* take the amount of data to send into account */
		*sp = htons(min(w, so->so_snd.sb_cc)/tp->t_maxseg);
		optlen += TCPOLEN_PEERWIN;
	}
#endif


 	hdrlen += optlen;

#ifdef already_accounted_for 
	/*
	 * Adjust data length if insertion of options will
	 * bump the packet length beyond the t_maxseg length.
	 */
	if (len > tp->t_maxseg - optlen) {
#ifdef LIMIT_BURST
		/* 
		 * We assume that the leftovers (caused by options) from each
		 * segment will add up to at most one segment. So we adjust
		 * sendalot exactly once per burst.
		 */
		if (!tp->burst_pending)
			sendalot++;
#else
		sendalot = 1;
#endif
		len = tp->t_maxseg - optlen;
		flags &= ~TH_FIN;
	}
#endif /* already_accounted_for */


#ifdef DIAGNOSTIC
 	if (max_linkhdr + hdrlen > MHLEN)
		panic("tcphdr too big");
#endif

	/*
	 * Grab a header mbuf, attaching a copy of data to
	 * be transmitted, and initialize the header from
	 * the template for sends on this connection.
	 */
	if (len) {
		if (tp->t_force && len == 1)
			tcpstat.tcps_sndprobe++;
		else if (SEQ_LT(tp->snd_nxt, tp->snd_max)) {
			tcpstat.tcps_sndrexmitpack++;
			tcpstat.tcps_sndrexmitbyte += len;
		} else {
			tcpstat.tcps_sndpack++;
			tcpstat.tcps_sndbyte += len;
		}
#ifdef notyet
		if ((m = m_copypack(so->so_snd.sb_mb, off,
		    (int)len, max_linkhdr + hdrlen)) == 0) {
			error = ENOBUFS;
			goto out;
		}
		/*
		 * m_copypack left space for our hdr; use it.
		 */
		m->m_len += hdrlen;
		m->m_data -= hdrlen;
#else
		MGETHDR(m, M_DONTWAIT, MT_HEADER);
		if (m == NULL) {
			error = ENOBUFS;
			goto out;
		}
		m->m_data += max_linkhdr;
		m->m_len = hdrlen;
		if (len <= MHLEN - hdrlen - max_linkhdr) {
			m_copydata(so->so_snd.sb_mb, off, (int) len,
			    mtod(m, caddr_t) + hdrlen);
			m->m_len += len;
		} else {
			m->m_next = m_copy(so->so_snd.sb_mb, off, (int) len);
			if (m->m_next == 0) {
				(void) m_free(m);
				error = ENOBUFS;
				goto out;
			}
		}
#endif
		/*
		 * If we're sending everything we've got, set PUSH.
		 * (This will keep happy those implementations which only
		 * give data to the user when a buffer fills or
		 * a PUSH comes in.)
		 */
		if (off + len == so->so_snd.sb_cc)
			flags |= TH_PUSH;
	} else {
		if (tp->t_flags & TF_ACKNOW)
			tcpstat.tcps_sndacks++;
		else if (flags & (TH_SYN|TH_FIN|TH_RST))
			tcpstat.tcps_sndctrl++;
		else if (SEQ_GT(tp->snd_up, tp->snd_nxt))
			tcpstat.tcps_sndurg++;
		else
			tcpstat.tcps_sndwinup++;

		MGETHDR(m, M_DONTWAIT, MT_HEADER);
		if (m == NULL) {
			error = ENOBUFS;
			goto out;
		}
		m->m_data += max_linkhdr;
		m->m_len = hdrlen;
	}
	m->m_pkthdr.rcvif = (struct ifnet *)0;
	ti = mtod(m, struct tcpiphdr *);
	if (tp->t_template == 0)
		panic("tcp_output");
	bcopy((caddr_t)tp->t_template, (caddr_t)ti, sizeof (struct tcpiphdr));

	/*
	 * Fill in fields, remembering maximum advertised
	 * window for use in delaying messages about window sizes.
	 * If resending a FIN, be sure not to use a new sequence number.
	 */
	if (flags & TH_FIN && tp->t_flags & TF_SENTFIN && 
	    tp->snd_nxt == tp->snd_max)
		tp->snd_nxt--;
	/*
	 * If we are doing retransmissions, then snd_nxt will
	 * not reflect the first unsent octet.  For ACK only
	 * packets, we do not want the sequence number of the
	 * retransmitted packet, we want the sequence number
	 * of the next unsent octet.  So, if there is no data
	 * (and no SYN or FIN), use snd_max instead of snd_nxt
	 * when filling in ti_seq.  But if we are in persist
	 * state, snd_max might reflect one byte beyond the
	 * right edge of the window, so use snd_nxt in that
	 * case, since we know we aren't doing a retransmission.
	 * (retransmit and persist are mutually exclusive...)
	 */
	if (len || (flags & (TH_SYN|TH_FIN)) || tp->t_timer[TCPT_PERSIST])
		ti->ti_seq = htonl(tp->snd_nxt);
	else
		ti->ti_seq = htonl(tp->snd_max);
#ifdef SACK
	if (!tp->sack_disable && sack_rxmit) {
		/* 
		 * sendalot must ideally be 0 here.  If not, it means that
		 * the existence of extra options has caused the flag to be
		 * set.  We keep track of this fact until the end of this 
		 * function and loop back to again:.
		 */
		ti->ti_seq = htonl(p->rxmit);
		p->rxmit += len;
		tp->snd_cwnd -= len; /* XXX I think this is needed */
	}
#endif
	ti->ti_ack = htonl(tp->rcv_nxt);
	if (optlen) {
		bcopy((caddr_t)opt, (caddr_t)(ti + 1), optlen);
		ti->ti_off = (sizeof (struct tcphdr) + optlen) >> 2;
	}
	ti->ti_flags = flags;

#ifdef MOBILE_HOST
#ifdef PERFECT_ELN
	if (perfect_eln_enable) {
	     if ((tp->t_flags & TF_SEND_ELN) && (SEQ_LEQ(tp->rcv_nxt, tp->eln_end_seq)))
		  ti->ti_x2 |= TH_ELN;
	     else
		  tp->t_flags &= ~TF_SEND_ELN;
	}
#endif /* PERFECT_ELN */
#endif /* MOBILE_HOST */

#ifdef FAST_START
	/*
	 * Turn on the TH_FS bit in the header if fast start is enabled AND we are in 
	 * fast start mode AND this is not a retransmission AND there is a non-zero 
	 * overlap between this segment and the sequence space spanned by
	 * (fs_startseq,fs_endseq).
	 */
	if (FS(tp) && ntohl(ti->ti_seq) >= tp->snd_max && 
	    ntohl(ti->ti_seq) < tp->fs_endseq && ntohl(ti->ti_seq)+len > tp->fs_startseq) 
		ti->ti_x2 |= TH_FS;

#endif

	/*
	 * Calculate receive window.  Don't shrink window,
	 * but avoid silly window syndrome.
	 */
	if (win < (long)(so->so_rcv.sb_hiwat / 4) && win < (long)tp->t_maxseg)
		win = 0;
	if (win > (long)TCP_MAXWIN << tp->rcv_scale)
		win = (long)TCP_MAXWIN << tp->rcv_scale;
	if (win < (long)(tp->rcv_adv - tp->rcv_nxt))
		win = (long)(tp->rcv_adv - tp->rcv_nxt);
	ti->ti_win = htons((u_short) (win>>tp->rcv_scale));

	/*
	 * If no urgent pointer is outstanding, then we pull the
	 * urgent pointer to the left edge of the send window so
	 * that it doesn't drift into the send window on sequence
	 * number wraparound.  Otherwise, if the urgent pointer
	 * points into/after this packet, add it in.
	 */
	if (SEQ_GT(tp->snd_up, tp->snd_una)) {
		if (SEQ_GT(tp->snd_up, tp->snd_nxt)) {
			int urp = tp->snd_up - tp->snd_nxt;
			/*
			 * Internally we store the urgent pointer as the first
			 * byte of non-urgent data.  But in the packet, the
			 * urgent pointer is supposed to be the last byte of
			 * urgent data.  If the user specified TF_STDURG then
			 * use this behavior, otherwise use the old method.
			 */
			if (tp->t_flags & TF_STDURG)
				urp--;
			if (urp > 65535)
				urp = 65535;
			ti->ti_urp = htons((u_short)urp);
			ti->ti_flags |= TH_URG;
		}
	} else
		tp->snd_up = tp->snd_una;

	/*
	 * Put TCP length in extended header, and then
	 * checksum extended header and data.
	 */
	if (len + optlen)
		ti->ti_len = htons((u_short)(sizeof (struct tcphdr) +
		    optlen + len));
	ti->ti_sum = in_cksum(m, (int)(hdrlen + len));

	/*
	 * In transmit state, time the transmission and arrange for
	 * the retransmit.  In persist state, just set snd_max.
	 */
	if (tp->t_force == 0 || tp->t_timer[TCPT_PERSIST] == 0) {
		tcp_seq startseq = tp->snd_nxt;

		/*
		 * Advance snd_nxt over sequence space of this segment.
		 */
		if (flags & (TH_SYN|TH_FIN)) {
			if (flags & TH_SYN)
				tp->snd_nxt++;
			if (flags & TH_FIN) {
				tp->snd_nxt++;
				tp->t_flags |= TF_SENTFIN;
			}
		}
		
#ifdef SACK
		if (!tp->sack_disable) {
			if (sack_rxmit && (p->rxmit != tp->snd_nxt)) {
				goto timer;
			}
		}
#endif
		tp->snd_nxt += len;
		if (SEQ_GT(tp->snd_nxt, tp->snd_max)) {
			tp->snd_max = tp->snd_nxt;

#ifdef TCP_VERBOSE
			tp->bytecount += len;
			tp->packetcount++;
			tp->totalbytecount += (len + hdrlen);
#endif TCP_VERBOSE

#ifdef TCP_STATS
		    if ((tp->t_flags & TF_ENABLE_STATS) && (tcp_stats_flags &
			    TCP_STATS_SND_MAX))
			    tcp_stats_record(TCP_STATS_SND_MAX,tp);
#endif

			/*
			 * Time this transmission if not a retransmission and
			 * not currently timing anything.
			 */
			if (tp->t_rtt == 0) {
				tp->t_rtt = 1;
				tp->t_rtseq = startseq;
				tcpstat.tcps_segstimed++;
			}

#ifdef TCP_VERBOSE
		} else {
			tp->totalbytecount += (len + hdrlen);
			if (len) {
				tp->retransmissions++;
			} else {
				tp->packetcount++;
			}
#endif TCP_VERBOSE
		}

		/*
		 * Set retransmit timer if not currently set,
		 * and not doing an ack or a keep-alive probe.
		 * Initial value for retransmit timer is smoothed
		 * round-trip time + 2 * round-trip time variance.
		 * Initialize shift counter which is used for backoff
		 * of retransmit time.
		 */
#ifdef SACK
 timer:

#ifdef TCP_VERBOSE
	if (sack_rxmit && (p->rxmit != (tp->snd_nxt - len))) {
		if (SEQ_GT(tp->snd_nxt, tp->snd_max)) {
			tp->bytecount += len;
			tp->packetcount++;
			tp->totalbytecount += (len + hdrlen); 
		} else {
			tp->totalbytecount += (len + hdrlen);
			if (len) {
				tp->retransmissions++;
			} else {
				tp->packetcount++;
			}
		}
	}
#endif TCP_VERBOSE

		if (!tp->sack_disable && sack_rxmit &&
		    tp->t_timer[TCPT_REXMT] == 0 &&
		    tp->snd_nxt != tp->snd_max) {
			tp->t_timer[TCPT_REXMT] = tp->t_rxtcur;
			if (tp->t_timer[TCPT_PERSIST]) {
				tp->t_timer[TCPT_PERSIST] = 0;
				tp->t_rxtshift = 0;
			}
		}
#endif
		if (tp->t_timer[TCPT_REXMT] == 0 &&
#ifdef FAST_START
		    (!FS_FLR_FRT(tp) || tp->t_timer[TCPT_RESET] == 0) &&
#endif
		    tp->snd_nxt != tp->snd_una) {
			tp->t_timer[TCPT_REXMT] = tp->t_rxtcur;
#ifdef FAST_START
			if (FS_FLR_FRT(tp))
				tcp_set_reset_timer(tp);
#endif
			if (tp->t_timer[TCPT_PERSIST]) {
				tp->t_timer[TCPT_PERSIST] = 0;
				tp->t_rxtshift = 0;
			}
		}
	} else 
		if (SEQ_GT(tp->snd_nxt + len, tp->snd_max)) {
			tp->snd_max = tp->snd_nxt + len;

#ifdef TCP_VERBOSE
			tp->totalbytecount += (len + hdrlen);
			tp->packetcount++;
#endif TCP_VERBOSE

#ifdef TCP_STATS
			if ((tp->t_flags & TF_ENABLE_STATS) && 
			    (tcp_stats_flags & TCP_STATS_SND_MAX))
				tcp_stats_record(TCP_STATS_SND_MAX,tp);
#endif
		}
	
	/*
	 * Trace.
	 */
	if (so->so_options & SO_DEBUG)
		tcp_trace(TA_OUTPUT, tp->t_state, tp, ti, 0);

	/*
	 * Fill in IP length and desired time to live and
	 * send to IP level.  There should be a better way
	 * to handle ttl and tos; we could keep them in
	 * the template, but need a way to checksum without them.
	 */
	m->m_pkthdr.len = hdrlen + len;
#ifdef TUBA
	if (tp->t_tuba_pcb)
		error = tuba_output(m, tp);
	else
#endif
    {
	/* Perhaps we should set/clear this in the template??? -dab */
	if (tcp_pmtu && ((tp->t_flags & TF_NO_PMTU) == 0) &&
	    (rt = tp->t_inpcb->inp_route.ro_rt) &&
	    (rt->rt_flags & RTF_UP) &&
	    ((rt->rt_rmx.rmx_locks & RTV_MTU) == 0))
		((struct ip *)ti)->ip_off |= IP_DF;
	((struct ip *)ti)->ip_len = m->m_pkthdr.len;
	((struct ip *)ti)->ip_ttl = tp->t_inpcb->inp_ip.ip_ttl;	/* XXX */
	((struct ip *)ti)->ip_tos = tp->t_inpcb->inp_ip.ip_tos;	/* XXX */
#ifdef ACKSFIRST
	/* 
	 * If this is a pure ack, set IPTOS_LOWDELAY. The interface layer
	 * will then schedule this at a higher priority than other packets.
	 * This is useful in asymmetric networks with bidirectional traffic.
	 * 
	 * For questions/comments, please contact:
	 * Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
	 * http://www.cs.berkeley.edu/~padmanab
	 */
	if (ti->ti_flags & TH_ACK && len == 0)
		((struct ip *)ti)->ip_tos |= IPTOS_LOWDELAY;
#endif
#ifdef TCP_USER_CONTROL
	if (tp->t_xflags & TXF_LOW_DELAY)
		((struct ip *)ti)->ip_tos |= IPTOS_LOWDELAY;
#endif
#if BSD >= 43
	error = ip_output(m, tp->t_inpcb->inp_options, &tp->t_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);
#else
	error = ip_output(m, (struct mbuf *)0, &tp->t_inpcb->inp_route, 
			  so->so_options & SO_DONTROUTE);
#endif /* BSD 43 */

#ifdef TCP_VERBOSE
	if (tp->starttime == 0)
		tp->starttime = current_time();
#endif TCP_VERBOSE

    }

#ifdef FAST_START
	if (len > 0 && !error)
		tp->t_snd_idle = 0;
	if (tp->snd_max == tp->snd_una && FS_RESET_PEND(tp)) {
		TAILQ_REMOVE(&tcp_faststart, tp, t_faststart);
		tp->t_xflags &= ~TXF_FS_RESET_PEND;		
		tp->t_faststart.tqe_prev = 0;
	}
#endif

#ifdef TCP_STANDARD
	if (!error)
		tp->num_segs_since_ack = 0;
#endif

	if (error) {
out:
		switch (error) {
		case EMSGSIZE:
			/*
			 * The Path MTU must have changed.  Re-get
			 * the mtu information, and resend.
			 * XXX Should we check for a valid route???
			 */
			tp->snd_nxt -= len;
			win = tp->t_maxseg;
			len = tp->t_inpcb->inp_route.ro_rt->rt_rmx.rmx_mtu -
						sizeof(struct tcpiphdr);
			tcp_changemss(tp, len);
			if (tp->t_maxseg < win)
				goto again;
			/* XXX FALL THROUGH if maxseg didn't get smaller! */

		case ENOBUFS:
			tcp_quench(tp->t_inpcb, 0);
			/*
			 * If we can't send, make sure there is something
			 * to get us going again later.  Persist state
			 * is not necessarily right, but it is close enough.
			 */
			if (tp->t_timer[TCPT_REXMT] == 0 &&
#ifdef FAST_START
			    (!FS_FLR_FRT(tp) || tp->t_timer[TCPT_RESET] == 0)&&
#endif
			    tp->t_timer[TCPT_PERSIST] == 0) {
				tp->t_rxtshift = 0;
				tcp_setpersist(tp);
			}
			error = 0;
			break;

		case EHOSTUNREACH:
		case ENETDOWN:
			if (TCPS_HAVERCVDSYN(tp->t_state)) {
				tp->t_softerror = error;
				error = 0;
			}
			break;

		default:
			break;
		}

		return (error);
	}
	tcpstat.tcps_sndtotal++;

	/*
	 * Data sent (as far as we can tell).
	 * If this advertises a larger window than any other segment,
	 * then remember the size of the advertised window.
	 * Any pending ACK has now been sent.
	 */
	if (win > 0 && SEQ_GT(tp->rcv_nxt+win, tp->rcv_adv))
		tp->rcv_adv = tp->rcv_nxt + win;
	tp->last_ack_sent = tp->rcv_nxt;
	tp->t_flags &= ~TF_ACKNOW;
	tcp_delack_done(tp);

#ifdef TCP_STATS
	if ((tp->t_flags & TF_ENABLE_STATS) && 
	    (tcp_stats_flags & TCP_STATS_LAST_ACK_SENT))
		tcp_stats_record(TCP_STATS_LAST_ACK_SENT,tp);
#endif		


#ifdef SACK
        /* 
	 * Only send out 1 seg at a time in fast retransmit/recovery mode.
	 */
	if (!tp->sack_disable && tp->t_xflags & TXF_FR_ACTIVE)
	        sendalot = 0;
#endif

#ifdef ER
        /* 
	 * Only send out 1 seg at a time in enhanced recovery.
	 */
	if (tp->t_xflags & TXF_ER_ACTIVE) {
	        sendalot = 0;
		printf("sent out ER data %d\n", tp->snd_nxt - len - tp->iss);
#ifdef TCP_STATS
		if ((tp->t_flags & TF_ENABLE_STATS) && 
		    (tcp_stats_flags & TCP_STATS_ER))
			tcp_stats_record(TCP_STATS_ER, tp);
#endif
	}
	
#endif

#ifdef TCP_SESSION
	/* explicitly make sure that we sent at most 1 segment */
	if (SESSION(tp) && SESSION_OUTPUT(tp))
		return (0);
	/* sanity check */
	else if (SESSION(tp) && sendalot)
		printf("In tcp_output(): error: sendalot is set though output was not forced by session\n");
#endif

#ifdef LIMIT_BURST
	if (sendalot--)
#else
	if (sendalot)
#endif
#ifdef SACK
	{
		if (!tp->sack_disable && sack_rxmit)  
			sack_rxmit = 0;
#endif
		goto again;

#ifdef SACK
	}
#endif
	return (0);
}

void
tcp_setpersist(tp)
	register struct tcpcb *tp;
{
	register t = ((tp->t_srtt >> 2) + tp->t_rttvar) >> 1;

	if (tp->t_timer[TCPT_REXMT])
		panic("tcp_output REXMT");
#ifdef FAST_START
	if (FS_FLR_FRT(tp) && tp->t_timer[TCPT_RESET])
		panic("In tcp_setpersist(): TCPT_RESET is set");
#endif	
	/*
	 * Start/restart persistance timer.
	 */
	TCPT_RANGESET(tp->t_timer[TCPT_PERSIST],
	    t * tcp_backoff[tp->t_rxtshift],
	    TCPTV_PERSMIN, TCPTV_PERSMAX);
	if (tp->t_rxtshift < TCP_MAXRXTSHIFT)
		tp->t_rxtshift++;
}

