/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Author: Venkata N. Padmanabhan (padmanab@cs.berkeley.edu)
 */

#ifndef TCP_STATS_H 
#define TCP_STATS_H


/* Constants */
#define TCP_STATS_BUFFER_SIZE  32768
#define TCP_STATS_BUFFER_THRESH (TCP_STATS_BUFFER_SIZE - 50)
 
/* Flags */
#define TCP_STATS_SND_UNA      0x0001
#define TCP_STATS_SND_NXT      0x0002
#define TCP_STATS_ISS          0x0004
#define TCP_STATS_SND_WND      0x0008
#define TCP_STATS_SND_MAX      0x0010
#define TCP_STATS_SND_CWND     0x0020
#define TCP_STATS_SND_SSTHRESH 0x0040
#define TCP_STATS_RXTCUR       0x0080
#define TCP_STATS_DUPACK_RXT   0x0100
#define TCP_STATS_COARSE_TIM   0x0200
#define TCP_STATS_COARSE_RXT   0x0400
#define TCP_STATS_SMART_RXT    0x0800
#define TCP_STATS_ELN_RXT      0x1000
#define TCP_STATS_DUPACK_EVENT 0x2000
#define TCP_STATS_DELACK       0x4000
#define TCP_STATS_RTT          0x8000
#define TCP_STATS_SRTT         0x10000
#define TCP_STATS_RTTVAR       0x20000
#define TCP_STATS_SRTT_EXACT   0x40000
#define TCP_STATS_RCV_NXT      0x80000
#define TCP_STATS_LAST_ACK_SENT 0x100000
#define TCP_STATS_IRS          0x200000
#define TCP_STATS_ER           0x400000
#define TCP_STATS_ALL          0xffffff	/* to record everything */

/* Variables */
extern unsigned long tcp_stats_flags;

/* Functions */
#ifdef KERNEL
extern void tcp_stats_init(register struct tcpcb *tp, unsigned long flags);
extern void tcp_stats_record(unsigned long flags, register struct tcpcb *tp);
extern void tcp_stats_dump();
#endif

#endif  /* TCP_STATS_H */
