/*
 * Copyright (c) 1996, 1997 Berkeley Software Design, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that this notice is retained,
 * the conditions in the following notices are met, and terms applying
 * to contributors in the following notices also apply to Berkeley
 * Software Design, Inc.
 *
 *	BSDI tcp_subr.c,v 2.11 1997/01/16 14:06:35 karels Exp
 */

/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)tcp_subr.c	8.2 (Berkeley) 5/24/95
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/errno.h>
#include <sys/kernel.h>

#include <net/route.h>
#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

/* patchable/settable parameters for tcp */
extern	int tcp_mssdflt;
extern	int tcp_rttdflt;
extern	int tcp_do_rfc1323;
#ifdef SACK
extern  int tcp_do_sack;
#endif
#ifdef  ER
extern  int tcp_er_enable;
#endif

extern	struct inpcb *tcp_last_inpcb;

#ifdef TCP_STATS
#include <netinet/tcp_stats.h>

extern       struct timeval time;
#endif

#ifdef TCP_SESSION
#include <netinet/tcp_session.h>
#endif

/*
 * Tcp initialization
 */
void
tcp_init()
{

	tcp_iss = random();	/* wrong, but better than a constant */
	tcb.inp_next = tcb.inp_prev = &tcb;
	if (max_protohdr < sizeof(struct tcpiphdr))
		max_protohdr = sizeof(struct tcpiphdr);
	if (max_linkhdr + sizeof(struct tcpiphdr) > MHLEN)
		panic("tcp_init");
#ifdef FINE_GRAINED_TSTAMP
	{
		struct timeval curtime;
		
		microtime(&curtime);
		exact_time_at_slowtick = EXACT_TIME(&curtime);
		exact_time_at_fasttick = EXACT_TIME(&curtime);
	}
#endif

#ifdef FAST_START
	TAILQ_INIT(&tcp_faststart);
	if (tcp_faststart.tqh_first && (tcp_faststart.tqh_first == tcp_faststart.tqh_first->t_faststart.tqe_next))
		printf("something wrong in tcp_init()\n");
#endif
#ifdef TCP_SESSION
	session_init();
#endif
}

/*
 * Create template to be used to send tcp packets on a connection.
 * Call after host entry created, allocates an mbuf and fills
 * in a skeletal tcp/ip header, minimizing the amount of work
 * necessary when the connection is used.
 */
struct tcpiphdr *
tcp_template(tp)
	struct tcpcb *tp;
{
	register struct inpcb *inp = tp->t_inpcb;
	register struct mbuf *m;
	register struct tcpiphdr *n;

	if ((n = tp->t_template) == 0) {
		m = m_get(M_DONTWAIT, MT_HEADER);
		if (m == NULL)
			return (0);
		m->m_len = sizeof (struct tcpiphdr);
		n = mtod(m, struct tcpiphdr *);
	}
	n->ti_next = n->ti_prev = 0;
	n->ti_x1 = 0;
	n->ti_pr = IPPROTO_TCP;
	n->ti_len = htons(sizeof (struct tcpiphdr) - sizeof (struct ip));
	n->ti_src = inp->inp_laddr;
	n->ti_dst = inp->inp_faddr;
	n->ti_sport = inp->inp_lport;
	n->ti_dport = inp->inp_fport;
	n->ti_seq = 0;
	n->ti_ack = 0;
	n->ti_x2 = 0;
	n->ti_off = 5;
	n->ti_flags = 0;
	n->ti_win = 0;
	n->ti_sum = 0;
	n->ti_urp = 0;
	return (n);
}

/*
 * Send a single message to the TCP at address specified by
 * the given TCP/IP header.  If m == 0, then we make a copy
 * of the tcpiphdr at ti and send directly to the addressed host.
 * This is used to force keep alive messages out using the TCP
 * template for a connection tp->t_template.  If flags are given
 * then we send a message back to the TCP which originated the
 * segment ti, and discard the mbuf containing it and any other
 * attached mbufs.
 *
 * In any case the ack and sequence number of the transmitted
 * segment are as specified by the parameters.
 */
int
tcp_respond(tp, ti, m, ack, seq, flags)
	struct tcpcb *tp;
	register struct tcpiphdr *ti;
	register struct mbuf *m;
	tcp_seq ack, seq;
	int flags;
{
	register int tlen, optlen = 0;
	int win = 0;
	struct route *ro = 0;

	if (tp) {
		win = sbspace(&tp->t_inpcb->inp_socket->so_rcv);
		ro = &tp->t_inpcb->inp_route;
	}
	if (m == 0) {
		m = m_gethdr(M_DONTWAIT, MT_HEADER);
		if (m == NULL)
			return(ENOBUFS);
#ifdef TCP_COMPAT_42
		tlen = 1;
#else
		tlen = 0;
#endif
		m->m_data += max_linkhdr;
		*mtod(m, struct tcpiphdr *) = *ti;
		ti = mtod(m, struct tcpiphdr *);
		flags = TH_ACK;
	} else {
		m_freem(m->m_next);
		m->m_next = 0;
		m->m_data = (caddr_t)ti;
		m->m_len = sizeof (struct tcpiphdr);
		tlen = 0;
#define xchg(a,b,type) { type t; t=a; a=b; b=t; }
		xchg(ti->ti_dst.s_addr, ti->ti_src.s_addr, u_long);
		xchg(ti->ti_dport, ti->ti_sport, u_short);
#undef xchg
	}
	ti->ti_next = ti->ti_prev = 0;
	ti->ti_x1 = 0;
	ti->ti_seq = htonl(seq);
	ti->ti_ack = htonl(ack);
	ti->ti_x2 = 0;
	if ((flags & TH_SYN) == 0) {
		if (tp)
			ti->ti_win = htons((u_short) (win >> tp->rcv_scale));
		else
			ti->ti_win = htons((u_short)win);
		ti->ti_off = sizeof (struct tcphdr) >> 2;
		tlen += sizeof (struct tcphdr);
	} else
		tlen += ti->ti_off << 2;
	ti->ti_len = htons((u_short)tlen);
	tlen += sizeof (struct ip);
	m->m_len = tlen;
	m->m_pkthdr.len = tlen;
	m->m_pkthdr.rcvif = (struct ifnet *) 0;
	ti->ti_flags = flags;
	ti->ti_urp = 0;
	ti->ti_sum = 0;
	ti->ti_sum = in_cksum(m, tlen);
	((struct ip *)ti)->ip_len = tlen;
	((struct ip *)ti)->ip_ttl = ip_defttl;
	return(ip_output(m, NULL, ro, 0, NULL));
}

/*
 * Create a new TCP control block, making an
 * empty reassembly queue and hooking it to the argument
 * protocol control block.
 */
struct tcpcb *
tcp_newtcpcb(inp)
	struct inpcb *inp;
{
	register struct tcpcb *tp;

	tp = malloc(sizeof(*tp), M_PCB, M_NOWAIT);
	if (tp == NULL)
		return ((struct tcpcb *)0);
	bzero((char *) tp, sizeof(struct tcpcb));
	tp->seg_next = tp->seg_prev = (struct tcpiphdr *)tp;
	tp->t_maxseg = tcp_mssdflt;
	tp->t_peermaxseg = tcp_mssdflt;

	tp->t_flags = tcp_do_rfc1323 ? (TF_USE_SCALE|TF_SEND_TSTMP) : 0;
	tp->t_xflags = 0;
	tp->t_dupacks = 0;

#ifdef SACK
	if (tcp_do_sack)
		tp->t_flags |= TF_SACK_PERMIT;
	else 
	        tp->sack_disable = 1;
#endif

#ifdef ER
	if (tcp_er_enable)	/* enable ER; may be disabled using sockopt */
		tp->t_xflags |= TXF_ER;
	else
		tp->t_xflags &= ~TXF_ER;
#endif

#ifdef ELN
	tp->eln_last_rxmit = tp->iss;
#endif
	
	tp->t_inpcb = inp;
	/*
	 * Init srtt to TCPTV_SRTTBASE (0), so we can tell that we have no
	 * rtt estimate.  Set rttvar so that srtt + 4 * rttvar gives
	 * reasonable initial retransmit time (tcp_rttdflt seconds).
	 */
	tp->t_srtt = TCPTV_SRTTBASE;
#ifdef FINE_GRAINED_TSTAMP
	/* 
	 * XXX It is okay to set both these to zero because proper 
	 * initialization will be done in tcp_xmit_timer_exact() before these 
	 * values are used to set a timer.
	 */
	tp->t_srtt_exact = TCPTV_SRTTBASE_EXACT;
	tp->t_rttvar_exact = 0; 
	tp->t_rxtcur_exact = tp->t_srtt_exact + (tp->t_rttvar_exact << 2);
#endif
#ifdef FAST_START
	tp->t_snd_idle = 0;
	tp->t_timer[TCPT_RESET] = 0;
	tp->fs_cwnd = 0;
#endif

#ifdef TCP_STANDARD
	tp->num_segs_since_ack = 0;
#endif

#ifdef TCP_SESSION
	tp->session = NULL;
#endif

	tp->t_rttvar = tcp_rttdflt * PR_SLOWHZ /* / 4 << 2 */;
	tp->t_rttmin = TCPTV_MIN;
	TCPT_RANGESET(tp->t_rxtcur, TCP_REXMTVAL(tp),
		      TCPTV_MIN, TCPTV_REXMTMAX);
	tp->snd_cwnd = TCP_MAXWIN << TCP_MAX_WINSHIFT;
	tp->snd_ssthresh = TCP_MAXWIN << TCP_MAX_WINSHIFT;

#if defined(TCP_STATS) && defined(TOMH) 
	tp->last_cwnd = 0;
#endif

#ifdef SMART
	tp->smart_snd_ceil = 0;
	tp->smart_last_time_rexmt.tv_sec=tp->smart_last_time_rexmt.tv_usec=0;
#endif

/* 
 * XXX Only non-zero fields need to be initialized because of the bzero().
 */
#ifdef DPC
	tp->peer_mss = 0;
#endif

#ifdef ACC
	tp->delack = 2;                /* delayed ack factor */
	tp->max_delack = 0;            /* max. delayed ack factor */
	tp->peer_win_in_segs = 0;      /* size of peer's window (in segs) */
	tp->min_acks_per_win = 3;      /* minimum # of acks per window */
#endif

#ifdef LIMIT_BURST
	tp->maxburst = 0;
	tp->burst_pending = 0;
#endif

#ifdef TCP_USER_CONTROL
	tp->snd_cwnd_scale = 0;
#endif

#ifdef TCP_STATS
	if ((tp->t_flags & TF_ENABLE_STATS) && (tcp_stats_flags &
						(TCP_STATS_SND_CWND |
						 TCP_STATS_SND_MAX  |
						 TCP_STATS_RXTCUR   |
						 TCP_STATS_SND_SSTHRESH |
						 TCP_STATS_DELACK |
						 TCP_STATS_SRTT |
						 TCP_STATS_RTTVAR)))
		tcp_stats_record(TCP_STATS_SND_CWND | TCP_STATS_SND_MAX |
				 TCP_STATS_RXTCUR | TCP_STATS_SND_SSTHRESH |
				 TCP_STATS_DELACK | TCP_STATS_SRTT |
				 TCP_STATS_RTTVAR, tp);
#endif
	inp->inp_ip.ip_ttl = ip_defttl;
	inp->inp_ppcb = (caddr_t)tp;

#ifdef TCP_VERBOSE
	tp->starttime = 0;
	tp->bytecount = tp->totalbytecount = tp->bytesreceived = 0;
	tp->retransmissions = tp->packetcount = tp->ackcount = 0;
#endif TCP_VERBOSE

	return (tp);
}

/*
 * Drop a TCP connection, reporting
 * the specified error.  If connection is synchronized,
 * then send a RST to peer.
 */
struct tcpcb *
tcp_drop(tp, errno)
	register struct tcpcb *tp;
	int errno;
{
	struct socket *so = tp->t_inpcb->inp_socket;

	if (TCPS_HAVERCVDSYN(tp->t_state)) {
		tp->t_state = TCPS_CLOSED;
		(void) tcp_output(tp);
		tcpstat.tcps_drops++;
	} else
		tcpstat.tcps_conndrops++;
	if (errno == ETIMEDOUT && tp->t_softerror)
		errno = tp->t_softerror;
	so->so_error = errno;
	return (tcp_close(tp));
}

/*
 * Close a TCP control block:
 *	discard all space held by the tcp
 *	discard internet protocol block
 *	wake up any sleepers
 */
struct tcpcb *
tcp_close(tp)
	register struct tcpcb *tp;
{
	register struct tcpiphdr *t;
	struct inpcb *inp = tp->t_inpcb;
	struct socket *so = inp->inp_socket;
	register struct mbuf *m;
#ifdef SACK
	struct sackhole *p, *q;
#endif
#ifdef RTV_RTT
	register struct rtentry *rt;

	/*
	 * If we sent enough data to get some meaningful characteristics,
	 * save them in the routing entry.  'Enough' is arbitrarily 
	 * defined as the sendpipesize (default 4K) * 16.  This would
	 * give us 16 rtt samples assuming we only get one sample per
	 * window (the usual case on a long haul net).  16 samples is
	 * enough for the srtt filter to converge to within 5% of the correct
	 * value; fewer samples and we could save a very bogus rtt.
	 *
	 * Don't update the default route's characteristics and don't
	 * update anything that the user "locked".
	 */
	if (SEQ_LT(tp->iss + so->so_snd.sb_hiwat * 16, tp->snd_max) &&
	    (rt = inp->inp_route.ro_rt) &&
	    ((struct sockaddr_in *)rt_key(rt))->sin_addr.s_addr != INADDR_ANY) {
		register u_long i;

		if ((rt->rt_rmx.rmx_locks & RTV_RTT) == 0) {
			i = tp->t_srtt *
			    (RTM_RTTUNIT / (PR_SLOWHZ * TCP_RTT_SCALE));
			if (rt->rt_rmx.rmx_rtt && i)
				/*
				 * filter this update to half the old & half
				 * the new values, converting scale.
				 * See route.h and tcp_var.h for a
				 * description of the scaling constants.
				 */
				rt->rt_rmx.rmx_rtt =
				    (rt->rt_rmx.rmx_rtt + i) / 2;
			else
				rt->rt_rmx.rmx_rtt = i;
		}
		if ((rt->rt_rmx.rmx_locks & RTV_RTTVAR) == 0) {
			i = tp->t_rttvar *
			    (RTM_RTTUNIT / (PR_SLOWHZ * TCP_RTTVAR_SCALE));
			if (rt->rt_rmx.rmx_rttvar && i)
				rt->rt_rmx.rmx_rttvar =
				    (rt->rt_rmx.rmx_rttvar + i) / 2;
			else
				rt->rt_rmx.rmx_rttvar = i;
		}
		/*
		 * update the pipelimit (ssthresh) if it has been updated
		 * already or if a pipesize was specified & the threshhold
		 * got below half the pipesize.  I.e., wait for bad news
		 * before we start updating, then update on both good
		 * and bad news.
		 */
		if ((rt->rt_rmx.rmx_locks & RTV_SSTHRESH) == 0 &&
		    (i = tp->snd_ssthresh) && rt->rt_rmx.rmx_ssthresh ||
		    i < (rt->rt_rmx.rmx_sendpipe / 2)) {
			/*
			 * convert the limit from user data bytes to
			 * packets then to packet data bytes.
			 */
			i = (i + tp->t_maxseg / 2) / tp->t_maxseg;
			if (i < 2)
				i = 2;
			i *= (u_long)(tp->t_maxseg + sizeof (struct tcpiphdr));
			if (rt->rt_rmx.rmx_ssthresh)
				rt->rt_rmx.rmx_ssthresh =
				    (rt->rt_rmx.rmx_ssthresh + i) / 2;
			else
				rt->rt_rmx.rmx_ssthresh = i;
		}
	}
#endif /* RTV_RTT */
	/* free the reassembly queue, if any */
	t = tp->seg_next;
	while (t != (struct tcpiphdr *)tp) {
		t = (struct tcpiphdr *)t->ti_next;
		m = REASS_MBUF((struct tcpiphdr *)t->ti_prev);
		remque(t->ti_prev);
		m_freem(m);
	}
#ifdef SACK
	/*
	 * Free SACK holes.
	 */
	q = p = tp->snd_holes;
	while (p != 0) {
		q = p->next;
		free(p, M_PCB);
		p = q;
	}
#endif

#ifdef FAST_START
	/* remove this connection from list of fast start connections */
	if (tp->t_faststart.tqe_prev) {
		TAILQ_REMOVE(&tcp_faststart, tp, t_faststart);
		tp->t_xflags &= ~TXF_FS_RESET_PEND;
		if (tcp_faststart.tqh_first && (tcp_faststart.tqh_first == tcp_faststart.tqh_first->t_faststart.tqe_next))
			printf("something wrong in tcp_close()\n");
		tp->t_faststart.tqe_prev = 0;
	}
#endif

#ifdef TCP_SESSION
	/* remove this connection from its session */
	if (SESSION(tp) && tp->session)
		session_delconn(tp);
#endif
	if (tp->t_template)
		(void) m_free(dtom(tp->t_template));
	tcp_delack_done(tp);	/* just in case */
	if (tp->t_state == TCPS_TIME_WAIT)
		tcp_cancel2msl(inp, tp);
	free(tp, M_PCB);
	inp->inp_ppcb = 0;

	/*
	 * Flush any remaining data.  We might have data in the output queue
	 * if connection was reset, and the socket might not be closed
	 * for a while if user does not notice.
	 */
	sbdrop(&so->so_snd, (int)so->so_snd.sb_cc);
	soisdisconnected(so);

	/* clobber input pcb cache if we're closing the cached connection */
	if (inp == tcp_last_inpcb)
		tcp_last_inpcb = &tcb;
	if (inp->inp_hlist.le_prev)	/* XXX */
		LIST_REMOVE(inp, inp_hlist);
	in_pcbdetach(inp);
	tcpstat.tcps_closed++;
	return ((struct tcpcb *)0);
}

void
tcp_drain()
{

}

/*
 * Notify a tcp user of an asynchronous error;
 * store error as soft error, but wake up user
 * (for now, won't do anything until can select for soft error).
 */
void
tcp_notify(inp, error)
	struct inpcb *inp;
	int error;
{
	register struct tcpcb *tp = (struct tcpcb *)inp->inp_ppcb;
	register struct socket *so = inp->inp_socket;

	/*
	 * If we are hooked up. do not report errors directly,
	 * but record them as soft errors in case we time out.
	 * If connection hasn't completed, has retransmitted several times,
	 * and receives a second error, give up now.  This is better
	 * than waiting a long time to establish a connection that
	 * can never complete.
	 */
#if 0
	if (tp->t_state == TCPS_ESTABLISHED &&
	     (error == EHOSTUNREACH || error == ENETUNREACH ||
	      error == EHOSTDOWN)) {
		tp->t_softerror = error;
		return;
	} else
#endif
	if (tp->t_state < TCPS_ESTABLISHED && tp->t_rxtshift >= 3 &&
	    tp->t_softerror)
		so->so_error = error;
	else 
		tp->t_softerror = error;
	wakeup((caddr_t) &so->so_timeo);
	sorwakeup(so);
	sowwakeup(so);
}

void
tcp_ctlinput(cmd, sa, ip)
	int cmd;
	struct sockaddr *sa;
	register struct ip *ip;
{
	register struct tcphdr *th;
	extern struct in_addr zeroin_addr;
	extern u_char inetctlerrmap[];
	void (*notify) __P((struct inpcb *, int)) = tcp_notify;
	int nmatch;

	if (cmd == PRC_QUENCH)
		notify = tcp_quench;
	else if (cmd == PRC_IFNEWADDR)
		in_mrejoin(&tcb);
	else if (cmd == PRC_MSGSIZE)
		notify = tcp_mtudisc;
	else if (!PRC_IS_REDIRECT(cmd) &&
		 ((unsigned)cmd > PRC_NCMDS || inetctlerrmap[cmd] == 0))
		return;
	if (ip) {
		th = (struct tcphdr *)((caddr_t)ip + (ip->ip_hl << 2));
		nmatch = in_pcbnotify(&tcb, sa, th->th_dport, ip->ip_src,
		    th->th_sport, cmd, notify);
		if (nmatch == 0 && syn_cache_count &&
		    (inetctlerrmap[cmd] == EHOSTUNREACH ||
		    inetctlerrmap[cmd] == ENETUNREACH ||
		    inetctlerrmap[cmd] == EHOSTDOWN))
			syn_cache_unreach(ip, th);
	} else
		(void) in_pcbnotify(&tcb, sa, 0, zeroin_addr, 0, cmd, notify);
}

/*
 * When a source quench is received, close congestion window
 * to one segment.  We will gradually open it again as we proceed.
 */
void
tcp_quench(inp, errno)
	struct inpcb *inp;
	int errno;
{
	struct tcpcb *tp = intotcpcb(inp);

	if (tp->t_xflags & TXF_NO_QUENCH)
		return;
	if (tp) {
		tp->snd_cwnd = tp->t_maxseg;
#ifdef TCP_STATS
		if ((tp->t_flags & TF_ENABLE_STATS) && (tcp_stats_flags &
			(TCP_STATS_SND_CWND | TCP_STATS_SND_MAX)))
		  tcp_stats_record(TCP_STATS_SND_CWND | TCP_STATS_SND_MAX, tp);

#endif
	      }
		
}

/*
 * When a Destination Unreachable/Fragmentation Needed and DF set
 * is received, this routine is called for every TCP connection that
 * has a problematic destination:
 *	If current route isn't route-to-host then purge it and get new route;
 *		icmp_mtudisc() has already created a new route for us to use
 *	Change t_maxseg
 *	Retransmit dropped segment
 * Note that some of these connections might NOT need to retransmit;
 * we accept that there may be a few spurious retransmissions.
 */
void
tcp_mtudisc(inp, errno)
	struct inpcb *inp;
	int errno;
{
	struct tcpcb *tp = intotcpcb(inp);
	struct route *ro = &(inp->inp_route);
	int usable_mtu;

	if (tp == NULL)
		return;

	/*
	 * Note that we only re-route if we already had a route;
	 * this should avoid creating a route if SO_DONTROUTE was set
	 */
	if (ro->ro_rt == NULL)
		return;

	/*
	 * Make sure we have a route-to-host for this destination.
	 */
	if ((ro->ro_rt->rt_flags & RTF_HOST) == 0) {
		in_rtchange(inp, 0);
		rtcalloc(ro);
		if (ro->ro_rt == NULL) {
			printf("tcp_mtudisc: no new route?\n");
			tcp_changemss(tp, TCP_MSS);
			return;
		}
		if ((ro->ro_rt->rt_flags & RTF_HOST) == 0) {
			printf("tcp_mtudisc: new route not to host?\n");
			usable_mtu = ro->ro_rt->rt_ifp->if_mtu
					- sizeof(struct tcpiphdr);
			tcp_changemss(tp, min(usable_mtu, TCP_MSS));
			return;
		}
	}

	usable_mtu = ro->ro_rt->rt_rmx.rmx_mtu - sizeof(struct tcpiphdr);
	tcp_changemss(tp, usable_mtu);

	/*
	 * If there is more than 1 segment of unacknowledged data,
	 * force a retransmission of the (probably) lost datagrams
	 */
	if (SEQ_GT(tp->snd_una, tp->snd_nxt + tp->t_maxseg)) {
		tp->snd_nxt = tp->snd_una;
		(void) tcp_output(tp);
	}
}

/*
 * Called when the MSS should change to reflect a new MTU value.
 */
void
tcp_changemss(tp, usable_mtu)
	register struct tcpcb *tp;
	u_int usable_mtu;
{
	u_int win;
	int newmaxseg;

	newmaxseg = tcp_maxseg(tp, min(usable_mtu, tp->t_peermaxseg));

	if (tp->t_maxseg == newmaxseg)
		return;				/* no change */

	tp->t_maxseg = newmaxseg;

	/*
	 * Readjust things that depend upon t_maxseg
	 * (code copied from tcp_timer.c in the REXMT case)
	 */
	win = min(tp->snd_wnd, tp->snd_cwnd) / 2 / tp->t_maxseg;
	if (win < 2)
		win = 2;
	tp->snd_cwnd = tp->t_maxseg;
	tp->snd_ssthresh = win * tp->t_maxseg;
	tp->t_dupacks = 0;
}

/*
 * Check to see if our Path MTU information is out of date.
 * If so, try again to use a higher MTU and see if things have
 * improved.
 */

tcp_agepathmtu(inp, rt)
	struct inpcb *inp;
	register struct rtentry *rt;
{
	unsigned int usable_mtu;
	extern unsigned int mtu_table[];
	unsigned int *mtup, rt_mtu = rt->rt_rmx.rmx_mtu;
	time_t expire_at;

	/* Find the next higher MTU plateau */
	for (mtup = mtu_table; *mtup; mtup++) {
		if (mtup[1] <= rt_mtu)
			break;
	}
	usable_mtu = *mtup;

	/* Don't try an MTU greater than the if_mtu! */
	if (usable_mtu > rt->rt_ifp->if_mtu)
		usable_mtu = rt->rt_ifp->if_mtu;

	if (usable_mtu > rt_mtu) {
		expire_at = time.tv_sec + pmtu_probe;
		rt->rt_rmx.rmx_mtu = usable_mtu;

		usable_mtu -= sizeof(struct tcpiphdr);
		tcp_changemss(intotcpcb(inp), usable_mtu);
	} else {
		expire_at = time.tv_sec + pmtu_expire;
	}
	rt->rt_flags &= ~RTF_PROBEMTU;
	rt_timer_add(rt, expire_at, icmp_mtuexpire, NULL);
}
