/*	BSDI elf.h,v 1.2 1996/01/04 05:38:24 donn Exp	*/

#ifndef _SYS_ELF_H_
#define _SYS_ELF_H_ 1

/*
 * Structures of ELF binary files.
 * The information here is derived from the System V ABI section 4
 * and from the SVr4 API ELF library man pages.
 */


typedef unsigned long Elf32_Addr;	/* program addresses */
typedef unsigned short Elf32_Half;	/* 16 bit field */
typedef unsigned long Elf32_Off;	/* file offsets */
typedef long Elf32_Sword;		/* signed integers */
typedef unsigned long Elf32_Word;	/* 32-bit field */


/*
 * The ELF header.
 */

#define	EI_NIDENT	16

typedef struct {
	unsigned char	e_ident[EI_NIDENT];
	Elf32_Half	e_type;
	Elf32_Half	e_machine;
	Elf32_Word	e_version;
	Elf32_Addr	e_entry;
	Elf32_Off	e_phoff;	/* program header table offset */
	Elf32_Off	e_shoff;	/* section header table offset */
	Elf32_Word	e_flags;
	Elf32_Half	e_ehsize;	/* ELF header size in bytes */
	Elf32_Half	e_phentsize;	/* program header size in bytes */
	Elf32_Half	e_phnum;	/* number of program headers */
	Elf32_Half	e_shentsize;	/* section header size in bytes */
	Elf32_Half	e_shnum;	/* number of section headers */
	Elf32_Half	e_shstrndx;	/* SHT index of section name strings */
} Elf32_Ehdr;

/*
 * Symbolic names for indices into the e_ident field.
 * Why wasn't e_ident declared as a structure?
 */
#define	EI_MAG0		0		/* the magic number */
#define	EI_MAG1		1
#define	EI_MAG2		2
#define	EI_MAG3		3
#define	EI_CLASS	4		/* 32 vs. 64 bits */
#define	EI_DATA		5		/* byte ordering */
#define	EI_VERSION	6		/* ELF header version */
#define	EI_PAD		7

#define	ELFMAG0		0x7f
#define	ELFMAG1		0x45		/* ASCII 'E' */
#define	ELFMAG2		0x4c		/* ASCII 'L' */
#define	ELFMAG3		0x46		/* ASCII 'F' */

#define	ELFCLASSNONE	0
#define	ELFCLASS32	1
#define	ELFCLASS64	2

#define	ELFDATANONE	0
#define	ELFDATA2LSB	1		/* little endian */
#define	ELFDATA2MSB	2		/* big endian */

#define	EV_NONE		0
#define	EV_CURRENT	1		/* the current version */

#define	EM_NONE		0
#define	EM_M32		1
#define	EM_SPARC	2
#define	EM_386		3
#define	EM_68K		4
#define	EM_88K		5
#define	EM_860		7
#define	EM_MIPS		8
#define	EM_MIPS_RS4_BE	10
#define	EM_SPARC64	11
#define	EM_PARISC	15
#define	EM_PPC		20

#define	ET_NONE		0
#define	ET_REL		1		/* ELF relocatable file type */
#define	ET_EXEC		2		/* ELF executable file type */
#define	ET_DYN		3		/* ELF shared object file type */
#define	ET_CORE		4		/* ELF core file type */
#define	ET_LOPROC	0xff00		/* reserved for implementations ... */
#define	ET_HIPROC	0xffff		/* ... through this value */


/*
 * ELF section headers.
 */

/*
 * The section headers appear in an array;
 * the index of an element is its 'section header table index'.
 * Certain fields contain section header table indices;
 * the following reserved values may appear in those fields
 * even though they don't refer to actual sections.
 */
#define	SHN_UNDEF	0
#define	SHN_LORESERVE	0xff00
#define	SHN_LOPROC	0xff00
#define	SHN_HIPROC	0xff1f
#define	SHN_ABS		0xfff1
#define	SHN_COMMON	0xfff2
#define	SHN_HIRESERVE	0xffff

typedef struct {
	Elf32_Word	sh_name;	/* index into section header strings */
	Elf32_Word	sh_type;
	Elf32_Word	sh_flags;
	Elf32_Addr	sh_addr;	/* load address */
	Elf32_Off	sh_offset;	/* offset in file */
	Elf32_Word	sh_size;
	Elf32_Word	sh_link;	/* hack dependent on section type */
	Elf32_Word	sh_info;	/* hack dependent on section type */
	Elf32_Word	sh_addralign;
	Elf32_Word	sh_entsize;	/* size of entries in section */
} Elf32_Shdr;

/*
 * Quick reminder about the sh_link and sh_info hacks...
 * SHT_DYNAMIC: sh_link = SHT index for string table
 * SHT_HASH: sh_link = SHT index for symbol table
 * SHT_REL, SHT_RELA: sh_link = SHT index for symbol table,
 *	sh_info = SHT index for relocated section
 * SHT_SYMTAB, SHT_DYNSYM: sh_link = SHT index for string table,
 *	sh_link = symbol table index of last local symbol + 1
 */

#define	SHT_NULL	0
#define	SHT_PROGBITS	1		/* information unique to the program */
#define	SHT_SYMTAB	2		/* full symbol table */
#define	SHT_STRTAB	3		/* string table */
#define	SHT_RELA	4		/* relocations with addends */
#define	SHT_HASH	5		/* symbol hash table */
#define	SHT_DYNAMIC	6		/* dynamic linking information */
#define	SHT_NOTE	7		/* information about the file */
#define	SHT_NOBITS	8		/* the empty section */
#define	SHT_REL		9		/* relocations without addends */
#define	SHT_SHLIB	10		/* reserved */
#define	SHT_DYNSYM	11		/* reduced, loaded symbol table */
#define	SHT_LOPROC	0x70000000
#define	SHT_HIPROC	0x7fffffff
#define	SHT_LOUSER	0x80000000
#define	SHT_HIUSER	0xffffffff

#define	SHF_WRITE	0x1		/* section in a writable segment */
#define	SHF_ALLOC	0x2		/* this section is loaded */
#define	SHF_EXECINSTR	0x4		/* this section is executable */
#define	SHF_MASKPROC	0xf0000000

/*
 * For reference, the names of the reserved sections...
 * A = SHF_ALLOC, W = SHF_WRITE, X = SHF_EXECINSTR
 *
 * .bss (SHT_NOBITS/AW): BSS segment, zero-filled data
 * .comment (SHT_PROGBITS): version control information
 * .data (SHT_PROGBITS/AW): the usual data segment
 * .data1: same as .data
 * .debug* (SHT_PROGBITS): DWARF symbolic debugging information
 * .dynamic (SHT_DYNAMIC/AW): the _DYNAMIC[] array
 * .dynstr (SHT_STRTAB/A): dynamic linking string table
 * .dynsym (SHT_DYNSYM/A): dynamic linking symbol table
 * .fini (SHT_PROGBITS/AX): global destructors
 * .got (SHT_PROGBITS/AW): global offset table
 * .hash (SHT_HASH/A): symbol hash table
 * .init (SHT_PROGBITS/AX): global constructors
 * .interp (SHT_PROGBITS/A?): interpreter pathname
 * .line (SHT_PROGBITS): DWARF line number debugging information
 * .note (SHT_NOTE): variable length note records
 * .plt (SHT_PROGBITS/AX): procedure linkage table
 * .rel<section-name> (SHT_REL/A?): relocation records
 * .rela<section-name> (SHT_RELA/A?): relocation records with addends
 * .rodata (SHT_PROGBITS/A): read-only data
 * .rodata1: same as .rodata
 * .shstrtab (SHT_STRTAB): section name string table
 * .strtab (SHT_STRTAB/A?): string table
 * .symtab (SHT_SYMTAB/A?): symbol table
 * .text (SHT_PROGBITS/AX): the usual text segment
 */

/*
 * The symbol table.
 */

typedef struct {
	Elf32_Word	st_name;	/* index into the string table */
	Elf32_Addr	st_value;
	Elf32_Word	st_size;	/* size of the object */
	unsigned char	st_info;	/* type and binding attributes */
	unsigned char	st_other;
	Elf32_Half	st_shndx;	/* SHT index */
} Elf32_Sym;

/* macro for packing st_info */
#define	ELF32_ST_INFO(bind, type) \
	(((bind) << 4) | ((type) & 0xf))
/* macros for unpacking st_info */
#define	ELF32_ST_BIND(info) \
	((info) >> 4)
#define	ELF32_ST_TYPE(info) \
	((info) & 0xf)

/* values for 'bind' field */
#define	STB_LOCAL	0x0
#define	STB_GLOBAL	0x1
#define	STB_WEAK	0x2
#define	STB_LOPROC	0xd
#define	STB_HIPROC	0xf

/* values for 'type' field */
#define	STT_NOTYPE	0x0
#define	STT_OBJECT	0x1
#define	STT_FUNC	0x2
#define	STT_SECTION	0x3
#define	STT_FILE	0x4
#define	STT_LOPROC	0xd
#define	STT_HIPROC	0xf

/*
 * Relocation entries.
 * The 'rela' form is for systems that
 * don't put the addend bits at the relocated offset in the section.
 */

typedef struct {
	Elf32_Addr	r_offset;
	Elf32_Word	r_info;
} Elf32_Rel;

typedef struct {
	Elf32_Addr	r_offset;
	Elf32_Word	r_info;
	Elf32_Sword	r_addend;
} Elf32_Rela;

/* macro for packing r_info */
#define	ELF32_R_INFO(index, type) \
	(((index) << 8) | ((type) & 0xff))
/* macros for unpacking r_info */
#define	ELF32_R_SYM(info) \
	((info) >> 8)
#define	ELF32_R_TYPE(info) \
	((info) & 0xff)


/*
 * ELF program headers.
 */

typedef struct {
	Elf32_Word	p_type;
	Elf32_Off	p_offset;
	Elf32_Addr	p_vaddr;	/* virtual address after loading */
	Elf32_Addr	p_paddr;	/* physical address after loading */
	Elf32_Word	p_filesz;	/* size of segment in file */
	Elf32_Word	p_memsz;	/* size of segment in memory */
	Elf32_Word	p_flags;
	Elf32_Word	p_align;
} Elf32_Phdr;

#define	PT_NULL		0
#define	PT_LOAD		1
#define	PT_DYNAMIC	2
#define	PT_INTERP	3		/* interpreter name */
#define	PT_NOTE		4
#define	PT_SHLIB	5		/* reserved, unused */
#define	PT_PHDR		6		/* PHT segment, if loaded */
#define	PT_LOPROC	0x70000000
#define	PT_HIPROC	0x7fffffff

#define	PF_X		0x1
#define	PF_W		0x2
#define	PF_R		0x4
#define	PF_MASKPROC	0xf0000000



/*
 * An 'auxiliary vector' is passed on the stack to ELF linker/loaders.
 * An auxv_t is an element of the vector; an AT_NULL element terminates it.
 * The information comes from p 3-29 of the Power PC ABI.
 */
typedef struct {
	int	a_type;
	union {
		long	a_val;
		void	*a_ptr;
		void	(*a_fcn)();
	} a_un;
} auxv_t;

#define	AT_NULL		0
#define	AT_IGNORE	1
#define	AT_EXECFD	2		/* file descriptor for loaded file */
#define	AT_PHDR		3		/* PHT address in loaded file */
#define	AT_PHENT	4		/* PHT element size */
#define	AT_PHNUM	5		/* PHT vector element count */
#define	AT_PAGESZ	6
#define	AT_BASE		7		/* loader's 'base address' */
#define	AT_FLAGS	8
#define	AT_ENTRY	9		/* loaded file's entry point */



#include <machine/elf.h>

#ifdef KERNEL

struct text;
int exec_elf_binary __P((struct text *));

#endif

#endif
